/* This is the collection of routines and event handlers for the RDB Profiler feature to be added
   to AppView.

   The Profiler generates a table of the different resource types within an application, and the
   number of records and bytes used in each resource type.  This helps users to find out which
   parts of an application are using the most space.

   In AppView's case it'd probably be bitmaps and strings from the About box.
*/


#include <Pilot.h>
#include "AppView.h"
#include "ProfilerHandler.h"
#include "Resources.h"
#include "Callbacks.h"


static VoidHand profileH;
/* The following pointer is valid and locked for the entire time the form is open. */
static ProfileTypePtr profile;
static UInt firstRow;


void tableDrawFunction(VoidPtr table, Word row, Word column, RectanglePtr bounds) {
  /* custom draw function - row data is index into the profile, 
     item data is column number */

  ULong index;
  ULong resType;
  Word font;
  ProfileTypePtr profile;
  Char buffer[12];

  CALLBACK_PROLOGUE

  /*  I had hoped this function would be able to access the global profile pointer.
      It can't - but it can access the global profile handle.
      I suspect it's either a side effect of callback prologue, or our callback is being
      executed in a different memory space, so the pointer doesn't translate the same way. */
  WinEraseRectangle(bounds, 0);

  profile = MemHandleLock(profileH);
  if (profile) {
    index = TblGetRowData(table, row);
    switch (TblGetItemInt(table, row, column)) {
    case 0:
      StrNCopy(buffer, (CharPtr)&profile->entry[index].type, 4);
      buffer[4] = 0;
      break;

    case 1:
      StrIToA(buffer, profile->entry[index].numRecords);
      break;

    case 2:
      StrIToA(buffer, profile->entry[index].numBytes);
      break;
    }
    
    WinDrawChars(buffer, StrLen(buffer), bounds->topLeft.x, bounds->topLeft.y);

    MemHandleUnlock(profileH);
  }

  CALLBACK_EPILOGUE

}


void displayProfileTable(ProfileTypePtr profile) {
  FormPtr frm = FrmGetActiveForm();
  TablePtr table;
  UInt numRows;
  UInt x;

  table = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ProfileTable));
  numRows = TblGetNumberOfRows(table);

  for (x = 0; x < numRows; x++) {
    TblSetRowSelectable(table, x, false);
    if (x + firstRow < profile->numEntries) {
      TblSetRowUsable(table, x, true);

      TblSetRowData(table, x, x);

      TblSetItemStyle(table, x, 0, customTableItem);
      TblSetItemInt(table, x, 0, 0);

      TblSetItemStyle(table, x, 1, customTableItem);
      TblSetItemInt(table, x, 1, 1);

      TblSetItemStyle(table, x, 2, customTableItem);
      TblSetItemInt(table, x, 2, 2);
    } else {
      TblSetRowUsable(table, x, false);
    }
  }

  for (x = 0; x < 3; x++) {
    TblSetColumnUsable(table, x, true);
    TblSetCustomDrawProcedure(table, x, tableDrawFunction);
  }

  TblDrawTable(table);
}    


Int compareTableEntries(const VoidPtr a, const VoidPtr b, Long other) {
  TableEntryPtr entry_a, entry_b;

  entry_a = (TableEntryPtr)a;
  entry_b = (TableEntryPtr)b;

  if (entry_a->numBytes < entry_b->numBytes) {
    return 1;
  } else {
    return -1;
  }
}


UInt getEntryIndex(VoidHand table, ULong type) {
  ProfileTypePtr tableP = MemHandleLock(table);
  Boolean done = false;
  UInt result;


  if (tableP) {
    int x;
    for (x = 0; !done && (x < tableP->numEntries); x++) {
      if (tableP->entry[x].type == type) {
	done = true;
	result = x;
      }
    }
    if (!done) {
      UInt tableCapacity = (MemHandleSize(table) - sizeof(UInt)) / sizeof(TableEntry);
      if (tableCapacity <= tableP->numEntries) {
	MemHandleUnlock(table);
	MemHandleResize(table, MemHandleSize(table) + 10*sizeof(TableEntry));
	tableP = MemHandleLock(table);
      }
      tableP->entry[tableP->numEntries].type = type;
      tableP->entry[tableP->numEntries].numRecords = 0;
      tableP->entry[tableP->numEntries].numBytes = 0;
      result = tableP->numEntries;
      tableP->numEntries += 1;
      done = true;
    }
    MemHandleUnlock(table);
  }
  if (done) {
    return result;
  } else {
    doAbout();
    return 0;
  }
}      


void makeProfile(UInt cardNum, LocalID localID) {
  ProfileTypePtr profileP = MemHandleLock(profileH);
  DmOpenRef db;
  if (profileP) {
    profileP->numEntries = 0;
    MemHandleUnlock(profileH);
  }

  db = DmOpenDatabase(cardNum, localID, dmModeShowSecret);
  if (db) {
    UInt totalItems = DmNumResources(db);
    UInt index = 0;
    UInt x;
    
    for (x = 0; x < totalItems; x++) {
      VoidHand resourceH = DmGetResourceIndex(db, x);
      ULong resType;
      UInt resNum;
      TableEntryPtr entry;
      UInt index;

      if (resourceH) {
	
	DmResourceInfo(db, x, &resType, &resNum, 0);
	
	index = getEntryIndex(profileH, resType);
	profileP = MemHandleLock(profileH);
	if (profileP) {
	  entry = &(profileP->entry[index]);
	  entry->numRecords++;
	  entry->numBytes += MemHandleSize(resourceH);
	  MemHandleUnlock(profileH);
	} else {
	  doAbout();
	}
	DmReleaseResource(resourceH);
      } else {
	doAbout();
      }
    }
    DmCloseDatabase(db);
  }
  profileP = MemHandleLock(profileH);
  if (profileP) {
    SysQSort(profileP->entry, profileP->numEntries, sizeof(TableEntry), compareTableEntries, 0);
    MemHandleUnlock(profileH);
  } else {
    doAbout();
  }
}


void refreshProfile(UInt appIdx) {
  AppEnumPtr apps;

  if (profile) {
    MemHandleUnlock(profileH);
    profile = 0;
  }
  apps = MemHandleLock(global.enumAppList);

  if (apps) {
    makeProfile(apps[appIdx].cardNum, apps[appIdx].localID);
    MemHandleUnlock(global.enumAppList);
  }

  profile = MemHandleLock(profileH);
  if (profile) {
    displayProfileTable(profile);
  }
}


Boolean ProfilerHandleEvent(EventPtr event) {
  Boolean handled;
  FormPtr frm = FrmGetActiveForm();
  ListPtr list;
  AppEnumPtr apps;

  ProfileTypePtr profile;

  CALLBACK_PROLOGUE

  handled = false;
  switch (event->eType) {
  case menuEvent:
    if (event->data.menu.itemID == MenuAbout) {
      doAbout();
      handled = true;
    }
    break;

  case frmOpenEvent:
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ApplicationList));

    LstSetListChoices(list, NULL, global.numApps);
    LstSetDrawFunction(list, appListDrawFun);

    profileH = MemHandleNew(sizeof(ProfileType));
    apps = MemHandleLock(global.enumAppList);
    if (apps) {
      makeProfile(apps[global.currentApp].cardNum, apps[global.currentApp].localID);
      MemHandleUnlock(global.enumAppList);
    }

    profile = MemHandleLock(profileH);
    firstRow = 0;

    FrmDrawForm(frm);
    if (profile) {
      displayProfileTable(profile);
    }

    updatePopup();

    handled = true;
    break;


  case frmCloseEvent:
    if (profileH) {
      if (profile) {
	MemHandleUnlock(profileH);
      }
      MemHandleFree(profileH);
    }
    /* It's important to return false - otherwise the form won't get cleaned up by
       the default handler */
    handled = false;
    break;


  case popSelectEvent:
    switch (event->data.popSelect.controlID) {
      /*
    case ProfileSortPopup:
      list = event->data.popSelect.listP;
      CtlSetLabel(event->data.popSelect.controlP, 
		  LstGetSelectionText(list, event->data.popSelect.selection));
      handled = true;
      break;
      */
    
    case AppPopup:
      refreshProfile(global.currentApp);
      handled = true;
    }
    break;

  case keyDownEvent:
    switch (event->data.keyDown.chr) {
    case pageUpChr:
      refreshProfile(global.currentApp);
      handled = true;
      break;

    case pageDownChr:
      refreshProfile(global.currentApp);
      handled = true;
      break;
    }
    break;
  }

  CALLBACK_EPILOGUE

  return handled;
}
