///////////////////////////////////////////////////////////////////////////////
// PalmOS Application to view the Caltrain Schedule
//
// Copyright info:
//    This is free software; you can redistribute it and/or modify
//    it as you like.
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
//  FileName:
//    Caltrain.h
// 
//  Description:
//    This is the main source file for the Caltrain Schedule application. 
//    It is derived from the "Tex 2 Hex" 1.0 sample source written by
//    Andrew Howlett (howlett@iosphere.net) provided with his
//    ASDK tutorial.
//
//  ToDo:
//    - Need to add the cost of a trip.
//    - Add copy of row to clipboard w/ headings, turn on item select: this
//      will select all cells in the row.
//    - Need a conduit which will read the web page and create a DB of
//      the station names and times.
//
//  History:
//    10-Feb-1999  - Created as Tex2Hex
//    4/3/2000     - Used as the base for Caltrain Schedule by Rob Baynes
//    4/21/2000    - Version 0.1 ready for release
//    6/2/2000     - Mark E. Wilson (MEW), use binary data
//    9/19/2000    - Updated to new Sept. 24, 2000 schedule.  Rob
///////////////////////////////////////////////////////////////////////////////

#include <Pilot.h>
#include "Caltrain.h"


///////////////////////////////////////////////////////////////////////////////
// Constants
//
#define c_DBCreator     'CLTN'  // A unique app DB ID from www.Palm.com
#define c_DBType        'Data'   
#define c_NumDirs       2       // Two directions North and South
#define c_NumDays       3       // Three days Mon-Fri, Sat, Sun
#define c_Northbound    0       // Index of North
#define c_Southbound    1       // Index of South
#define c_MaxTrains     39      // Max number of trains (is on Mon-Fri)
#define c_Stations      32      // Max number of stations
#define c_NoSelection   100     // Popup lists have not been selected
#define c_TrainCol      0       // The column for the train numbers
#define c_OriginCol     1       // The column for the origin time
#define c_DestCol       2       // The column for the destination time


///////////////////////////////////////////////////////////////////////////////
// Prototypes
//
static int      StartApplication( void );
static void     EventLoop( void );
static void     StopApplication( void );
static Boolean  CaltrainEventHandler( EventPtr event );
static Boolean  AboutEventHandler( EventPtr event );
static void     LoadTable( void );
static void     TableDrawItemCallback( VoidPtr table, 
                                       Word row, 
                                       Word column, 
                                       RectanglePtr bounds );
static Boolean  OpenDatabase( void );
static void     CloseDatabase( void );


///////////////////////////////////////////////////////////////////////////////
// Globals  
//
DmOpenRef   g_DB;                           // reference to open DB
char        g_DBName[] = "CaltrainDB";      // to store user prefs
Word        g_FromStation = c_NoSelection;  // defaults to unset value
Word        g_ToStation = c_NoSelection;    // defaults to unset value
Word        g_Day = 0;                      // defaults to zero = mon-fri
Word        g_Scroll = 0;                   // zero = top of table


// Our caltrain global data
// NOTE: web data is from 2/6/2000 schedule
// See http://www.transitinfo.org/cgi-bin/sched?C=CT&R=&M=A
#include "Glob_data.h"


///////////////////////////////////////////////////////////////////////////////
// Pilot Main
//
DWord PilotMain( Word cmd, Ptr cmdPBP, Word launchFlags )
{
    int error;

    if( cmd == sysAppLaunchCmdNormalLaunch )
    {
        error = StartApplication();     // Application start code
        if( error )  return error;

        EventLoop();                    // Event loop

        StopApplication();              // Application stop code
    }

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Start Application
//
static int StartApplication( void )
{
    int error;

    // Open the DB and get user prefs
    error = OpenDatabase();
    if( error )  return error;

    FrmGotoForm( resFormIDCaltrain );  // go to our main form
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Stop Application
//
static void StopApplication( void )
{
    FormPtr formP;

    // Save user prefs and close the DB
    CloseDatabase();

    // Free the active form
    formP = FrmGetActiveForm();     // free the memory from our main form
    FrmDeleteForm( formP );
}


///////////////////////////////////////////////////////////////////////////////
// Event Loop
//
static void EventLoop( void )
{
    short     err;
    int       formID;
    FormPtr   formP;
    EventType event;

    do
    {
        EvtGetEvent( &event, sysTicksPerSecond/2 );

        if( SysHandleEvent( &event ))  continue;
        if( MenuHandleEvent( (void *)0, &event, &err ))  continue;

        if( event.eType == frmLoadEvent )
        {
            formID = event.data.frmLoad.formID;
            formP = FrmInitForm( formID );
            FrmSetActiveForm( formP );

            switch( formID ) 
            {
            case resFormIDCaltrain:
                FrmSetEventHandler( formP, 
                                    (FormEventHandlerPtr)CaltrainEventHandler );
                break;

            case resFormIDAbout:
                FrmSetEventHandler( formP, 
                                    (FormEventHandlerPtr)AboutEventHandler );
                break;
            }
        }

        FrmDispatchEvent( &event );

     } while( event.eType != appStopEvent );
}


///////////////////////////////////////////////////////////////////////////////
// Caltrain Form Event Handler
//
static Boolean CaltrainEventHandler( EventPtr event )
{
    FormPtr         formP;
    ListPtr         listP;
    TablePtr        tableP;
	UInt            i, numRows;
    int             handled = 0;
    SWord           yoff;
    Short           val, min, max, page;

    switch( event->eType ) 
    {
    case frmOpenEvent:
        formP = FrmGetActiveForm();
        FrmDrawForm( formP );

        // Set the drop boxes if there were any prefs read from the DB
        if( g_FromStation != c_NoSelection )
        {
            listP = FrmGetObjectPtr( formP, 
                    FrmGetObjectIndex( formP, resFromStationList ));
            LstSetSelection( listP, g_FromStation );
            CtlSetLabel( FrmGetObjectPtr( formP, 
                         FrmGetObjectIndex( formP, resFromTrigger )),
                         LstGetSelectionText( listP, g_FromStation ));
        }

        if( g_ToStation != c_NoSelection )
        {
            listP = FrmGetObjectPtr( formP, 
                    FrmGetObjectIndex( formP, resToStationList ));
            LstSetSelection( listP, g_ToStation );
            CtlSetLabel( FrmGetObjectPtr( formP, 
                         FrmGetObjectIndex( formP, resToTrigger )),
                         LstGetSelectionText( listP, g_ToStation ));
        }

        listP = FrmGetObjectPtr( formP, 
                FrmGetObjectIndex( formP, resDayList ));
        LstSetSelection( listP, g_Day );
        CtlSetLabel( FrmGetObjectPtr( formP, 
                     FrmGetObjectIndex( formP, resDayTrigger )),
                     LstGetSelectionText( listP, g_Day ));


        // Init the table
        tableP = FrmGetObjectPtr( formP, 
                 FrmGetObjectIndex( formP, resTimeTable ));
        numRows = TblGetNumberOfRows( tableP );

        // Setup the rows of the table
        for( i = 0; i < numRows; i++ )
        {
            TblSetItemStyle( tableP, i, c_TrainCol, customTableItem ); 
            TblSetItemStyle( tableP, i, c_OriginCol, customTableItem );
            TblSetItemStyle( tableP, i, c_DestCol, customTableItem ); 
            TblSetRowSelectable( tableP, i, false );
        }

        // Make sure the columns are usable - so they draw
        TblSetColumnUsable( tableP, c_TrainCol, true );
        TblSetColumnUsable( tableP, c_OriginCol, true );
        TblSetColumnUsable( tableP, c_DestCol, true );
        TblHasScrollBar( tableP, true );

        // Set our custom table item draw callback
        TblSetCustomDrawProcedure( tableP, c_TrainCol, TableDrawItemCallback );
        TblSetCustomDrawProcedure( tableP, c_OriginCol, TableDrawItemCallback );
        TblSetCustomDrawProcedure( tableP, c_DestCol, TableDrawItemCallback );

        // Draw the outline of the table
        WinDrawGrayRectangleFrame( simpleFrame, &(tableP->bounds) );
        for( i = 1; i < numRows; i++ )
        {
            yoff = tableP->bounds.topLeft.y + 
                   ( i * tableP->rowAttrs->height ) - 1;
            WinDrawGrayLine( tableP->bounds.topLeft.x, yoff,
                             tableP->bounds.extent.x,  yoff );
        }

        // If there were any prefs read from the DB (in OpenDatabase) then
        // this will reset the display to just how the user left it.
        LoadTable();

        handled = 1;
        break;
    

    // Handle the popup lists
	case popSelectEvent:
        switch (event->data.popSelect.listID)
		{
		case resFromStationList:
            if( g_FromStation != event->data.popSelect.selection )
            {
                g_FromStation = event->data.popSelect.selection;
                LoadTable();
            }
		    break;

		case resToStationList:
            if( g_ToStation != event->data.popSelect.selection )
            {
                g_ToStation = event->data.popSelect.selection;
                LoadTable();
            }
		    break;

        case resDayList:
            if( g_Day != event->data.popSelect.selection )
            {
                g_Day = event->data.popSelect.selection;
                g_Scroll = 0;
                LoadTable();
            }
            break;
        } 
        break;


    // Handle the scroll bar (should scroll the table)
    case sclRepeatEvent:
        if( g_Scroll != event->data.sclRepeat.newValue )
        {
            g_Scroll = event->data.sclRepeat.newValue;
            LoadTable();
        }
        break;

    case sclExitEvent:
        if( g_Scroll != event->data.sclExit.newValue )
        {
            g_Scroll = event->data.sclExit.newValue;
            handled = 1;
            LoadTable();
        }
        break;


    // Handle menu selections
    case menuEvent:
        switch( event->data.menu.itemID )
        {
        case resMenuItemCopy:
//debugrob // copy table text ?
            break;

        case resMenuItemAbout:  // Show the about form
            FrmGotoForm( resFormIDAbout );
            break;

        case resMenuItemInfo:
            FrmAlert( resAlertIDInfo );
            break;
        }
        handled = 1;
        break;


    // Make the physical up/down button scroll the list
    case keyDownEvent:
        if( event->data.keyDown.modifiers & commandKeyMask ) // physical key
        {
            if( event->data.keyDown.chr == chrPageUp )  //MEW - use pageUpChr for Palm-GCC dev
            {
                formP = FrmGetActiveForm();
                SclGetScrollBar( FrmGetObjectPtr( formP, 
                                 FrmGetObjectIndex( formP, resTableScrollBar)),
                                 &val, &min, &max, &page );
                if( val > min )
                {
                    g_Scroll--;
                    LoadTable();
                }
            }
            else if( event->data.keyDown.chr == chrPageDown )  //MEW - use pageDownChr for Palm-GCC dev
            {
                formP = FrmGetActiveForm();
                SclGetScrollBar( FrmGetObjectPtr( formP, 
                                 FrmGetObjectIndex( formP, resTableScrollBar)),
                                 &val, &min, &max, &page );
                if( val < max )
                {
                    g_Scroll++;
                    LoadTable();
                }
            }
        }
        break;


    case nilEvent:
        handled = 1;
        break;

    default:
        break;
    }

    return handled;
}


///////////////////////////////////////////////////////////////////////////////
// About Form Event Handler
//
static Boolean AboutEventHandler( EventPtr event )
{
    FormPtr   formP;
    int       handled = 0;

    switch( event->eType ) 
    {
    case frmOpenEvent:
        formP = FrmGetActiveForm();
        FrmDrawForm( formP );
        handled = 1;
        break;
 

    case ctlSelectEvent:
        // Did use select OK on about form ?
        if( event->data.ctlSelect.controlID == resAboutOkBut )
        {
            // Yes, so go back to the main form
            FrmGotoForm( resFormIDCaltrain );  
            handled = true;
        }
        break;


    case nilEvent:
        handled = 1;
        break;

    default:
        break;
    }

    return handled;
}


///////////////////////////////////////////////////////////////////////////////
// Load table data
//
static void LoadTable( void )
{
    FormPtr         formP;
    TablePtr        tableP;
    UInt            i, numRows, dir;
    SWord           yoff;

    // Which direction do they want to go ?
    if( g_FromStation == c_NoSelection || g_ToStation == c_NoSelection )
        return;  // no selections made yet

    if( g_ToStation >= g_FromStation )
        dir = c_Southbound;
    else
        dir = c_Northbound;

    // Get the active form
    formP = FrmGetActiveForm();
    tableP = FrmGetObjectPtr( formP, 
             FrmGetObjectIndex( formP, resTimeTable ));
    numRows = TblGetNumberOfRows( tableP );

    for( i = 0; i < numRows; i++ )
    {
        if( i >= g_TrainCounts[ g_Day ] )
            break;

        // Train number column = 0
        TblSetItemInt( tableP, i, c_TrainCol, // MEW - Int not Ptr...
                       g_Trains[ dir ][ g_Day ][ i + g_Scroll ] );

        // Origin station time
        TblSetItemInt( tableP, i, c_OriginCol, 
            g_Stations[ dir ][ g_Day ][ g_FromStation ][ i + g_Scroll ] );

        // Dest. station time
        TblSetItemInt( tableP, i, c_DestCol, 
            g_Stations[ dir ][ g_Day ][ g_ToStation ][ i + g_Scroll ] );
    }

    // Cause the table to redraw
    TblDrawTable( tableP );


    // Draw the outline of the rows
    for( i = 1; i < numRows; i++ )
    {
        yoff = tableP->bounds.topLeft.y + 
               ( i * tableP->rowAttrs->height ) - 1;
        WinDrawGrayLine( tableP->bounds.topLeft.x, yoff,
                         tableP->bounds.extent.x,  yoff );
    }


    // Update the scroll bar
    SclSetScrollBar( FrmGetObjectPtr( formP, 
                     FrmGetObjectIndex( formP, resTableScrollBar )), 
                     g_Scroll,                          // current value
                     0,                                 // min
                     g_TrainCounts[ g_Day ] - numRows,  // max
                     numRows );                         // page size
}


///////////////////////////////////////////////////////////////////////////////
// Table custom item draw callback.
// NOTE: You can't call stdio funcs here: you get a runtime 'unallocated 
//       chunks' error.
//       You also can't access global data, you get a 'low mem access' error.
// :(
// MEW - Huh?  Hacked to build strings from binary data
static void TableDrawItemCallback( VoidPtr table, 
                                   Word row, 
                                   Word column, 
                                   RectanglePtr bounds )
{
    TablePtr tp = (TablePtr)table;
    RectangleType rect;
    Word wd;
    char str[20];
    Boolean IsTime = false;


    // Erase previous label
    rect = *bounds; 
    rect.topLeft.x += 1;
    rect.topLeft.y += 1;
    rect.extent.y  -= 1;
    WinEraseRectangle( &rect, 0 );


    // Get the text of the item to draw (set in LoadTable)
    wd = tp->items[ ( row * tp->numColumns ) + column ].intValue;
    if( wd == 0 )
        return;


    if ((wd & TIME_AM) || (wd & TIME_PM))
    {
      // it's a time!
      int Time, Hour, Minute;

      IsTime = true;
      // mask off the flags
      Time = wd & ~(TIME_AM | TIME_PM);
      Hour = Time >> 8;
      Minute = Time & 0xFF;
      if (Minute == 99)
      {
        StrCopy(str, "--  "); // N/A
      }
      else
      {
        char min[8];
        StrIToA(str, Hour);
        StrIToA(min, Minute);
        StrCat(str, ":");
        if (Minute < 10)
        {
          StrCat(str, "0");
        }
        StrCat(str, min);
        if (wd & TIME_AM)
        {
          StrCat(str, "a");
        }
        else
        {
          StrCat(str, "p");
        }
      }
    }
    else
    {
      // it's a train number
      int Train;

      // mask off the flags
      Train = wd & ~(EXPRESS_TRAIN | DISCOUNT_TRAIN | GAME_TRAIN);
      StrIToA(str, Train);

      if (wd & EXPRESS_TRAIN)
      {
        StrCat(str, "(x)");
      }
      else if (wd & DISCOUNT_TRAIN)
      {
        StrCat(str, "(d)");
      }
      else if (wd & GAME_TRAIN)
      {
        StrCat(str, "(A)");  // 'A' ??
      }
    }

    // Draw new label 
    FntSetFont( stdFont );
    if (IsTime)
    {
      SWord width;
      width = FntCharsWidth(str, StrLen(str));  // MEW- rt aligned looks better
      WinDrawChars( str, StrLen( str ), 
                    bounds->topLeft.x + 2 + rect.extent.y - width, 
                    bounds->topLeft.y );
    }
    else
    {
      WinDrawChars( str, StrLen( str ), 
                    bounds->topLeft.x + 2, 
                    bounds->topLeft.y );
    }
}


///////////////////////////////////////////////////////////////////////////////
// Open our database to get user preferences
//
static Boolean OpenDatabase( void )
{
    UInt          index = 0;
    VoidHand      recH;
    Ptr           recP, tmpP;


    // First try to open our DB
    g_DB = DmOpenDatabaseByTypeCreator( c_DBType, c_DBCreator, dmModeReadWrite );
    if( !g_DB ) 
    {
        // It did not exist, so create it for the first time.
        if( DmCreateDatabase( 0, g_DBName, c_DBCreator, c_DBType, false )) 
            return 1;

        g_DB = DmOpenDatabaseByTypeCreator( c_DBType, c_DBCreator, dmModeReadWrite );

        // Create a new record and write some defaults
        recH = DmNewRecord( g_DB, &index, sizeof(Word)*4 );
        recP = MemHandleLock( recH );

        // The four globals we need to save are Word(s) which are 2 bytes ea.
        DmWrite( recP, 0, &g_FromStation, 2 );
        DmWrite( recP, 2, &g_ToStation, 2 );
        DmWrite( recP, 4, &g_Day, 2 );
        DmWrite( recP, 6, &g_Scroll, 2 );

        MemPtrUnlock( recP ); 
        DmReleaseRecord( g_DB, index, true );  
    }
    else // the DB was opened, so set our user prefs from it
    {
        recH = DmQueryRecord( g_DB, 0 );
        recP = MemHandleLock( recH );
        tmpP = recP;

        MemMove( &g_FromStation, tmpP, 2 );
        
        tmpP += 2;
        MemMove( &g_ToStation, tmpP, 2 );

        tmpP += 2;
        MemMove( &g_Day, tmpP, 2 );

        tmpP += 2;
        MemMove( &g_Scroll, tmpP, 2 );

        MemPtrUnlock( recP ); 
    }

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Close our database to save user preferences
//
static void CloseDatabase( void )
{
    VoidHand      recH;
    Ptr           recP;

    // Get the record and lock it
    recH = DmQueryRecord( g_DB, 0 );
    recP = MemHandleLock( recH );

    // The four globals we need to save are Word(s) which are 2 bytes ea.
    DmWrite( recP, 0, &g_FromStation, 2 );
    DmWrite( recP, 2, &g_ToStation, 2 );
    DmWrite( recP, 4, &g_Day, 2 );
    DmWrite( recP, 6, &g_Scroll, 2 );

    // Release DB record and close it.
    MemPtrUnlock( recP ); 
    DmReleaseRecord( g_DB, 0, true );
    DmCloseDatabase( g_DB );
}


// end of file