#include "EditForm.h"
#include "Database.h"
#include "Utils.h"
#include "blowfish.h"
#include "MainRsc.h"

#include <System/Graffiti.h>

static void EditFormInit(void);
static Boolean EditValidateFormat(void);

static void EditRetrieveData(FieldPtr);
static void EditRetrieveCryptoPadData(FieldPtr);
static void EditRetrieveMemoPadData(FieldPtr);

static void EditSaveData(FieldPtr);
static void EditSaveCryptoPadData(Char *);
static void EditSaveCryptoPadDataInMemoPad(Char *);
static void EditSaveMemoPadDataInCryptoPad(Char *);

static void EditFormDisplayMaxSizeLabel(void);
static Err EditSortMemo(DmOpenRef, UInt16*);

static void EditFormUpdateScrollBar(void);
static void EditFormScrollLines(Int16, Boolean);
static void EditFormScrollPage(WinDirectionType);

static void EditSelectCategory(void);

static Boolean EditFormUpdateDisplay(Int16);
static void EditFormSetTitle (void);

static Boolean EditFormHandleMenuEvent(UInt16);


/******************************************
 * EditFormInit:
 ******************************************
 */

static void EditFormInit(void)
{
	FormPtr formP;
	FieldPtr fieldP;
	FieldAttrType fieldAttr;
	UInt32 recordID;

	formP = FrmGetActiveForm();
	fieldP = FrmGetObjectPtr(formP,
	FrmGetObjectIndex(formP, fldEditText));
	
	/* Show or hide the lines of the edit field
	* according to the global variable set in
	* the Preferences form.
	*/
	FldGetAttributes(fieldP, &fieldAttr);
	
	if (showLines == linesHidden)
		fieldAttr.underlined = noUnderline;
	else fieldAttr.underlined = grayUnderline;

	FldSetAttributes (fieldP, &fieldAttr);

	/* Attaches the correct menu bar depending
	 * upon the currentPasswordP.
	 */

	if (currentPasswordP != 0) FrmSetMenu(formP, mnuEditMenuNoPwd);

	/* Updates the currently selected font */

	if (useSingleFont == separateFont) {
		DmRecordInfo (currentDBRef, currentRecord, 0, &recordID, 0);
		SetOrGetRecordFont(currentDBRef, recordID, getFont, fontField);
	}

	FldSetFont(fieldP, currentFieldFont);

	EditRetrieveData(fieldP);

	currentForm = frmEditForm;
}

/******************************************
 * EditValidateFormat:
 * Checks that the memo contains at least
 * two lines to perform encryption.
 ******************************************
 */

static Boolean EditValidateFormat(void)
{
	FieldPtr fieldP;
	Char * textP;

	UInt16 nOffset;

	fieldP = GetActiveFormObjectPtr(fldEditText);
	textP = FldGetTextPtr(fieldP);

	/* Finds the offset to start encryption */

	nOffset = 0;
	while (nOffset <= StrLen(textP)) {
		if (textP[nOffset++] == '\n') break;
	}

	return (nOffset < StrLen(textP));
}

/******************************************
 * EditRetrieveCryptoPadData:
 * Get the data from the current record
 * and start to edit in place.
 ******************************************
 */

static void EditRetrieveCryptoPadData(FieldPtr fieldP)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	RecordType record;

	UInt16 nOffset, nLength;

	if (DmNumRecords(dbCryptoPadRef) > 0) {
		handle = DmGetRecord(dbCryptoPadRef, currentRecord);

		/* Get the text length */

		packedP = MemHandleLock(handle);
		UnpackRecord(&record, packedP);
		nOffset = sizeof(record.flag) +
					sizeof(record.passwdLen) +
					sizeof(record.textLen) +
					record.passwdLen + 1;
		nLength = StrLen(record.textP) + 1;
		MemHandleUnlock(handle);

		/* Start editing the record in place */

		FldSetText(fieldP, handle, nOffset, nLength);
	}
}


/******************************************
 * EditRetrieveMemoPadData:
 * Get the data from the current record
 * and start to edit in place.
 ******************************************
 */

static void EditRetrieveMemoPadData(FieldPtr fieldP)
{
	MemHandle handle;

	if (DmNumRecords(dbMemoPadRef) > 0) {
		handle = DmGetRecord(dbMemoPadRef, currentRecord);
		FldSetTextHandle(fieldP, handle);
	}
}

/******************************************
 * EditRetrieveData:
 * Get the data from the current record
 * and start to edit in place.
 ******************************************
 */

static void EditRetrieveData(FieldPtr fieldP)
{
	FormPtr formP;
	ControlPtr triggerP;
	UInt16 recAttributes;


	if (currentRecord != noRecordSelected) {
		if (currentDB == memoPadDB) {
			EditRetrieveMemoPadData (fieldP);
			EditFormDisplayMaxSizeLabel();
		} else EditRetrieveCryptoPadData (fieldP);

		formP = FrmGetActiveForm();
	
		triggerP = FrmGetObjectPtr(formP,
				FrmGetObjectIndex(formP, popEditCategory));

		/* Updates the currently selected category */
		DmRecordInfo(currentDBRef, currentRecord, &recAttributes, 0, 0);
		currentCategory = (recAttributes & dmRecAttrCategoryMask);

		// If the record is secret and secret records are hidden, then the record isn't
		// accounted for by MemosInCategory.  Adjust it, and EditFormSaveData will 
		// adjust when done with the record.
		if ((PrivateRecordVisualStatus == hidePrivateRecords) && (recAttributes & dmRecAttrSecret))
			memosInCategory++;

		// Set the edit form title
		EditFormSetTitle ();

		CategoryGetName(dbMemoPadRef, currentCategory, categoryName);
		CategorySetTriggerLabel(triggerP, categoryName);

		FldSetScrollPosition (fieldP, editScrollPosition);
		EditFormUpdateScrollBar();	
	}
}

/******************************************
 * EditSaveCryptoPadData:
 * Saves the data of the current record
 * in the CryptoPad database, and remove 
 * the record from MemoPadDB.
 ******************************************
 */

static void EditSaveCryptoPadData(Char * textP)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	UInt16 nOffset, nLength;

	handle = DmGetRecord(dbCryptoPadRef, currentRecord);
	packedP = MemHandleLock(handle);

	if (globalPasswordP != 0) {
		currentPasswordP = globalPasswordP;
	}

	/* Updates the text length */
	nOffset = sizeof(packedP->flag) + sizeof(packedP->passwdLen);
	nLength = StrLen(textP);
	DmWrite(packedP, nOffset, &nLength, sizeof(nLength));

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, currentRecord, true);

	/* Encrypt the record if the password is not empty */
	if (currentPasswordP != 0) {
		EncryptRecord(currentRecord, currentPasswordP);
		RemoveCurrentPassword();
	}
}

/******************************************
 * EditSaveCryptoPadDataInMemoPad:
 * Saves the data of the current record
 * in the MemoPad database, and remove 
 * the record from CryptoPadDB.
 ******************************************
 */

static void EditSaveCryptoPadDataInMemoPad(Char * textP)
{
	MemPtr		p;
	MemHandle handle;
	UInt16 index;

	UInt16 recMemoAttributes;	
	UInt16 recCryptoAttributes;
	UInt16 nLength;			
	UInt32 cryptoRecordID;

	index = currentRecord;

	currentRecord = dmMaxRecordIndex;
	currentDB = memoPadDB;

	/* Create a new record in the MemoPad database */
	nLength = StrLen(textP);
	handle = DmNewRecord(dbMemoPadRef, &currentRecord, nLength + 1);
	ErrFatalDisplayIf(handle == 0, "Unable to add record to the database.");

	/* Lock down the block containing the new record */
	p = MemHandleLock(handle);

	/* Write the field's text to the database record */
	DmWrite(p, 0, textP, StrLen(textP) + 1);

	/* Unlock the block of the new record */
	MemPtrUnlock(p);

	/* Set the category of the record */
	DmRecordInfo(dbCryptoPadRef, index, &recCryptoAttributes, &cryptoRecordID, 0);

	DmReleaseRecord(dbMemoPadRef, currentRecord, true);

	DmSetRecordInfo(dbMemoPadRef, currentRecord, &recCryptoAttributes, 0);
	
	// Create the record in the Font database.
	if (useSingleFont == separateFont) {
		UInt32 memoRecordID;

		DmRecordInfo (dbMemoPadRef, currentRecord, 0, &memoRecordID, 0);
		CreateFontRecord(dbMemoPadRef, memoRecordID);
		SetOrGetRecordFont(dbMemoPadRef, memoRecordID, setFont, fontField);

		// Delete the record in the Font database.
		DeleteFontRecord(dbCryptoPadRef, cryptoRecordID);
	}

	/* Delete the record from CryptoPadDB */
	DmRemoveRecord(dbCryptoPadRef, index);
}

/******************************************
 * EditSaveMemoPadDataInCryptoPad:
 * Saves the data of the current record
 * in the CryptoPad database, and remove 
 * the record from MemoPadDB.
 ******************************************
 */

static void EditSaveMemoPadDataInCryptoPad(Char * textPtr)
{
	MemHandle handle;
	RecordType newRecord;

	UInt16 recMemoAttributes;	
	UInt16 recCryptoAttributes;
	UInt16 index;
	UInt32 memoRecordID;

	PackedRecordPtr packedP;
	UInt16 nOffset, textLength, passwordLength;

	/* Creates a new record at the end of the database */
	index = currentRecord;

	currentRecord = dmMaxRecordIndex;
	currentDB = cryptoPadDB;

	if (globalPasswordP != 0) currentPasswordP = globalPasswordP;

	handle = DmNewRecord(dbCryptoPadRef, &currentRecord, sizeof(PackedRecordType));
	ErrFatalDisplayIf(handle == 0, "Unable to add record to the database.");

	/* Current record now contains the record index.
	 * Sets an initial value for the record.
	 */
	
	passwordLength = StrLen(currentPasswordP);
	textLength = StrLen(textPtr);

	newRecord.flag      = recDecrypted;
	newRecord.passwdLen =  passwordLength;
	newRecord.textLen   = textLength;
	newRecord.passwordP = currentPasswordP;
	newRecord.textP = textPtr;

	PackRecord(&newRecord, handle);
	
	/* Set the category of the record */
	DmRecordInfo(dbMemoPadRef, index, &recMemoAttributes, &memoRecordID, 0);
	
	DmReleaseRecord(dbCryptoPadRef, currentRecord, true);

	DmSetRecordInfo(dbCryptoPadRef, currentRecord, &recMemoAttributes, 0);
	
	if (useSingleFont == separateFont) {
		UInt32 cryptoRecordID;

		DmRecordInfo (dbCryptoPadRef, currentRecord, 0, &cryptoRecordID, 0);
		CreateFontRecord(dbCryptoPadRef, cryptoRecordID);
		SetOrGetRecordFont(dbCryptoPadRef, cryptoRecordID, setFont, fontField);
	}

	/* Delete the record from MemoPadDB */
	DmDeleteRecord (dbMemoPadRef, index);
	DmMoveRecord (dbMemoPadRef, index, DmNumRecords (dbMemoPadRef));

	/* Encrypt the record if the password is not empty */
	EncryptRecord(currentRecord, currentPasswordP);
	RemoveCurrentPassword();
}


/******************************************
 * EditSaveData:
 * Saves the data to the current record
 * and release the record.
 ******************************************
 */

static void EditSaveData(FieldPtr fieldP)
{
	Char * textP;

	UInt16 recAttributes;
	UInt32 recordID;

	Boolean dirty;
	Boolean empty;

	textP = FldGetTextPtr(fieldP);

	// Find out if the field has been modified.
	dirty = FldDirty (fieldP);
	empty = (*textP == '\0');

	FldReleaseFocus (fieldP);

	// Release any free space in the memo field.
	FldCompactText(fieldP);

	// Clear the handle value in the field, otherwise the handle
	// will be free when the form is disposed of.
	FldSetTextHandle(fieldP, 0);


	// If there's data in an existing record, mark it dirty if 
	// necessary and release it.
	DmRecordInfo (currentDBRef, currentRecord, &recAttributes, &recordID, 0);

	if (!empty) {
		if (PrivateRecordVisualStatus == hidePrivateRecords && (recAttributes & dmRecAttrSecret))
				memosInCategory--;
		DmReleaseRecord(currentDBRef, currentRecord, dirty);
		
		if ((currentDB == cryptoPadDB) && (currentPasswordP == 0)) {	
				EditSaveCryptoPadDataInMemoPad(textP);
				dirty = true;
		}
		if ((currentDB == memoPadDB) && (currentPasswordP != 0)) {
				EditSaveMemoPadDataInCryptoPad(textP);
				dirty = true;
		}
		if ((currentDB == cryptoPadDB) && (currentPasswordP != 0)) EditSaveCryptoPadData(textP);

		// Move the current record to the correct sort position.
		if (dirty)
			EditSortMemo(currentDBRef, &currentRecord);

		if (moveRecordOnTop) {
			DmMoveRecord (currentDBRef, currentRecord, 0);
			currentRecord = 0;
		}
	}

	// If the record is empty, delete it.
	else
		{
		// Delete the record in the Font database.
		if ((useSingleFont == separateFont) 
				&& (currentDB == cryptoPadDB))
			DeleteFontRecord(currentDBRef, recordID);

		if (dirty)
			{
			DmDeleteRecord (currentDBRef, currentRecord);
			DmMoveRecord (currentDBRef, currentRecord, DmNumRecords (currentDBRef));
			}
		else DmRemoveRecord(currentDBRef, currentRecord);

		if (currentPasswordP != 0) RemoveCurrentPassword();

		memosInCategory--;
		currentForm = frmMainForm;
		currentRecord = noRecordSelected;
	}
}

/******************************************
 * EditSortMemo:
 * Sorts the memo in one or the other
 * database according to the current sort Order.
 * This function taken from the
 * 'Memo' 3COM sample application.
 ******************************************
 */

Err EditSortMemo(DmOpenRef dbP, UInt16* indexP)
{

	Err err;
	Int16 sortOrder;
	UInt16 index;
	UInt16 attributes;
	UInt32 uniqueID;
	MemHandle recordH;
	MemHandle h;
	Boolean dontMove;

	MemPtr cmp;
	MemPtr recordP;

	if (currentSortOrder != soAlphabetic)
		return (0);

	/* Check if the record is already in the correct position */
	recordP = MemHandleLock(DmQueryRecord (dbP, *indexP));		

	if (*indexP > 0) {

		/* This record wasn't deleted and deleted records are at the end of the
		 * database so the prior record may not be deleted!
		 */

		h = DmQueryRecord (dbP, *indexP-1);

		if (! h) dontMove = false;
		else {
			cmp = MemHandleLock(h);

			if (dbP == dbMemoPadRef) {
				dontMove = (CompareMemoPadRecords(cmp, recordP,
								(Int16) currentSortOrder) == -1);
			} else {
				dontMove = (CompareCryptoPadRecords((PackedRecordPtr) cmp, 
					(PackedRecordPtr) recordP, (Int16) currentSortOrder) == -1);
			}

			MemPtrUnlock(cmp);
		}

	} else dontMove = true;

	if (dontMove && (*indexP+1 < DmNumRecords (dbP))) {

		DmRecordInfo(dbP, *indexP+1, &attributes, NULL, NULL);
		if ( ! (attributes & dmRecAttrDelete) ) {
			cmp = MemHandleLock(DmQueryRecord (dbP, *indexP+1));
			if (dbP == dbMemoPadRef) {
				dontMove &= (CompareMemoPadRecords(recordP, cmp,
							(Int16) currentSortOrder) == -1);
			} else {
				dontMove &= (CompareCryptoPadRecords((PackedRecordPtr) recordP, 
					(PackedRecordPtr) cmp, (Int16) currentSortOrder) == -1);
			}

			MemPtrUnlock(cmp);
		}
	}

	MemPtrUnlock(recordP);

	if (dontMove) return (0);


	/* Since the routine that determines the records sort position uses a 
	 * binary search algorythm we need to remove the record from the database 
	 * before we can determine its new position.  We will also save and restore the 
	 * record's attributes and unique ID.
	 */

	DmRecordInfo(dbP, *indexP, &attributes, &uniqueID, NULL);

	err = DmDetachRecord(dbP, *indexP, &recordH);
	if (err) return err;
	
	recordP = MemHandleLock(recordH);
	
	if (dbP == dbMemoPadRef) {
		index = DmFindSortPosition(dbP, recordP, 0,
			(DmComparF *) CompareMemoPadRecords, (Int16) currentSortOrder);
	} else {
		index = DmFindSortPosition(dbP, (PackedRecordPtr) recordP, 0,
			(DmComparF *) CompareCryptoPadRecords, (Int16) currentSortOrder);
	}

	MemPtrUnlock(recordP);
	
	err = DmAttachRecord(dbP, &index, recordH, 0);
	if (err) return err;

	DmSetRecordInfo(dbP, index, &attributes, &uniqueID);	

	*indexP = index;
	
	return (err);
}

/******************************************
 * EncryptRecord:
 * Encrypt the current record
 ******************************************
 */

void EncryptRecord(UInt16 recordNum, Char * passwordP)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	RecordType record;
	UInt16 nOffset;

	handle = DmGetRecord(dbCryptoPadRef, recordNum);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	/* Finds the offset to start encryption */

	nOffset = 0;
	while (nOffset <= StrLen(record.textP)) {
		if (record.textP[nOffset++] == '\n') break;
	}

	if (nOffset < StrLen(record.textP)) {
		RecordType newRecord;

		/* Set the flag to reflect the status of the record */

		newRecord.flag      = recEncrypted;
		newRecord.passwdLen = StrLen(passwordP);
		newRecord.textLen   = StrLen(record.textP);

		/* Store the encrypted password into the record */

		newRecord.passwordP = MemPtrNew(newRecord.passwdLen + 1);
		Blowfish(passwordP, newRecord.passwordP, passwordP,
									newRecord.passwdLen, BF_ENCRYPT);

		/* Store the encrypted text in the record */

		newRecord.textP = MemPtrNew(StrLen(record.textP) + 1);
		StrNCopy(newRecord.textP, record.textP, nOffset);
		Blowfish(record.textP + nOffset, newRecord.textP + nOffset,
								passwordP,
								newRecord.textLen - nOffset,
								BF_ENCRYPT);

		/* Replaces the record by the new record */

		PackRecord(&newRecord, handle);

		/* Deallocates memory */

		MemPtrFree(newRecord.passwordP);
		MemPtrFree(newRecord.textP);
	}

	/* Release the record to the data manager */

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, recordNum, true);
}

/******************************************
 * EditFormUpdateScrollBar:
 * Updates the size and position of
 * the scrollbar.
 ******************************************
 */

static void EditFormUpdateScrollBar(void)
{
	FieldPtr fieldP;
	ScrollBarPtr scrollBarP;

	UInt16 nPosition, nMax;
	UInt16 textHeight, fieldHeight;

	fieldP = GetActiveFormObjectPtr(fldEditText);
	scrollBarP = GetActiveFormObjectPtr(sclEditScrollBar);

	FldGetScrollValues(fieldP, &nPosition, &textHeight, &fieldHeight);

		// On occasion, such as after deleting a multi-line selection of text,
		// the display might be the last few lines of a field followed by some
		// blank lines.  To keep the current position in place and allow the user
		// to "gracefully" scroll out of the blank area, the number of blank lines
		// visible needs to be added to max value.  Otherwise the scroll position
		// may be greater than maxValue, get pinned to maxvalue in SclSetScrollBar
		// resulting in the scroll bar and the display being out of sync.
	if (textHeight > fieldHeight) nMax = textHeight - fieldHeight + FldGetNumberOfBlankLines (fieldP);
	else if (nPosition) nMax = nPosition;
	else nMax = 0;

	SclSetScrollBar(scrollBarP, nPosition, 0, nMax, fieldHeight - 1);
}

/******************************************
 * EditFormScrollLines:
 * Scrolls the text from nLines upwards
 * (negative) or downwards (positive).
 ******************************************
 */

static void EditFormScrollLines(Int16 nLines, Boolean redraw)
{
	FieldPtr fieldP;
	Int16 blankLines;

	fieldP = GetActiveFormObjectPtr(fldEditText);
	blankLines = FldGetNumberOfBlankLines(fieldP);

	if (nLines < 0) FldScrollField(fieldP, -nLines, winUp);
	else FldScrollField(fieldP, nLines, winDown);

	/* If there are blank lines at the end, FldScrollField
	 * makes the blank lines disappear. Therefore, we have
	 * got to update the scrollbar.
	 */

	if (blankLines || redraw) {
		EditFormUpdateScrollBar();
		}
}

/******************************************
 * EditFormScrollPage:
 * Scrolls the text one full page in
 * the given direction.
 ******************************************
 */

static void EditFormScrollPage(WinDirectionType direction)
{
	FieldPtr fieldP;
	Int16 nLines;

	Int16 seekDirection;
	UInt16 category;
	UInt16 recordNum;
	UInt32 uniqueID;

	
	fieldP = GetActiveFormObjectPtr(fldEditText);
	if (FldScrollable(fieldP, direction)) {
		nLines = FldGetVisibleLines(fieldP) - 1;

		if (direction == winUp) nLines = -nLines;
		EditFormScrollLines(nLines, true);

		return;
	} 

	if (currentDB == cryptoPadDB) return;

	// Move to the next or previous memo.
	if (direction == winUp)	{
		seekDirection = dmSeekBackward;
		editScrollPosition = maxFieldTextLen;		
	} else {
		seekDirection = dmSeekForward;
		editScrollPosition = 0;
	}

	if (showAllCategories) category = dmAllCategories;
	else category = currentCategory;

	recordNum = currentRecord;
	
	DmSeekRecordInCategory (dbMemoPadRef, &recordNum, 1, seekDirection, category);
	  	
	if (recordNum == currentRecord) return;

	SndPlaySystemSound (sndInfo);

	// Saving the current record may cause it to move if the records are 
	// sorted alphabeticly.
	DmRecordInfo (dbMemoPadRef, recordNum, NULL, &uniqueID, NULL);
	EditSaveData(fieldP);
	DmFindRecordByID (dbMemoPadRef, uniqueID, &currentRecord);

	EditRetrieveData(fieldP);
	FldGrabFocus (fieldP);

	// Set the correct font.
	if (useSingleFont == separateFont) {
		SetOrGetRecordFont(currentDBRef, uniqueID, getFont, fontField);
		FldSetFont(fieldP, currentFieldFont);
	}
}

/******************************************
 * EditSelectCategory:
 * Invokes the standard category UI
 ******************************************
 */

static void EditSelectCategory(void)
{
	Boolean categoryUpdated;
	Boolean categoryChanged;
	UInt16 category;
	FormPtr formP;
	UInt16 recAttributes;
	UInt32 recordID;

	DmRecordInfo(currentDBRef, currentRecord, &recAttributes, &recordID, 0);
	category = (recAttributes & dmRecAttrCategoryMask);


	/* Process the category popup list */
	formP = FrmGetActiveForm();
	categoryUpdated = CategorySelect(dbMemoPadRef, formP,
			popEditCategory, lstEditCategory, false,
			&category, categoryName, 1, 0);

	categoryChanged = (category != (recAttributes & dmRecAttrCategoryMask));

	if (categoryChanged) {
		/* Updates the tableFont of the record */
		if (useSingleFont == separateFont) {
			GetTableFontOfCategory(category);
			SetOrGetRecordFont(currentDBRef, recordID, setFont, fontTable);
		}

		/* Updates the record category */
		DmRecordInfo(currentDBRef, currentRecord, &recAttributes, 0, 0);
		recAttributes &= ~dmRecAttrCategoryMask;
		recAttributes |= category | dmRecAttrDirty;
		DmSetRecordInfo(currentDBRef, currentRecord, &recAttributes, 0);
	}

	if (categoryUpdated || categoryChanged) {
		/* Updates the category trigger and
		 * changes the current category to
		 * make the record visible in the
		 * main form.
		 */
		ChangeCategory(category);

		// If the record is secret and secret records are hidden, then the record isn't
		// accounted for by MemosInCategory.  Adjust it, and EditFormSaveData will 
		// adjust when done with the record.
		if ((PrivateRecordVisualStatus == hidePrivateRecords) && (recAttributes & dmRecAttrSecret))
			memosInCategory++;
		EditFormSetTitle();	
	}
}

/******************************************
 * EditFormUpdateDisplay:
 * Updates the category that was changed
 * from the Details form.
 ******************************************
 */

static Boolean EditFormUpdateDisplay(Int16 updateCode)
{
	FormPtr formP;
	FieldPtr fieldP;
	UInt16 category;
	

	if (updateCode == updateCategoryChanged) {
		UInt16 recAttributes;

		DmRecordInfo(currentDBRef, currentRecord, &recAttributes, 0, 0);
		category = (recAttributes & dmRecAttrCategoryMask);

		ChangeCategory(category);

		CategoryGetName(dbMemoPadRef, currentCategory, categoryName);
		CategorySetTriggerLabel(GetActiveFormObjectPtr(popEditCategory),
						categoryName);

		// If the record is secret and secret records are hidden, then the record isn't
		// accounted for by MemosInCategory.  Adjust it, and EditViewSaveRecord will 
		// adjust when done with the record.
		if (PrivateRecordVisualStatus == hidePrivateRecords && (recAttributes & dmRecAttrSecret))
			memosInCategory++;
		EditFormSetTitle ();
		
		return (true);
		}

	if (updateCode == updateFontChanged) {
		UInt32 recordID;
		
		formP = FrmGetActiveForm ();
		fieldP = FrmGetObjectPtr(formP,
		FrmGetObjectIndex(formP, fldEditText));
	
		/* Updates the currently selected font */
		FldSetFont(fieldP, currentFieldFont);

		if (useSingleFont == separateFont) {
			DmRecordInfo (currentDBRef, currentRecord, 0, &recordID, 0);
			SetOrGetRecordFont(currentDBRef, recordID, setFont, fontField);
		}

		EditFormUpdateScrollBar();
		return (true);
		}
	
	return (false);
}


/******************************************
 * EditFormSetTitle :
 * Sets the title of the edit view
 * with the number of records.
 * If the edit view is visible, the new title is drawn.
 * The title parts are referenced as follows:
 * [text 1][arg 1][text 2][arg 2]
 ******************************************
 */
static void EditFormSetTitle (void)
{
	Char * titleTemplateP;
	Char * strP;
	Char * titleP;
	FormPtr frm;
	char posStr [digitsForRecordPosition];
	char totalStr [digitsForRecordPosition];
	UInt16 pos;
	UInt16 length;
	UInt16 category;

	/* Format as strings, the memo's postion within its category, and 
	* the total number of memos in the category.
	*/
	if (showAllCategories)
		category = dmAllCategories;
	else 
		category = currentCategory;

	if (currentDB == memoPadDB) 
		pos = DmPositionInCategory (dbMemoPadRef, currentRecord, category);
	else pos = DmNumRecordsInCategory (dbMemoPadRef, category) +
		DmPositionInCategory (dbCryptoPadRef, currentRecord, category);
	StrIToA (posStr, pos+1);

	StrIToA (totalStr, memosInCategory);


	/* Get the title template string.  It contains '#' chars which we replace
	* with the position of currentRecord within currentCategory and with
	* the total count of records in currentCategory.
	*/
	titleTemplateP = MemHandleLock (DmGetResource ('tSTR', strEditTitleMemo));

	/* Now get space for the new title */
	length = StrLen (posStr) + StrLen (totalStr) + StrLen (titleTemplateP);


	titleP = MemPtrNew (length);
	ErrFatalDisplayIf (!titleP, "Out of memory");


	strP = StrChr (titleTemplateP, '#');
	ErrNonFatalDisplayIf (!strP, "Missing '#' in MemoEdit'tStr' strEditTitleMemo");

	length = strP - titleTemplateP;
	StrNCopy (titleP, titleTemplateP, length);
	titleP[length] = 0;
	StrCat (titleP, posStr);

	strP++;
	StrCat (titleP, strP);
	StrCat (titleP, totalStr);

	frm = FrmGetFormPtr (frmEditForm);

	FrmSetTitle (frm, titleP);

	MemPtrUnlock(titleTemplateP);
}

/******************************************
 * EditFormDisplayMaxSizeLabel:
 * Show or hide the 4k max size label.
 ******************************************
 */

static void EditFormDisplayMaxSizeLabel(void)
{
	FormPtr formP;
	UInt16 fieldIndex;
	FieldPtr fieldP;

	formP = FrmGetActiveForm();
	fieldIndex = FrmGetObjectIndex(formP, fldEditText);
	fieldP = FrmGetObjectPtr(formP, fieldIndex);
	
	/* Display the label if the size of the memo
	 * is over 4k otherwise hide it.
	 */
	if (FldGetTextLength(fieldP) <= 4097)
		FrmHideObject (formP, FrmGetObjectIndex(formP, lblEditMaxSize));
	else FrmShowObject (formP, FrmGetObjectIndex(formP, lblEditMaxSize));
}

/******************************************
 * EditFormHandleMenuEvent:
 * Process menu commands.
 ******************************************
 */

static Boolean EditFormHandleMenuEvent(UInt16 itemID)
{
	Boolean handled = true;

	FieldPtr fieldP;
	EventType event;
	FormPtr formP;
	UInt32 uniqueID;
	Char * textP;

	switch (itemID) {
		case itemRecordNewMemo:
			fieldP = GetFocusObjectPtr();
			if (fieldP) {
				EditSaveData(fieldP);
				CreateRecord();
				editScrollPosition = 0;
				if (currentRecord != noRecordSelected) {
					EditRetrieveData(fieldP);
					FldGrabFocus(fieldP);
				} else FrmGotoForm(frmMainForm);
			}
			break;

		case itemRecordDeleteMemo:
			if (DetailsFormDeleteCurrentRecord ())
				FrmGotoForm(frmMainForm);
			break;

		case itemRecordEncryptMemo:
			/* Prompts for the password if
			 * the format of the record is
			 * valid for encryption.
			 */
			ErrFatalDisplayIf(currentPasswordP != 0, "Password should be NULL");

			if (EditValidateFormat()) {
				if ((prefsGlobalPasswordP != 0) && (globalPasswordP == 0)) {
					event.eType = globalPasswordStatusEvent;
					event.data.generic.datum[1] = queryGlobalPassword;
					event.data.generic.datum[2] = recDecrypted;

					FrmPopupForm(frmQueryGlobalPasswordForm);
					EvtAddEventToQueue(&event);
					handled = true;
					break;
				}

				event.eType = memoStatusEvent;
				event.data.generic.datum[1] = recDecrypted;

				FrmPopupForm(frmPasswordForm);
				EvtAddEventToQueue(&event);

			} else FrmAlert(altInvalidFormatAlert);
			break;

		case itemRecordRemovePasswd:
			/* Remove the password protection
			 * so that the memo is not
			 * automatically encrypted on exit.
			 */

//			ErrFatalDisplayIf(currentPasswordP == 0, "Password should not be NULL");
			RemoveCurrentPassword();

			FrmSetMenu(FrmGetActiveForm(), mnuEditMenu);
			EditFormDisplayMaxSizeLabel();
			FrmAlert(altRemovePasswdAlert);
			break;

		case itemOptionsPhoneLookup:
		case itemOptionsGoTopOfPage:
		case itemOptionsGoBottomOfPage:
			fieldP = GetFocusObjectPtr();
			if (!fieldP)  break;
			if (itemID == itemOptionsPhoneLookup) {
				PhoneNumberLookup (fieldP);
				EditFormUpdateScrollBar();
			}
			else if (itemID == itemOptionsGoTopOfPage)	{
				FldSetInsPtPosition(fieldP, 0);
				EditFormUpdateScrollBar();
				}
			else if (itemID == itemOptionsGoBottomOfPage)	{
				FldSetInsPtPosition(fieldP, FldGetTextLength(fieldP));
				EditFormUpdateScrollBar();
				}
			break;

		case itemOptionsFont:
			currentFieldFont = SelectFont (currentFieldFont);
			break;

		case itemRecordBeamMemo:
			if ((currentDB == memoPadDB) ||
				((currentDB == cryptoPadDB) && (currentPasswordP == 0))) {
				formP = FrmGetActiveForm();
				fieldP = FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, fldEditText));
				if (FldGetTextLength(fieldP) > 0) {
					EditSaveData(fieldP);
					MemoSendRecord(currentDBRef, currentRecord);
				
				// Redisplay the record.  If the IR loopback mechanism sends the 
				// record to this app the goto action code closes all forms and 
				// send a frmGotoEvent.  Load the record again only if the form 
				// still exits.
				if (formP) EditRetrieveData(fieldP);
				} else FrmAlert(NoDataToBeamAlert);
			} else FrmAlert(altBeamCryptoDataAlert);
			break;

		case itemOptionsAbout:
			formP = FrmInitForm(frmAboutForm);
			FrmDoDialog(formP);
			FrmDeleteForm(formP);
			break;

		default:
			handled = false;
	}

	return handled;
}

Boolean EditFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;

	FormPtr formP;
	UInt16 fieldIndex;
	FieldPtr fieldP;
	Char * textP;
	UInt16 startPos, endPos;

	switch (eventP->eType) {
		/* A notification to immediately encrypt
		 * the record has been made. Close the
		 * form and let the frmCloseEvent handling
		 * code do the work.
		 */
		case memoProcessEvent:
			ErrFatalDisplayIf(currentPasswordP == 0, "Password shouldn't be NULL");
			FrmGotoForm(frmMainForm);
			handled = true;
			break;

		/* Notification was received to update the display */
		case frmUpdateEvent:
			handled = EditFormUpdateDisplay(eventP->data.frmUpdate.updateCode);
			break;

		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				/* Saves the memo and return to the main form */
				case btnEditTerminate:
					FrmGotoForm(frmMainForm);
					handled = true;
					break;

				/* Show the details attributes */
				case btnEditDetails:
					FrmPopupForm(frmDetailsForm);
					handled = true;
					break;

				/* Changes the category of the current memo */
				case popEditCategory:
					EditSelectCategory();
					handled = true;
					break;

			}
			break;

		/* The text in the field has changed */
		case fldChangedEvent:
			EditFormUpdateScrollBar();
			handled = true;
			break;

		/* the scrollbars has been tapped on the screen */
		case sclRepeatEvent:
			EditFormScrollLines(eventP->data.sclRepeat.newValue -
					eventP->data.sclRepeat.value, false);
			/* Let the system repeat the event */
			break;

		case keyDownEvent:
			if (TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)) {
				FrmGotoForm(frmMainForm);
				handled = true;
			}

			else if (EvtKeydownIsVirtual(eventP)) {
				if (eventP->data.keyDown.chr == vchrPageUp) {
					EditFormScrollPage(winUp);
					handled = true;
				}
	
				else if (eventP->data.keyDown.chr == vchrPageDown) {
					EditFormScrollPage(winDown);
					handled = true;
				}
			
				// Send data key presed?
				else if (eventP->data.keyDown.chr == vchrSendData) {
					EditFormHandleMenuEvent(itemRecordBeamMemo);
					handled = true;
				}
			}	
			
			else if (currentPasswordP == 0) EditFormDisplayMaxSizeLabel();
			break;

		case menuEvent:
			handled = EditFormHandleMenuEvent(eventP->data.menu.itemID);
			break;

		// Add the buttons that we want available on the command bar, based on the current context
		case menuCmdBarOpenEvent:
			formP = FrmGetActiveForm();
			FldGetSelection(FrmGetObjectPtr(formP, 
			FrmGetObjectIndex(formP, fldEditText)) , &startPos, &endPos);
		
			// if there's some highlighted text
			if (startPos != endPos) {
				// (Note that we're adding buttons from right to left)
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarPasteBitmap, menuCmdBarResultMenuItem, sysEditMenuPasteCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarCopyBitmap, menuCmdBarResultMenuItem, sysEditMenuCopyCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarCutBitmap, menuCmdBarResultMenuItem, sysEditMenuCutCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarUndoBitmap, menuCmdBarResultMenuItem, sysEditMenuUndoCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, bmpBarPhoneBitmap, menuCmdBarResultMenuItem,itemOptionsPhoneLookup, 0);
			} 
			// there's no highlighted text
			else {
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarDeleteBitmap, menuCmdBarResultMenuItem, itemRecordDeleteMemo, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarPasteBitmap, menuCmdBarResultMenuItem, sysEditMenuPasteCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarUndoBitmap, menuCmdBarResultMenuItem, sysEditMenuUndoCmd, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, BarBeamBitmap, menuCmdBarResultMenuItem, itemRecordBeamMemo, 0);
				MenuCmdBarAddButton(menuCmdBarOnLeft, bmpBarFontBitmap, menuCmdBarResultMenuItem, itemOptionsFont, 0);
			}

			// tell the field package to not add buttons automatically; we've done it all ourselves.
			eventP->data.menuCmdBarOpen.preventFieldButtons = true;
		
			// don't set handled to true; this event must fall through to the system.
			break;

		case frmOpenEvent:
			formP = FrmGetActiveForm();
			fieldIndex = FrmGetObjectIndex(formP, fldEditText);
			fieldP = FrmGetObjectPtr(formP, fieldIndex);

			EditFormInit();
			FrmDrawForm(formP);
			FrmSetFocus(formP, fieldIndex);

			EvtGetEvent(eventP, evtWaitForever);
			if (eventP->eType == keyDownEvent) {
				Boolean capsLock, numLock, autoShifted;
				UInt16 tempShift;

				GrfGetState(&capsLock, &numLock, &tempShift, &autoShifted);

				GrfSetState(capsLock, numLock, false);
				EvtAddEventToQueue(eventP);
			}
			handled = true;
			break;
			
		case frmGotoEvent:
			formP = FrmGetActiveForm ();
			fieldIndex = FrmGetObjectIndex(formP, fldEditText);
			fieldP = FrmGetObjectPtr(formP, fieldIndex);
			
			currentRecord = eventP->data.frmGoto.recordNum;

			if (currentDB == cryptoPadDB)
				if (isEncrypted(currentRecord)) {
					currentPasswordP = globalPasswordP;
					DecryptRecord(currentRecord, currentPasswordP);
				}
			EditFormInit();
			FldSetScrollPosition (fieldP, eventP->data.frmGoto.matchPos);
			FldSetSelection(fieldP, eventP->data.frmGoto.matchPos, 
			eventP->data.frmGoto.matchPos + eventP->data.frmGoto.matchLen);
			EditFormUpdateScrollBar();
			FrmDrawForm (formP);
			FrmSetFocus(formP, fieldIndex);
			handled = true;
			break;

		case frmCloseEvent:
		case frmSaveEvent:
			formP = FrmGetActiveForm();

			if (FldGetTextHandle (FrmGetObjectPtr (formP, FrmGetObjectIndex (formP, fldEditText)))) {
		
				fieldIndex = FrmGetObjectIndex(formP, fldEditText);
				fieldP = FrmGetObjectPtr(formP, fieldIndex);
				EditSaveData(fieldP);
			}
			/* Let the system handle the event */
			break;
	}

	return handled;
}
