/*
 *   $Id: graph.c,v 1.4 2000/10/29 16:17:20 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/


#include <PalmOS.h>

#include "defuns.h"
#include "MathLib.h"
#include "konvert.h"
#include "funcs.h"
#include "calcDB.h"
#include "calcrsc.h"
#include "prefs.h"
#include "calc.h"
#include "graph.h"
#include "grprefs.h"
#include "stack.h"
#include "grsetup.h"
#include "grtaps.h"
#include "varmgr.h"


static struct {
	Boolean active;
	Int32 position;
	Int16 funcnum;
	double start;
	double stop;
	double step;
}graphState;

/* Screen preferences in pixels */
/* Please note, tha ymax is de facto smaller then ymin on the screen */
struct {
	/* The Y screen max is the smallest value */
	/*0,0   1,0                                     */
	/*0,1                                        */
	/*0,2[beginning of graph                     */
	 
	Coord xmin,ymax;
	Coord width,height;
	/* Cached coeficients for graph2screen computations */
	/* CacheTrans MUST be called when any of the ScrPrefs
	 * and graphPrefs attributes is changed */
	double xtrans,ytrans;
}ScrPrefs;

Tgraph_curves graphCurves[MAX_GRFUNCS];

/***********************************************************************
 *
 * FUNCTION:     graph_setcolor
 * 
 * DESCRIPTION:  Sets foreground color to the color of funcnum and
 *               background to white
 *
 * PARAMETERS:   funcnum - (-1) - sets foreground color to black
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
graph_setcolor(Int16 funcnum)
{
	if (palmOS35)
	  WinPushDrawState();
	
	if (colorDisplay) {
		RGBColorType rgb;
		
		rgb.r=rgb.g=rgb.b=255; /* White */
		WinSetBackColor(WinRGBToIndex(&rgb));
		if (funcnum == -1) {
			rgb.r=rgb.g=rgb.b=0; /* Black */
			WinSetForeColor(WinRGBToIndex(&rgb));
		}
		else 
		  WinSetForeColor(graphColors[funcnum]);
	}
}

/***********************************************************************
 *
 * FUNCTION:     graph_unsetcolor
 * 
 * DESCRIPTION:  Restores color settings after graph_setcolor
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
graph_unsetcolor(void)
{
	if (palmOS35)
	  WinPopDrawState();
}

/***********************************************************************
 *
 * FUNCTION:     graph_is_onscreen
 * 
 * DESCRIPTION:  If the coordinates fit on the screen
 *
 * PARAMETERS:   x,y - coordinates of the original function
 *
 * RETURN:       true - fits, false - outside
 *      
 ***********************************************************************/
static Boolean
graph_is_onscreen(double x,double y)
{
	if (!finite(x) || !finite(y))
	  return false;
	if (x>=graphPrefs.xmin && x<=graphPrefs.xmax &&
		y>=graphPrefs.ymin && y<=graphPrefs.ymax)
	  return true;
	return false;
}

/***********************************************************************
 *
 * FUNCTION:     graph_xgr2scr
 * 
 * DESCRIPTION:  Convert the 'x' or 'y' coordinate from the function to
 *               screen coordinates
 * 
 * PARAMETERS:   x or y
 *
 * RETURN:       coordinate
 *      
 ***********************************************************************/
Coord 
graph_xgr2scr(double x)
{
	return ScrPrefs.xmin+(Coord)round((x-graphPrefs.xmin)*ScrPrefs.xtrans);
}

Coord 
graph_ygr2scr(double y)
{
	return ScrPrefs.ymax+(Coord)round((graphPrefs.ymax-y)*ScrPrefs.ytrans);
}

/***********************************************************************
 *
 * FUNCTION:     graph_xscr2gr
 * 
 * DESCRIPTION:  Convert a screen 'x' (or y) coordinate to a graph coordinate
 *
 * PARAMETERS:   x or y
 *
 * RETURN:       coordinate
 *      
 ***********************************************************************/
double 
graph_xscr2gr(Coord x)
{
	return graphPrefs.xmin+(double)(x-ScrPrefs.xmin)/(double)ScrPrefs.xtrans;
}

double 
graph_yscr2gr(Coord y)
{
	return graphPrefs.ymin+(double)((ScrPrefs.ymax+ScrPrefs.height)-y)/ScrPrefs.ytrans;
//	return graphPrefs.ymax-(double)(y-ScrPrefs.ymax)/(double)ScrPrefs.ytrans;
}

/***********************************************************************
 *
 * FUNCTION:     graph_isec_y
 * 
 * DESCRIPTION:  Find an intersection of a line with a vertical line 
 *               - it is used when drawing lines to points outside of
 *               the drawing area
 *
 * PARAMETERS:   x1,y1 - point 1
 *               x2,y2 - point 2
 *               x3 - the vertical line
 * RETURN:       y3 - the point on the vertical line
 *      
 ***********************************************************************/
static double
graph_isec_y(double x1,double y1,double x2,double y2,double x3)
{
	double sidea,sideb,parta;
	
	sidea = y2-y1;
	sideb = x2-x1;
	parta = x2 - x3;
	
	return y2-(sidea*parta)/sideb;
}

#define graph_isec_x(x1,y1,x2,y2,y3) (graph_isec_y(y1,x1,y2,x2,y3))

/***********************************************************************
 *
 * FUNCTION:     graph_outscreen
 * 
 * DESCRIPTION:  Find a point, where a line should be drawn instead of
 *               the point outside of drawing area
 *
 * PARAMETERS:   onx,ony - points in the drawing area
 *               offx,offy - points outside drawing area
 *
 * RETURN:       newx,newy - new points on the border of drawing area
 *      
 ***********************************************************************/
static void
graph_outscreen(double *newx,double *newy,
				double onx,double ony,double offx,double offy)
{
	double tmp;

	tmp = graph_isec_x(onx,ony,offx,offy,graphPrefs.ymin);
	if (offy<graphPrefs.ymin && tmp>=graphPrefs.xmin && tmp<=graphPrefs.xmax) {
		*newx = tmp;
		*newy = graphPrefs.ymin;
		return;
	}
	tmp = graph_isec_x(onx,ony,offx,offy,graphPrefs.ymax);
	if (offy>graphPrefs.ymax && tmp>=graphPrefs.xmin && tmp<=graphPrefs.xmax) {
		*newx = tmp;
		*newy = graphPrefs.ymax;
		return;
	} 
	
	tmp = graph_isec_y(onx,ony,offx,offy,graphPrefs.xmin);
	if (offx<graphPrefs.xmin && tmp>=graphPrefs.ymin && tmp<=graphPrefs.ymax) {
		*newx = graphPrefs.xmin;
		*newy = tmp;
		return;
	}
	tmp = graph_isec_y(onx,ony,offx,offy,graphPrefs.xmax);
	if (offx>graphPrefs.xmax && tmp>=graphPrefs.ymin && tmp<=graphPrefs.ymax) {
		*newx = graphPrefs.xmax;
		*newy = tmp;
		return;
	} 
	*newx = onx;
	*newy = ony;
}

/***********************************************************************
 *
 * FUNCTION:     graph_new_point
 * 
 * DESCRIPTION:  Adds a new point to graph. Depending on the selected
 *               style this function chooses way of drawing, colors etc.
 *
 * PARAMETERS:   x,y - coordinates of a new point
 *               funcnum - function number, where the point belongs to
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_new_point(double x,double y,Int16 funcnum)
{
	double oldx = graphCurves[funcnum].x;
	double oldy = graphCurves[funcnum].y;

	if (!graphCurves[funcnum].valid &&
		graph_is_onscreen(x,y)) {
		WinDrawLine(graph_xgr2scr(x),graph_ygr2scr(y),
					graph_xgr2scr(x),graph_ygr2scr(y));
	}	
	else if (graphCurves[funcnum].valid) {
		if (graph_is_onscreen(x,y) && graph_is_onscreen(oldx,oldy)) { 
			WinDrawLine(graph_xgr2scr(oldx),graph_ygr2scr(oldy),
						graph_xgr2scr(x),graph_ygr2scr(y));
		}
		else if (graph_is_onscreen(x,y)) { 
			/* First point on screen */
			double newx,newy;
			graph_outscreen(&newx,&newy,x,y,oldx,oldy);
			WinDrawLine(graph_xgr2scr(x),graph_ygr2scr(y),
						graph_xgr2scr(newx),graph_ygr2scr(newy));
		}
		else if (graph_is_onscreen(oldx,oldy)) { 
			/* Second point on screen */
			double newx,newy;
			graph_outscreen(&newx,&newy,oldx,oldy,x,y);
			WinDrawLine(graph_xgr2scr(oldx),graph_ygr2scr(oldy),
						graph_xgr2scr(newx),graph_ygr2scr(newy));
		}
	}
	graphCurves[funcnum].x = x;
	graphCurves[funcnum].y = y;
	graphCurves[funcnum].valid = true;
}
/* Initialize cache variables to
 * avoid this time-expensive computation in the future */		
void
cacheTrans(void)
{		
	ScrPrefs.xtrans=(double)ScrPrefs.width/(graphPrefs.xmax-graphPrefs.xmin);
	ScrPrefs.ytrans=(double)ScrPrefs.height/(graphPrefs.ymax-graphPrefs.ymin);
}
		
/* Initializes Graph gadget and shows the basic grid */
static void 
InitGraphScreen(FormPtr frm,UInt16 graphId)
{
	RectangleType bounds;
	Coord x,y;

	graph_setcolor(-1);
	
	gadget_bounds(graphId,&bounds);
	WinEraseRectangle(&bounds,0);
	
	ScrPrefs.xmin=bounds.topLeft.x;
	ScrPrefs.width=bounds.extent.x;
	ScrPrefs.ymax=bounds.topLeft.y;
	ScrPrefs.height=bounds.extent.y;
	cacheTrans();

/* Now draw the x axis */
	x = graph_xgr2scr(0.0);
	y = graph_ygr2scr(0.0);
	if (x>bounds.topLeft.x && x<bounds.topLeft.x+bounds.extent.x)
	  WinDrawLine(x,
		      bounds.topLeft.y,
		      x,
		      bounds.topLeft.y+bounds.extent.y);
/* Now draw the y axis */
	if (y>bounds.topLeft.y && y<bounds.topLeft.y+bounds.extent.y)
	  WinDrawLine(bounds.topLeft.x,
		      y,
		      bounds.topLeft.x+bounds.extent.x,
		      y);

	
/* Draw the rectangle around the drawing area */
	WinDrawLine(bounds.topLeft.x-1,bounds.topLeft.y-1,bounds.topLeft.x-1,
		    bounds.topLeft.y+bounds.extent.y);
	WinDrawLine(bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y-1,bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y+bounds.extent.y);
	WinDrawLine(bounds.topLeft.x,bounds.topLeft.y-1,
		    bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y-1);
	WinDrawLine(bounds.topLeft.x-1,bounds.topLeft.y+bounds.extent.y,
		    bounds.topLeft.x+bounds.extent.x,
		    bounds.topLeft.y+bounds.extent.y);
	
	graph_unsetcolor();
}

/***********************************************************************
 *
 * FUNCTION:     graph_get_value
 * 
 * DESCRIPTION:  Return a real value of a given function with
 *               a given parameter
 *
 * PARAMETERS:   origstack - the compiled function
 *               param - argument of function
 *
 * RETURN:       result of function
 *               NaN - error, not defined etc.
 *      
 ***********************************************************************/
double
graph_get_value(CodeStack *origstack,double param)
{
	Trpn argument;
	CodeStack *stack;
	CError err;
	double result;
	
	stack = stack_new(origstack->size);
	stack_copy(stack,origstack);

	argument.type = real;
	argument.allocsize = 0;
	argument.u.realval = param;	
	set_func_argument(&argument);
	err = stack_compute(stack);
	decr_func_argument();

	result = NaN;
	if (!err) {
		err = stack_get_val(stack,&result,real);
		if (err)
		  result = NaN;
	}
	stack_delete(stack);
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     graph_get_vals
 *
 * DESCRIPTION:  Return the coordinates on the resulting graph (in graph
 *               coordinates) of the function given by funcnum and
 *               from the parameter 'param'. 
 *
 * PARAMETERS:   fnum - the function number
 *               param - parameter value
 *
 * RETURN:       true - the resulting 'x' and 'y' coordinates are valid
 *               false - an error occured - e.g. function not defined etc.
 *      
 ***********************************************************************/
Boolean
graph_get_vals(Int16 fnum,double param,double *x,double *y)
{
	if (graphPrefs.functype == graph_param) {
		if (!graphCurves[fnum].stack1 ||
			!graphCurves[fnum].stack2)
		  return false;
		*x = graph_get_value(graphCurves[fnum].stack1,param);
		*y = graph_get_value(graphCurves[fnum].stack2,param);
		if (finite(*x) && finite(*y))
		  return true;
		return false;
	}
	else if (graphPrefs.functype == graph_func) {
		if (!graphCurves[fnum].stack1)
		  return false;
		*x = param;
		*y = graph_get_value(graphCurves[fnum].stack1,param);
		if (finite(*y))
		  return true;
		return false;		
	}
	else { /* graph_pol */
		double r;
		
		if (!graphCurves[fnum].stack1)
		  return false;
		r = graph_get_value(graphCurves[fnum].stack1,param);
		if (!finite(r))
		  return false;
		*x = r*cos(param);
		*y = r*sin(param);
		
		return true;
	}
}

/***********************************************************************
 *
 * FUNCTION:     graph_zoom_out
 *
 * DESCRIPTION:  Zoom out by making a window of a same center but
 *               twice as high and twice as wide
 *
 * PARAMETERS:   frm, objectId - id of the graph gadget
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_zoom_out(FormPtr frm,Int16 objectId)
{
	double width = graphPrefs.xmax - graphPrefs.xmin;
	double height = graphPrefs.ymax - graphPrefs.ymin;
	
	graphPrefs.xmin-=width/2;
	graphPrefs.xmax+=width/2;
	
	graphPrefs.ymin-=height/2;
	graphPrefs.ymax+=height/2;
	
	FrmUpdateForm(tdFormGraph,frmUpdateVars);
}


/***********************************************************************
 *
 * FUNCTION:     graph_init_cache
 * 
 * DESCRIPTION:  Initializes the field of precompiled functions with
 *               NULLs
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_init_cache(void)
{
	Int16 i;
	for (i=0;i<MAX_GRFUNCS;i++) {
		graphCurves[i].stack1 = NULL;
		graphCurves[i].stack2 = NULL;
	}	  
}

/***********************************************************************
 *
 * FUNCTION:     graph_clean_cache
 * 
 * DESCRIPTION:  Frees memory occupied by precompiled functions
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_clean_cache(void)
{
	Int16 i;
	for (i=0;i<MAX_GRFUNCS;i++) {
		if (graphCurves[i].stack1) {
			stack_delete(graphCurves[i].stack1);
			graphCurves[i].stack1 = NULL;
		}
			if (graphCurves[i].stack2) {
			stack_delete(graphCurves[i].stack2);
			graphCurves[i].stack2 = NULL;
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:     graph_compile_cache
 * 
 * DESCRIPTION:  Compiles functions that should be drawn and saves
 *               for later use
 *
 * PARAMETERS:   None
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
graph_compile_cache(void)
{
	int i;
	char *name;
	CodeStack *stack;
	CError err;
	
	graph_clean_cache();
	if (graphPrefs.functype==graph_func ||
		graphPrefs.functype==graph_polar) {
		for (i=0;i<MAX_GRFUNCS;i++) {
			name = grsetup_get_fname(i);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack1 = stack;
		}
	} else {
		/* Parametric */
		for (i=0;i<MAX_GRFUNCS;i++) {
			name  = grsetup_get_fname(i*2);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack1 = stack;
			
			name  = grsetup_get_fname(i*2+1);
			if (StrLen(name)==0)
			  continue;
			stack = db_read_function(name,&err);
			if (err)
			  continue;
			graphCurves[i].stack2 = stack;
		}
	}		
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw_start
 * 
 * DESCRIPTION:  Initialize the drawing procedure, that later occurs 
 *               using nilEvent in background
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw_start(void)
{
	int i;
		
	for (i=0;i<MAX_GRFUNCS;i++) 
	  graphCurves[i].valid = false;
	
	graph_compile_cache();
	
	InitGraphScreen(FrmGetActiveForm(),graphGadget);
	
	graphState.active = true;
	graphState.position = 0;
	graphState.funcnum = 0;
	if (graphPrefs.functype == graph_func) {
		graphState.start = graphPrefs.xmin;
		graphState.stop = graphPrefs.xmax;
		graphState.step = (graphPrefs.xmax - graphPrefs.xmin)/ScrPrefs.width;
	} 
	else if (graphPrefs.functype == graph_polar) {
		graphState.start = graphPrefs.fimin;
		graphState.stop = graphPrefs.fimax;
		graphState.step = graphPrefs.fistep;
	}
	else if (graphPrefs.functype == graph_param) {
		graphState.start = graphPrefs.tmin;
		graphState.stop = graphPrefs.tmax;
		graphState.step = graphPrefs.tstep;
	}
	calc_nil_timeout(0);
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw_stop, graph_draw_resum
 * 
 * DESCRIPTION:  Stop/resume drawing functions
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw_resume()
{
	graphState.active = true;
	calc_nil_timeout(0);
}

static void
graph_draw_stop()
{
	if (graphState.active)
	  calc_nil_timeout(evtWaitForever);
	graphState.active = false;
}

/***********************************************************************
 *
 * FUNCTION:     graph_draw
 * 
 * DESCRIPTION:  Draws a piece of a function, called from nilEvent
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
graph_draw(void)
{
	Int32 i;
	double start;
	double x,y;

	if (!graphState.active)
	  return;
	if (graphState.funcnum>=MAX_GRFUNCS) {
		graph_draw_stop();
		return;
	}

	/* Check, if we have all functions to draw graph */
	if (!graphCurves[graphState.funcnum].stack1 ||
		(graphPrefs.functype==graph_param && !graphCurves[graphState.funcnum].stack2)) {
		graphState.funcnum++;
		return;
	}

	graph_setcolor(graphState.funcnum);
	/* Should we draw all points at once, or step slowly to the end? */
	start = graphState.start + graphState.step * graphState.position;

	for (i=0;i < DRAW_AT_ONCE && start <= graphState.stop;i+=1) {
		graph_get_vals(graphState.funcnum,
					   start,
					   &x,&y);
		graph_new_point(x,y,graphState.funcnum);
		start = graphState.start+graphState.step*(graphState.position+i);
	}
	graph_unsetcolor();
	
	graphState.position+=i;
	if (start >= graphState.stop) {
		/* Reached end of drawing area/parameters */
		graphState.position = 0;
		graphState.funcnum++;
	}
}

Boolean 
GraphFormHandleEvent(EventPtr event)
{
	Boolean    handled = false;
	FormPtr    frm=FrmGetActiveForm();
	Int16 controlId;
	double position;

	switch (event->eType) {
	 case winEnterEvent:
		/* We still recieve nilEvent when in the menu, disable the
		 * nilEvent when in Menu */
		if (event->data.winEnter.enterWindow == 
			(WinHandle) FrmGetFormPtr(tdFormGraph))
		  graph_draw_resume();
		else 
		  graph_draw_stop();
		break;
	 case nilEvent:
		graph_draw();
		handled = true;
		break;
	 case penDownEvent:		  
		handled=grtaps_tap(event->screenX,event->screenY);
		break;
	 case ctlSelectEvent:
		controlId=event->data.ctlSelect.controlID;
		  if (chooseForm(controlId)) {
			  handled=true;
			  break;
		  }
		handled=true;
		switch (controlId) {			
		 case graphZoomOut:
			graph_zoom_out(frm,graphGadget);
			break;
		 case graphPrefButton:
			FrmPopupForm(grPrefForm);
			break;
		 case graphSetupButton:
			FrmPopupForm(grSetupForm);
			break;
		 case graphTrack:
			if (CtlGetValue(GetObjectPtr(graphTrack))) {				
				if (!grtaps_track_but()) 
				  grtaps_track_pause();
				else 
				  grtaps_track_start();
			}
			else /* User unchecked the button */
			  grtaps_track_pause();
			graph_draw_resume();
			break;
		 case graphTrackGoto:
			if (varmgr_get_double(&position,"Jump to position")) {
				grtaps_track_manual(position,track_set);
			}
			handled = true;
			break;
		 default:
			handled=false;
		}
		break;
	 case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmRedrawUpdateCode) {
			FrmDrawForm(FrmGetActiveForm());
			graph_draw_start();
			handled = true;
		}
		if (event->data.frmUpdate.updateCode == frmUpdateVars) {
			graph_draw_start();
			grtaps_track_reset(graphState.start);
			handled = true;
		}
		break;
	 case keyDownEvent:
		if (TxtCharIsHardKey(event->data.keyDown.modifiers,event->data.keyDown.chr) 
			&& !(event->data.keyDown.modifiers & poweredOnKeyMask)) {
			GotHardKey();
			handled=true;
		}
		else if (CtlGetValue(GetObjectPtr(graphTrack)) &&
				 !(event->data.keyDown.modifiers & poweredOnKeyMask))
		  switch (event->data.keyDown.chr) {
/*		   // The SysHandleEvent doesn't pass this throught,
           // I'll have to create something to disable/enable
           // calling SysHandleEvent for these buttons
		   case vchrHard3:
			  grtaps_track_manual(graphState.step*5,track_add);
			  handled = true;
			  break; */
		   case vchrPageDown:
			  grtaps_track_manual(graphState.step,track_add);
			  handled = true;
			  break;
		   case vchrPageUp:
			  grtaps_track_manual(-graphState.step,track_add);
			  handled = true;
			  break;
/*		   case vchrHard2:
			  grtaps_track_manual(-graphState.step*5,track_add);
			  handled = true;
			  break; */
		  }
		break;
	 case frmCloseEvent:
		graph_draw_stop();
		graph_clean_cache();
		handled = false;
		break;
	 case frmOpenEvent:
		calcPrefs.form=frm->formId;		
		FrmDrawForm(frm);
		
		CtlSetValue(GetObjectPtr(tdBGra),1);
		graph_init_cache();
		graph_draw_start();
		
		grtaps_track_reset(graphState.start);
				
		handled=true;
		break;
	 case menuEvent:
		if (chooseForm(event->data.menu.itemID)) {
			handled=true;
			break;
		  }
		switch (event->data.menu.itemID) {
		 case graphPrefButton:
			FrmPopupForm(grPrefForm);
			handled = true;
			break;
		 case graphTableButton:
			FrmPopupForm(grTableForm);
			handled = true;
			break;
		 case graphSetupButton:
			FrmPopupForm(grSetupForm);
			handled = true;
			break;
		 case tdBPref:
			FrmPopupForm(prefForm);
			handled = true;
			break;
		}
		break;
	}
	  
	return handled;
}
