/*
 *   $Id: grsetup.c,v 1.8 2000/11/14 23:15:44 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include <string.h>

#include "calcDB.h"
#include "konvert.h"
#include "grsetup.h"
#include "grprefs.h"
#include "graph.h"
#include "varmgr.h"
#include "calcrsc.h"
#include "calc.h"

static Int16 firstVisible = 0;
static Int16 selectedRow;

static char *tblFuncs[] = {"Y1","Y2","Y3","Y4","Y5","Y6"};
static char *tblPols[] = {"r1","r2","r3","r4","r5","r6"};
static char *tblPars[] = {"X1","Y1","X2","Y2","X3","Y3","X4","Y4","X5","Y5","X6","Y6"};

enum {
	COL_FLIST = 2,
	COL_FUNC = 3
};

static const char *
grsetup_param_name(void) IFACE;
static const char *
grsetup_param_name(void)
{
	if (graphPrefs.functype == graph_func)
	  return "x";
	else if (graphPrefs.functype == graph_polar)
	  return "fi";
	else /* Parametric */
	  return "t";
}

const char *
grsetup_fn_descr(Int16 i) IFACE;
const char *
grsetup_fn_descr(Int16 i)
{
	if (graphPrefs.functype == graph_func)
	  return tblFuncs[i];
	else if (graphPrefs.functype == graph_polar)
	  return tblPols[i];
	else 
	  return tblPars[i];
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_def_name
 * 
 * DESCRIPTION:  Return name of function, that should be created specially
 *               for graphing purposes (the 'User' selection)
 *
 * PARAMETERS:   row - the row number in a grsetup table
 *
 * RETURN:       Pointer to static function name 
 *      
 ***********************************************************************/
static const char * grsetup_def_name(Int16 row) IFACE;
static const char *
grsetup_def_name(Int16 row)
{
	static char result[MAX_FUNCNAME+1];
	char tmp[2];
	
	StrCopy(tmp,"1");
	switch (graphPrefs.functype) {
	 case graph_func:
		StrCopy(result,"z_grafun");
		tmp[0]+=(char)row;
		StrCat(result,tmp);
		break;
	 case graph_polar:
		StrCopy(result,"z_grapol");
		tmp[0]+=(char)row;
		StrCat(result,tmp);
		break;
	 case graph_param:
		StrCopy(result,"z_grapar");
		StrCat(result,row%2?"b":"a");
		tmp[0]+=(char)row/2;
		StrCat(result,tmp);
		break;
	}
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_get_fname
 * 
 * DESCRIPTION:  Get a name of a function for the given row, that is
 *               already saved in Preferences
 *
 * PARAMETERS:   row - number of row in the displayed table
 *
 * RETURN:       !!! Pointer to a graphPrefs structure, so
 *               functions can modify the name directly
 *      
 ***********************************************************************/
char * grsetup_get_fname(Int16 row) IFACE;
char *
grsetup_get_fname(Int16 row)
{
	switch (graphPrefs.functype) {
	 case graph_func:
		return graphPrefs.funcFunc[row];
	 case graph_polar:
		return graphPrefs.funcPol[row];
	 case graph_param:
	 default:
		return graphPrefs.funcPar[row/2][row%2];
	}
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_tbl_contents
 * 
 * DESCRIPTION:  Callback function for drawing the function name or
 *               contents of the user defined function
 *
 * PARAMETERS:   see PalmOS Reference
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
grsetup_tbl_contents(void *table,Int16 row,Int16 column, 
					 RectangleType *bounds) IFACE;
static void
grsetup_tbl_contents(void *table,Int16 row,Int16 column, 
					 RectangleType *bounds)
{
	char *name;
	const char *usrname;
	char *text;
	CError err;
	RectangleType rp;
	
	row += firstVisible;
	name = grsetup_get_fname(row); /* name is now a pointer to graphPrefs */
	usrname = grsetup_def_name(row);

	WinGetClip(&rp);
	WinSetClip(bounds);
	if (StrLen(name)==0)
	  ; /* Do not draw nothing */
	else if (StrCompare(usrname,name)==0) {
		/* Special graph function, show contents */
		err = db_func_description(usrname,&text,NULL);
		if (!err) {
			WinDrawChars(text,StrLen(text),bounds->topLeft.x,bounds->topLeft.y);
			MemPtrFree(text);			
		}
		else 
		  /* Somone probably deleted it, reset the value */
		  StrCopy(name,"");
	} else {
		/* Normal function, show ordinary name */
		char tmpname[MAX_FUNCNAME+5];
		StrCopy(tmpname,name);
		StrCat(tmpname,"()");
		WinDrawChars(tmpname,StrLen(tmpname),bounds->topLeft.x,bounds->topLeft.y);
	}
	WinSetClip(&rp);
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_tbl_name
 * 
 * DESCRIPTION:  Display the name ('Y1','r1' etc.) of a function,
 *               use a function color if on color display
 *
 * PARAMETERS:   See PalmOS Reference for callback functions
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grsetup_tbl_name(void *table,Int16 row,Int16 column, 
				 RectangleType *bounds) IFACE;
static void
grsetup_tbl_name(void *table,Int16 row,Int16 column, 
				  RectangleType *bounds)
{
	char text[6];
	Coord x;

	StrCopy(text,grsetup_fn_descr(row+firstVisible));
	StrCat(text,":");
	if (colorDisplay) {
		WinPushDrawState();
		if (graphPrefs.functype==graph_param) 
		  WinSetTextColor(graphColors[(row+firstVisible)/2]);
		else
		  WinSetTextColor(graphColors[row+firstVisible]);
	} 
	x = bounds->topLeft.x+bounds->extent.x-FntCharsWidth(text,StrLen(text))-2;
	WinDrawChars(text,StrLen(text),x,bounds->topLeft.y);

	if (colorDisplay)
	  WinPopDrawState(); 
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_tbl_init
 * 
 * DESCRIPTION:  Initialize the on-screen table 
 *
 * PARAMETERS:   None
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grsetup_tbl_init(void) IFACE;
static void
grsetup_tbl_init(void)
{
	Int16 i;
	TablePtr table;
	UInt16 rows;
	
	table = GetObjectPtr(grSetupTable);
	rows = TblGetNumberOfRows(table);
	
	for (i=0;i<rows;i++) {
		TblSetItemStyle(table,i,0,customTableItem);
		TblSetItemStyle(table,i,1,customTableItem);
		TblSetItemStyle(table,i,2,customTableItem);
		TblSetItemStyle(table,i,3,customTableItem);
	}
	TblSetColumnUsable(table,2,true);
	TblSetColumnUsable(table,3,true);
	
	TblSetCustomDrawProcedure(table,3,grsetup_tbl_contents);
	TblSetCustomDrawProcedure(table,2,grsetup_tbl_name);
	
	/* We will effectively remember the last state of graph_param */
	/* because others do not have scrolling table */
	if (graphPrefs.functype != graph_param)
	  firstVisible = 0;
	
	/* Setup the up/down pointers */
	if (graphPrefs.functype == graph_param) {
		FormPtr frm = FrmGetActiveForm();
		FrmShowObject(frm,FrmGetObjectIndex(frm,grSetupDown));
		FrmShowObject(frm,FrmGetObjectIndex(frm,grSetupUp));
	}
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_handle_pendown
 * 
 * DESCRIPTION:  Simulates a button in all cells of table
 *
 * PARAMETERS:   bounds of a cell in table, where the user tapped
 *
 * RETURN:       true - user lifted the pen inside bounds of table 
 *                      (button pressed)
 *               false - user lifted pen outside bounds of a field
 *      
 ***********************************************************************/
Boolean
grsetup_handle_pendown(RectangleType *bounds)
{
	Boolean penDown;
	Boolean inrect = true;
	Coord x,y;
	
	WinInvertRectangle(bounds,0);
	do {
		EvtGetPen(&x,&y,&penDown);
		if (RctPtInRectangle(x,y,bounds) && !inrect) {
			inrect = true;
			WinInvertRectangle(bounds,0);
		} else if (!RctPtInRectangle(x,y,bounds) && inrect) {
			inrect = false;
			WinInvertRectangle(bounds,0);
		}		
	}while(penDown);
	if (inrect)
	  WinInvertRectangle(bounds,0);
	return inrect;
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_func_tap
 * 
 * DESCRIPTION:  Triggered when user tapped on the function name, show
 *               a varmgr_edit window
 *
 * PARAMETERS:   row - where the function name was pressed
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
grsetup_func_tap(Int16 row) IFACE;
static void
grsetup_func_tap(Int16 row)
{
	char usrname[MAX_FUNCNAME+1]; /* grsetup_def_name returns static variable,
								  * that can change e.g. on form closure
								  * (and update on Debug ROM).
								  * DO not store a pointer
								  */
	char tmptxt[MAX_RSCLEN];
	char *grname;
	
	grname = grsetup_get_fname(row);	
	StrCopy(usrname,grsetup_def_name(row));
	
	if (StrLen(grname)==0) {
		SysCopyStringResource(tmptxt,grSetupStrFunc);
		if (varmgr_edit(usrname,tmptxt,function,false,
						grsetup_param_name())) {
			StrCopy(grname,usrname);
			FrmUpdateForm(grSetupForm,frmUpdateVars);
		}
		return;
	}		
	if (StrCompare(usrname,grname)==0) {
		SysCopyStringResource(tmptxt,grSetupStrFunc);
		if (varmgr_edit(usrname,tmptxt,function,false,
						grsetup_param_name())) {
			StrCopy(grname,usrname);
			FrmUpdateForm(grSetupForm,frmUpdateVars);
		}
	}
	else {
		SysCopyStringResource(tmptxt,varFuncModTitle);
		varmgr_edit(grname,tmptxt,function,false,
					grsetup_param_name());
	}
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_scroll
 * 
 * DESCRIPTION:  Scrolls a table (parameter functions do not fit on 1 
 *               screen)
 *
 * PARAMETERS:   dir - which direction shall we scroll
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
grsetup_scroll(WinDirectionType dir) IFACE;
static void
grsetup_scroll(WinDirectionType dir)
{
	TablePtr table;		
	Int16 tmp;
	
	if (graphPrefs.functype!=graph_param)
	  return;
	table = GetObjectPtr(grSetupTable);
	if (dir == winUp) {
		tmp=firstVisible - TblGetNumberOfRows(table);
		if (tmp<0)
		  tmp = 0;
	}
	else /* winDown */ {
		tmp = firstVisible + TblGetNumberOfRows(table);
		if (tmp > TblGetNumberOfRows(table))
		  tmp = TblGetNumberOfRows(table);
	}
	/* Don't redraw if we did not move */
	if (tmp==firstVisible)
	  return;
	firstVisible = tmp;
	TblMarkTableInvalid(table);
	TblRedrawTable(table);
}

/***********************************************************************
 *
 * FUNCTION:     grsetup_flist_tap
 * 
 * DESCRIPTION:  User tapped the 'X:'/'Y:' etc. Show a popup list of 
 *               defined functions with 'Off' and 'User' at the beginning
 *               Make a table searchable, Off and User have space as 
 *               a 1st character to allow incremental search
 *
 * PARAMETERS:   row, where the 'button' was pressed
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void grsetup_flist_tap(Int16 row) IFACE;
static void
grsetup_flist_tap(Int16 row)
{
	dbList *dblist;
	char **displist;
	ListPtr list;
	RectangleType bounds;
	char *fname;
	Int16 i,pos=0;
	char tmptxt[MAX_RSCLEN],txtuser[MAX_RSCLEN],txtnone[MAX_RSCLEN];

	dblist = db_get_list(function);
	displist = MemPtrNew(sizeof(*displist)*(dblist->size+2));
	
	/* displist will contain dynamic and static items, but the dynamic
	 * will be released by db_delete_list with dblist */
	txtnone[0] = ' ';
	SysCopyStringResource(txtnone+1,grSetupNone);
	txtuser[0] = ' ';
	SysCopyStringResource(txtuser+1,grSetupUser);	
	displist[0] = txtnone;
	displist[1] = txtuser;	
	memcpy((void *)&displist[2],(void *)dblist->list,dblist->size*sizeof(*displist));

	/* Fname is actually pointer to graphPref structure */
	fname = grsetup_get_fname(row);
	if (StrLen(fname)==0)
	  pos = 0;
	else if (StrCompare(fname,grsetup_def_name(row))==0)
	  pos = 1;
	else
	  for (i=0;i<dblist->size;i++)
		if (StrCompare(fname,dblist->list[i])==0)
		  pos = i+2;
	
	TblGetBounds(GetObjectPtr(grSetupTable),&bounds);
	list = GetObjectPtr(grSetupList);
	list->attr.search = true;
	LstSetListChoices(list,displist,dblist->size+2);
	LstSetSelection(list,pos);
	LstMakeItemVisible(list,pos);
	LstSetPosition(list,bounds.topLeft.x,bounds.topLeft.y);
	LstSetHeight(list,dblist->size+2>6?6:dblist->size+2);
	
	i = LstPopupList(list);
	
	if (i!=noListSelection && i!=pos) {
		if (i==0) 
		  StrCopy(fname,"");
		else if (i==1) { /* The 'User' function get always edit window */
			SysCopyStringResource(tmptxt,grSetupStrFunc);
			varmgr_edit(grsetup_def_name(row),tmptxt,function,false,
						grsetup_param_name());
			StrCopy(fname,grsetup_def_name(row));
		}
		else
		  StrCopy(fname,displist[i]);
		FrmUpdateForm(grSetupForm,frmUpdateVars);
	}
	
	db_delete_list(dblist);
	MemPtrFree(displist);
}

/***********************************************************************
 *
 * FUNCTION:     GraphSetupHandleEvent
 * 
 * DESCRIPTION:  Event handler for graphic Setup form, where
 *               the functions, that should be drawn are selected
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean 
GraphSetupHandleEvent(EventPtr event)
{
	Boolean    handled = false;
	FormPtr    frm=FrmGetActiveForm();
	Int16 formId;
	Int16 controlId;
	RectangleType bounds;
	Int16 row,col;

	switch (event->eType) {
	 case tblEnterEvent:
		row = event->data.tblSelect.row;
		col = event->data.tblSelect.column;
		TblGetItemBounds(GetObjectPtr(grSetupTable),
						 row,col,&bounds);
		if (grsetup_handle_pendown(&bounds)) {
			selectedRow = row + firstVisible;
			switch (col) {
			 case COL_FLIST:
				grsetup_flist_tap(selectedRow);
				break;
			 case COL_FUNC:
				grsetup_func_tap(selectedRow);
				break;
			}
		}
		handled = true;
		break;
	 case keyDownEvent:
		switch (event->data.keyDown.chr) {
		 case pageDownChr:
			grsetup_scroll(winDown);
			handled = true;
			break;
		 case pageUpChr:
			grsetup_scroll(winUp);
			handled = true;
			break;
		}
		break;
	 case ctlSelectEvent:
		controlId=event->data.ctlSelect.controlID;
		switch (controlId) {
		 case grSetupDone:
			FrmReturnToForm(0);
			FrmUpdateForm(tdFormGraph,frmUpdateVars);
			handled = true;
			break;
		 case grSetupDown:
			grsetup_scroll(winDown);
			handled = true;
			break;
		 case grSetupUp:
			grsetup_scroll(winUp);
			handled = true;
			break;
		}
		break;
	 case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode == frmUpdateVars) {
			TablePtr tbl;
			
			tbl = GetObjectPtr(grSetupTable);
			TblMarkRowInvalid(tbl,selectedRow-firstVisible);
			TblRedrawTable(tbl);
			handled = true;
		}		  
		break;
	 case frmCloseEvent:
		handled = false;
		break;
	 case frmOpenEvent:
		formId=frm->formId;
		
		grsetup_tbl_init();

		FrmDrawForm(frm);
		handled=true;
		break;
	}
	  
	return handled;
}
