/*
 *   $Id: memo.c,v 1.5 2000/11/25 15:16:08 ondrap Exp $
 * 
 *   Scientific Calculator for Palm's and PilmPilot's.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
 * 
*/

#include <PalmOS.h>
#include <string.h>

#include "konvert.h"
#include "calcDB.h"
#include "display.h"
#include "memo.h"
#include "stack.h"
#include "calcrsc.h"
#include "calc.h"
#include "prefs.h"
#include "funcs.h"
#include "result.h"

static DmOpenRef MemopadDB;
static UInt16 memoCategory;
static UInt16 *indexList;

/***********************************************************************
 *
 * FUNCTION:     strslashcat
 * 
 * DESCRIPTION:  Do a string 'quote-safe'
 *
 * PARAMETERS:   str1 - destination string
 *               str2 - source string
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
strslashcat(char *str1,char *str2)
{
	str1+=StrLen(str1);
	while ((*str1++=*str2++))
	  if (*str2=='\"' || *str2=='\\')
		*(str1++)='\\';
}

/***********************************************************************
 *
 * FUNCTION:     memo_dump_db
 * 
 * DESCRIPTION:  Create a string dump of an EasyCalc's database
 *               containing definitions of variables and functions
 *
 * PARAMETERS:   maxdump - maximum size of a dump
 *
 * RETURN:       pointer to a newly allocated memo-dump
 *      
 ***********************************************************************/
static char *
memo_dump_db(Int16 maxdump,rpntype type)
{
	char *text;
	dbList *list;
	Int16 i;
	char *dump;
	char parameter[MAX_FUNCNAME+1];
	CError err;
	Trpn item;
	
	dump = MemPtrNew(maxdump);
	StrCopy(dump,"");

	list = db_get_list(type);
	
	/* Dump the functions */
	for (i=0;i<list->size;i++) {
		if (list->type[i]==function) {
			err=db_func_description(list->list[i],&text,parameter);
			if (err)
			  continue;
			if (StrLen(dump)+StrLen(text)+StrLen(list->list[i])+10 > maxdump) {
				MemPtrFree(text);
				break;
			}
			/* Define the parameter */
			StrCat(dump,"defparamn(\"");
			StrCat(dump,parameter);
			StrCat(dump,"\")\n");
			
			StrCat(dump,list->list[i]);
			StrCat(dump,"()=\"");
			strslashcat(dump,text);
			StrCat(dump,"\"\n");
			MemPtrFree(text);
		}		
		else if (list->type[i]==variable) {
			item = db_read_variable(list->list[i],&err);
			if (err)
			  continue;
			/* Do not work with strings now */
			if (item.type == string) {
				rpn_delete(item);
				continue;
			}
			  
			text = display_default(item);
			if (StrLen(dump)+StrLen(text)+StrLen(list->list[i])+5 > maxdump) {
				rpn_delete(item);
				MemPtrFree(text);
				break;
			}
			StrCat(dump,list->list[i]);
			StrCat(dump,"=");
			StrCat(dump,text);
			StrCat(dump,"\n");
			rpn_delete(item);
			MemPtrFree(text);
		}
	}
	StrCat(dump,"defparamn(\"x\")\n");
	db_delete_list(list);
	return dump;
}

/***********************************************************************
 *
 * FUNCTION:     memo_dump
 * 
 * DESCRIPTION:  Dump an EasyCalc db to memopad
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
memo_dump(rpntype type)
{
	static UInt16 index = dmMaxRecordIndex;
	char *text;
	UInt16 mode = dmModeReadWrite;
	MemPtr recP;
	MemHandle recH;
	char *header = "EasyCalc data structures\n";
	UInt16 offset = 0;
	Char null = '\0';
	UInt16 category;
	UInt16 attr;
	
	if((MemopadDB = DmOpenDatabaseByTypeCreator('DATA', 'memo', mode)) == NULL)
	  return;
	
	text = memo_dump_db(MAX_DUMP,type);
	
	recH = DmNewRecord(MemopadDB, &index, StrLen(header)+StrLen(text)+1);
	if(recH != NULL) { 
		recP = MemHandleLock(recH); 
		DmWrite(recP,offset,header,StrLen(header));
		DmWrite(recP,offset + StrLen(header),text,StrLen(text));
		DmWrite(recP,offset + StrLen(header) + StrLen(text),&null,1);		
		MemHandleUnlock(recH);
		
		DmRecordInfo(MemopadDB, index, &attr, NULL, NULL);
		attr &= ~dmRecAttrCategoryMask;
		
		category = CategoryFind(MemopadDB, "EasyCalc");
		if (category == dmAllCategories)
		  attr |= dmUnfiledCategory;
		else
		  attr |= category;		
		
		DmSetRecordInfo(MemopadDB, index, &attr, NULL);
		
		DmReleaseRecord(MemopadDB,index,true);		
	}
	
	DmCloseDatabase(MemopadDB);		
	MemPtrFree(text);
}

static void
memo_draw_item(Int16 itemNum,RectangleType *bounds,
			  Char **itemText)
{
	MemHandle recH;
	char *text;
	Int16 titlelen,i;
	
	recH = DmQueryRecord(MemopadDB,indexList[itemNum]);
	text = MemHandleLock(recH);
	
	for (i=0,titlelen=0;text[i]!='\0' && text[i]!='\n';i++,titlelen++)
	  ;
	WinDrawChars(text,titlelen,bounds->topLeft.x,bounds->topLeft.y);
	
	MemHandleUnlock(recH);
}

static Boolean
memo_import_memo(UInt16 itemNum)
{
	MemHandle recH;
	char *text;
	char *input;
	Int16 i,j,memolen;
	CodeStack *stack;
	CError err;
	Trpn result;
	
	
	recH = DmQueryRecord(MemopadDB,indexList[itemNum]);
	memolen = MemHandleSize(recH);
	input = MemPtrNew(memolen+1);
	text = MemHandleLock(recH);
	memcpy(input,text,memolen);
	input[memolen]='\0';
	MemHandleUnlock(recH);
	
	/* Skip the first line */
	for (i=0;input[i]!='\n' && i<memolen;i++)
	  ;
	i++;
	while (i<memolen) {
		for (j=i;input[j]!='\n' && j<memolen;j++)
		  ;
		input[j] = '\0';
		
		if (StrLen(input+i) == 0)
		  break;
		stack = text_to_stack(input+i,&err);
		if (!err) {
			err = stack_compute(stack);
			if (!err) {
				result = stack_pop(stack);
				set_ans_var(result);
				result_set(result);
				rpn_delete(result);
			}
			stack_delete(stack);
        }		
		if (err) {
			MemPtrFree(input);
			return false;
		}
		i=j+1;
	}
	
	MemPtrFree(input);
	return true;
}

static void
memo_init_list(void)
{
	UInt16 mode = dmModeReadOnly;
	ListPtr list;
	UInt16 index,i;
	UInt16 numRecords;
	MemHandle recH;
	
	MemopadDB = DmOpenDatabaseByTypeCreator('DATA', 'memo', mode);
	ErrFatalDisplayIf(MemopadDB==NULL,"Error opening MemoDB");
	
	memoCategory = CategoryFind(MemopadDB, "EasyCalc");
	numRecords = DmNumRecordsInCategory(MemopadDB,memoCategory);
	
	list = GetObjectPtr(memoImportList);	
	
	LstSetListChoices(list,NULL,numRecords);
	LstSetDrawFunction(list,memo_draw_item);

	/* Allocate at least 1 byte to make later the MemPtrFree happy */
	indexList = MemPtrNew(sizeof(*indexList)*numRecords + 1);

	index = 0;
	i=0;
	while ((recH = DmQueryNextInCategory(MemopadDB,&index,memoCategory))) {
		indexList[i] = index;
		i++;
		index++;
	}
}

static void
memo_destroy_list(void)
{
	DmCloseDatabase(MemopadDB);
	MemPtrFree(indexList);
}

/***********************************************************************
 *
 * FUNCTION:     MemoImportHandleEvent
 * 
 * DESCRIPTION:  Event handler for memopad import dialog
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean
MemoImportHandleEvent(EventPtr event)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 controlID;
	Int16 i;

	switch (event->eType) {
	 case ctlSelectEvent:
		controlID = event->data.ctlSelect.controlID;
		if (controlID == memoImportCancel) {
			memo_destroy_list();
			FrmReturnToForm(0);
			handled = true;
			break;
		}
		else if (controlID == memoImportOK) {
			i = LstGetSelection(GetObjectPtr(memoImportList));
			if (i != noListSelection) {
				if (!memo_import_memo(i)) 
				  FrmAlert(memoImportError);
				memo_destroy_list();
				FrmReturnToForm(0);
				FrmUpdateForm(calcPrefs.form,frmUpdateResult);
			}
			handled = true;
			break;
		}
		break;
	 case frmOpenEvent:
		frm = FrmGetActiveForm();
		
		memo_init_list();
		
		FrmDrawForm(frm);
		
		handled = true;
		break;		
	 case frmCloseEvent:
		memo_destroy_list();
		/* FrmHandleEvent must handle this event itself */
		handled = false;
		break;
	}
	
	return handled;
}
