/*
 *   $Id: mathem.c,v 1.3 2000/11/24 10:32:24 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
 *
 *  2000-10-11 - Rafael R. Sevilla <dido@pacific.net.ph>
 *               Added code to compute the Euler gamma function and modified
 *               the factorial code to use it for non-integral arguments.
 *  2000-10-13 - Rafael R. Sevilla <dido@pacific.net.ph>
 *               Added hooks for complex hyperbolic functions, and complex
 *               inverse circular and hyperbolic functions.
 */

#include <PalmOS.h>

#include "MathLib.h"
#include "konvert.h"
#include "funcs.h"
#include "mathem.h"
#include "stack.h"
#include "defuns.h"
#include "prefs.h"
#include "complex.h"
#include "display.h"

#define MAX_FACT 33

/***********************************************************************
 *
 * FUNCTION:     math_bin
 * 
 * DESCRIPTION:  Binary aritmetics
 *
 * PARAMETERS:   On stack - 2 arguments, convertible to integer
 *
 * RETURN:       On stack - 1 result
 *      
 ***********************************************************************/
CError 
math_bin(Int16 funcnum,CodeStack *stack)
{
	UInt32 arg1,arg2;
	UInt32 result;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,integer)))
	  return err;	
	switch (funcnum) {
	 case FUNC_AND:
		result=arg1 & arg2;
		break;
	 case FUNC_OR:
		result=arg1 | arg2;
		break;
	 case FUNC_XOR: 
		result=arg1 ^ arg2;
		break;
	 case FUNC_SHL:
		result=arg1 << arg2;
		break;
	 case FUNC_SHR:
		result=arg1 >> arg2;
		break; 
	 default:
		return c_internal;
	}
	stack_add_val(stack,&result,integer);
	return c_noerror;		
}

/***********************************************************************
 *
 * FUNCTION:     int_oper
 * 
 * DESCRIPTION:  Integer operations
 *
 * PARAMETERS:   On stack - 2 arguments, convertible to integer
 *
 * RETURN:       On stack - 1 type integer
 *      
 ***********************************************************************/
static CError 
int_oper(Int16 funcnum,CodeStack *stack) MLIB;
static CError 
int_oper(Int16 funcnum,CodeStack *stack)
{
	UInt32 arg1,arg2;
	UInt32 tmp;
	UInt32 result;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,integer)))
	  return err;
	switch (funcnum) {
	 case FUNC_PLUS:
		result=arg1+arg2;
		break;
	 case FUNC_MINUS:
		result=arg1-arg2;
		break;
	 case FUNC_DIVIDE:
		if (!arg2)
		  return c_divbyzero;
		result=arg1/arg2;
		break;
	 case FUNC_MOD:
		if (!arg2)
		  return c_divbyzero;
		result = arg1 % arg2;
		break;
	 case FUNC_MULTIPLY:
		result=arg1*arg2;
		break;
	 case FUNC_POWER:
		if (arg2==0) {
			result=1;
			break;
		}
		if (arg2>300) /* The user might have entered a negative number */
		  return c_badarg;
		
		tmp=arg1;
		while (--arg2)
		  tmp*=arg1;
		result=tmp;
		break;
	 default:
		return c_badfunc;
	}       
	stack_add_val(stack,&result,integer);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     real_oper
 * 
 * DESCRIPTION:  Operations on real numbers
 *
 * PARAMETERS:   On stack - 2 numbers convertible to real
 *
 * RETURN:       On stack - 1 real number
 *      
 ***********************************************************************/
static CError
real_oper(Int16 funcnum,CodeStack *stack) MLIB;
static CError
real_oper(Int16 funcnum,CodeStack *stack)
{
	double arg1,arg2;
	double result;
	CError err;

	if ((err=stack_get_val2(stack,&arg1,&arg2,real)))
	  return err;
	switch (funcnum) {
	 case FUNC_PLUS:
		result=arg1+arg2;
		break;
	 case FUNC_MINUS:
		result=arg1-arg2;
		break;
	 case FUNC_DIVIDE:
		if (arg2==0)
		  return c_divbyzero;
		result=arg1/arg2;
		break;
	 case FUNC_MOD:
		if (arg2==0)
		  return c_divbyzero;
		result =fmod(arg1,arg2);
		break;
	 case FUNC_MULTIPLY:
		result=arg1*arg2;
		break;
	 case FUNC_POWER:
	   /* Raise a negative number to a nonintegral power to fall to
	      complex math. */
	   if (arg1 < 0.0 && (!IS_ZERO(arg2-round(arg2)))) {
	     stack_add_val(stack, &arg2, real);
	     stack_add_val(stack, &arg1, real);
	     return(cplx_math(funcnum, stack));
	   }

	   result=pow(arg1,arg2);
	   break;
	 default:
		return c_badfunc;
	}       
	stack_add_val(stack,&result,real);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     math_oper
 * 
 * DESCRIPTION:  Generic function, that selects using some criteria
 *               which function (cplx_oper,real_oper,int_oper) 
 *               should perform the requested operation
 * 
 * NOTE:         Doesn't modify the stack
 * 
 * PARAMETERS:   Detects type of 2 arguments
 *
 * RETURN:       Adds 1 object on stack
 *      
 ***********************************************************************/
CError 
math_oper(Int16 funcnum,CodeStack *stack)
{
	CError err;
	rpntype type1, type2;
	
	err = stack_item_type(stack,&type1,1);
	if (err)
	  return err;
	err = stack_item_type(stack,&type2,0);
	if (err)
	  return err;

	/* if both arguments are integer, the user probably wanted 
	 * integer operations */
	if (dispPrefs.forceInteger || (type1==integer && type2==integer))
	  return int_oper(funcnum,stack);
	else if (dispPrefs.forceComplex || type1==complex || type2==complex)
	  return cplx_oper(funcnum,stack);
	else 
	  return real_oper(funcnum,stack);
}

/* Negate function, uses appropriate functions for types
 * Input - 1 real/integer, Output - 1 real/integer */
/***********************************************************************
 *
 * FUNCTION:     math_neg
 * 
 * DESCRIPTION:  Negates a number, preserving a type of argument
 *
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError 
math_neg(Int16 funcnum,CodeStack *stack)
{
	CError err=0;
	rpntype type;
	
	err=stack_item_type(stack,&type,0);
	if (err)
	  return err;
	
	if (funcnum!=FUNC_NEGATE)
	  return c_internal;
	if (type==real) {
		double darg;
		
		err=stack_get_val(stack,&darg,real);
		if (err)
		  return err;
		darg = -darg;
		return stack_add_val(stack,&darg,real);
	}
	else if (type==integer) {
		UInt32 iarg;
		
		err=stack_get_val(stack,&iarg,integer);
		if (err) 
		  return err;
		iarg = -iarg;
		return stack_add_val(stack,&iarg,integer);		
	}
	else if (type==complex) {
		Complex carg;
		
		err = stack_get_val(stack,&carg,complex);
		if (err)
		  return err;
		carg.real = -carg.real;
		carg.imag = -carg.imag;
		return stack_add_val(stack,&carg,complex);
	}
	else
	  return c_badarg;
}

/* Simulation of logb(), as the MathLib's one returns bad result */
static double logbin(double arg) MLIB;
static double
logbin(double arg) {
	return log(arg)/log(2);
}

static double toreal(double arg) MLIB;
static double toreal(double arg)
{
	return arg;
}					 

static double fpart(double arg) MLIB;
static double
fpart(double arg) {
	return arg-trunc(arg);
}
	  

/* Read the const names in mathem.h, they poInt16 in this array */
struct {
	double (*func)(double);
}math_func[]={
	{sin},{cos},{tan},{sinh},{cosh},{tanh},
	{asin},{acos},{atan},{asinh},{acosh},{atanh},
	{exp},{log},{log10},{sqrt},{cbrt},{fabs},{logbin},
	{trunc},{round},{fpart},{toreal}
	
};

/***********************************************************************
 *
 * FUNCTION:     math_arctrigo
 * 
 * DESCRIPTION:  Perform computation of functions, that do 
 *               degree/radian/grad conversion on output
 *               asin,acos,atan,asinh,acosh,atanh
 *
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError
math_arctrigo(Int16 funcnum,CodeStack *stack)
{
	double arg,res;
	CError err;
	rpntype type;

	err = stack_item_type(stack,&type,0);
	if (err)
	  return err;

	/* complex is defined for these operations */
	if ((funcnum==MATH_ASIN || funcnum==MATH_ACOS ||
	     funcnum==MATH_ATAN ||
	     funcnum==MATH_ASINH || funcnum==MATH_ACOSH ||
	     funcnum==MATH_ATANH) &&
	    (type==complex || dispPrefs.forceComplex))
	  return  cplx_math(funcnum,stack);

	err=stack_get_val(stack,&arg,real);
	if (err)
	  return err;

	/* Fall to complex results for asin and acos if argument is
	   abs(arg) is greater than 1. */
	if (fabs(arg) > 1.0 && (funcnum == MATH_ASIN || funcnum == MATH_ACOS)) {
	  stack_add_val(stack, &arg, real);
	  return(cplx_math(funcnum, stack));
	}
	
	res=math_func[funcnum].func(arg);
	if (calcPrefs.trigo_mode==degree)
	  res=(360/(2*M_PIl))*res;
	else if (calcPrefs.trigo_mode==grad)
	  res=(400/(2*M_PIl))*res;
	
	stack_add_val(stack,&res,real);
	return c_noerror;
}

#define GAMMPARM 5
#define SQRT2PI 2.5066282746310005024

/* Euler Gamma function.  Uses Lanczos' approximation as described
   in Numerical Recipes.  Applies the gamma function reflection formula for
   arguments less than 1. */
double euler_gamma(double z)
{
  double x, tmp, ser, gam;
  int j;
  static double gammacoeffs[] = { 76.18009173, -86.50532033, 24.01409822,
				  -1.231739516, 0.120858003e-2, -0.536382e-5 };

  x = (z < 1) ? (1-z) : z;
  x--;
  tmp = x + GAMMPARM + 0.5;
  tmp = pow(tmp, x+0.5)*exp(-tmp);
  ser = 1.0;
  for (j=0; j<6; j++) {
    x++;
    ser += gammacoeffs[j]/x;
  }
  gam = tmp*ser*SQRT2PI;
  /* apply the reflection formula if needed */
  if (z < 1)
    gam = M_PIl/(gam*sin(M_PIl*(1-z)));
  return(gam);
}

/***********************************************************************
 *
 * FUNCTION:     math_gamma
 * 
 * DESCRIPTION:  Euler gamma function
 *
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 * 
 ***********************************************************************/
CError math_gamma(Int16 funcnum, CodeStack *stack)
{
	double arg;
	double res;	
	CError err;

	err=stack_get_val(stack,&arg,real);
	if (err)
	  return err;
 	/* return error if arg > MAX_FACT or if the argument is a negative
	   integer. */
	if (arg>MAX_FACT)
	  return c_range;
	if (IS_ZERO(arg-round(arg)) && arg <= 0)
	  return c_badarg;	/* negative int or too big */
	res = euler_gamma(arg);
	stack_add_val(stack,&res,real);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     math_fact
 * 
 * DESCRIPTION:  Factorial
 *               This now uses the Euler gamma function if non integral
 *               arguments are passed.  Error only if the argument is a
 *               negative number.
 *
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError math_fact(Int16 funcnum,CodeStack *stack)
{	
	double arg;
	double res=1.0;	
	CError err;
		
	err=stack_get_val(stack,&arg,real);
	if (err) 
	  return err;
	
 	/* return error if arg > MAX_FACT or if the argument is a negative
	   integer. */
	if (arg>MAX_FACT)
	  return c_range;
	if (IS_ZERO(arg-round(arg)) && arg < 0)
	  return c_badarg;	/* negative factorial or too big */

	if (!IS_ZERO(arg-round(arg))) {
	        res=euler_gamma(arg+1.0);
		stack_add_val(stack,&res,real);
		return c_noerror;
	}
	
	/* go back to basics and use the usual computations for integral
	   arguments.  */
	while (arg>0.0) {
		res*=arg;
		arg = arg - 1.0;
	}
	
	stack_add_val(stack,&res,real);
	return c_noerror;
}


/***********************************************************************
 *
 * FUNCTION:     math_trigo
 * 
 * DESCRIPTION:  Perform computation of functions, that do 
 *               degree/radian/grad conversion on input
 *               sin,cos,tan,sinh,cosh,tanh
 *
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError 
math_trigo(Int16 funcnum,CodeStack *stack)
{
	double arg,res;
	CError err;
	rpntype type;
	
	err = stack_item_type(stack,&type,0);
	if (err)
	  return err;

	/* complex is defined for these operations */
	if ((funcnum==MATH_SIN || funcnum==MATH_COS ||
	     funcnum==MATH_TAN ||
	     funcnum==MATH_SINH || funcnum==MATH_COSH ||
	     funcnum==MATH_TANH) &&
	    (type==complex || dispPrefs.forceComplex))
	  return  cplx_math(funcnum,stack);

	err=stack_get_val(stack,&arg,real);
	if (err)
	  return err;
	
	if (calcPrefs.trigo_mode==degree)
	  arg=(2*M_PIl/360)*arg;
	else if (calcPrefs.trigo_mode==grad)
	  arg=(2*M_PIl/400)*arg;

	res=math_func[funcnum].func(arg);

	stack_add_val(stack,&res,real);
	return c_noerror;
}

/***********************************************************************
 *
 * FUNCTION:     math_math
 * 
 * DESCRIPTION:  Perform computation of functions, that do not
 *               need any input/output conversion
 *              
 * PARAMETERS:   On stack - 1 number
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError
math_math(Int16 funcnum,CodeStack *stack)
{
	double arg,res;
	CError err;
	rpntype type;
	
	err = stack_item_type(stack,&type,0);
	if (err)
	  return err;
	
	/* complex is defined for these operations */
	if ((funcnum==MATH_ABS || funcnum==MATH_EXP ||
		funcnum==MATH_SQRT || funcnum==MATH_LN || funcnum==MATH_LOG || 
		funcnum==MATH_LOG2) &&
		(type==complex || dispPrefs.forceComplex))
	  return  cplx_math(funcnum,stack);

	err=stack_get_val(stack,&arg,real);
	if (err)
	  return err;

	if (arg < 0.0 && (funcnum==MATH_SQRT || funcnum == MATH_LN ||
			  funcnum==MATH_LOG || funcnum == MATH_LOG2)) {
	  stack_add_val(stack, &arg, real);
	  return(cplx_math(funcnum, stack));
	}
	res=math_func[funcnum].func(arg);
	stack_add_val(stack,&res,real);
	return c_noerror;
}

/* (int) functions - returns value with rpntype=integer*/
CError 
math_int(Int16 funcnum,CodeStack *stack)
{
	UInt32 arg;
	CError err;
	
	err=stack_get_val(stack,&arg,integer);
	if (err)
	  return err;
	stack_add_val(stack,&arg,integer);
	
	return c_noerror;
}

/* Find a point, wher f(x)=0. This function works
 * with a precision of F_IS_ZERO
 */

CError
math_fsolve(Int16 funcnum,CodeStack *stack)
{
	double min,max;
	double resmin,resmax,res;
	double tmp;
	CodeStack *fcstack,*origstack;
	Trpn funcarg;
	CError err;
	Trpn tmp1;
	
	tmp1 = stack_pop(stack);
	if (tmp1.type==funcempty) {
		origstack = db_read_function(tmp1.u.funcname,&err);
		rpn_delete(tmp1);
		if (err)
		  return err;
	}
	else if (tmp1.type==string) {
		origstack = text_to_stack(tmp1.u.stringval,&err);
		rpn_delete(tmp1);
		if (err)
		  return err;
	}
	else {
		rpn_delete(tmp1);
		return c_badarg;
	}
	
	if ((err=stack_get_val2(stack,&min,&max,real))) {
		stack_delete(origstack);
		return err;
	}
 
	fcstack = stack_new(origstack->size);
	funcarg.type = real;
	funcarg.allocsize = 0;	
	
/*	if (max<min)
	  return c_badarg;
*/	
	
	set_func_argument(&funcarg);	
	funcarg.u.realval = min;
	stack_copy(fcstack,origstack);
	err = stack_compute(fcstack);
	decr_func_argument();

	if (err) {
		stack_delete(fcstack);stack_delete(origstack);
		return c_compimp;
	}
	err=stack_get_val(fcstack,&resmin,real);
	if (err) {
		stack_delete(fcstack);stack_delete(origstack);
		return err;
	}
		
	funcarg.u.realval = max;
	set_func_argument(&funcarg);
	stack_copy(fcstack,origstack);
	err = stack_compute(fcstack);
	decr_func_argument();

	if (err) {		
		stack_delete(fcstack);stack_delete(origstack);
		return c_compimp;
	}
	
	err=stack_get_val(fcstack,&resmax,real);
	if (err) {
		stack_delete(fcstack);stack_delete(origstack);
		return err;
	}
	
	if (resmin*resmax>0.0) {
		stack_delete(fcstack);stack_delete(origstack);
		return c_compimp;
	}
	
	if (resmin==0.0)
	  max=tmp=min;
	else if (resmax==0.0)
	  min=tmp=max;

	while (!F_IS_ZERO(max-min)) { 
		/* Choose half */
		tmp = (max+min)/2.0;
		
		funcarg.u.realval = tmp;
		set_func_argument(&funcarg);
		stack_copy(fcstack,origstack);
		err=stack_compute(fcstack);
		decr_func_argument();
		if (err) {
			stack_delete(fcstack);stack_delete(origstack);
			return err;
		}
		err=stack_get_val(fcstack,&res,real);
		if (err) {
			stack_delete(fcstack);stack_delete(origstack);
			return err;
		}
		
		if (!finite(res)) {
			stack_delete(fcstack);stack_delete(origstack);
			return c_compimp;
		}

		if (res*resmin<0.0) {
			resmax = res;
			max = tmp;
		}
		else if (res*resmax<0.0){
			resmin = res;
			min = tmp;
		}
		else 
		  break;
	}
	
	/* It's possible we didn't compute a 'zero' but something else,
	 * if the result isn't near 0 return error */
	if (fabs(res)>1.0)
	  return c_compimp;

	stack_add_val(stack,&tmp,real);
	stack_delete(fcstack);stack_delete(origstack);
	return c_noerror;
}

 
CError
math_convcoord(Int16 funcnum,CodeStack *stack)
{
	double arg1,arg2;
	double res;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,real)))
	  return err;
	
	if (funcnum==MATH_PTORX || funcnum==MATH_PTORY) {
		if (calcPrefs.trigo_mode==degree)
		  arg2=(2*M_PIl/360)*arg2;
		else if (calcPrefs.trigo_mode==grad)
		  arg2=(2*M_PIl/400)*arg2;	
	}

	switch (funcnum) {
	 case MATH_RTOPR:
//		res = sqrt(arg1*arg1+arg2*arg2);
		res = hypot(arg1,arg2);
		break;
	 case MATH_RTOPD:
		if (arg1==0.0 && arg2==0.0) {
			res = 0.0;
			break;
		}
		res = atan(arg2/arg1);		
		if (calcPrefs.trigo_mode==degree)
		  res=(360/(2*M_PIl))*res;
		else if (calcPrefs.trigo_mode==grad)
		  res=(400/(2*M_PIl))*res;
		break;
	 case MATH_PTORX:
		res = cos(arg2)*arg1;
		break;
	 case MATH_PTORY:
		res = sin(arg2)*arg1;
		break;
	}
	stack_add_val(stack,&res,real);
	
	return c_noerror;	
}

/***********************************************************************
 *
 * FUNCTION:     math_comb
 * 
 * DESCRIPTION:  Compute ncr(combination) and npr(permutation)
 *
 * PARAMETERS:   On stack 2 arguments
 *
 * RETURN:       On stack 1 argument
 *      
 ***********************************************************************/
CError
math_comb(Int16 funcnum,CodeStack *stack)
{
	UInt32 arg1,arg2;
	UInt32 i;
	double res=1.0;
	CError err;
	
	if ((err=stack_get_val2(stack,&arg1,&arg2,integer)))
	  return err;
	
	if (arg2>arg1)
	  return c_badarg;
	if (arg1-arg2>MAX_FACT && arg2>MAX_FACT)
	  return c_range;
	
	for (i=arg1;i>arg1-arg2;i--)
	  res*=i;
	
	if (funcnum==MATH_COMB) {
		for (i=1;i<=arg2;i++)
		  res/=(double)i;
	}
	
	stack_add_val(stack,&res,real);
	
	return c_noerror;
}
