/*
 *   $Id: specfun.c,v 1.1.1.1 2000/10/27 17:16:57 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PalmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *   This file Copyright (C) 2000 Rafael R. Sevilla
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact Ondrej at 'ondrap@penguin.cz'.
 *  You can contact me at 'dido@pacific.net.ph'.
 *
 */

#include <PalmOS.h>
#include "MathLib.h"
#include "konvert.h"
#include "funcs.h"
#include "mathem.h"
#include "stack.h"
#include "prefs.h"
#include "display.h"
#include "specfun.h"

extern double euler_gamma(double z);

#define MAX_ITER 200

/***********************************************************************
 *
 * FUNCTION:     gser
 * 
 * DESCRIPTION:  Compute incomplete gamma function, power series
 *
 * PARAMETERS:   a, x
 *
 * RETURN:       P(a,x)
 *      
 ***********************************************************************/
static double gser(double a, double x)
{
  double gln, ap, sum, del;
  int n;

  gln = log(euler_gamma(a));
  if (x == 0)
    return(0);
  ap = a;
  sum = 1.0/a;
  del = sum;
  for (n=0; n<MAX_ITER; n++) {
    ap++;
    del*=x/ap;
    sum+=del;
    if (fabs(del < fabs(sum)*EPSILON))
      break;
  }
  /* Screw it if the series failed to converge after the maximum number
     of iterations.  We need to determine a way of signalling to the user
     that the iteration failed to converge--next version maybe. */
  return(sum*exp(-x + a*log(x) - gln));
}

/***********************************************************************
 *
 * FUNCTION:     gcf
 * 
 * DESCRIPTION:  Compute complementary incomplete gamma function,
 *               continued fraction rep.
 *
 * PARAMETERS:   a, x
 *
 * RETURN:       Q(a,x)
 *      
 ***********************************************************************/
double gcf(double a, double x)
{
  double gln, g_old, g, a0, a1, b0, b1, fac, ana, anf;
  int n;

  gln = log(euler_gamma(a));
  g_old = 0;
  a0 = 1;
  a1 = x;
  b0 = 0;
  b1 = 1;
  fac = 1;
  g = 1;
  for (n=1; n<=MAX_ITER; n++) {
    ana = (double)n - a;
    a0 = (a1 + a0*ana)*fac;
    b0 = (b1 + b0*ana)*fac;
    anf = n*fac;
    a1 = x*a0+anf*a1;
    b1 = x*b0+anf*b1;
    if (a1 != 0.0) {
      fac = 1.0/a1;
      g = b1*fac;
      if (fabs((g-g_old)/g) < EPSILON)
	break;
      g_old = g;
    }
  }
  return(exp(-x + a*log(x) - gln)*g);
}

/***********************************************************************
 *
 * FUNCTION:     incgamma
 * 
 * DESCRIPTION:  Compute incomplete gamma function
 *
 * PARAMETERS:   a, x
 *
 * RETURN:       P(a,x)
 *      
 ***********************************************************************/
static double incgamma(double a, double x) SPECFUN;
static double incgamma(double a, double x)
{
  if (x < a+1)
    return(gser(a, x));
  return(1.0-gcf(a, x));
}

/***********************************************************************
 *
 * FUNCTION:     erf
 * 
 * DESCRIPTION:  Compute error function
 *
 * PARAMETERS:   x
 *
 * RETURN:       erf(x)
 *      
 ***********************************************************************/
static double erf(double x)
{
  if (x < 0)
    return(-incgamma(0.5, x*x));
  return(incgamma(0.5, x*x));
}

/***********************************************************************
 *
 * FUNCTION:     betacf
 * 
 * DESCRIPTION:  Compute incomplete beta function, continued fraction
 *
 * PARAMETERS:   a, b, x
 *
 * RETURN:       ibeta(a, b, x)
 *      
 ***********************************************************************/
static double betacf(double a, double b, double x) SPECFUN;
static double betacf(double a, double b, double x)
{
  double am, bm, az, qab, qap, qam, bz, em, tem, ap, bp, d, app, bpp, aold;
  int m;

  am = bm = az = 1;
  qab = a+b;
  qap = a+1;
  qam = a-1;
  bz = 1-qab*x/qap;
  for (m=1; m<=MAX_ITER; m++) {
    em = m;
    tem = em + em;
    d = em*(b-m)*x/((qam+tem)*(a+tem));
    ap = az + d*am;
    bp = bz + d*bm;
    d = -(a+em)*(qab+em)*x/((a+tem)*(qap+tem));
    app = ap + d*az;
    bpp = bp + d*bz;
    aold = az;
    am = ap/bpp;
    bm = bp/bpp;
    az = app/bpp;
    bz = 1.0;
    if (fabs(az - aold) < EPSILON*fabs(az))
      break;
  }
  return(az);
}

/***********************************************************************
 *
 * FUNCTION:     ibeta
 * 
 * DESCRIPTION:  Compute incomplete beta function
 *
 * PARAMETERS:   a, b, x
 *
 * RETURN:       ibeta(a, b, x)
 *      
 ***********************************************************************/
static double incbeta(double a, double b, double x) SPECFUN;
static double incbeta(double a, double b, double x)
{
  double bt;

  bt = (x == 0 || x == 1) ? 0 :
    exp(log(euler_gamma(a+b)) - log(euler_gamma(a)) -
	    log(euler_gamma(b)) + a*log(x) + b*log(1-x));
  if (x < (a+1)/(a+b+2))		/* use continued fraction directly */
    return(bt*betacf(a, b, x)/a);
  /* use continued fraction after making symmetry transform */
  return(1-bt*betacf(b, a, 1-x)/b);
}

/***********************************************************************
 *
 * FUNCTION:     math_gambeta
 * 
 * DESCRIPTION:  Compute Gamma, Beta, and related functions
 *
 * PARAMETERS:   On stack - up to three numbers, depending on function
 *
 * RETURN:       On stack - 1 number
 *      
 ***********************************************************************/
CError
math_gambeta(Int16 funcnum, CodeStack *stack)
{
  double z, w, v, res;
  CError err;

  err = stack_get_val(stack, &z, real);
  if (err)
    return(err);

  switch (funcnum) {
  case MATH_BETA:
    err=stack_get_val(stack, &w, real);
    if (err)
      return(err);
    if (z < 0 || w < 0)
      return(c_badarg);
    v = z+w;
    res = exp(log(euler_gamma(z)) + log(euler_gamma(w)) -
	      log(euler_gamma(z+w)));
    break;
  case MATH_IGAMMA:
    err=stack_get_val(stack, &w, real);
    if (err)
      return(err);
    if (z < 0 || w <= 0)
      return(c_badarg);
    res = incgamma(w, z);
    break;
  case MATH_ERF:
    res = erf(z);
    break;
  case MATH_ERFC:
    res = 1 - erf(z);
    break;
  case MATH_IBETA:
    err = stack_get_val2(stack, &w, &v, real);
    if (err)
      return(err);
    if (z < 0 || z > 1 || w <= 0 || v <= 0)
      return(c_badarg);
    res = incbeta(w, v, z);
    break;
  default:
    return(c_internal);
  }
  stack_add_val(stack, &res, real);
  return(c_noerror);
}

/***********************************************************************
 *
 * FUNCTION:     polevl
 * 
 * DESCRIPTION:  Evaluate a polynomial given an array of its coefficients

 * PARAMETERS:   x, coeffs, degree N
 *
 * RETURN:       value of polynomial at x
 *      
 ***********************************************************************/
double polevl(double x, double *coef, int N )
{
  double p;
  int i;

  p = coef[N-1];
  for (i=N-2; i>=0; i--)
    p = p*x + coef[i];
  return(p);
}

