/*
 * Grouper v1.4
 *
 * Grouper--Card Puzzle/Game
 * Copyright  1998, 1999 Daniel McCarty
 *
 * (for legal issues, see below)
 * (for revisions, see history.txt)
 *
 * Version History
 * 1.4  + added undo feature
 *      + modified code and structure to allow localizations
 *      + localized German version
 *      + localized French version
 *      + localized Spanish version
 * 1.3  + fixed tap-and-drag-away card invert bug
 *      + added feature to popup preferences when a new game begins
 * 1.2  + fixed fast-tapping during card dealing bug
 *      + added different shaped card marks
 *      + fixed crash bug: handling penUp from menu events like cards
 *
 * 1.0  + initial release (see history.txt, readme.txt)
 *
 *
 * Grouper 1.4 public code release: 99.02.22
 *
 * Disclaimers and other legal stuff:
 *  This code is provided AS-IS; there is NO WARRANTY, express or implied.
 *  The user assumes FULL RESPONSIBILITY AND LIABILITY for any use of
 *  this code.
 *
 *  Copyright (c)1998 Daniel McCarty
 *  All Rights Reserved by Daniel McCarty, except as noted below:
 *
 *  Permission is hereby granted to the public, free of charge, to
 *  distribute this package complete and unmodified, without restriction.
 *
 *  Modified versions may be distributed with this notice intact, provided
 *  that said versions are clearly marked as modified code.  The above
 *  provisions shall also apply to any such modified packages.
 *
 * NOTE: These terms are intended to protect _me_, not make this package any
 *       less useful to you. If you have a special need or use for this code,
 *       please contact me and I'm sure we can work something out.
 *
 *       Warmest regards,
 *       Daniel.
 *       mcdan@csi.com
 */

#include <Pilot.h>
#include <DLServer.h>						// for getting user-name

#include "callback.h"
#include "grouper.h"


#define AppID			 'dmGR'				// registered with Palm/3Com
#define AppVer			 1

#define NumPicks		 3						// pick per group
#define NumProps		 4						// properties per card
#define NumGrid		 12					// number of spots on the grid

#define NumGameSounds		10
#define MaxNotesPerSound	4

#if EMULATION_LEVEL == EMULATION_NONE
	#define DifficultyFactor 27			// number of cards per level
#else
	#define DifficultyFactor 12			// makes debugging easier
#endif

#if EMULATION_LEVEL == EMULATION_NONE
	#define TextDelay    2
	#define FrameDelay   10					// assuming 10 frames/sec
	#define FramesPerSec 6
#else
	#define TextDelay		1					// since emulation is so slow,
	#define FrameDelay	1					// "remove" these wait times
	#define FramesPerSec	1
#endif	// EMULATION_LEVEL == EMULATION_NONE

// status location definitions
#define StatusX	1
#define StatusY	146
#define StatusDX	108
#define StatusDY	12

// high score location definitions
#define MaxHighScores				5
#define MaxNameLength				9
#define FirstHighScoreY				28
#define HighScoreHeight				12
#define HighScoreNameColumnX		15
#define HighScoreLevelColumnX		84		// right-aligned
#define HighScoreScoreColumnX		123	// right-aligned
#define HighScoreWinColumnX		127
#define HighScoreFlagpoleX			4
#define HighScoreFlagpoleY			19
#define HighScoreFlagpoleDY		54
#define HighScoreFlagpoleDX		2

// scoring definitions
#define ScoreUndo						-7		// -points for an undo
#define ScoreHint						-2		// -points for a hint
#define ScoreGroup					5		// points for a group
#define ScoreWin						15		// points for winning the game
#define ScoreX							141
#define ScoreY							1

// misc. location/size definitions
#define AboutDeckX				63
#define AboutDeckY				49
#define PrefsSampleX				48
#define PrefsSampleY				72
#define ListPatternSize			8
#define NumAnimationBitmaps	4

// misc. defines
#define DialogButton0		0
#define DialogButton1		1
#define NoCard					-1				// grid spot doesn't have a card
#define MenuHandyman			0				// menu index for pulldown and item
#define UpdateRedrawMask	0x1
#define UpdateNewGameMask	0x2

// this is how we cram 81 styles into 1 byte
// using a bit array is possible (see card.ba and grouper.ba), but isn't
// cost effective at this time; so for now we'll stick with shifts and macros
#define StyleColorMask		0x0003
#define StyleMarksMask		0x000c
#define StyleMarkColorMask	0x0030
#define StyleMarkTypeMask	0x00c0

// deck macros
#define CardsInDeck					81
#define DeckFullIndex				-1
#define DeckFull(deck)				(deck.index == DeckFullIndex)
#define DeckEmpty(deck)				(deck.index == deck.numCards - 1)
#define SetDeckFull(deck)			(deck.index = DeckFullIndex)
#define SetDeckEmpty(deck)			(deck.index = deck.numCards - 1)
#define CardsLeft(deck)				(deck.numCards - deck.index - 1)
// same macros as above, but for pointers--is there a better way?
#define DeckFullP(deckP)			(deckP->index == DeckFullIndex)
#define DeckEmptyP(deckP)			(deckP->index == deckP->numCards - 1)
#define SetDeckFullP(deckP)		(deck->index = DeckFullIndex)
#define CardsLeftP(deckP)			(deckP->numCards - deckP->index - 1)

// card macros
#define GetStyleMarkType(card)	((card & StyleMarkTypeMask) >> 6)
											// 2 hi bits in the style byte
#define GetStyleMarkColor(card)	((card & StyleMarkColorMask) >> 4)
											// next two bits
#define GetStyleMarks(card)		(((card & StyleMarksMask) >> 2) + 1)
#define GetStyleColor(card)		(card & StyleColorMask)
											// 2 lo bits in the style byte

//  .------------.
//  | STRUCTURES |
//  '------------'
typedef enum {
	modeStart = 0,
	modeHint,
	modeButWhy,
	modeEnd
} ModeType;

typedef enum {
	soundShuffle = 0,
	soundDeal,
	soundSelect,
	soundUnselect,
	soundNotAGroup,
	soundGrouper,
	soundGameOver,
	soundYouWin,
	soundNewHighScore,
	soundError
} GameSoundsType;

typedef enum {
	msgClear = 0,								// clear current message
	msgWelcome,									// Welcome to Grouper!
	msgNotAGroup,								// Not a group
	msgYouWin,									// You Win!!!
	msgGameOver,								// No groups: Game Over
	msgGrouper									// Grouper!
} MessageType;

typedef struct {
	Long frequency;							// we don't need an amplitude value
	UInt duration;								// because of the Pilot's weak speaker
} NoteType;

typedef struct {
	UInt numNotes;
	NoteType notes[MaxNotesPerSound];
} SoundType;

typedef struct {
	int grid[NumPicks];
	int count;
} SelectType;

typedef struct {
	Boolean undoAvailable;
	Boolean redrawDeck;
	int grid[NumPicks];
	SByte card[NumPicks];
} UndoType;

typedef struct {
	int index;									// index into grid[] (current hint)
	int grid[NumPicks];						// grid position of hint
	Boolean fUsed[NumPicks];				// flag whether the hint has been used

	int numGroups;								// total number of current groups
} HintType;

typedef struct {
	Byte score;
	Byte reasonWhy;							// reason why last group was invalid
	Boolean fSound;							// fSound is stored here because it
													// defaults to the sound pref at start
	ModeType mode;								// game must always be in _some_ mode
	HintType hint;								// current hint info
	SelectType selection;					// current card selection
} GameType;

typedef struct {
	SByte card;

	Boolean fSelected;
	RectangleType spot;
} GridType;

typedef struct {
	SByte card;
	Boolean fSelected;
} GridLiteType;								// 2 bytes instead of 11

typedef struct {
	Byte score;
	Byte level;
	Boolean fWin;
	char name[MaxNameLength + 1];
} HighScoreType;

typedef Byte CardType;						// unsigned char

typedef struct {
	SByte index;
	Byte numCards;
	CardType card[CardsInDeck];
} DeckType;

typedef DeckType *DeckPtr;

typedef struct {
	Boolean fGameInProgress;
	Byte score;
	UndoType undo;
	DeckType deck;
	GridLiteType grid[NumGrid];

	Byte difficulty;
	Byte grayPattern;
	Byte lastHighScore;

	HighScoreType highScores[MaxHighScores];
} PrefType;


//  .---------.
//  | GLOBALS |
//  '---------'
// thanks to Dave MacLeod and the fretboard source for Hz values:
SoundType gSounds[NumGameSounds] = {
	4, {{ 200, 10}, { 300, 10}, {250, 10}, { 200, 10}},// soundShuffle
	1, {{ 200, 10}},		// soundDeal, copy of click sound from SoundPrv.h
	1, {{ 600, 10}},												// soundSelect
	1, {{ 400, 10}},												// soundUnselect
	2, {{ 494, 70}, { 440, 70}},								//	soundNotAGroup
	2, {{ 659, 70}, { 784, 70}},								// soundGrouper
	4, {{ 784, 90}, {1047, 90}, {784, 90}, { 523, 90}},// soundGameOver
	4, {{ 523, 90}, { 659, 90}, {784, 90}, {1047, 90}},// soundYouWin
	4, {{1047, 90}, { 523, 90}, {784, 90}, {1047, 90}},// soundNewHighScore
	1, {{ 500, 70}}		// soundError, copy of error beep from SoundPrv.h
};

GameType			gGame;						// game status struct
PrefType			gPrefs;						// app preferences, of course
DeckType			gDeck;						// 81 card array and its index
GridType			gGrid[NumGrid];			// array of the 12 grid spots
UndoType			gUndo;						// undo information
HighScoreType	gHighScore;					// current high score, if any
Char				gString[32];				// util global for localized strings

int				gAnimate = 0;				// animation status indicator


//  .------------.
//  | PROTOTYPES |
//  '------------'
// Palm-defined prototypes
static Word StartApplication(void);
static void StopApplication(void);
static void EventLoop(void);
static Boolean AppEventHandler(EventPtr event);
static Boolean MainEventHandler(EventPtr event);
static Boolean PrefsEventHandler(EventPtr event);
static Boolean NewHighScoreEventHandler(EventPtr event);
static Boolean HighScoreEventHandler(EventPtr event);
static Boolean AboutEventHandler(EventPtr event);

// grid/card functions
static void GameInit();
static void GameInitFromSave();
static void GameStart(void);
static void GameFinish(void);
static Boolean GameWon();
static void DealCard(int spot);
static void CardPenUp(int spot);
static void GridInvert(int index);
static int GetCardSelection(SWord x, SWord y);
static void SelectGridSpot(Boolean fSelect, int spot);

// hint functions
static void FindHint();
static void ShowHint(Boolean fShow);
static void ShowHintsFound();

// utility functions
static void Undo();
static void SetMode(Word mode);
static void AddBit(int *bucket, int bit);
static VoidPtr GetObjectPtr(int objectID);
static Word AlertEx(Word alertID, Word stringID);
static Boolean IsGroup(int card1, int card2, int card3);
static CharPtr GetHighScoreName();
static void PlayGameSound(GameSoundsType soundToPlay);
static CharPtr GetString(Word idString);

// draw functions
static void DrawButWhyForm(void);
static void StatusStockMsg(Word msgID);
static void StatusMsg(char *msg, FontID font);
static void DrawHighScores(void);
static void UpdateScore(int increment);
static void ListDrawPatterns (Word itemNum, RectanglePtr bounds, CharPtr *itemsText);


#include "card.c"								// card-related functions


//  .------------------.
//  | AND THEY'RE OFF! |
//  '------------------'
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
	Word error;

	if (cmd == sysAppLaunchCmdNormalLaunch) {
		error = StartApplication();
		if (error) return error;

		EventLoop();
		StopApplication();
	} else {
		return sysErrParamErr;
	}

	return 0;
}


static Word StartApplication(void) {
	int i;
	int rows[3] = {17, 60, 103};
	int cols[4] = { 1, 36,  71, 106};
	SystemPreferencesType sysPrefs;

	// randomize from 1/1/1904 (another way to do this is with TimGetTicks())
	SysRandom(TimGetSeconds());

	if (PrefGetAppPreferencesV10(AppID, AppVer, &gPrefs, sizeof(PrefType)) == NULL) {
		// initialize preferences
		gPrefs.difficulty = 0;				// easy
		gPrefs.grayPattern = 4;				// gray
		gPrefs.lastHighScore = MaxHighScores;
													// no last high score
		for (i = 0; i < MaxHighScores; i++)
			gPrefs.highScores[i].score = 0;
	}

	// make the sound dependent on the system preferences gameSound
	// used instead of PrefGetPreference() for OS 1.0 compatibility
	PrefGetPreferences(&sysPrefs);
	gGame.fSound = (sysPrefs.gameSoundLevelV20 == slOn);

	// create the playing grid
	for (i = 0; i < NumGrid; i++) {
		gGrid[i].card = NoCard;
		gGrid[i].fSelected = false;

		RctSetRectangle(&gGrid[i].spot, cols[i % 4], rows[i / 4], CardX, CardY);
	}

	FrmGotoForm(idFormMain);
	return 0;
}


static void StopApplication(void) {
	int i;

	gPrefs.fGameInProgress = (gGame.mode == modeHint || gGame.mode == modeButWhy);
	if (gPrefs.fGameInProgress) {
		// write save-game prefs
		gPrefs.score = gGame.score;

		MemMove(&gPrefs.deck, &gDeck, sizeof (DeckType));
		MemMove(&gPrefs.undo, &gUndo, sizeof (UndoType));

		for (i = 0; i < NumGrid; i++) {
			gPrefs.grid[i].card = gGrid[i].card;
			gPrefs.grid[i].fSelected = gGrid[i].fSelected;
		}
	}

	// write the current preference information
	PrefSetAppPreferencesV10(AppID, AppVer, &gPrefs, sizeof(PrefType));
}


static void EventLoop(void) {
	int delay;
	Word error;
	EventType event;

	do {
		switch (gAnimate) {
		case idFormNewHighScore:
			delay = FrameDelay;
			break;
		case idFormAbout:
			// cycle through 4 cards / sec
			delay = FrameDelay * FramesPerSec / 4;
			break;
		default:
			delay = evtWaitForever;
			break;
		}
		EvtGetEvent(&event, delay);

		if(!SysHandleEvent(&event))
			if(!MenuHandleEvent(NULL, &event, &error))
				if(!AppEventHandler(&event))
					FrmDispatchEvent(&event);
	} while(event.eType != appStopEvent);
}


static Boolean AppEventHandler(EventPtr event) {
	int     formID;
	FormPtr frm;
	Boolean handled = false;

	if (event->eType == frmLoadEvent) {
		formID = event->data.frmLoad.formID;
		frm = FrmInitForm(formID);
		FrmSetActiveForm(frm);

		switch (formID) {
		case idFormMain:
			FrmSetEventHandler(frm, MainEventHandler);
			break;
		case idFormPrefs:
			FrmSetEventHandler(frm, PrefsEventHandler);
			break;
		case idFormNewHighScore:
			FrmSetEventHandler(frm, NewHighScoreEventHandler);
			break;
		case idFormHighScores:
			FrmSetEventHandler(frm, HighScoreEventHandler);
			break;
		case idFormAbout:
			FrmSetEventHandler(frm, AboutEventHandler);
			break;
		}
		handled = true;
	}

	return handled;
}


static Boolean MainEventHandler (EventPtr event) {
	SWord x, y;
	Boolean penDown;
	Boolean handled = false;
	int i, selected, selectedDown;

	FormPtr frm;

	CALLBACK_PROLOGUE							// see callback.h

	switch (event->eType) {
	case frmOpenEvent:
		frm = FrmGetActiveForm();
		FrmDrawForm(frm);

		if (gPrefs.fGameInProgress && AlertEx(idAlertNewGame, idStringStartup) == DialogButton0)
			GameInitFromSave();
		else {
			GameInit();
			StatusStockMsg(msgWelcome);
		}

		handled = true;
		break;

	case winEnterEvent:
		// this forces the form menu to set itself back to the correct
		// value after getting focus again.  Apparently, the form rechecks
		// its menu structure every time it receives focus.
		SetMode(gGame.mode);
		handled = true;
		break;

	case frmUpdateEvent:
		// event is sent by the Prefs form; redraw the playing area
		if (event->data.frmUpdate.updateCode & UpdateRedrawMask)
			for (i = 0; i < NumGrid; i++)
				if (gGrid[i].card != NoCard)
					DrawCard(gDeck.card[gGrid[i].card], gGrid[i].spot.topLeft.x, gGrid[i].spot.topLeft.y, gGrid[i].fSelected);
		if (event->data.frmUpdate.updateCode & UpdateNewGameMask) {
			// redraw the deck with the different number of cards
			if (gGame.mode == modeStart || 
				(AlertEx(idAlertNewGame, idStringNewGame) == DialogButton0)) {
				GameInit();
			}
		}

		handled = true;
		break;

	case menuEvent:
		switch (event->data.menu.itemID) {
		case idMenuItemHandyman:
			switch (gGame.mode) {
			case modeStart:
				GameStart();
				break;
			case modeHint:
				ShowHint(true /*fShow*/);
				SysTaskDelay(FrameDelay * 2);
				ShowHint(false /*fShow*/);
				break;
			case modeButWhy:
				DrawButWhyForm();
				break;
			case modeEnd:
				GameFinish();
				GameInit();
				break;
			}
			break;
		case idMenuItemUndo:
			if (gUndo.undoAvailable) {
				// this fixes a problem when the menu is activated by a graffiti
				// accelerator char...the problem being that when the command
				// rect disappears, it restores the pixels behind it when it was
				// activated--here, it restores the pixels from .5 seconds ago,
				// which is from the _wrong_ screen
				MenuEraseStatus(NULL);
				Undo();
			} else
				PlayGameSound(soundError);
			break;
		case idMenuItemNewGame:
			if (gGame.mode == modeStart)
				PlayGameSound(soundError);
			else {
				// don't ask, don't tell
				// if (AlertEx(idAlertNewGame, idStringNewGame) == DialogButton0)

				MenuEraseStatus(NULL);
				// give the user a chance to change the prefs for the new game
				FrmPopupForm(idFormPrefs);
				GameInit();
			}
			break;
		case idMenuItemHighScores:
			FrmPopupForm(idFormHighScores);
			break;
		case idMenuItemPrefs:
			FrmPopupForm(idFormPrefs);
			break;
		case idMenuItemQuickTips:
			FrmHelp(idHelpQuickTips);
			break;
		case idMenuItemHowToPlay:
			FrmHelp(idHelpHowToPlay);
			break;
		case idMenuItemAbout:
			FrmPopupForm(idFormAbout);
			break;
		}
		handled = true;
		break;

	case ctlEnterEvent:
	case ctlExitEvent:
		// leave Enter and Exit events unhandled (so OS can redraw button)
		if (gGame.mode == modeHint)
			// invert the card on EnterEvent, and revert it on ExitEvent
			ShowHint(event->eType == ctlEnterEvent);
		break;

	case ctlSelectEvent:
		switch (gGame.mode) {
		case modeStart:
			GameStart();
			break;
		case modeHint:
			ShowHint(false /*fShow*/);
			break;
		case modeButWhy:
			DrawButWhyForm();
			break;
		case modeEnd:
			GameFinish();
			GameInit();
			break;
		}
		handled = true;
		break;

	case penDownEvent:
		if (gGame.mode == modeStart)
			break;

		selected = GetCardSelection(event->screenX, event->screenY);
		if (selected != NoCard) {
			selectedDown = selected;
			GridInvert(selected);

			do {
				EvtGetPen (&x, &y, &penDown);
			} while (penDown);
			selected = GetCardSelection(x, y);

			// this won't happen on a penDown because selectedDown = -1
			// if the pen came up on the same card it went down on
			if (selected == selectedDown)
				CardPenUp(selected);
			else
				GridInvert(selectedDown);

			handled = true;
		} else
			handled = false;
		break;
	}

	CALLBACK_EPILOGUE							// see callback.h

	return handled;
}


static Boolean PrefsEventHandler(EventPtr event) {
	int		updateMask,
				difficulty;
	Boolean	handled = false;

	FormPtr    frm;
	ListPtr    lst;
	ControlPtr ctl;

	static PrefType tempPrefs;

	CALLBACK_PROLOGUE							// see callback.h

	switch (event->eType) {
	case frmOpenEvent:
		frm = FrmGetActiveForm();
		ctl = GetObjectPtr(idCheckSound);
		CtlSetValue(ctl, gGame.fSound);

		FrmSetControlGroupSelection(frm, idGroupDiff, gPrefs.difficulty + idPushDiff1);

		// initialize the pattern list and update its label
		lst = GetObjectPtr(idListPatterns);
		LstSetDrawFunction(lst, ListDrawPatterns);
		LstSetSelection(lst, gPrefs.grayPattern);
		LstMakeItemVisible(lst, gPrefs.grayPattern);
		ctl = GetObjectPtr(idPopTrigBack);
		CtlSetLabel(ctl, LstGetSelectionText(lst, LstGetSelection(lst)));

		FrmDrawForm(frm);
		DrawPatternSample(colorGray, PrefsSampleX, PrefsSampleY);

		// make local copy of the prefs
		MemMove(&tempPrefs, &gPrefs, sizeof(PrefType));

		handled = true;
		break;

	case popSelectEvent:
		lst = GetObjectPtr(idListPatterns);
		gPrefs.grayPattern = LstGetSelection(lst);
		DrawPatternSample(colorGray, PrefsSampleX, PrefsSampleY);

		// let OS handle label redraw
		break;

	case ctlSelectEvent:
		switch (event->data.ctlSelect.controlID) {
		case idCheckSound:
			gGame.fSound = !gGame.fSound;
			if (gGame.fSound)
				// play a sample sound when sound is enabled
				PlayGameSound(soundGrouper);
			handled = true;
			break;

		case idPushDiff1:
		case idPushDiff2:
		case idPushDiff3:
			frm = FrmGetActiveForm();
			difficulty = FrmGetControlGroupSelection(frm, idGroupDiff);
			gPrefs.difficulty = FrmGetObjectId(frm, difficulty) - idPushDiff1;
			handled = true;
			break;

		case idButtonOK:
			// only perform a redraw if necessary (it looks sloppy otherwise)
			updateMask = 0;
			if (gPrefs.grayPattern != tempPrefs.grayPattern)
				updateMask |= UpdateRedrawMask;
			if (gPrefs.difficulty != tempPrefs.difficulty)
				updateMask |= UpdateNewGameMask;
			FrmUpdateForm(idFormMain, updateMask);

			// fall through.....aaaaaiiiiieeeeeeeeeeeeeeeee!
		case idButtonCancel:
			// the prefs copying was done in reverse so that when GetPattern()
			// is called it can retrieve the most current value in gPrefs
			if (event->data.ctlSelect.controlID == idButtonCancel)
				MemMove(&gPrefs, &tempPrefs, sizeof(PrefType));
			FrmReturnToForm(idFormMain);
			handled = true;
			break;
		}
	}

	CALLBACK_EPILOGUE							// see callback.h

	return handled;
}


static Boolean NewHighScoreEventHandler(EventPtr event) {
	int i;
	static int counter = 0;
	Boolean handled = false;

	FormPtr   frm;
	VoidHand  bmpHandle;
	BitmapPtr bmpPtr;
	static CharPtr nameP;
	static Handle  nameH;
	RectangleType flagpole = {{HighScoreFlagpoleX,  HighScoreFlagpoleY },
									  {HighScoreFlagpoleDX, HighScoreFlagpoleDY}};

	CALLBACK_PROLOGUE							// see callback.h


	switch (event->eType) {
	case frmOpenEvent:
		frm = FrmGetActiveForm();

		nameH = (Handle) MemHandleNew(MaxNameLength + 1);
		nameP = MemHandleLock((VoidHand) nameH);
		StrCopy(nameP, GetHighScoreName());
		MemHandleUnlock((VoidHand) nameH);

		// set the field to edit the name.
		FldSetTextHandle(GetObjectPtr(idFieldNewHighScore), nameH);

		FrmDrawForm(frm);
		WinDrawRectangle(&flagpole, NoDiam);

		// we could use GrfSetState() and GsiSetShiftState() here, but
		// is it really necessary for just a high score form?
		FrmSetFocus(frm, FrmGetObjectIndex(frm, idFieldNewHighScore));
		PlayGameSound(soundNewHighScore);

		gAnimate = idFormNewHighScore;
		handled = true;
		break;

	case ctlSelectEvent:
		gAnimate = 0;
		switch (event->data.ctlSelect.controlID) {
		case idButtonOK:
			// add the score unless the user removed the name; if so
			// they probably didn't want the score entered...so don't!
			if (nameP[0] != '\0')	{
				// find high score position
				i = MaxHighScores;
				while (i > 0 && gHighScore.score > gPrefs.highScores[i - 1].score)
					i--;

				MemMove(&gPrefs.highScores[i + 1], &gPrefs.highScores[i],
						  (MaxHighScores - i - 1) * sizeof (HighScoreType));

				// record this new score as the last one entered.
				gPrefs.lastHighScore = i;
				MemMove(&gPrefs.highScores[i], &gHighScore, sizeof (HighScoreType));
				StrCopy(gPrefs.highScores[i].name, nameP);
			}

			FrmPopupForm(idFormHighScores);
			break;

		case idButtonCancel:
			// don't save this game or popup the high score form if they cancel

			// we shouldn't really do this; the right way is to call
			// SetMode(), but SetMode() won't work here because the
			// main form must be the active form
			gGame.mode = modeStart;			
			FrmUpdateForm(idFormMain, UpdateNewGameMask);
			break;
		}
		FrmReturnToForm(idFormMain);
		handled = true;
		break;

	case nilEvent:
		// cycle through animation bitmaps
		bmpHandle = DmGetResource(bitmapRsc, idBitmapAnim1 + counter);
		bmpPtr = MemHandleLock(bmpHandle);
		WinDrawBitmap(bmpPtr, HighScoreFlagpoleX + 1, HighScoreFlagpoleY);
		MemHandleUnlock(bmpHandle);

		// there are 4 bitmaps we cycle through to produce the animation
		counter = (++counter % NumAnimationBitmaps);
		handled = true;
		break;
	}

	CALLBACK_EPILOGUE							// see callback.h

	return handled;
}


static Boolean HighScoreEventHandler(EventPtr event) {
	int i;
	Boolean  handled = false;

	CALLBACK_PROLOGUE							// see callback.h

	switch (event->eType) {
	case frmOpenEvent:
		FrmDrawForm(FrmGetActiveForm());
		DrawHighScores();

		handled = true;
		break;

	case ctlSelectEvent:
		switch (event->data.ctlSelect.controlID) {
		case idButtonOK:
			// we only want to reset if they won and cleared the board
			if (gGame.mode == modeEnd && gPrefs.highScores[gPrefs.lastHighScore].fWin) {
				// we shouldn't really do this; the right way is to call
				// SetMode(), but SetMode() won't work here because the
				// main form must be the active form
				gGame.mode = modeStart;
				FrmUpdateForm(idFormMain, UpdateNewGameMask);
			}

			FrmReturnToForm(idFormMain);
			handled = true;
			break;

		case idButtonClear:
			// if there are no high scores, do nothing
			if (gPrefs.lastHighScore == MaxHighScores)
				PlayGameSound(soundError);
			else if (AlertEx(idAlertNewGame, idStringClear) == DialogButton0) {
				for (i = 0; i < MaxHighScores; i++)
					gPrefs.highScores[i].score = 0;
				gPrefs.lastHighScore = MaxHighScores;

				FrmGotoForm(idFormHighScores);
			}
			handled = true;
			break;
		}
		break;
	}

	CALLBACK_EPILOGUE							// see callback.h

	return handled;
}


static Boolean AboutEventHandler(EventPtr event) {
	Boolean  handled = false;
	static DeckType aboutDeck;

	FormPtr  frm;

	CALLBACK_PROLOGUE							// see callback.h

	switch (event->eType) {
	case frmOpenEvent:
		FrmDrawForm(FrmGetActiveForm());

		aboutDeck.numCards = CardsInDeck;
		DeckInit(&aboutDeck);

		gAnimate = idFormAbout;
		handled = true;
		break;

	case ctlSelectEvent:
		gAnimate = 0;

		FrmReturnToForm(idFormMain);
		handled = true;
		break;

	case nilEvent:
		// cycle through cards in the deck
		DrawCard(aboutDeck.card[++aboutDeck.index % CardsInDeck],
					AboutDeckX, AboutDeckY, false);
		if (DeckEmpty(aboutDeck))
			SetDeckFull(aboutDeck);
		handled = true;
		break;
	}

	CALLBACK_EPILOGUE							// see callback.h

	return handled;
}


//  .----------------------------------.
//  | GRID/CARD FUNCTIONS (see card.c) |
//  '----------------------------------'
static void GameInit() {
	int i;

	// if the grid still has cards in it erase them now
	// and remove any current selections
	SetDeckEmpty(gDeck);
	for (i = 0; i < NumGrid; i++) {
		if (gGrid[i].card != NoCard) {
			DealCard(i);
			gGrid[i].fSelected = false;
		}
	}

	// clear the score, status bar
	UpdateScore(-gGame.score);
	StatusStockMsg(msgClear);

	// initialize the deck
	gDeck.numCards = (gPrefs.difficulty + 1) * DifficultyFactor;
	DeckInit(&gDeck);
	PlayGameSound(soundShuffle);
	DrawDeck(true /*fDraw*/, &gDeck);
	DrawCardsLeft(&gDeck);

	// initialize selection status
	gGame.selection.count = 0;
	gGame.selection.grid[0] = gGame.selection.grid[1] = gGame.selection.grid[2] = 0;

	gUndo.undoAvailable = false;

	SetMode(modeStart);
}


static void GameInitFromSave() {
	int i, picks = 0;

	// intialize the status bar
	StatusStockMsg(msgClear);

	// initialize from preferences
	gGame.score = gPrefs.score;
	MemMove(&gDeck, &gPrefs.deck, sizeof (DeckType));
	MemMove(&gUndo, &gPrefs.undo, sizeof (UndoType));

	// recreate playing grid
	for (i = 0; i < NumGrid; i++) {
		gGrid[i].card = gPrefs.grid[i].card;
		if (gGrid[i].card != NoCard) {
			gGrid[i].fSelected = gPrefs.grid[i].fSelected;
			if (gGrid[i].fSelected) {
				picks++;
				SelectGridSpot(true, i);
			} else
				DrawCard(gDeck.card[gGrid[i].card], gGrid[i].spot.topLeft.x, gGrid[i].spot.topLeft.y, gGrid[i].fSelected);
		}
	}

	// draw the deck with the correct number of cards left
	if (!DeckEmpty(gDeck)) {
		DrawDeck(true /*fDraw*/, &gDeck);
		DrawCardsLeft(&gDeck);
	}
	UpdateScore(0);

	// initialize current hint
	FindHint();

	if (picks == NumPicks) {
		// call IsGroup to fill reasonWhy with reason why it's not a group
		IsGroup(gGrid[gGame.selection.grid[0]].card, gGrid[gGame.selection.grid[1]].card, gGrid[gGame.selection.grid[2]].card);
		PlayGameSound(soundNotAGroup);
		StatusStockMsg(msgNotAGroup);
		// well, it obviously wasn't a group, or the 3 picks still wouldn't
		// be selected!
		SetMode(modeButWhy);		
	} else {
		ShowHintsFound();
		SetMode(modeHint);
	}
}


static void GameStart(void) {
	int i;

	// initialize playing grid, setup game
	do {
		// ensure that the game starts with at least 1 group available
		// and that no selections from the previous game remain
		for (i = 0; i < NumGrid; i++)
			DealCard(i);

		FindHint();
		if (!gGame.hint.numGroups)
			DeckInit(&gDeck);
	} while (!gGame.hint.numGroups);
	SetMode(modeHint);
	ShowHintsFound();
}


static void GameFinish(void) {
	int i = 0;

	// see if this is a high score
	i = MaxHighScores;
	while (i > 0 && gGame.score > gPrefs.highScores[i - 1].score)
		i--;
	if (i == MaxHighScores)
		return;

	// set the current high score info
	gHighScore.level = gPrefs.difficulty;
	gHighScore.score = gGame.score;
	gHighScore.fWin = GameWon();

	FrmPopupForm(idFormNewHighScore);
}


static Boolean GameWon() {
	int i = 0;
	Boolean fWon = true;

	// test for a won position (empty grid)
	while (fWon && i < NumGrid) {
		fWon = (gGrid[i].card == NoCard);
		i++;
	}

	return fWon;
}


static void DealCard(int spot) {
	if (DeckEmpty(gDeck)) {
		// once we're past the first redraw we don't want to do it again
		gUndo.redrawDeck = false;

		// end-of-deck situation
		gGrid[spot].card = NoCard;
		// erase card
		WinEraseRectangle(&gGrid[spot].spot, CardDiam);
		// erase selection rectangle
		WinEraseRectangleFrame(rectangleFrame, &gGrid[spot].spot);
	} else {
		gGrid[spot].card = ++gDeck.index;

		DrawCardsLeft(&gDeck);
		if (DeckEmpty(gDeck)) {
			gUndo.redrawDeck = true;
			DrawDeck(false /*fDraw*/, &gDeck);
		}
		// if we add card animation, put it here
		//DrawDealtCard(spot);
		DrawCard(gDeck.card[gDeck.index], gGrid[spot].spot.topLeft.x, gGrid[spot].spot.topLeft.y, gGrid[spot].fSelected);
	}
	PlayGameSound(soundDeal);
}


static void CardPenUp(int spot) {
	// this code was originally in the main form event handler
	int i, gridSpot;

	// if 3 cards are already selected, deselect the last one
	if (!gGrid[spot].fSelected && gGame.selection.count == NumPicks)
		SelectGridSpot(false, gGame.selection.grid[NumPicks - 1]);

	StatusStockMsg(msgClear);
	// reverse current selection
	SelectGridSpot(!gGrid[spot].fSelected, spot);
	if (gGrid[spot].fSelected)
		PlayGameSound(soundSelect);
	else
		PlayGameSound(soundUnselect);

	if (gGame.selection.count == NumPicks) {
		if (IsGroup(gGrid[gGame.selection.grid[0]].card, gGrid[gGame.selection.grid[1]].card, gGrid[gGame.selection.grid[2]].card)) {
			StatusStockMsg(msgGrouper);
			UpdateScore(ScoreGroup);
			PlayGameSound(soundGrouper);
			// pause a half second
			SysTaskDelay(FrameDelay * FramesPerSec / 2);

			for (i = 0; i < NumPicks; i++) {
				gUndo.grid[i] = gGame.selection.grid[0];
				gUndo.card[i] = gGrid[gGame.selection.grid[0]].card;

				SysTaskDelay(FrameDelay * 2);
				gridSpot = gGame.selection.grid[0];
				SelectGridSpot(false, gridSpot);
				DealCard(gridSpot);
			}
			gUndo.undoAvailable = true;

			FindHint();
			if (!gGame.hint.numGroups) {
				// let the user look at what's left, to find out why the
				// why the cards don't make a group or to undo the last group
				SetMode(modeEnd);

				if (GameWon()) {
					UpdateScore(ScoreWin);				// bonus points for win
					StatusStockMsg(msgYouWin);
					PlayGameSound(soundYouWin);
					GameFinish();
				} else {
					StatusStockMsg(msgGameOver);
					PlayGameSound(soundGameOver);
				}
			}
			else
				ShowHintsFound();
		} else {
			StatusStockMsg(msgNotAGroup);
			PlayGameSound(soundNotAGroup);
			SetMode(modeButWhy);
		}
	} else {
		// game mode should always be in hint mode when user is
		// picking cards, unless it's in modeButWhy
		if (gGame.mode != modeHint)
			SetMode(modeHint);

		// if we're at 0 selections, display the last number of hints found
		if (!gGame.selection.count)
			ShowHintsFound();
	}
}


static void GridInvert(int index) {
	RctInsetRectangle(&gGrid[index].spot, 1);
	WinInvertRectangle(&gGrid[index].spot, CardDiam);
	RctInsetRectangle(&gGrid[index].spot, -1);
}


static int GetCardSelection(SWord x, SWord y) {
	int i = 0;
	Boolean found = false;

	// don't allow selections when the game is over
	if (gGame.mode == modeEnd)
		return NoCard;

	while (!found && i < NumGrid) {
		// if gGrid[i].card == -1 the spot is empty (end-of-deck situation)
		found = RctPtInRectangle(x, y, &gGrid[i].spot) && (gGrid[i].card != NoCard);
		i++;
	}

	if (!found)
		return NoCard;
	else
		return i - 1;
}


static void SelectGridSpot(Boolean fSelect, int spot) {
	// update the grid information
	gGrid[spot].fSelected = fSelect;
	if (fSelect) {
		if (gGame.selection.count < NumPicks)
			gGame.selection.count++;
		gGame.selection.grid[gGame.selection.count - 1] = spot;
	} else {
		gGame.selection.count--;
		if (spot == gGame.selection.grid[0]) {
			// look for a deselection in the 1st card; move other 2 cards "up"
			gGame.selection.grid[0] = gGame.selection.grid[1];
			gGame.selection.grid[1] = gGame.selection.grid[2];
		} else if (spot == gGame.selection.grid[1])
			// look for a deselection in the 2nd card; move 3rd card "up"
			gGame.selection.grid[1] = gGame.selection.grid[2];
	}
	DrawCard(gDeck.card[gGrid[spot].card], gGrid[spot].spot.topLeft.x, gGrid[spot].spot.topLeft.y, fSelect);
}


//  .----------------.
//  | HINT FUNCTIONS |
//  '----------------'
static void FindHint() {
	int i, j, k;
	int found = false;

	gGame.hint.index = -1;
	for (i = 0; i < NumGrid - 2; i++) {
		for (j = i + 1; j < NumGrid - 1; j++) {
			for (k = j + 1; k < NumGrid; k++) {
				if (gGrid[i].card == NoCard || gGrid[j].card == NoCard || gGrid[k].card == NoCard)
					continue;					// don't use non-cards in a hint
				if (IsGroup(gGrid[i].card, gGrid[j].card, gGrid[k].card)) {
					if (!found) {				// store only the 1st hint we find
						gGame.hint.grid[0] = i;
						gGame.hint.grid[1] = j;
						gGame.hint.grid[2] = k;
						gGame.hint.index = 0;
					}
					found++;
				}
			}
		}
	}

	if (found) {
		// randomize the hint array (this prevents users from knowing that
		// the next hint is always below the current one in the grid)
		for (i = 0; i < NumPicks; i++) {
			j = (UInt)SysRandom(0) % NumPicks;

			k = gGame.hint.grid[i];
			gGame.hint.grid[i] = gGame.hint.grid[j];
			gGame.hint.grid[j] = k;

			gGame.hint.fUsed[i] = false;	// reset "used" flag
		}
	}
	gGame.hint.numGroups = found;
/* UPDATE: a hint grid is unnecessary (and a waste...see below)
 *         C(12, 3) = 220, so at worst that's still not bad at all
 *
 * (old attempt)
 * The hint grid stores the attempts previously made so a match attempt
 * from card1, card5, card8 isn't attempted with card8, card1 and card5.
 * This reduces the number of match attempts by 1100 (P(12, 3) - C(12, 3)).
 */
}


static void ShowHint(Boolean fShow) {
	if (fShow)
		// skip cards that are already selected (if any)
		while (gGrid[gGame.hint.grid[gGame.hint.index]].fSelected)
			gGame.hint.index = (++gGame.hint.index % NumPicks);
	else {
		// only take away points if this is the first time for this hint
		if (!gGame.hint.fUsed[gGame.hint.index]) {
			gGame.hint.fUsed[gGame.hint.index] = true;
			UpdateScore(ScoreHint);
		}
	}

	GridInvert(gGame.hint.grid[gGame.hint.index]);
}


static void ShowHintsFound() {
	char string[32];							// Note: 10+ groups _is_ possible

	StrIToA(string, gGame.hint.numGroups);
	StrCat(string, " ");
	if (gGame.hint.numGroups == 1)
		GetString(idStringGroupFound);
	else
		GetString(idStringGroupsFound);
	StrCat(string, gString);
	if (gGame.hint.numGroups > 1)
		StrCat(string, "!");
	else
		StrCat(string, ".");

	StatusMsg(string, stdFont);
}


//  .-------------------.
//  | UTILITY FUNCTIONS |
//  '-------------------'
static void Undo() {
	int i, spot;

	// remove any current selections
	while (gGame.selection.count > 0)
		SelectGridSpot(false, gGame.selection.grid[0]);

	for (i = 0; i < NumPicks; i++) {
		spot = gUndo.grid[i];
		gGrid[spot].card = gUndo.card[i];
		DrawCard(gDeck.card[gGrid[spot].card], gGrid[spot].spot.topLeft.x, gGrid[spot].spot.topLeft.y, gGrid[spot].fSelected);

		// this is a special case where the deck has disappeared because there
		// are no more cards.  We have to replace it and work from there.
		if (gUndo.redrawDeck) {
			gUndo.redrawDeck = false;
			DrawDeck(true /*fDraw*/, &gDeck);
			gDeck.index -= 1;
		}
		else if (!DeckEmpty(gDeck)) {
			gDeck.index -= 1;
			DrawCardsLeft(&gDeck);
		}
		PlayGameSound(soundDeal);
	}

	gUndo.undoAvailable = false;
	UpdateScore(ScoreUndo);

	if (gGame.mode == modeEnd)
		SetMode(modeHint);
	FindHint();
	ShowHintsFound();
}


static void SetMode(Word mode) {
	MenuBarPtr mb = MenuGetActiveMenu();
	ControlPtr ctl = GetObjectPtr(idButtonHandyman);

	// we can't change the menu item without it being initialized first
	if (!mb) {
		mb = MenuInit(idMenuGrouper);
		MenuSetActiveMenu(mb);
	}

	switch (mode) {
	case modeStart:
		GetString(idStringDeal);
		break;
	case modeHint:
		GetString(idStringHint);
		break;
	case modeButWhy:
		GetString(idStringButWhy);
		break;
	case modeEnd:
		GetString(idStringOK);
		break;
	}
	gGame.mode = mode;
	CtlSetLabel(ctl, gString);
	StrCopy(mb->menus[MenuHandyman].items[MenuHandyman].itemStr, gString);
}


static void AddBit(int *bucket, int bit) {
	switch (bit) {
		case 0:
			*bucket |= 1;
			break;
		case 1:
			*bucket |= 2;
			break;
		case 2:
			*bucket |= 4;
			break;
	}
}


static VoidPtr GetObjectPtr(int objectID) {
	// as many have noted, this function should've been in the API
	FormPtr frm;

	frm = FrmGetActiveForm();
	return (FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, objectID)));
}


static Word AlertEx(Word alertID, Word stringID) {
	int buttonHit;
	VoidHand  strHandle;
	CharPtr strPtr;

	strHandle = DmGetResource(strRsc, stringID);
	strPtr = MemHandleLock(strHandle);

	buttonHit = FrmCustomAlert(alertID, strPtr, NULL, NULL);
	MemPtrUnlock(strPtr);

	return buttonHit;
}


static Boolean IsGroup(int card1, int card2, int card3) {
/*
 * My theory is this: we create a bit bucket.  bit0 holds style value
 * 00, bit1 holds style value 01 and bit2 holds style value 10.  If
 * all values are different, the bucket = 7 (111); if all the values are
 * the same, the bucket will be 4 (100) or 2 (010) or 1 (001).  In these
 * four cases we have a group.  Otherwise, we don't have a group.
 *
 */
	int i = 0, j,
		 bucket,
		 cards[NumPicks] = {card1, card2, card3};
	Boolean fIsGroup;

	gGame.reasonWhy = 0;
	// we could run this loop only until we found a non-match, but to
	// fill reasonWhy with the group specifics, we have to run through
	// all categories
	for (i = 0; i < NumProps; i++) {
		bucket = 0;
		for (j = 0; j < NumPicks; j++) {
			switch (i) {
			case 0:
				AddBit(&bucket, GetStyleColor(gDeck.card[cards[j]]));
				break;
			case 1:
				AddBit(&bucket, GetStyleMarks(gDeck.card[cards[j]]) - 1);
				break;
			case 2:
				AddBit(&bucket, GetStyleMarkColor(gDeck.card[cards[j]]));
				break;
			case 3:
				AddBit(&bucket, GetStyleMarkType(gDeck.card[cards[j]]));
				break;
			}
		}
		fIsGroup = (bucket == 7 || bucket == 4 || bucket == 2 || bucket == 1);

		gGame.reasonWhy |= !fIsGroup << i;
	}

	return (gGame.reasonWhy ? false : true);
}


static CharPtr GetHighScoreName() {
	// again, thanks to Roger Flores and the reptoids Source
	CharPtr firstSpace;
	CharPtr name = MemPtrNew(dlkUserNameBufSize);

	if (!name)
		return NULL;

	// try and use the name last entered
	if (gPrefs.lastHighScore != MaxHighScores) {
		StrCopy(name, gPrefs.highScores[gPrefs.lastHighScore].name);
	} else {
		// try and use the user's name
		StrCopy(name, "");
		DlkGetSyncInfo(NULL, NULL, NULL, name, NULL, NULL);

		// just use the first name
		firstSpace = StrChr(name, spaceChr);
		if (firstSpace)
		 	*firstSpace = '\0';
		// truncate the string to ensure it's not too long
		name[MaxNameLength] = '\0';
	}
	MemPtrFree(name);

	return name;
}


static void PlayGameSound(GameSoundsType soundToPlay) {
	int i;
	SndCommandType sound;
	static long lastSound;

	if (gGame.fSound) {
		sound.cmd = sndCmdFreqDurationAmp;
		sound.param3 = sndDefaultAmp;			// this is constant
		for (i = 0; i < gSounds[soundToPlay].numNotes; i++) {
			// note: put pause code in here if necessary (freq = 0)
			sound.param1 = gSounds[soundToPlay].notes[i].frequency;
			sound.param2 = gSounds[soundToPlay].notes[i].duration;
			SndDoCmd(0, &sound, true);			// true: noWait
		}
	}
}


static CharPtr GetString(Word idString) {
	SysCopyStringResource(gString, idString);

	return gString;
}

//  .----------------.
//  | DRAW FUNCTIONS |
//  '----------------'
static void DrawButWhyForm(void) {
// We've changed this form half a dozen times and we're not any happier
// with the end result than we were with the first one.  (Well, maybe a
// _little_ happier.  All values are hard-coded in; it makes little sense
// to declare a bunch of constants since this code is likely to change
// half a dozen times again.
	Char status[2][16];
	Int i, j, k, pattern, reasonWhyBit;
	Int cols[NumPicks] = {64, 84, 104};
	Int rows[NumProps] = {31, 53,  61, 76};
	RectangleType square = {{ 0,  0}, { MarkDY, MarkDY}};

	FontID  oldFont;
	FormPtr frm, oldForm;

	// setup form info
	oldForm = FrmGetActiveForm();
	frm = FrmInitForm(idFormButWhy);
	FrmSetActiveForm(frm);
	FrmDrawForm(frm);

	// draw top/bottom horizontal lines
	WinDrawLine(4, 28, 151, 28);
	WinDrawLine(4, 88, 151, 88);

	StrCopy(status[0], GetString(idStringPass));
	StrCopy(status[1], GetString(idStringFail));

	// draw graphical card representation
	oldFont = FntSetFont(boldFont);
	for (i = 0; i < NumProps; i++) {
		if (i)									// draw dividing gray lines
			WinDrawGrayLine(4, 28 + i * 15, 151,  28 + i * 15);

		for (j = 0; j < NumPicks; j++) {
			// reset rect square before each use; it may have been altered
			RctSetRectangle(&square, cols[j], rows[i], MarkDY, MarkDY);
			switch (i) {
			case 0:
				// create a 10x10 colored rect
				pattern = GetStyleColor(gDeck.card[gGrid[gGame.selection.grid[j]].card]);
				DrawPatternSample(pattern, square.topLeft.x, square.topLeft.y);
				break;
			case 1:
				square.extent.y = 2;			// create n 10x2 rects

				for (k = 0; k < GetStyleMarks(gDeck.card[gGrid[gGame.selection.grid[j]].card]); k++) {
					square.topLeft.y = rows[i] - k * 3;
					WinDrawRectangle(&square, 0);
				}
				break;
			case 2:
				// create a 10x10 colored rect
				pattern = GetStyleMarkColor(gDeck.card[gGrid[gGame.selection.grid[j]].card]);
				DrawPatternSample(pattern, square.topLeft.x, square.topLeft.y);
				break;
			case 3:
				// create a square, ellipse or mini-peanut for "spot type"
				WinSetPattern(cardPattern[8 /*black*/]);
				square.extent.x = square.extent.y = MarkDY;
				k = GetStyleMarkType(gDeck.card[gGrid[gGame.selection.grid[j]].card]);

				switch (k) {
				case markSquare:
					WinFillRectangle(&square, NoDiam);
					break;
				case markEllipse:
					square.topLeft.y++;
					square.extent.y -= 2;
					WinFillRectangle(&square, DiamEllipse);
					break;
				case markPeanut:
					square.extent.x = square.extent.y = MarkPeanutDY;
					WinFillRectangle(&square, NoDiam);

					square.topLeft.x += MarkPeanutDiff;
					square.topLeft.y += MarkPeanutDiff;
					WinFillRectangle(&square, NoDiam);
					break;
				}
				break;
			}
		}

		// test the fail bit
		reasonWhyBit = (gGame.reasonWhy >> i) & 1;
		WinDrawChars(status[reasonWhyBit], StrLen(status[reasonWhyBit]),
			120, 30 + i * 15);
	}
	FntSetFont(oldFont);

	// show form. remove form. return to current form.
	FrmDoDialog(frm);
	FrmDeleteForm(frm);
	FrmSetActiveForm(oldForm);
}


static void UpdateScore(int increment) {
	FontID oldFont;
	char string[4] = "   ";

	oldFont = FntSetFont(boldFont);

	// clear the old score
	StrIToA(string, gGame.score);
	WinEraseChars(string, StrLen(string), ScoreX, ScoreY);

	// don't allow a negative score
	gGame.score = max(0, gGame.score + increment);
	StrIToA(string, gGame.score);
	WinDrawChars(string, StrLen(string), ScoreX, ScoreY);

	FntSetFont(oldFont);
}


static void StatusStockMsg(Word msgID) {
	switch (msgID) {
	case msgClear:
		StatusMsg("", stdFont);
		break;
	case msgWelcome:
		StatusMsg(GetString(idStringWelcome), stdFont);
		break;
	case msgNotAGroup:
		StatusMsg(GetString(idStringNotAGroup), stdFont);
		break;
	case msgYouWin:
		StatusMsg(GetString(idStringYouWin), boldFont);
		break;
	case msgGameOver:
		StatusMsg(GetString(idStringGameOver), stdFont);
		break;
	case msgGrouper:
		StatusMsg(GetString(idStringGrouper), boldFont);
		break;
	}
}


static void StatusMsg(char *msg, FontID font) {
	int i;
	FontID oldFont;
	RectangleType StatusRect = {{StatusX, StatusY}, {StatusDX, StatusDY}};

	WinEraseRectangle(&StatusRect, 0);
	WinDrawRectangleFrame(popupFrame, &StatusRect);
	if (!StrLen(msg))
		return;									// effectively a msg erase

	// show message one letter at a time, reverse-in
	oldFont = FntSetFont(font);
	for (i = 1; i <= StrLen(msg); i++) {
		char *piece = (char *)(msg + StrLen(msg) - i);
		WinDrawChars(piece, i, StatusRect.topLeft.x + 3, StatusRect.topLeft.y);

		SysTaskDelay(TextDelay);
	}
	FntSetFont(oldFont);
}


static void DrawHighScores(void) {
	// thanks to Roger Flores and the reptoids source for most of this code
	Int i, y;
	Char levels[3][16];

	FontID oldFont;
	VoidHand bmpHandle;
	BitmapPtr bmpPtr;

	StrCopy(levels[0], GetString(idStringEasy));
	StrCopy(levels[1], GetString(idStringMedium));
	StrCopy(levels[2], GetString(idStringHard));

	// remember the font	
	oldFont = FntSetFont(boldFont);

	// draw the titles of the columns
	GetString(idStringName);
	WinDrawChars(gString, StrLen(gString), HighScoreNameColumnX,
		FirstHighScoreY - HighScoreHeight);
	GetString(idStringLevel);
	WinDrawChars(gString, StrLen(gString), HighScoreLevelColumnX -
		FntCharsWidth(gString, StrLen(gString)),
		FirstHighScoreY - HighScoreHeight);
	GetString(idStringScore);
	WinDrawChars(gString, StrLen(gString), HighScoreScoreColumnX -
		FntCharsWidth(gString, StrLen(gString)),
		FirstHighScoreY - HighScoreHeight);

	// separator line
	WinDrawLine(HighScoreNameColumnX, FirstHighScoreY - 1,
		HighScoreScoreColumnX, FirstHighScoreY - 1);

	// initialize the "Win!" bitmap
	bmpHandle = DmGetResource(bitmapRsc, idBitmapWin);
	bmpPtr = MemHandleLock(bmpHandle);

	// draw each high score
	for (i = 0; i < MaxHighScores; i++) {
		y = FirstHighScoreY + i * HighScoreHeight;

		// differentiate the last high score by choosing a different font.
		if (i == gPrefs.lastHighScore)
			FntSetFont(boldFont);
		else
			FntSetFont(stdFont);

		// display the score number
		StrIToA(gString, i + 1);
		StrCat(gString, ". ");
		WinDrawChars(gString, StrLen(gString), HighScoreNameColumnX - 
			FntCharsWidth(gString, StrLen(gString)), y);

		if (gPrefs.highScores[i].score > 0) {
			WinDrawChars(gPrefs.highScores[i].name, StrLen(gPrefs.highScores[i].name),
				HighScoreNameColumnX, y);
			WinDrawChars(levels[gPrefs.highScores[i].level],
				StrLen(levels[gPrefs.highScores[i].level]), HighScoreLevelColumnX
				- FntCharsWidth(levels[gPrefs.highScores[i].level],
				StrLen(levels[gPrefs.highScores[i].level])), y);
			StrIToA(gString, gPrefs.highScores[i].score);
			WinDrawChars(gString, StrLen(gString), HighScoreScoreColumnX - 
				FntCharsWidth(gString, StrLen(gString)), y);

			// draw a windicator(tm) if they won the game
			if (gPrefs.highScores[i].fWin)
				WinDrawBitmap(bmpPtr, HighScoreWinColumnX, y);
		}
	}
	MemHandleUnlock(bmpHandle);
	FntSetFont(oldFont);
}


static void ListDrawPatterns(Word itemNum, RectanglePtr bounds, CharPtr *itemsText) {
	RectangleType spot = {{bounds->topLeft.x, bounds->topLeft.y + 1},
								 {ListPatternSize,   ListPatternSize}};

	CALLBACK_PROLOGUE							// see callback.h

	WinSetPattern(cardPattern[itemNum + PatternGray]);
	WinFillRectangle(&spot, 0);
	WinDrawRectangleFrame(rectangleFrame, &spot);

	WinDrawChars(itemsText[itemNum], StrLen(itemsText[itemNum]), bounds->topLeft.x + 12, bounds->topLeft.y);

	CALLBACK_EPILOGUE							// see callback.h
}