/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm_common.h"
#include "gutenpalmRsc.h"
#include "gutenpalm.h"
#include "gutenpalm_prefs.h"
#include "gutenpalm_cache.h"
#include "gutenpalm_bookmarks.h"



/* Function prototypes */
static UInt16   StartApplication(void);
static void     StopApplication(void);
static void     EventLoop(void);
static Boolean  ApplicationHandleEvent(EventPtr);


/* GutenPalm globals */
MemHandle       gDoclistHandle;
UInt16          gNumDocuments;
Document        gCurrentDoc;
UInt16          gCurrentForm;
GPlmPrefs       gPrefs;
GPlmCacheState  *gCache;


UInt32
PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
  UInt16 err;

  if (cmd == sysAppLaunchCmdNormalLaunch)
    {
      err = StartApplication();
      if (err)
        return err;

      EventLoop();
      StopApplication();
    }
  else
    {
      return sysErrParamErr;
    }

  return 0;
}


static UInt16
StartApplication(void)
{
  UInt16        prefBufSize;
  DmSearchStateType state;
  Err           theErr;
  UInt16        card;
  UInt16        currDBnum = 0;
  UInt16        currDocListSize = 10;
  LocalID       currDB = 0;
  DocList       *doclist;
  Char          tmpName[dmDBNameLength];
  Int16         resumeIndex = -1; /* For resuming an already open doc */


  /* Get GutenPalm preferences */
  MemSet(&gPrefs, sizeof(GPlmPrefs), 0);
  prefBufSize = sizeof(GPlmPrefs);
  if (PrefGetAppPreferences(id_to_int(GPLM_CREATOR_ID), PREFS_MAIN, &gPrefs,
                            &prefBufSize, true) == noPreferenceFound)
    {
      gPrefs.alwaysRememberPos = true;
    }

  /* Get GutenPalm zTXT cache state, or create if nonexistant */
  prefBufSize = sizeof(GPlmCacheState) * gPrefs.numCachedDBs;
  if (prefBufSize == 0)
    {
      /* Cache does not exist */
      gCache = NULL;
    }
  else
    {
      /* Cache entries exist */
      gCache = (GPlmCacheState *)MemPtrNew(prefBufSize);
      PrefGetAppPreferences(id_to_int(GPLM_CREATOR_ID), PREFS_CACHE, gCache,
                            &prefBufSize, true);
    }

  /* Clear gCurrentDoc before first usage */
  MemSet(&gCurrentDoc, sizeof(Document), 0);

  /* Initialize the DocList structure */
  gDoclistHandle = MemHandleNew(MIN_DOCLIST_SIZE);
  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  doclist->numDocuments = 0;

  /*  Grab a listing of all available zTXT documents */
  theErr = DmGetNextDatabaseByTypeCreator(true, &state,
                              id_to_int(ZTXT_TYPE_ID),
                              id_to_int(GPLM_CREATOR_ID), false, &card,
                              &currDB);
  while ((theErr != dmErrCantFind) && currDB)
    {
      /* Grow the DocList if out of room */
      if (doclist->numDocuments >= currDocListSize)
        {
          MemHandleUnlock(gDoclistHandle);
          MemHandleResize(gDoclistHandle, MIN_DOCLIST_SIZE +
                          (sizeof(struct DocEntryType) * currDocListSize));
          currDocListSize += 10;
          doclist = (DocList *)MemHandleLock(gDoclistHandle);
        }
      DmDatabaseInfo(card, currDB, tmpName, NULL, NULL, NULL, NULL, NULL,
                     NULL, NULL, NULL, NULL, NULL);

      /* Find location to put new DB in alphabetical order */
      for (currDBnum = 0; ((currDBnum < doclist->numDocuments) &&
                           (StrNCompare(doclist->documents[currDBnum].name,
                                        tmpName, dmDBNameLength) < 0));
           currDBnum++);

      if (currDBnum < doclist->numDocuments)
        {
          /* Inserting into middle of list, so adjust entries */
          MemMove(&(doclist->documents[currDBnum+1]),
                  &(doclist->documents[currDBnum]),
                  sizeof(struct DocEntryType) * (doclist->numDocuments -
                                                 currDBnum));
        }
      /* Store info about the document */
      StrNCopy(doclist->documents[currDBnum].name, tmpName, dmDBNameLength);
      doclist->documents[currDBnum].cache =
                     cacheExists(doclist->documents[currDBnum].name);
      doclist->numDocuments++;

      /* Get the next DB */
      theErr = DmGetNextDatabaseByTypeCreator(false, &state,
                                  id_to_int(ZTXT_TYPE_ID),
                                  id_to_int(GPLM_CREATOR_ID), false, &card,
                                  &currDB);
    }

  gNumDocuments = doclist->numDocuments;

  /* If there is an open document to return to, do so */
  if (gPrefs.resume)
    {
      for (resumeIndex = 0; ((resumeIndex < doclist->numDocuments) &&
                             (StrNCompare(doclist->documents[resumeIndex].name,
                                          gPrefs.resumeDBName,
                                          dmDBNameLength) != 0));
           resumeIndex++);

      /* Doc to resume no longer exists so delete cache and bookmarks */
      if (resumeIndex == doclist->numDocuments)
        {
          cacheDelete(gPrefs.resumeDBName);
          gPrefs.resume = false;
          gPrefs.resumeDBName[0] = '\0';

          /* Proceed normally */
          FrmGotoForm(MainForm);
          MemHandleUnlock(gDoclistHandle);
          return 0;
        }
      gCurrentDoc.dbIndex = resumeIndex;
      FrmGotoForm(DocViewForm);
    }
  else
    FrmGotoForm(MainForm);

  MemHandleUnlock(gDoclistHandle);
  return 0;
}


/* The main event loop */
static void
EventLoop(void)
{
  UInt16        err;
  EventType     e;

  do {
    EvtGetEvent(&e, evtWaitForever);
    if (! SysHandleEvent (&e))
      if (! MenuHandleEvent (NULL, &e, &err))
        if (! ApplicationHandleEvent (&e))
          FrmDispatchEvent (&e);
  } while (e.eType != appStopEvent);
}


static Boolean
ApplicationHandleEvent(EventPtr e)
{
  FormPtr       frm;
  UInt16        formId;
  Boolean       handled = false;

  if (e->eType == frmLoadEvent)
    {
      formId = e->data.frmLoad.formID;
      frm = FrmInitForm(formId);
      FrmSetActiveForm(frm);

      switch(formId)
        {
          case MainForm:
            FrmSetEventHandler(frm, MainFormHandleEvent);
            handled = true;
            break;
          case DocViewForm:
            FrmSetEventHandler(frm, DocViewFormHandleEvent);
            handled = true;
            break;
        }
    }

  return handled;
}


/* Close up shop.  If temp DB or input DB are still around, get rid of them */
static void
StopApplication(void)
{
  MemHandleFree(gDoclistHandle);

  /* If in DocViewForm, then save current location */
  if (gCurrentForm == DocViewForm)
    {
      gPrefs.resume = true;
      StrNCopy(gPrefs.resumeDBName, gCurrentDoc.dbName, dmDBNameLength);
      bmrkAdd(BMRK_RESUME, gCurrentDoc.curPos);

      /* Close temp database, but don't delete it */
      MemHandleUnlock(gCurrentDoc.curRecordHandle);
      DmReleaseRecord(gCurrentDoc.tempRef, gCurrentDoc.curRecordIndex, false);
      DmCloseDatabase(gCurrentDoc.tempRef);
    }
  else
    gPrefs.resume = false;

  /* Close the bookmark record if it's open */
  if (gCurrentDoc.bmrkRecordHandle)
    {
      MemHandleUnlock(gCurrentDoc.bmrkRecordHandle);
      DmReleaseRecord(gCurrentDoc.dbRef, gCurrentDoc.bookmarkRecord, false);
    }

  /* Close the zTXT database if it's open */
  if (gCurrentDoc.dbRef)
    DmCloseDatabase(gCurrentDoc.dbRef);

  /* Save GutenPalm preferences and state */
  PrefSetAppPreferences(id_to_int(GPLM_CREATOR_ID), PREFS_MAIN, GPLM_VERSION,
                        &gPrefs, sizeof(GPlmPrefs), true);

  /* Save GutenPalm cache state if cached data exists */
  if (gPrefs.numCachedDBs > 0)
    PrefSetAppPreferences(id_to_int(GPLM_CREATOR_ID), PREFS_CACHE,
                          GPLM_VERSION, gCache,
                          sizeof(GPlmCacheState) * gPrefs.numCachedDBs, true);

  FrmSaveAllForms();
  FrmCloseAllForms();
}
