/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm_common.h"
#include "gutenpalm.h"
#include "gutenpalm_bookmarks.h"



/* Add a bookmark to the bookmark list in a currently open zTXT.
 * This function assumes that the zTXT DB is open and that appropriate
 * values in gCurrentDoc have been set */
void
bmrkAdd(Char *newTitle, UInt32 newOffset)
{
  UInt16        newRecordIndex;
  MemHandle     recHandle;
  zTXT_record0  *record0;
  UInt16        i;
  GPlmMark      *mark;


  /* First check to make sure there isn't a bookmark with the same name */
  for (i = 0; i < gCurrentDoc.numBookmarks; i++)
    {
      mark = ((GPlmMark *)gCurrentDoc.bmrkRecordPtr) + i;
      if (StrNCompare(mark->title, newTitle, MAX_BMRK_LENGTH) == 0)
        return;
    }

  /* Open record #0 */
  recHandle = DmGetRecord(gCurrentDoc.dbRef, 0);
  record0 = (zTXT_record0 *)MemHandleLock(recHandle);

  if (gCurrentDoc.bookmarkRecord == 0)
    {
      newRecordIndex = dmMaxRecordIndex;
      gCurrentDoc.bmrkRecordHandle = DmNewRecord(gCurrentDoc.dbRef,
                                                 &newRecordIndex,
                                                 sizeof(GPlmMark));
      gCurrentDoc.bookmarkRecord = newRecordIndex;
      gCurrentDoc.numBookmarks = 0;
      DmWrite(record0, (UInt32)(&(record0->bookmarkRecord)) -
              (UInt32)record0, &gCurrentDoc.bookmarkRecord, sizeof(UInt16));
    }
  else
    {
      MemHandleUnlock(gCurrentDoc.bmrkRecordHandle);
      gCurrentDoc.bmrkRecordHandle = DmResizeRecord(gCurrentDoc.dbRef,
                          gCurrentDoc.bookmarkRecord,
                          sizeof(GPlmMark) * (gCurrentDoc.numBookmarks + 1));
    }
  gCurrentDoc.bmrkRecordPtr = MemHandleLock(gCurrentDoc.bmrkRecordHandle);

  DmWrite(gCurrentDoc.bmrkRecordPtr,
          sizeof(GPlmMark) * gCurrentDoc.numBookmarks,
          &newOffset, sizeof(UInt32));
  DmWrite(gCurrentDoc.bmrkRecordPtr,
          (sizeof(GPlmMark) * gCurrentDoc.numBookmarks) + sizeof(UInt32),
          newTitle, MAX_BMRK_LENGTH);

  gCurrentDoc.numBookmarks++;
  DmWrite(record0, (UInt32)(&(record0->numBookmarks)) -
          (UInt32)record0, &gCurrentDoc.numBookmarks, sizeof(UInt16));

  /* Release record #0 */
  MemHandleUnlock(recHandle);
  DmReleaseRecord(gCurrentDoc.dbRef, 0, true);
}


/* Delete a single bookmark and remove the bookmark record if empty */
void
bmrkDeleteMark(Char *title)
{
  UInt16        i;
  GPlmMark      *mark;
  MemHandle     recHandle;
  zTXT_record0  *record0;

  for (i = 0; i < gCurrentDoc.numBookmarks; i++)
    {
      mark = (GPlmMark *)gCurrentDoc.bmrkRecordPtr;
      mark += i;
      if (StrNCompare(mark->title, title, MAX_BMRK_LENGTH) == 0)
        {
          /* Bookmark found */
          if (i == gCurrentDoc.numBookmarks - 1)
            {
              /* Bookmark is the end entry */
              MemHandleUnlock(gCurrentDoc.bmrkRecordHandle);
              gCurrentDoc.bmrkRecordHandle = DmResizeRecord(gCurrentDoc.dbRef,
                          gCurrentDoc.bookmarkRecord,
                          sizeof(GPlmMark) * (gCurrentDoc.numBookmarks - 1));
              gCurrentDoc.bmrkRecordPtr =
                MemHandleLock(gCurrentDoc.bmrkRecordHandle);
            }
          else
            {
              /* Bookmark is amidst the bunch */
              /* I'm not sure if this is safe.  MemMove allows overlapping
               * areas, but the docs for DmWrite don't say if it does.  It
               * seems to work in my tests, however */
              DmWrite(gCurrentDoc.bmrkRecordPtr, sizeof(GPlmMark) * i,
                      (mark + 1), sizeof(GPlmMark) *
                      (gCurrentDoc.numBookmarks - i - 1));
                      
            }
          gCurrentDoc.numBookmarks--;

          /* Open record #0 */
          recHandle = DmGetRecord(gCurrentDoc.dbRef, 0);
          record0 = (zTXT_record0 *)MemHandleLock(recHandle);

          DmWrite(record0, (UInt32)(&(record0->numBookmarks)) -
                  (UInt32)record0, &gCurrentDoc.numBookmarks, sizeof(UInt16));

          if (gCurrentDoc.numBookmarks == 0)
            {
              MemHandleUnlock(gCurrentDoc.bmrkRecordHandle);
              DmReleaseRecord(gCurrentDoc.dbRef, gCurrentDoc.bookmarkRecord,
                              true);
              DmRemoveRecord(gCurrentDoc.dbRef, gCurrentDoc.bookmarkRecord);
              gCurrentDoc.bookmarkRecord = 0;
              gCurrentDoc.bmrkRecordHandle = 0;
              gCurrentDoc.bmrkRecordPtr = 0;
              DmWrite(record0, (UInt32)(&(record0->bookmarkRecord)) -
                      (UInt32)record0, &gCurrentDoc.bookmarkRecord,
                      sizeof(UInt16));
            }

          /* Release record #0 */
          MemHandleUnlock(recHandle);
          DmReleaseRecord(gCurrentDoc.dbRef, 0, true);

          return;
        }
    }
}
