/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm.h"
#include "gutenpalm_prefs.h"
#include "gutenpalm_cache.h"



/* Returns true if zTXT DB has an uncompressed cache available */
Boolean
cacheExists(Char *dbName)
{
  UInt16 i;

  if (gPrefs.numCachedDBs > 0)
    {
      for (i = 0; i < gPrefs.numCachedDBs; i++)
        if (StrNCompare(gCache[i].zTxtDB, dbName, dmDBNameLength) == 0)
          return true;
    }
  return false;
}


/* Returns LocalID of cache if it exists */
Char *
cacheGet(Char *dbName)
{
  UInt16 i;

  if (gPrefs.numCachedDBs > 0)
    {
      for (i = 0; i < gPrefs.numCachedDBs; i++)
        if (StrNCompare(gCache[i].zTxtDB, dbName, dmDBNameLength) == 0)
          return gCache[i].cache;
    }
  return NULL;
}


/* Adds a cache to the cache list */
void
cacheAdd(Char *cacheName, Char *zTxtName)
{
  GPlmCacheState *temp;
  UInt16         x;
  DocList        *doclist;

  /* If there is no cache, make one */
  if (gCache !=NULL)
    {
      temp = MemPtrNew((gPrefs.numCachedDBs + 1) * sizeof(GPlmCacheState));
      MemMove(temp, gCache, gPrefs.numCachedDBs * sizeof(GPlmCacheState));
      MemPtrFree(gCache);
      gCache = temp;
    }
  else
    gCache = MemPtrNew(sizeof(GPlmCacheState));

  /* Add the new cache to the cache list */
  StrNCopy(gCache[gPrefs.numCachedDBs].cache, cacheName, dmDBNameLength);
  StrNCopy(gCache[gPrefs.numCachedDBs].zTxtDB, zTxtName, dmDBNameLength);

  gPrefs.numCachedDBs++;

  /* Update the cache info in DocList */
  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  for (x = 0; x < doclist->numDocuments; x++)
    {
      if (StrNCompare(doclist->documents[x].name, zTxtName,
                      dmDBNameLength) == 0)
        {
          doclist->documents[x].cache = true;
          break;
        }
    }
  MemHandleUnlock(gDoclistHandle);
}


/* Delete a cache DB associated with a zTXT DB (dbName), if such exists */
void
cacheDelete(Char *dbName)
{
  UInt16        i;
  UInt16        x;
  DocList       *doclist;

  if (gPrefs.numCachedDBs > 0)
    {
      for (i = 0; i < gPrefs.numCachedDBs; i++)
        if (StrNCompare(gCache[i].zTxtDB, dbName, dmDBNameLength) == 0)
          {
            DmDeleteDatabase(0, DmFindDatabase(0, gCache[i].cache));

            /* Remove the cacheDB entry from the cache listing */
            if (i == gPrefs.numCachedDBs - 1)
              gPrefs.numCachedDBs--;
            else
              {
                MemMove(&(gCache[i]), &(gCache[i+1]),
                        (gPrefs.numCachedDBs - i - 1) *
                        sizeof(GPlmCacheState));
                gPrefs.numCachedDBs--;
              }

            /* Update the cache info in DocList */
            doclist = (DocList *)MemHandleLock(gDoclistHandle);
            for (x = 0; x < doclist->numDocuments; x++)
              {
                if (StrNCompare(doclist->documents[x].name, dbName,
                                dmDBNameLength) == 0)
                  {
                    doclist->documents[x].cache = false;
                    break;
                  }
              }
            MemHandleUnlock(gDoclistHandle);

            return;
          }
    }
}


/* Removes all entries from the cache list */
void
cacheClear(void)
{
  UInt16        i;
  DocList       *doclist;

  for (i = 0; i < gPrefs.numCachedDBs; i++)
    DmDeleteDatabase(0, DmFindDatabase(0, gCache[i].cache));
  gPrefs.numCachedDBs = 0;
  if (gCache)
    {
      MemPtrFree(gCache);
      gCache = NULL;
    }

  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  for (i = 0; i < doclist->numDocuments; i++)
    doclist->documents[i].cache = false;
  MemHandleUnlock(gDoclistHandle);
}
