/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm_common.h"
#include "zlib-palmos.h"
#include "gutenpalmRsc.h"
#include "gutenpalm.h"
#include "gutenpalm_prefs.h"
#include "gutenpalm_cache.h"
#include "gutenpalm_bookmarks.h"



/* Document form function prototypes */
static void     DocViewFormInit(void);
static void     DocViewCleanup(void);
static void     DecompressDocument(void);
static void     SkipPGLicense(void);
static void     DisplayCurrentPage(void);
static void     DisplayNextPage(void);
static void     DisplayPrevPage(void);
static void     UpdateLineBuffer(Char *, UInt16);
static void     DoLocationForm(void);
static void     JumpToBookmark(UInt16);
static void     DoAddBookmarkForm(void);
static void     DoDeleteBookmarksForm(void);
static void     BmrkDrawFunc(UInt16, RectanglePtr, Char *[]);
static Boolean  DelBmrkFormHandleEvent(EventPtr);



/* Event handler for the document display form */
Boolean
DocViewFormHandleEvent(EventPtr e)
{
  Boolean       handled = false;

  switch (e->eType)
    {
      case frmOpenEvent:
        DocViewFormInit();
        handled = true;
        break;

      case popSelectEvent:
        if (e->data.popSelect.controlID == BmrkPopup)
          {
            if (e->data.popSelect.selection < gCurrentDoc.numBookmarks)
              {
                JumpToBookmark(e->data.popSelect.selection);
              }
            else
              {
                /* User selected "Add bookmark" from list */
                DoAddBookmarkForm();
              }
            handled = true;
          }
        break;

      case keyDownEvent:             
        if (e->data.keyDown.chr == pageUpChr)
          {
            DisplayPrevPage();
            handled = true;
          }
        else if (e->data.keyDown.chr == pageDownChr)
          {
            DisplayNextPage();
            handled = true;
          }
        break;

      case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID)
          {
            case DoneButton:
              DocViewCleanup();
              FrmGotoForm(MainForm);
              handled = true;
              break;
          }
        break;

      case fldEnterEvent:
        switch(e->data.fldEnter.fieldID)
          {
            case PercentageField:
              DoLocationForm();
              handled = true;
              break;
          }
        break;

      case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID)
          {
            /* Add a bookmark */
            case DocViewFormMenu_AddBmrk:
              DoAddBookmarkForm();
              break;

            /* Display the form to delete one or more bookmarks */
            case DocViewFormMenu_DelBmrk:
              DoDeleteBookmarksForm();
              break;

            /* Display information about GutenPalm */
            case DocViewFormMenu_About:
              FrmAlert(About1Alert);
              FrmAlert(About2Alert);
              break;
          }
        handled = true;
        break;

      default:
        break;
    }

  return handled;
}


/* Cleanup form.  Close DBs.  Delete temporary database if option not set. */
static void
DocViewCleanup(void)
{
  /* Close the open record and the DB */
  MemHandleUnlock(gCurrentDoc.curRecordHandle);
  DmReleaseRecord(gCurrentDoc.tempRef, gCurrentDoc.curRecordIndex, false);
  DmCloseDatabase(gCurrentDoc.tempRef);

  if (!gPrefs.keepCache)
    {
      /* Delete the temporary DB.  It sucks up space */
      cacheDelete(gCurrentDoc.dbName);
    }

  if (gPrefs.alwaysRememberPos)
    bmrkAdd(BMRK_RESUME, gCurrentDoc.curPos);

  /* Close the bookmark record and the input database */
  if (gCurrentDoc.numBookmarks > 0)
    {
      MemHandleUnlock(gCurrentDoc.bmrkRecordHandle);
      DmReleaseRecord(gCurrentDoc.dbRef, gCurrentDoc.bookmarkRecord, true);
    }
  DmCloseDatabase(gCurrentDoc.dbRef);

  gCurrentForm = MainForm;

  /* Clear gCurrentDoc so other code doesn't get confused later */
  MemSet(&gCurrentDoc, sizeof(Document), 0);
}


/* Initialize the DocView form.  Initialize the decompression stream and
 * necessary buffers
 */
static void
DocViewFormInit(void)
{
  static RectangleType progressRec = {{70,80}, {30, 11}};
  FormPtr       frm;
  ListPtr       list;
  UInt16        initStatus;
  DocList       *doclist;
  MemHandle     recHandle;
  zTXT_record0  *record0;
  UInt32        numRecords;
  GPlmMark      *resumeMark;


  /* Set current form indicator */
  gCurrentForm =  DocViewForm;

  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  StrNCopy(gCurrentDoc.dbName, doclist->documents[gCurrentDoc.dbIndex].name,
           dmDBNameLength);
  MemHandleUnlock(gDoclistHandle);
  gCurrentDoc.dbID = DmFindDatabase(0, gCurrentDoc.dbName);

  /* Open the zTXT database */
  gCurrentDoc.dbRef = DmOpenDatabase(0, gCurrentDoc.dbID, dmModeReadWrite);

  /* Open record #0 */
  recHandle = DmGetRecord(gCurrentDoc.dbRef, 0);
  record0 = (zTXT_record0 *)MemHandleLock(recHandle);

  /* Get the zTXT info from record #0 */
  gCurrentDoc.version = record0->version;
  gCurrentDoc.size = record0->size;
  gCurrentDoc.numRecords = record0->numRecords;
  gCurrentDoc.recordSize = record0->recordSize;
  gCurrentDoc.numBookmarks = record0->numBookmarks;
  gCurrentDoc.bookmarkRecord = record0->bookmarkRecord;

  /* Don't need record #0 anymore */
  MemHandleUnlock(recHandle);
  DmReleaseRecord(gCurrentDoc.dbRef, 0, false);

  if (cacheExists(gCurrentDoc.dbName))
    {
      /* Use cached data instead */
      StrNCopy(gCurrentDoc.tempName, cacheGet(gCurrentDoc.dbName),
               dmDBNameLength);
      gCurrentDoc.tempID = DmFindDatabase(0, gCurrentDoc.tempName);
      DmDatabaseSize(0, gCurrentDoc.tempID, &numRecords, NULL, NULL);
      gCurrentDoc.numTmpRecords = (UInt16)numRecords;
      gCurrentDoc.tempRef = DmOpenDatabase(0, gCurrentDoc.tempID,
                                           dmModeReadOnly);
    }
  else
    {
      /* Cache does not exist, or cache data not valid
       * Load via normal means
       */

      /* Initialize some stuff before decompression */
      gCurrentDoc.curRecordIndex = 1;
      gCurrentDoc.curRecordHandle = DmGetRecord(gCurrentDoc.dbRef, 1);
      gCurrentDoc.curRecordPtr = MemHandleLock(gCurrentDoc.curRecordHandle);

      /* Allocate the decompression buffer */
      gCurrentDoc.zBuffer = MemPtrNew(gCurrentDoc.recordSize);

      /* These values must be set before using Zlib */
      gCurrentDoc.zstream.zalloc = Z_NULL;
      gCurrentDoc.zstream.zfree = Z_NULL;
      gCurrentDoc.zstream.opaque = Z_NULL;
      gCurrentDoc.zstream.next_in = gCurrentDoc.curRecordPtr;
      gCurrentDoc.zstream.next_out = gCurrentDoc.zBuffer;
      gCurrentDoc.zstream.avail_out = gCurrentDoc.recordSize;
      if (gCurrentDoc.numRecords == 1)
        gCurrentDoc.zstream.avail_in = gCurrentDoc.size;
      else
        gCurrentDoc.zstream.avail_in = gCurrentDoc.recordSize;

      /* With large documents this can take a while.
       * Let the user know what's happening */
      WinDrawChars("Decompressing...", 16, 45, 60);
      WinEraseRectangle(&progressRec, 0);

      /* Initialize the decompression stream */
      initStatus = inflateInit2(&gCurrentDoc.zstream, MAXWBITS);

      /* Decompress the entire document
       *  Other low-memory methods may get added later,
       *  but it will be hard to allow foreward/backward movement
       */
      DecompressDocument();

      /* Close the decompression stream and free the buffer */
      initStatus = inflateEnd(&gCurrentDoc.zstream);
      MemPtrFree(gCurrentDoc.zBuffer);

      /* Stick the uncompressed data into the cache */
      cacheAdd(gCurrentDoc.tempName, gCurrentDoc.dbName);
    }

  /* Open bookmarks record if there is one */
  if (gCurrentDoc.numBookmarks > 0)
    {
      gCurrentDoc.bmrkRecordHandle = DmGetRecord(gCurrentDoc.dbRef,
                                                 gCurrentDoc.bookmarkRecord);
      gCurrentDoc.bmrkRecordPtr = MemHandleLock(gCurrentDoc.bmrkRecordHandle);
    }

  /* See if there is a resume bookmark */
  if  (gCurrentDoc.numBookmarks > 0)
    {
      resumeMark = (GPlmMark *)gCurrentDoc.bmrkRecordPtr;
      resumeMark += (gCurrentDoc.numBookmarks - 1);
      if (StrNCompare(resumeMark->title, BMRK_RESUME,
                      MAX_BMRK_LENGTH) == 0)
        gPrefs.resume =  true;
    }
  if (gPrefs.resume)
    {
      /* Resume last location in document */
      resumeMark = (GPlmMark *)gCurrentDoc.bmrkRecordPtr;
      resumeMark += (gCurrentDoc.numBookmarks - 1);

      gCurrentDoc.curRecordIndex = resumeMark->offset / gCurrentDoc.recordSize;
      gCurrentDoc.curRecordHandle = DmGetRecord(gCurrentDoc.tempRef,
                                                gCurrentDoc.curRecordIndex);
      gCurrentDoc.curRecordPtr = MemHandleLock(gCurrentDoc.curRecordHandle);
      gCurrentDoc.curPage.offset = resumeMark->offset % gCurrentDoc.recordSize;
      gCurrentDoc.curPage.recIndex = gCurrentDoc.curRecordIndex;
      bmrkDeleteMark(BMRK_RESUME);

      if ((gCurrentDoc.curRecordIndex == 0) &&
          (gCurrentDoc.curPage.offset == 0))
        gCurrentDoc.startOfDoc = true;
      else
        gCurrentDoc.startOfDoc = false;

      gPrefs.resume = false;
    }
  else
    {
      /* Display the first page of the document */
      gCurrentDoc.curRecordIndex = 0;
      gCurrentDoc.curRecordHandle = DmGetRecord(gCurrentDoc.tempRef, 0);
      gCurrentDoc.curRecordPtr = MemHandleLock(gCurrentDoc.curRecordHandle);
      gCurrentDoc.startOfDoc = true;

      /* Skip the Project Gutenberg license if option is set */
      if ((gPrefs.skipLicense) && (gCurrentDoc.size > 16384))
        SkipPGLicense();
    }

  /* Display the form and set up the bookmark list */
  frm = FrmGetActiveForm();
  list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, BmrkList));
  /* Plus 1 for the entry to add a bookmark */
  LstSetListChoices(list, NULL, gCurrentDoc.numBookmarks + 1);
  LstSetDrawFunction(list, (ListDrawDataFuncPtr) BmrkDrawFunc);
  FrmDrawForm(frm);

  DisplayCurrentPage();
}


/* Skip Project Gutenberg License */
static void
SkipPGLicense(void)
{
  UInt16        x;
  UInt16        y;
  Boolean       foundLicense = false;

  /* End of license should be at least this far into document */
  OpenTmpRecord((UInt16)(8192 / gCurrentDoc.recordSize));
  y = 8192 % gCurrentDoc.recordSize;
  x = 0;

  /* Search at most 8192 characters for the license */
  while (x < 8192)
    {
      if  (StrNCompare((Char *)(gCurrentDoc.curRecordPtr + y),
                       "*END*", 5) == 0)
        {
          /* Found the end of license marker */
          x = 0;
          while ((*((Char *)(gCurrentDoc.curRecordPtr + y)) != '\n') &&
                 (x < 160))
            {
              x++;
              y++;
              if (y >= gCurrentDoc.recordSize)
                {
                  OpenTmpRecord(gCurrentDoc.curRecordIndex + 1);
                  y = 0;
                }
            }
          y++;
          if (y >= gCurrentDoc.recordSize)
            {
              OpenTmpRecord(gCurrentDoc.curRecordIndex + 1);
              y = 0;
            }
          foundLicense = true;
          x = 8192;
          break;
        }
      else
        {
          /* No license yet, move forward */
          x++;
          y++;
          if (y + 5 >= gCurrentDoc.recordSize)
            {
              OpenTmpRecord(gCurrentDoc.curRecordIndex + 1);
              y = 0;
            }
        }
    }
  if (foundLicense)
    {
      /* Found the end of the license */
      gCurrentDoc.curPage.offset = y;
      gCurrentDoc.curPage.recIndex = gCurrentDoc.curRecordIndex;
      gCurrentDoc.startOfDoc = false;
    }
  else
    {
      /* Couldn't find license, so go back to beginning */
      OpenTmpRecord(0);
      gCurrentDoc.curPage.offset = 0;
      gCurrentDoc.curPage.recIndex = 0;
    }
}


/* Display one page from the current position in the document */
static void
DisplayCurrentPage()
{
  static RectangleType eraseRec = {{0, 0}, {160, 144}};
  UInt16        curCharOffset;
  Char          curLine[101];
  UInt16        numChars;
  UInt16        y = 0;
  UInt16        a = 0;
  UInt16        max_y;


  WinEraseRectangle(&eraseRec, 0);

  OpenTmpRecord(gCurrentDoc.curPage.recIndex);
  curCharOffset = gCurrentDoc.curPage.offset;

  gCurrentDoc.curPos = ((((UInt32)gCurrentDoc.curPage.recIndex) *
                         ((UInt32)gCurrentDoc.recordSize)) +
                        ((UInt32)gCurrentDoc.curPage.offset));

  max_y = 145 - FntLineHeight();
  while ((y < max_y) && (!gCurrentDoc.endOfDoc))
    {
      UpdateLineBuffer(curLine, curCharOffset);

      numChars = a = FntWordWrap(curLine, 160);
      if (curLine[numChars-1] == '\n')
        a--;
      WinDrawChars(curLine, a, 0, y);

      y += FntLineHeight();
      curCharOffset += numChars;
      if (curCharOffset >= gCurrentDoc.recordSize)
        {
          /* Past the end of the current record */
          OpenTmpRecord(gCurrentDoc.curRecordIndex + 1);
          curCharOffset -= gCurrentDoc.recordSize;
        }

      /* If last record AND offset == end of document */
      if ((gCurrentDoc.curRecordIndex == gCurrentDoc.numTmpRecords - 1) &&
          (curCharOffset >= (gCurrentDoc.size % gCurrentDoc.recordSize)))
        {
          /* All done */
          gCurrentDoc.endOfDoc = true;
        }
    }

  gCurrentDoc.nextPage.offset = curCharOffset;
  gCurrentDoc.nextPage.recIndex = gCurrentDoc.curRecordIndex;
  OpenTmpRecord(gCurrentDoc.curPage.recIndex);

  if (!gCurrentDoc.endOfDoc)
    {
      StrPrintF(curLine, "%d%%",
                (UInt16)((gCurrentDoc.curPos * 100) / gCurrentDoc.size));
    }
  else
    StrCopy(curLine, "100%");
  SetFieldTextAndDraw(PercentageField, curLine);
}


/* Updates the line buffer, moving to a new record if necessary */
static void
UpdateLineBuffer(Char *curLine, UInt16 curOffset)
{
  UInt16        recEnd;
  UInt32        maxCopy;
  UInt16        record;


  /* Figure out how big the current record is */
  if (gCurrentDoc.curRecordIndex < gCurrentDoc.numTmpRecords - 1)
    recEnd = gCurrentDoc.recordSize;
  else
    recEnd = (gCurrentDoc.size - ((gCurrentDoc.numTmpRecords - 1) *
                                  gCurrentDoc.recordSize));

  /* Text left in the current record */
  if (curOffset < recEnd)
    {
      maxCopy = recEnd - curOffset;
      if (maxCopy > 100)
        maxCopy = 100;
      StrNCopy(curLine, gCurrentDoc.curRecordPtr + curOffset,
               (UInt16)maxCopy);
      curLine[maxCopy] = '\0';

      /* Ran into the record boundry.  Get text from the next record */
      if ((maxCopy < 100) &&
          (gCurrentDoc.curRecordIndex < (gCurrentDoc.numTmpRecords - 1)))
        {
          record = gCurrentDoc.curRecordIndex;
          OpenTmpRecord(record + 1);
          StrNCopy(&(curLine[maxCopy]), gCurrentDoc.curRecordPtr,
                   100 - maxCopy);
          curLine[100] = '\0';
          OpenTmpRecord(record);
        }
    }
}


/* Display next page in document */
static void
DisplayNextPage()
{
  if (!gCurrentDoc.endOfDoc)
    {
      gCurrentDoc.startOfDoc = false;
      gCurrentDoc.curPage.offset = gCurrentDoc.nextPage.offset;
      gCurrentDoc.curPage.recIndex = gCurrentDoc.nextPage.recIndex;
      DisplayCurrentPage();
    }
}


/* Display previous page in document */
static void
DisplayPrevPage()
{
  UInt16	numLines = 13;
  UInt16	charOffset;
  UInt16	origOffset;
  Char		*buff;
  Char		*begin;

  if (!gCurrentDoc.startOfDoc)
    {
      gCurrentDoc.endOfDoc = false;
      OpenTmpRecord(gCurrentDoc.curPage.recIndex);
      charOffset = origOffset = gCurrentDoc.curPage.offset;

      /* Find out how many characters to go backwards.
       * This check is okay because we can't have an offset of zero and be
       * at the beginning of the document AND reach this point
       */
      if (gCurrentDoc.curPage.offset != 0)
        {
          FntWordWrapReverseNLines(gCurrentDoc.curRecordPtr, 160,
                                   &numLines, &charOffset);
        }
      else
        {
          gCurrentDoc.curPage.recIndex--;
          OpenTmpRecord(gCurrentDoc.curPage.recIndex);
          charOffset = gCurrentDoc.curPage.offset = gCurrentDoc.recordSize - 1;
          FntWordWrapReverseNLines(gCurrentDoc.curRecordPtr, 160,
                                   &numLines, &charOffset);
        }

      if (charOffset != 0)
        {
          /* In the middle of text.  Just page backwards */
          gCurrentDoc.curPage.offset = charOffset;
        }
      else if (gCurrentDoc.curPage.recIndex == 0)
        {
          /* Beginning of document */
          gCurrentDoc.curPage.offset = 0;
          gCurrentDoc.startOfDoc = true;
        }
      else
        {
          /* Wrapped to beginning of a record */

          /* Make a buffer to hold text from two records */
          buff = (Char *)MemPtrNew(801);

          /* Get some text from previous record */
          OpenTmpRecord(gCurrentDoc.curPage.recIndex - 1);
          begin = gCurrentDoc.curRecordPtr + (gCurrentDoc.recordSize -
                                         (800 - gCurrentDoc.curPage.offset));
          StrNCopy(buff, begin, 800 - gCurrentDoc.curPage.offset);
          /* Get remaining text from current record */
          OpenTmpRecord(gCurrentDoc.curPage.recIndex);
          StrNCopy(buff + (800 - gCurrentDoc.curPage.offset),
                   gCurrentDoc.curRecordPtr, gCurrentDoc.curPage.offset + 1);

          buff[800]='\0';
          numLines = 13;
          charOffset = 799;
          /* Rewrap on new buffer to get total amount moved */
          FntWordWrapReverseNLines(buff, 160, &numLines, &charOffset);
          MemPtrFree(buff);

          if (charOffset > 800-gCurrentDoc.curPage.offset)
            {
              /* Still in current record.
               * This happens sometimes if the word at the beginning of
               * the record was really long
               */
              gCurrentDoc.curPage.offset = (charOffset -
                                            (800-gCurrentDoc.curPage.offset));
            }
          else if (charOffset < 800-gCurrentDoc.curPage.offset)
            {
              /* Wrapped into previous record */
              gCurrentDoc.curPage.recIndex--;
              OpenTmpRecord(gCurrentDoc.curPage.recIndex);
              gCurrentDoc.curPage.offset = (gCurrentDoc.recordSize -
                                            (800 - gCurrentDoc.curPage.offset -
                                             charOffset));
            }
          else
            {
              /* Wrapped exactly to the beginning of a record */
              gCurrentDoc.curPage.offset = 0;
            }
        }

      DisplayCurrentPage();
    }
}


/* Displays the form to select a new location */
static void
DoLocationForm(void)
{
  FormPtr       prevFrm;
  FormPtr       frm;
  FieldPtr      newLoc;
  Char          loc[5];
  Char          yoda;
  UInt16        hitButton;
  UInt16        offset;
  UInt16        maxOffset;
  UInt32        newPos;
  UInt32        x;

  prevFrm = FrmGetActiveForm();
  frm = FrmInitForm(LocationForm);
  FrmSetActiveForm(frm);
  newLoc = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, NewLocField));

  if (!gCurrentDoc.endOfDoc)
    {
      StrPrintF(loc, "%d",
                (UInt16)((gCurrentDoc.curPos * 100) / gCurrentDoc.size));
    }
  else
    StrCopy(loc, "100");
  SetFieldText(NewLocField, loc);

  hitButton = FrmDoDialog(frm);
  StrNCopy(loc, FldGetTextPtr(newLoc), 4);
  loc[3] = '\0';
  x = StrAToI(loc);
  if (x < 0)
    x = 0;
  else if (x > 100)
    x = 100;

  if (prevFrm)
    FrmSetActiveForm(prevFrm);
  FrmDeleteForm(frm);

  /* 0% and 100% are special cases */
  if (hitButton == LocBeginButton)
    x = 0;
  else if (hitButton == LocEndButton)
    x = 100;

  if (hitButton != LocCancelButton)
    {
      if (x == 100)
        {
          /* Skip to the end of the document - 300 bytes */
          gCurrentDoc.curPage.recIndex = gCurrentDoc.numTmpRecords - 1;
          offset = (((Int16)(gCurrentDoc.size % gCurrentDoc.recordSize)) -
                    300);
          if (offset < 0)
            {
              if (gCurrentDoc.curPage.recIndex == 0)
                {
                  gCurrentDoc.curPage.offset = 0;
                  gCurrentDoc.startOfDoc = true;
                }
              else
                {
                  gCurrentDoc.curPage.recIndex--;
                  gCurrentDoc.curPage.offset = gCurrentDoc.recordSize + offset;
                }
            }
          else
            gCurrentDoc.curPage.offset = offset;

          OpenTmpRecord(gCurrentDoc.curPage.recIndex);
          yoda = *((Char *)(gCurrentDoc.curRecordPtr +
                            gCurrentDoc.curPage.offset));
          /* Roll back if offset is in the middle of a word */
          while ((gCurrentDoc.curPage.offset != 0) && !TxtCharIsSpace(yoda))
            {
              gCurrentDoc.curPage.offset--;
              yoda = *((Char *)(gCurrentDoc.curRecordPtr +
                                gCurrentDoc.curPage.offset));
            }
          gCurrentDoc.curPage.offset++;
          gCurrentDoc.startOfDoc = false;
          gCurrentDoc.endOfDoc = false;
        }
      else if (x == 0)
        {
          /* Jump to the very beginning of the doc */
          gCurrentDoc.curPage.recIndex = 0;
          gCurrentDoc.curPage.offset = 0;
          gCurrentDoc.startOfDoc = true;
          gCurrentDoc.endOfDoc = false;
        }
      else
        {
          /* Jump to some position inside the doc */
          newPos = ((StrAToI(loc) * gCurrentDoc.size) / 100);
          gCurrentDoc.curPage.recIndex = newPos / gCurrentDoc.recordSize;
          gCurrentDoc.curPage.offset = newPos % gCurrentDoc.recordSize;

          OpenTmpRecord(gCurrentDoc.curPage.recIndex);
          yoda = *((Char *)(gCurrentDoc.curRecordPtr +
                            gCurrentDoc.curPage.offset));
          if (gCurrentDoc.curPage.recIndex == gCurrentDoc.numTmpRecords - 1)
            maxOffset = gCurrentDoc.size - ((gCurrentDoc.numTmpRecords - 1) *
                                            gCurrentDoc.recordSize);
          else
            maxOffset = gCurrentDoc.recordSize;
          /* Move forward if offset is in the middle of a word */
          while ((gCurrentDoc.curPage.offset < maxOffset) &&
                 !TxtCharIsSpace(yoda))
            {
              gCurrentDoc.curPage.offset++;
              yoda = *((Char *)(gCurrentDoc.curRecordPtr +
                                gCurrentDoc.curPage.offset));
            }
          gCurrentDoc.curPage.offset++;
          if (gCurrentDoc.curPage.offset >= maxOffset)
            {
              if (maxOffset == gCurrentDoc.recordSize)
                {
                  /* Moved forward to end of record */
                  gCurrentDoc.curPage.offset=0;
                  gCurrentDoc.curPage.recIndex++;
                }
              else
                {
                  gCurrentDoc.curPage.offset = maxOffset - 1;
                  while ((gCurrentDoc.curPage.offset != 0) &&
                         !TxtCharIsSpace(yoda))
                    {
                      /* Moved to end of document, so move back a bit */
                      gCurrentDoc.curPage.offset--;
                      yoda = *((Char *)(gCurrentDoc.curRecordPtr +
                                        gCurrentDoc.curPage.offset));
                    }
                  gCurrentDoc.curPage.offset++;
                }
            }

          gCurrentDoc.startOfDoc = false;
          gCurrentDoc.endOfDoc = false;
        }

      DisplayCurrentPage();
    }
}


/* Displays the form to add a new bookmark */
static void
DoAddBookmarkForm(void)
{
  FormPtr       prevFrm;
  FormPtr       frm;
  FieldPtr      bmrkField;
  ListPtr       list;
  Char          title[MAX_BMRK_LENGTH + 1];
  UInt16        hitButton;
  UInt16        i;
  UInt16        x;

  prevFrm = FrmGetActiveForm();
  frm = FrmInitForm(AddBookmarkForm);
  FrmSetActiveForm(frm);
  bmrkField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, BmrkTitleField));

  OpenTmpRecord(gCurrentDoc.curPage.recIndex);
  MemSet(title, MAX_BMRK_LENGTH + 1, 0);
  for (i = gCurrentDoc.curPage.offset, x = 0;
       ((i < gCurrentDoc.recordSize) && (x < MAX_BMRK_LENGTH));
       i++, x++)
    {
      title[x] = *(Char *)(gCurrentDoc.curRecordPtr + i);
    }
  SetFieldText(BmrkTitleField, title);

  hitButton = FrmDoDialog(frm);
  StrNCopy(title, FldGetTextPtr(bmrkField), MAX_BMRK_LENGTH);
  title[MAX_BMRK_LENGTH] = '\0';

  if (prevFrm)
    FrmSetActiveForm(prevFrm);
  FrmDeleteForm(frm);

  if (hitButton == BmrkOKButton)
    {
      bmrkAdd(title, gCurrentDoc.curPos);
      frm = prevFrm;
      list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, BmrkList));
      LstSetListChoices(list, NULL, gCurrentDoc.numBookmarks + 1);
    }
}


/* Displays the form to let you fiddle with your bookmarks */
static void
DoDeleteBookmarksForm(void)
{
  FormPtr       prevFrm;
  FormPtr       frm;
  ListPtr       list;

  prevFrm = FrmGetActiveForm();
  frm = FrmInitForm(DeleteBookmarksForm);
  FrmSetActiveForm(frm);

  list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DeleteBookmarksList));
  LstSetListChoices(list, NULL, gCurrentDoc.numBookmarks);
  LstSetDrawFunction(list, (ListDrawDataFuncPtr) BmrkDrawFunc);

  FrmSetEventHandler(frm, DelBmrkFormHandleEvent);
  FrmDoDialog(frm);

  if (prevFrm)
    FrmSetActiveForm(prevFrm);
  FrmDeleteForm(frm);

  frm = prevFrm;
  list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, BmrkList));
  LstSetListChoices(list, NULL, gCurrentDoc.numBookmarks + 1);
}


/* Event handler for the delete bookmarks form */
static Boolean
DelBmrkFormHandleEvent(EventPtr e)
{
  Boolean       handled = false;
  FormPtr       frm;
  ListPtr       list;
  GPlmMark      *bmrk;

  switch (e->eType)
    {
      case ctlSelectEvent:
        if (e->data.ctlSelect.controlID == DelBmrksDeleteButton)
          {
            if (gCurrentDoc.numBookmarks > 0)
              {
                frm = FrmGetActiveForm();
                list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                                                       DeleteBookmarksList));
                bmrk = (((GPlmMark *)gCurrentDoc.bmrkRecordPtr) +
                        LstGetSelection(list));
                bmrkDeleteMark(bmrk->title);
                LstSetListChoices(list, NULL, gCurrentDoc.numBookmarks);
                LstDrawList(list);
              }
            handled = true;
          }
        break;

      default:
        break;
    }

  return handled;
}


/* Jumps to the offset associated with the specified bookmark */
static void
JumpToBookmark(UInt16 bmrkIndex)
{
  GPlmMark      *bmrk;

  bmrk = (GPlmMark *)gCurrentDoc.bmrkRecordPtr;
  bmrk += bmrkIndex;
  gCurrentDoc.curPage.recIndex = bmrk->offset / gCurrentDoc.recordSize;
  gCurrentDoc.curPage.offset = bmrk->offset % gCurrentDoc.recordSize;
  gCurrentDoc.endOfDoc = false;
  if ((gCurrentDoc.curPage.recIndex == 0) && (gCurrentDoc.curPage.offset == 0))
    gCurrentDoc.startOfDoc = true;
  else
    gCurrentDoc.startOfDoc = false;

  DisplayCurrentPage();
}


/* Decompress the entire zTXT document and store into a new database */
static void
DecompressDocument(void)
{
  Int16         status;
  Boolean       last_record = false;
  UInt16        outRecIndex;
  MemHandle     outRecHandle;
  MemPtr        outRecPtr;
  Char          progressStr[10];


  /* Create the DB to hold the uncompressed data */
  StrPrintF(gCurrentDoc.tempName, "cache-%d", gPrefs.numCachedDBs);
  DmCreateDatabase(0, gCurrentDoc.tempName, id_to_int(GPLM_CREATOR_ID),
                   id_to_int(ZTMP_TYPE_ID), false);
  gCurrentDoc.tempID = DmFindDatabase(0, gCurrentDoc.tempName);
  gCurrentDoc.tempRef = DmOpenDatabase(0, gCurrentDoc.tempID, dmModeReadWrite);
  outRecIndex = 0;


  /* Inflate the first block of data */
  status = inflate(&gCurrentDoc.zstream, Z_SYNC_FLUSH);

  /* Inflation loop */
  while (status == Z_OK)
    {
      /* Give progress indicator */
      StrPrintF(progressStr, "%d%%",
                (UInt16)(((UInt32)gCurrentDoc.curRecordIndex * 100)/
                         ((UInt32)gCurrentDoc.numRecords)));
      WinDrawChars(progressStr, StrLen(progressStr), 70, 80);

      /* output buffer is full.  Store it into a record */
      if (gCurrentDoc.zstream.avail_out == 0)
        {
          outRecHandle = DmNewRecord(gCurrentDoc.tempRef, &outRecIndex,
                                     gCurrentDoc.recordSize);
          outRecPtr = MemHandleLock(outRecHandle);
          DmWrite(outRecPtr, 0, gCurrentDoc.zBuffer, gCurrentDoc.recordSize);
          MemHandleUnlock(outRecHandle);
          DmReleaseRecord(gCurrentDoc.tempRef, outRecIndex, false);
          outRecIndex++;
          gCurrentDoc.zstream.next_out = gCurrentDoc.zBuffer;
          gCurrentDoc.zstream.avail_out = gCurrentDoc.recordSize;
        }

      /* input buffer is empty.  Grab some more */
      if ((gCurrentDoc.zstream.avail_in == 0) && !last_record)
        {
          MemHandleUnlock(gCurrentDoc.curRecordHandle);
          DmReleaseRecord(gCurrentDoc.dbRef, gCurrentDoc.curRecordIndex,
                          false);
          gCurrentDoc.curRecordIndex++;
          gCurrentDoc.curRecordHandle = DmGetRecord(gCurrentDoc.dbRef,
                                                 gCurrentDoc.curRecordIndex);
          gCurrentDoc.zstream.next_in =
                          MemHandleLock(gCurrentDoc.curRecordHandle);
          if (gCurrentDoc.curRecordIndex > gCurrentDoc.numRecords)
            {
              last_record = true;
              gCurrentDoc.zstream.avail_in = (UInt16)(gCurrentDoc.size -
                                                    (gCurrentDoc.numRecords *
                                                     gCurrentDoc.recordSize));
            }
          else
            gCurrentDoc.zstream.avail_in = gCurrentDoc.recordSize;
        }

      /* Inflate the next block of data */
      status = inflate(&gCurrentDoc.zstream, Z_SYNC_FLUSH);
    }

  /* No more input, but output may still be pending
   *   if so, write it out
   */
  if ((status == Z_STREAM_END) &&
      (gCurrentDoc.zstream.avail_out != gCurrentDoc.recordSize))
    {
      outRecHandle = DmNewRecord(gCurrentDoc.tempRef, &outRecIndex,
                                 (gCurrentDoc.recordSize -
                                  gCurrentDoc.zstream.avail_out) + 1);
      outRecPtr = MemHandleLock(outRecHandle);
      ((Char *)gCurrentDoc.zBuffer)[gCurrentDoc.recordSize -
                                    gCurrentDoc.zstream.avail_out] = '\0';
      DmWrite(outRecPtr, 0, gCurrentDoc.zBuffer,
              gCurrentDoc.recordSize - gCurrentDoc.zstream.avail_out);
      MemHandleUnlock(outRecHandle);
      DmReleaseRecord(gCurrentDoc.tempRef, outRecIndex, false);
      outRecIndex++;
    }

  /* Store number of records written */
  gCurrentDoc.numTmpRecords = outRecIndex;

  /* Done with the input DB.  Close open records */
  MemHandleUnlock(gCurrentDoc.curRecordHandle);
  DmReleaseRecord(gCurrentDoc.dbRef, gCurrentDoc.curRecordIndex, false);
}


/* This function displays the bookmark list in the Docview form */
static void
BmrkDrawFunc(UInt16 itemNum, RectanglePtr bounds, Char *data[])
{
  GPlmMark      *bmrk;
  Char          markText[MAX_BMRK_LENGTH + 1];
  Int16         numBytes;
  Int16         width;
  Boolean       trunc;

  if (itemNum >= gCurrentDoc.numBookmarks)
    {
      StrCopy(markText, "Add bookmark...");
    }
  else
    {
      bmrk = (GPlmMark *)gCurrentDoc.bmrkRecordPtr;
      bmrk += itemNum;
      StrNCopy(markText, bmrk->title, MAX_BMRK_LENGTH);
      markText[MAX_BMRK_LENGTH] = '\0';
    }

  numBytes = StrLen(markText);
  width = bounds->extent.x;
  trunc = true;
  FntCharsInWidth(markText, &width, &numBytes, &trunc);
  WinDrawChars(markText, numBytes, bounds->topLeft.x, bounds->topLeft.y);
}
