/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm_common.h"
#include "gutenpalmRsc.h"
#include "gutenpalm.h"
#include "gutenpalm_prefs.h"
#include "gutenpalm_cache.h"


/* Main form function prototypes */
static void     MainFormInit(void);
static void     ShowInfoForm(void);
static void     DeleteDocument(void);
static void     DoclistDrawFunc(UInt16, RectanglePtr, Char **);
static void     DoPreferencesForm(void);



/* Event handler for the main form (document listing) */
Boolean
MainFormHandleEvent (EventPtr e)
{
  Boolean       handled = false;
  DocList       *doclist;
  FormPtr       frm;

  switch (e->eType)
    {
      case frmOpenEvent:
        MainFormInit();
        handled = true;
        break;

      case lstSelectEvent:
        if (gNumDocuments)
          {
            if (gCurrentDoc.dbIndex == e->data.lstSelect.selection)
              FrmGotoForm(DocViewForm);
            else
              gCurrentDoc.dbIndex = e->data.lstSelect.selection;
            handled = true;
          }
        break;

      case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID)
          {
            /* Delete cache for current doc if it exists */
            case MainFormMenu_DelCache:
              if (gNumDocuments)
                {
                  doclist = (DocList *)MemHandleLock(gDoclistHandle);
                  if (cacheExists(
                            doclist->documents[gCurrentDoc.dbIndex].name))
                    {
                      cacheDelete(
                            doclist->documents[gCurrentDoc.dbIndex].name);
                      frm = FrmGetActiveForm();
                      LstDrawList(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                                                               DocumentList)));
                    }
                  MemHandleUnlock(gDoclistHandle);
                }
              break;

            /* Delete all cached data */
            case MainFormMenu_DelAllCaches:
              if ((gNumDocuments) && (gPrefs.numCachedDBs))
                {
                  if (FrmAlert(DeleteAllCacheAlert) == 0)
                    {
                      cacheClear();
                      frm = FrmGetActiveForm();
                      LstDrawList(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                                                              DocumentList)));
                    }
                }
              break;

            /* Show the preferences dialog */
            case MainFormMenu_Prefs:
              DoPreferencesForm();
              break;

            /* Display information about GutenPalm */
            case MainFormMenu_About:
              FrmAlert(About1Alert);
              FrmAlert(About2Alert);
              break;
          }
        handled = true;
        break;

      case ctlSelectEvent:
        if (gNumDocuments)
          {
            switch(e->data.ctlSelect.controlID)
              {
                case ReadButton:
                  FrmGotoForm(DocViewForm);
                  handled = true;
                  break;
                case InfoButton:
                  ShowInfoForm();
                  handled = true;
                  break;
                case DeleteButton:
                  DeleteDocument();
                  break;
              }
          }
        break;

      default:
        break;
    }

  return handled;
}


/* Displays the preferences form and sets options accordingly */
static void
DoPreferencesForm(void)
{
  FormPtr       prevFrm;
  FormPtr       frm;
  ControlPtr    skipLicense;
  ControlPtr    keepCache;
  ControlPtr    alwaysRemember;

  prevFrm = FrmGetActiveForm();
  frm = FrmInitForm(PrefsForm);
  skipLicense = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,PGLicenseCheckbox));
  keepCache = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, KeepCacheCheckbox));
  alwaysRemember = FrmGetObjectPtr(frm,
                           FrmGetObjectIndex(frm, AlwaysRememberCheckbox));

  FrmSetActiveForm(frm);

  CtlSetValue(skipLicense, gPrefs.skipLicense);
  CtlSetValue(keepCache, gPrefs.keepCache);
  CtlSetValue(alwaysRemember, gPrefs.alwaysRememberPos);

  if (FrmDoDialog(frm) == PrefsOKButton)
    {
      gPrefs.skipLicense = CtlGetValue(skipLicense);
      gPrefs.keepCache = CtlGetValue(keepCache);
      gPrefs.alwaysRememberPos = CtlGetValue(alwaysRemember);
    }

  if (prevFrm)
    FrmSetActiveForm(prevFrm);
  FrmDeleteForm(frm);
}


/* Deletes the selected document after user confirmation */
static void
DeleteDocument(void)
{
  UInt16        choice;
  UInt16        index;
  DocList       *doclist;
  FormPtr       frm;
  ListPtr       list;

  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  index = gCurrentDoc.dbIndex;
  choice = FrmCustomAlert(DeleteAlert, doclist->documents[index].name,
                          NULL, NULL);

  /* Yes pressed, delete document */
  if (choice == 0)
    {
      /* Kill the DB */
      DmDeleteDatabase(0, DmFindDatabase(0, doclist->documents[index].name));

      /* Kill the cache DB if it exists*/
      if (doclist->documents[index].cache)
        cacheDelete(doclist->documents[index].name);

      /* Remove the DB from the DocList */
      if (index == doclist->numDocuments - 1)
        doclist->numDocuments--;
      else
        {
          MemMove(&(doclist->documents[index]), &(doclist->documents[index+1]),
                  (doclist->numDocuments - index - 1) * sizeof(DocEntry));
          doclist->numDocuments--;
        }

      /* Update DocList display */
      frm = FrmGetActiveForm();
      list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DocumentList));
      LstSetListChoices(list, NULL, doclist->numDocuments);
      if (doclist->numDocuments > 0)
        LstSetSelection(list, 0);
      gCurrentDoc.dbIndex = 0;
      gNumDocuments = doclist->numDocuments;
      FrmDrawForm(frm);
    }

  MemHandleUnlock(gDoclistHandle);
}


/* Displays information about selected zTXT document */
static void
ShowInfoForm(void)
{
  FormPtr       prevFrm;
  FormPtr       frm;
  DocList       *doclist;
  DmOpenRef     dbRef;
  MemHandle     recHandle;
  zTXT_record0  *record0;
  UInt32        numBytes;
  Char          tempStr[dmDBNameLength + 1];

  doclist = (DocList *)MemHandleLock(gDoclistHandle);

  prevFrm = FrmGetActiveForm();
  frm = FrmInitForm(DocInfoForm);

  FrmSetActiveForm(frm);

  /* Set document title field */
  StrNCopy(tempStr, doclist->documents[gCurrentDoc.dbIndex].name,
           dmDBNameLength);
  SetFieldText(TitleField, tempStr);

  /* Open the zTXT database */
  dbRef = DmOpenDatabase(0,
              DmFindDatabase(0, doclist->documents[gCurrentDoc.dbIndex].name),
              dmModeReadOnly);

  /* Open record #0 */
  recHandle = DmGetRecord(dbRef, 0);
  record0 = (zTXT_record0 *)MemHandleLock(recHandle);

  /* Set document version field */
  StrPrintF(tempStr, "Version -- %d.%d", ((UInt8 *)&record0->version)[0],
            ((UInt8 *)&record0->version)[1]);
  SetFieldText(VersionField, tempStr);

  /* Set document size field */
  DmDatabaseSize(0, DmFindDatabase(0,
                 doclist->documents[gCurrentDoc.dbIndex].name), NULL,
                 NULL, &numBytes);
  StrPrintF(tempStr, "Size -- %ld", numBytes);
  SetFieldText(SizeField, tempStr);

  /* Set document uncompressed size field */
  StrPrintF(tempStr, "Uncomp. size -- %ld", record0->size);
  SetFieldText(UncompSizeField, tempStr);

  /* Set document recordSize field */
  StrPrintF(tempStr, "Record size -- %d", record0->recordSize);
  SetFieldText(RecordSizeField, tempStr);

  /* Set current location field */
  StrPrintF(tempStr, "Bookmarks -- %d", record0->numBookmarks);
  SetFieldText(NumBookmarksField, tempStr);

  /* Set cached status */
  StrCopy(tempStr, "Cached -- ");
  if (cacheExists(doclist->documents[gCurrentDoc.dbIndex].name))
    StrCat(tempStr, "Yes");
  else
    StrCat(tempStr, "No");
  SetFieldText(CacheField, tempStr);

  /* Don't need record #0 anymore */
  MemHandleUnlock(recHandle);
  DmReleaseRecord(dbRef, 0, false);

  /* Done with the input data */
  DmCloseDatabase(dbRef);
  MemHandleUnlock(gDoclistHandle);
  FrmDoDialog(frm);

  if (prevFrm)
    FrmSetActiveForm(prevFrm);
  FrmDeleteForm(frm);
}


static void
MainFormInit(void)
{
  FormPtr       frm;
  ListPtr       list;
  DocList       *doclist;

  /* Set current form */
  gCurrentForm = MainForm;

  /* Clear gCurrentDoc so nothing gets confused */
  MemSet(&gCurrentDoc, sizeof(Document), 0);

  /* Set up the document listing */
  frm = FrmGetActiveForm();
  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  if (doclist->numDocuments == 0)
    {
      FrmDrawForm(frm);
      return;
    }
  list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DocumentList));
  LstSetListChoices(list, NULL, doclist->numDocuments);
  MemHandleUnlock(gDoclistHandle);
  LstSetDrawFunction(list, (ListDrawDataFuncPtr) DoclistDrawFunc);

  FrmDrawForm(frm);
}


/* This function displays the lines in the DocList in the main form */
static void
DoclistDrawFunc(UInt16 itemNum, RectanglePtr bounds, Char *data[])
{
  DocList       *doclist;
  Char          *currDoc;

  doclist = (DocList *)MemHandleLock(gDoclistHandle);
  currDoc = doclist->documents[itemNum].name;
  if (doclist->documents[itemNum].cache)
    {
      WinDrawLine(bounds->topLeft.x+1, bounds->topLeft.y+4,
                  bounds->topLeft.x+3, bounds->topLeft.y+4);
      WinDrawLine(bounds->topLeft.x+1, bounds->topLeft.y+5,
                  bounds->topLeft.x+3, bounds->topLeft.y+5);
    }
  WinDrawChars(currDoc, StrLen(currDoc), bounds->topLeft.x+6,
               bounds->topLeft.y);
  MemHandleUnlock(gDoclistHandle);
}
