/*
** AutoLock - Locks your pilot after a specified idle time
** Copyright  2000 by Patrick Koppen <patrick@koppen.de>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
**  ChangeLog:
**
*/

#include <stdio.h>
#include <PalmOS.h>
#include "resource_ids.h"
#include "version.h"

#define AUTOLOCK_MIN_TIMEOUT 5		/* minimum timeout */
#define AUTOLOCK_RESET_TIMEOUT 5	/* timeout after reset */
#define AUTOLOCK_PREFS_VERSION 4	/* preference version */

#define SECAPPID 'secr'			/* security application id */

/* AutoLock preference structure */
typedef struct AutoLockPreferenceType
{
  Boolean	active;
  Boolean	chk_shortcut;
  Char 		shortcut[6];
  Boolean	chk_timeout;
  UInt32	timeout;
  Boolean	chk_reset;
}
AutoLockPreferenceType;

/* global preference variable */
static AutoLockPreferenceType AutoLock_prefs;

/* 
** Funktion: set_active_status
** Parameters: status 
**
** sets/unsets (status = true/false) active flags in preferences
**
*/

static void set_active_status (Boolean status)
{
  FormType *frm;
  FieldType *fldP;


  AutoLock_prefs.active = status;

  frm = FrmGetFormPtr (MainFormID);

  if (frm != NULL)
    {
      fldP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm, MainFormStatusFieldID));
      CtlSetLabel ((ControlType *) fldP,
		   AutoLock_prefs.active ? "Active    " : "Inactive   ");
      FldDrawField (fldP);
      FrmDrawForm (frm);
    }
}



/*
** Function: set_alarm
** Paramters: timeout
**
** schedules next alarm in 'timeout' seconds or deletes alarm if timeout=0.
**
*/

static void set_alarm (UInt32 timeout)
{
  UInt16 cardNo;
  LocalID dbID;
  UInt32 lock_time;
  Err err;

  SysCurAppDatabase (&cardNo, &dbID);
  lock_time = TimGetSeconds ();
  if (timeout > 0)
    lock_time += timeout;
  else
    lock_time = 0;

  err = AlmSetAlarm (cardNo, dbID, 0, lock_time, 0);

  return;
}

/*
** Function: unset_notify
** Parameters: none
**
** unregisters all notify events
**
*/

void unset_notify ()
{
  UInt16 cardNo;
  LocalID dbID;

  SysCurAppDatabase (&cardNo, &dbID);

  SysNotifyUnregister (cardNo, dbID, sysNotifySleepNotifyEvent,
		       sysNotifyNormalPriority);
  SysNotifyUnregister (cardNo, dbID, sysNotifyLateWakeupEvent,
		       sysNotifyNormalPriority);
  SysNotifyUnregister (cardNo, dbID, sysNotifyResetFinishedEvent,
		       sysNotifyNormalPriority);
  return;
}


/*
** Function: set_notify
** Parameters: none
**
** registers all notify events
**
*/

void set_notify ()
{
  UInt16 cardNo;
  LocalID dbID;

  SysCurAppDatabase (&cardNo, &dbID);

  SysNotifyRegister (cardNo, dbID, sysNotifySleepNotifyEvent,
		     NULL, sysNotifyNormalPriority, NULL);
  SysNotifyRegister (cardNo, dbID, sysNotifyLateWakeupEvent,
		     NULL, sysNotifyNormalPriority, NULL);
  SysNotifyRegister (cardNo, dbID, sysNotifyResetFinishedEvent,
		     NULL, sysNotifyNormalPriority, NULL);
  return;

}

/*
** Function: MainFormHandleEvent
** Parameters: event
**
** main formular event handler
**
*/

Boolean MainFormHandleEvent (EventPtr event)
{
  Boolean handled = false;
  EventType ev;


  switch (event->eType)
    {
    case ctlSelectEvent:	// A control button was pressed and released.
    case menuEvent:
      if (event->data.ctlSelect.controlID == MainFormQuitID
	  || event->data.ctlSelect.controlID == MainFormMenuQuitID)
	{
	  /* Quit-Button pressed */
	  ev.eType = appStopEvent;
	  EvtAddEventToQueue (&ev);
	  handled = true;
	}
      else if (event->data.ctlSelect.controlID == MainFormActivateID
	       || event->data.ctlSelect.controlID == MainFormMenuActivateID)
	{
	  /* Activate-Button pressed */
	  set_active_status (true);	/* set preference */
	  set_notify ();		/* register notifiers */
	  FrmAlert (ActivateAlertID);	/* display dialog */
	  handled = true;
	}
      else if (event->data.ctlSelect.controlID == MainFormDeactivateID
	       || event->data.ctlSelect.controlID == MainFormMenuDeactivateID)
	{
	  /* Deactivate-Button pressed */
	  set_active_status (false);	/* set preference */
	  set_alarm (0);		/* unschedules any alarm */
	  unset_notify ();		/* unregister notifiers */
	  FrmAlert (DeactivateAlertID);	/* display dialog */
	  handled = true;
	}
      else if (event->data.ctlSelect.controlID == MainFormMenuAboutID)
	{
	  /* About-Button pressed - display about dialog */
	  FrmCustomAlert (AboutAlertID, AUTOLOCK_VERSION, "", "");
	  handled = true;
	}
      else if (event->data.ctlSelect.controlID == MainFormMenuHelpID)
	{
	  /* About-Button pressed - display help dialog */
	  FrmAlert (HelpAlertID);
	  handled = true;
	}
      break;

    case frmOpenEvent:
      FrmDrawForm (FrmGetActiveForm ());
      handled = true;
      break;

    default:
      ;;
    }
  return handled;
}

/*
** Function: AutoLockSavePrefs
** Parameters: none
**
** save preferences
**
*/

void AutoLockSavePrefs ()
{
  FormType *frm;
  FieldType *fldP;
  ControlPtr ctlP;
  char *textP;
  MemHandle textH;

  UInt16 prefsSize;

  frm = FrmGetFormPtr (MainFormID);
  if (frm != NULL)
    {
      /* save shortcut */
      ctlP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm,
						 MainFormShortcutCHKID));
      AutoLock_prefs.chk_shortcut=CtlGetValue(ctlP);

      fldP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm,
						 MainFormShortcutFieldID));
      textH = FldGetTextHandle (fldP);
      if (textH != NULL)
	{
	  textP = MemHandleLock (textH);
	  StrCopy(AutoLock_prefs.shortcut,textP);
	  MemHandleUnlock (textH);
	}

      /* save timeout */
      ctlP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm,
						 MainFormTimeoutCHKID));
      AutoLock_prefs.chk_timeout=CtlGetValue(ctlP);

      fldP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm,
						 MainFormTimeoutFieldID));
      textH = FldGetTextHandle (fldP);
      if (textH != NULL)
	{
	  textP = MemHandleLock (textH);
	  AutoLock_prefs.timeout = StrAToI (textP);
	  if (AutoLock_prefs.timeout < AUTOLOCK_MIN_TIMEOUT)
	    AutoLock_prefs.timeout = AUTOLOCK_MIN_TIMEOUT;
	  MemHandleUnlock (textH);
	}

      /* save on_reset */
      ctlP = FrmGetObjectPtr (frm,
			      FrmGetObjectIndex (frm,
						 MainFormResetCHKID));
      AutoLock_prefs.chk_reset=CtlGetValue(ctlP);

    }


  prefsSize = sizeof (AutoLockPreferenceType);
  PrefSetAppPreferences (APPID, 0, AUTOLOCK_PREFS_VERSION,
			 &AutoLock_prefs, prefsSize, true);
  return;
}

/*
** Function: get_prefs
** Parameters: prefs
**
** read preferences
**
*/

void get_prefs (AutoLockPreferenceType *prefs)
{
  UInt16 prefsSize;
  Int16 ret;

  prefsSize = sizeof (AutoLockPreferenceType);
 
  ret = PrefGetAppPreferences (APPID, 0, prefs, &prefsSize, true);

  if ( ret == noPreferenceFound
	|| prefsSize != sizeof (AutoLockPreferenceType))
    {
      prefs->active = false;
      prefs->chk_shortcut = false;
      StrCopy(prefs->shortcut,"L");
      prefs->chk_timeout = false;
      prefs->timeout = 10800;	// 3 Hours
      prefs->chk_reset = false;
    }
}

/*
** Function: StartApplication
** Parameters: none
**
** load preferences and goto mainform
**
*/

static Err StartApplication (void)
{
  get_prefs (&AutoLock_prefs);
  FrmGotoForm (MainFormID);
  return 0;
}

/*
** Function: StopApplication
** Parameters: none
**
** save preferences and quit
**
*/

static void StopApplication (void)
{
  // Save prefs
  AutoLockSavePrefs ();

  // Send a frmSave event to all the open forms.
  FrmSaveAllForms ();

  // Close all the open forms.
  FrmCloseAllForms ();

  // Close the application's data file.
  //DmCloseDatabase (ApptDB);
}

/*
** Function: do_lock
** Parameters: none
**
** call security application to lock your palm
**
*/

static void do_lock ()
{
  DmSearchStateType stateInfop;
  UInt16 cardNo;
  LocalID dbID;

  if (!PrefGetPreference (prefDeviceLocked) && PwdExists ())
    {

      DmGetNextDatabaseByTypeCreator (true, &stateInfop,
				      sysFileTApplication, SECAPPID, true,
				      &cardNo, &dbID);
      SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdSystemLock, 0);
    }
  else
    {
      //SysSleep( false, false );
    }
  return;
}

/*
** Function: MainFormInit 
** Parameters: frm
**
**
*/

static void MainFormInit (FormPtr frm)
{
  FieldType *fldP;
  ControlPtr ctlP;
  MemHandle textH;
  MemHandle oldtextH;
  MemPtr textP;
  char buf[10];

  /* Set timeout field */
  ctlP = FrmGetObjectPtr (frm,
			  FrmGetObjectIndex (frm,
					     MainFormShortcutCHKID));
  CtlSetValue(ctlP,AutoLock_prefs.chk_shortcut);

  fldP = FrmGetObjectPtr (frm,
			  FrmGetObjectIndex (frm,
					     MainFormShortcutFieldID));
  oldtextH = FldGetTextHandle (fldP);
  if (oldtextH != NULL)
    {
      textH = MemHandleNew (MemHandleSize (oldtextH));
    }
  else
    {
      textH = MemHandleNew (sizeof (buf) + 1);
      StrCopy (buf, AutoLock_prefs.shortcut);
      textP = MemHandleLock (textH);
      MemMove (textP, buf, StrLen (buf) + 1);
      MemHandleUnlock (textH);
      FldSetTextHandle (fldP, textH);
      if (oldtextH != NULL)
	MemHandleFree (oldtextH);
    }

  /* Set timeout field */
  ctlP = FrmGetObjectPtr (frm,
			  FrmGetObjectIndex (frm,
					     MainFormTimeoutCHKID));
  CtlSetValue(ctlP,AutoLock_prefs.chk_timeout);
  
  fldP = FrmGetObjectPtr (frm,
			  FrmGetObjectIndex (frm,
					     MainFormTimeoutFieldID));
  oldtextH = FldGetTextHandle (fldP);
  if (oldtextH != NULL)
    {
      textH = MemHandleNew (MemHandleSize (oldtextH));
    }
  else
    {
      textH = MemHandleNew (sizeof (buf) + 1);
      StrIToA (buf, AutoLock_prefs.timeout);
      textP = MemHandleLock (textH);
      MemMove (textP, buf, StrLen (buf) + 1);
      MemHandleUnlock (textH);
      FldSetTextHandle (fldP, textH);
      if (oldtextH != NULL)
	MemHandleFree (oldtextH);
    }

  /* Set reset field */
  ctlP = FrmGetObjectPtr (frm,
			  FrmGetObjectIndex (frm,
					     MainFormResetCHKID));
  CtlSetValue(ctlP,AutoLock_prefs.chk_reset);

  /* Set (in)active field */
  CtlSetLabel (FrmGetObjectPtr (frm,
				FrmGetObjectIndex (frm,
						   MainFormStatusFieldID)),
	       AutoLock_prefs.active ? "Active" : "Inactive");

  return;
}

/*
** Function: ApplicationHandleEvent
** Parameters: event
**
** Application event handler
**
*/

static Boolean ApplicationHandleEvent (EventPtr event)
{
  FormPtr frm;
  UInt16 formId;
  Boolean handled = false;

  if (event->eType == frmLoadEvent)
    {
//Load the form resource specified in the event then activate it
      formId = event->data.frmLoad.formID;
      frm = FrmInitForm (formId);
      FrmSetActiveForm (frm);

// Set the event handler for the form.  The handler of the currently 
// active form is called by FrmDispatchEvent each time it is called 
      switch (formId)
	{
	case MainFormID:
	  FrmSetEventHandler (frm, MainFormHandleEvent);
	  MainFormInit(frm);
	  break;
	}
      handled = true;

    }

  return handled;
}

/*
** Function: handle_power_off
** Parameters: notify
**
** called when palm is turned off. schedules alarm if palm not already
** locked.
**
*/

void handle_power_off (SysNotifyParamType * notify,
		  AutoLockPreferenceType * prefsP)
{
  if (!PrefGetPreference (prefDeviceLocked) && PwdExists ()
	&& prefsP->active && prefsP->chk_timeout)
    {
      set_alarm (prefsP->timeout);
    }
}

/*
** Function: handle_power_on
** Parameters: notify
**
** called when palm is turned on. Deletes scheduled alarm.
**
*/

void handle_power_on (SysNotifyParamType * notify,
		AutoLockPreferenceType * prefsP)
{
  UInt16 cardNo;
  LocalID dbID;
  UInt32 lock_time;
  UInt32 refP;

  SysCurAppDatabase (&cardNo, &dbID);
  lock_time = AlmGetAlarm (cardNo, dbID, &refP);
  if (lock_time != 0 && lock_time > TimGetSeconds ())
    set_alarm (0);
}

static void
EventLoop (void)
{
  EventType event;
  UInt16 error;

  do
    {
      EvtGetEvent (&event, evtWaitForever);
      if (!SysHandleEvent (&event))
	if (!MenuHandleEvent (0, &event, &error))
	  if (!ApplicationHandleEvent (&event))
	    FrmDispatchEvent (&event);
    }
  while (event.eType != appStopEvent);
}

UInt32
PilotMain (UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
  Err err = 0;
  SysNotifyParamType *notify;
  AutoLockPreferenceType prefs;
  UInt32 value;

  /* Check Romversion */
  /* need 3.5 for sysAppLaunchCmdNotify ?? */
  value=0x03503000;
  err = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &value);

  if (err)
  {
	//FrmAlert (WrongRomVersionAlertID); 
	return 0;
  }

  if (launchCode == sysAppLaunchCmdNormalLaunch)
    {
      if ((err = StartApplication ()) == 0)
	{
	  EventLoop ();
	  StopApplication ();
	}
    }
  else if (launchCode == sysAppLaunchCmdNotify)
    {
      get_prefs(&prefs);

      notify = (SysNotifyParamType *) cmdPBP;
      if (notify->notifyType == sysNotifySleepNotifyEvent)
	{
	  //Power off
	  handle_power_off (notify, &prefs);
	}
      if (notify->notifyType == sysNotifyLateWakeupEvent)
	{
	  //Power on
	  handle_power_on (notify, &prefs);
	}
      if (notify->notifyType == sysNotifyResetFinishedEvent)
	{
	  if (prefs.active && prefs.chk_reset)
		set_alarm(AUTOLOCK_RESET_TIMEOUT);
	}
      if (prefs.active)
	set_notify ();
    }
  else if (launchCode == sysAppLaunchCmdSystemReset)
    {
      get_prefs(&prefs);

      if (prefs.active)
	set_notify ();
    }
  else if (launchCode == sysAppLaunchCmdDisplayAlarm)
    {
      get_prefs(&prefs);

      if (prefs.active)
	do_lock ();
    }

  return err;
}
