;**************************************************************************
;*  emuvideo.asm
;*			Video emulation routines for PalmBoy
;*
;*  2000-09-10	Bodo Wenzel	Got from runmode.asm
;*  2000-09-17	Bodo Wenzel	Now nearly full screen emulation
;*  2000-10-23	Bodo Wenzel	Interrupts and timer
;*  2000-11-18	Bodo Wenzel	Support of color devices
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== Notes ================================================================

;* Unlike documented for the real part, this emulation never
;  shows 153/0 in the last line to the cpu, but 255.
;* For speeding up the timings for LCD status modes 2 and 3
;  are shortened.
;real:	152 gap 153/0 line   0   line ... 143   vsync 144 gap ...
;emu:	152 gap 255   line   0   line ... 143   vsync 144 gap ...
;i_ly:	-10 gap 143   line 142   line ...  -1   vsync  -2 gap ...
;stat:	1       1          2/3/0      ... 2/3/0       1       ...
;
;* Unlike the real part the interrupts are just sampled at vsync
;  (for v-blank and joypad) and at the end of each line (for
;  LYC, timer, and serial[currently not emulated]).
;* Hsync interrupts are currently not emulated.
;
;* The display is built by different functions, each called for just one
;  line. In general every line function uses preconverted patterns and
;  masks. There are counters to manage the wrapping at borders, too.
;* The objects are drawn at last.
;* If data is written into video relevant memory, conversion takes place
;  immediate. Since memory contents change not as often as screen output is
;  made, this saves time.

;=== Constants ============================================================

vidWidthInBytes	equ	SCR_WIDTH*2

;=== Global variables =====================================================

	data

global	VidLineFunction.l		;address of next render function
global	VidBgCount1.w			;counter for loops, sum is 19
global	VidBgCount2.w
global	VidBgWiCount1.w			;if <0: no wrap before window
global	VidBgWiCount2.w			;if <0: only window
global	VidBgWiCount3.w			;if <0: window not visible

global	VidDisplayPtr.l			;pointer into Palm display

global	VidBgWiColorH.l			;addresses of color functions
global	VidBgWiColorL.l
global	VidBgWiTileLow.l		;pointer for tiles $00..$7f
global	VidBgWiTileHigh.l		;pointer for tiles $80..$ff
global	VidBgWiMask.40	;vidWidthInBytes	;mask buffer one line

vidMapIndicesSize	equ	2*MAP_WRAP*MAP_WRAP*2
global	VidMapIndices.l			;pointer to allocated MapIndices

	;both of pattern and mask are
	;  8 pixel lines in
	;    $100 possible tiles of
	;      1 shifted word as long
struct	VidBgRamType
	used.$100			;which tiles are used
	tile.$4000			;converted tiles: 2*8*$100*4
endstruct
global	VidBgRam.l			;pointer to allocated VidBgRamType
global	VidBgUsed.l
global	VidBgTile.l

global	VidBgScxPixel.w			;prehandled scroll
global	VidBgScxTile.w
global	VidBgMapStart.l			;pointer into background
global	VidBgMapOffset.w
global	VidBgTilePtr.l			;pointer to converted tiles
global	VidBgTileMax.l

	;both of pattern and mask are
	;  8 pixel lines in
	;    $100 possible tiles of
	;      1 shifted word as long
struct	VidWiRamType
	used.$100			;which tiles are used
	tile.$4000			;converted tiles: 2*8*$100*4
endstruct
global	VidWiRam.l			;pointer to allocated VidWiRamType
global	VidWiUsed.l
global	VidWiTile.l

global	VidWiWxPixel.w			;prehandled scroll
global	VidWiWxMask.w
global	VidWiMapStart.l			;pointer into window
global	VidWiMapOffset.w
global	VidWiTilePtr.l			;pointer to converted tiles
global	VidWiTileMax.l

	;(SCR_HEIGHT*8) lines with
	;  1 word offset to first background object
	;  1 word offset to first foreground object
	;40 objects of
	;  16 pixel lines with
	;    1 word screen horizontal offset
	;    1 shifted word as long mask
	;    1 shifted word as long pattern
	;    1 word next offset
	;    1 word prev offset
struct	VidObRamType
	dummy.w				;for offset -2
	first.576			;offsets: SCR_HEIGHT*8*2*2
	tiles.8960			;converted tiles: 40*16*(2+4+4+2+2)
endstruct
global	VidObRam.l			;pointer to allocated VidObRamType
global	VidObFirst.l
global	VidObTiles.l

global	VidObYOffset.w			;offset into offset table
global	VidObColor0H.l			;addresses of color functions
global	VidObColor0L.l
global	VidObColor1H.l
global	VidObColor1L.l
global	VidObCount.w			;prehandled height
global	VidObTilemask.w

global	VidIntMaskVbl.b			;mask for vblank int
global	VidIntMaskStat.b		;mask for STAT int
global	VidIntMaskSV.b			;mask for STAT, vblank
global	VidIntLycReload.b		;counter for STAT, LYC
global	VidIntLycCount.b
	align	2

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;Boolean VidInit(DmOpenRef dbP)
;------------------------------
;	Get memory for several purposes (in the given database). If
;	successful return true otherwise false.
proc	VidInit(dbP.l)
local	recId.w
beginproc
	clr.l	VidMapIndices(a5)
	clr.l	VidBgRam(a5)
	clr.l	VidWiRam(a5)
	clr.l	VidObRam(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),#vidMapIndicesSize.l)
	move.l	a0,d0
	beq	ViReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,VidMapIndices(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	move.l	#sizeof(VidBgRamType),d0
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),d0.l)
	move.l	a0,d0
	beq	ViReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,VidBgRam(a5)
	lea	VidBgRamType.used(a0),a1
	move.l	a1,VidBgUsed(a5)
	lea	VidBgRamType.tile(a0),a1
	move.l	a1,VidBgTile(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	move.l	#sizeof(VidWiRamType),d0
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),d0.l)
	move.l	a0,d0
	beq.s	ViReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,VidWiRam(a5)
	lea	VidWiRamType.used(a0),a1
	move.l	a1,VidWiUsed(a5)
	lea	VidWiRamType.tile(a0),a1
	move.l	a1,VidWiTile(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	move.l	#sizeof(VidObRamType),d0
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),d0.l)
	move.l	a0,d0
	beq.s	ViReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,VidObRam(a5)
	lea	VidObRamType.first(a0),a1
	move.l	a1,VidObFirst(a5)
	lea	VidObRamType.tiles(a0),a1
	move.l	a1,VidObTiles(a5)

	moveq	#true,d0
	bra.s	ViReturn

ViReturnError
	moveq	#false,d0
ViReturn
endproc

;--------------------------------------------------------------------------
;void VidExit(void)
;------------------
;	Release the used memory.
proc	VidExit()
beginproc
	move.l	VidMapIndices(a5),d0
	beq.s	VeNoMap

	systrap	MemPtrUnlock(d0.l)
	clr.l	VidMapIndices(a5)
VeNoMap

	move.l	VidBgRam(a5),d0
	beq.s	VeNoBg

	systrap	MemPtrUnlock(d0.l)
	clr.l	VidBgRam(a5)
VeNoBg

	move.l	VidWiRam(a5),d0
	beq.s	VeNoWi

	systrap	MemPtrUnlock(d0.l)
	clr.l	VidWiRam(a5)
VeNoWi

	move.l	VidObRam(a5),d0
	beq.s	VeNoOb

	systrap	MemPtrUnlock(d0.l)
	clr.l	VidObRam(a5)
VeNoOb
endproc

;--------------------------------------------------------------------------
;void VidSupport(void)
;---------------------
;	This function is called after the last screen line. The function
;	returns to the caller of the emulator if the emulation quits.
VidSupport
vsRegs		reg	d3/a0-a1
	movem.l	vsRegs,-(sp)		;save used registers

	move.w	#OPS_PER_LINE-1,d6

	add.w	#DIV_PER_LINE,T_DIV(a4)

	move.w	T_D_TIMA(a4),d0
	add.w	d0,T_TIMA(a4)
	bcc.s	VsTExit
	move.b	T_TMA(a4),d0
VsTLoop
	add.b	d0,T_TIMA(a4)
	bcs.s	VsTLoop
	bset	#2,INT_IF(a4)
VsTExit

	call	RunUserPoll()
	tst.b	d0
	beq.s	VsGoon			;continue emulation?

	movem.l	(sp)+,vsRegs
	addq.w	#4,sp			;ignore return to caller

	moveq	#USER_STOP,d0		;return to emulation caller
	rts
VsGoon
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	start a new screen
	move.w	#SCR_YMAX-1-(SCR_HEIGHT*8),LCD_I_LY(a4)
	lea	VidLineGap(pc),a2
	move.l	a2,V_LCD_LINE+2(a4)

	movem.l	(sp)+,vsRegs
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	check if interrupts have to be processed, and
;	return to emulation
	move.b	VidIntMaskSV(a5),d0
	or.b	VidIntMaskVbl(a5),d0
	or.b	d0,INT_IF(a4)
	jmp	INT_CHECK(a4)

;--------------------------------------------------------------------------
;void VidLineGap(void)
;---------------------
;	One line of nothing, just counting LY. After the end of vsync
;	prepare screen output.
VidLineGap
	move.w	LCD_I_LY(a4),d0
	subq.w	#1,d0
	cmp.w	#SCR_YMAX-1-153,d0
	beq.s	VlgNewPic		;vsync finished?
	move.w	d0,LCD_I_LY(a4)
	bra	VidLineQuitReturn
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	start a new picture, no need to check for on/off
VlgNewPic
	move.w	#SCR_YMAX-1+1,LCD_I_LY(a4)

	move.b	LCD_SCY(a4),d6
	add.w	d6,d6
	add.w	d6,d6
	move.b	d6,-(sp)
	and.w	#(MAP_WRAP-1)*MAP_WRAP,d6
	add.w	VidBgScxTile(a5),d6
	add.w	d6,d6
	move.w	d6,VidBgMapOffset(a5)
	move.w	(sp)+,d6
	and.w	#7*$100*4,d6
	move.l	VidBgTile(a5),a3
	lea	8*$100*4(a3),a2
	add.w	d6,a3
	move.l	a3,VidBgTilePtr(a5)
	move.l	a2,VidBgTileMax(a5)	;set up for background

	moveq	#0,d1
	move.w	d1,VidWiMapOffset(a5)
	move.b	LCD_WY(a4),d1
	move.w	d1,LCD_I_WY(a4)
	move.l	VidWiTile(a5),a3
	lea	8*$100*4(a3),a2
	move.l	a3,VidWiTilePtr(a5)
	move.l	a2,VidWiTileMax(a5)	;set up for window

	clr.w	VidObYOffset(a5)	;set up for objects

	bsr	GrayCopyBuffer
	move.l	a2,VidDisplayPtr(a5)

	move.l	VidLineFunction(a5),V_LCD_LINE+2(a4)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	return to emulation
	move.w	#OPS_PER_LINE-1,d6

	add.w	#DIV_PER_LINE,T_DIV(a4)

	move.w	T_D_TIMA(a4),d0
	add.w	d0,T_TIMA(a4)
	bcc.s	VlgTExit
	move.b	T_TMA(a4),d0
VlgTLoop
	add.b	d0,T_TIMA(a4)
	bcs.s	VlgTLoop
	bset	#2,INT_IF(a4)
VlgTExit

	move.b	VidIntLycReload(a5),VidIntLycCount(a5)
	bne.s	VlgLycExit
	btst	#6,LCD_STAT(a4)
	beq.s	VlgLycExit
	subq.b	#1,VidIntLycCount(a5)
	move.b	VidIntMaskStat(a5),d0
	or.b	d0,INT_IF(a4)
VlgLycExit

	jmp	INT_CHECK(a4)

;--------------------------------------------------------------------------
;void VidLineNone(void)
;----------------------
;	One line of empty screen.
VidLineNone
;a2	pointer into palm display

	move.l	VidDisplayPtr(a5),a2

	moveq	#vidWidthInBytes/4/2-1,d0
	moveq	#0,d1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	erase line
VlnLoop
	move.l	d1,(a2)+
	move.l	d1,(a2)+
	dbra	d0,VlnLoop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update pointer
	move.l	a2,VidDisplayPtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	LY stuff
VidLineQuit
	subq.w	#1,LCD_I_LY(a4)
	bcc.s	VidLineQuitReturn	;rendering not finished?

	lea	VidSupport(pc),a2
	move.l	a2,V_LCD_LINE+2(a4)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	return to emulation
VidLineQuitReturn
	move.w	#OPS_PER_LINE-1,d6

	add.w	#DIV_PER_LINE,T_DIV(a4)

	move.w	T_D_TIMA(a4),d0
	add.w	d0,T_TIMA(a4)
	bcc.s	VlqrTExit
	move.b	T_TMA(a4),d0
VlqrTLoop
	add.b	d0,T_TIMA(a4)
	bcs.s	VlqrTLoop
	bset	#2,INT_IF(a4)
VlqrTExit

	subq.b	#1,VidIntLycCount(a5)
	bcc.s	VlqrLycExit
	btst	#6,LCD_STAT(a4)
	beq.s	VlqrLycExit
	move.b	VidIntMaskStat(a5),d0
	or.b	d0,INT_IF(a4)
VlqrLycExit

	jmp	INT_CHECK(a4)

;--------------------------------------------------------------------------
;void VidLineOb(void)
;--------------------
;	One line of objects.
VidLineOb
;a0	pointer into palm display incl. offset
;a2	pointer into palm display
;a3	pointer to pixel line in object
;a6	pointer VidObFirst

	move.l	a0,-(sp)

	move.l	VidDisplayPtr(a5),a2

	moveq	#vidWidthInBytes/4/2-1,d0
	moveq	#0,d1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	erase line
VloLoop
	move.l	d1,(a2)+
	move.l	d1,(a2)+
	dbra	d0,VloLoop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update offsets and pointers
	move.l	a2,VidDisplayPtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects behind background/window
	lea	-vidWidthInBytes(a2),a2

	move.l	VidObFirst(a5),a6
	move.w	VidObYOffset(a5),d0
	move.w	(a6,d0.w),d1
	bmi.s	VloBgQuit		;no objects behind backgr./window?

VloBgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a3)+,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VloBgLoop		;more objects?
VloBgQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects in front of background/window
	addq.w	#2,d0
	move.w	(a6,d0.w),d1
	bmi.s	VloFgQuit		;no objects in front of bg./wi.?

VloFgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a3)+,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VloFgLoop		;more objects?
VloFgQuit

	addq.w	#2,d0
	move.w	d0,VidObYOffset(a5)

	move.l	(sp)+,a0
	bra	VidLineQuit

;--------------------------------------------------------------------------
;void VidLineBgCh(void)
;----------------------
;	One line of background, with checking for window.
VidLineBgCh
	move.w	LCD_I_WY(a4),d0
	dbmi	d0,VlbcBg		;window not reached?
	move.w	d0,LCD_I_WY(a4)

	lea	VidLineBgWi(pc),a2
	move.l	a2,V_LCD_LINE+2(a4)
	jmp	(a2)			;now render with window

VlbcBg
	move.w	d0,LCD_I_WY(a4)

	;falls through to VidLineBg

;--------------------------------------------------------------------------
;void VidLineBg(void)
;--------------------
;	One line of background.
VidLineBg
;a0	pointer to left tile pattern
;a1	pointer to right tile pattern
;a2	pointer into palm display
;a3	pointer to pixel line base
;a6	pointer to actual tile number in background

	move.l	a0,-(sp)
	move.l	a1,-(sp)

	move.l	VidDisplayPtr(a5),a2
	move.l	VidBgTilePtr(a5),a3
	move.l	VidBgMapStart(a5),a6
	add.w	VidBgMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgCount2(a5),d0
	move.w	VidBgCount1(a5),d2
	bmi.s	VlbLoop2		;no line wrap?
	bra.s	VlbNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlbLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbNext1
	dbra	d2,VlbLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlbLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlbLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update pointers
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbExit			;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbNewline
	move.w	d0,VidBgMapOffset(a5)

VlbExit
	move.l	a3,VidBgTilePtr(a5)

	move.l	(sp)+,a1
	move.l	(sp)+,a0
	bra	VidLineQuit

;--------------------------------------------------------------------------
;void VidLineBgWi(void)
;----------------------
;	One line of background and window.
VidLineBgWi
;a0	pointer to left tile pattern
;a1	pointer to right tile pattern
;a2	pointer into palm display
;a3	pointer to pixel line base
;a6	pointer to actual tile number in background

	move.l	a0,-(sp)
	move.l	a1,-(sp)

	move.l	VidDisplayPtr(a5),a2
	move.l	VidBgTilePtr(a5),a3

	move.w	VidBgWiCount2(a5),d0
	bmi	VlbwWi			;only window?

	move.l	VidBgMapStart(a5),a6
	add.w	VidBgMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount1(a5),d2
	bmi.s	VlbwLoop2		;no wrap before window?
	bra.s	VlbwNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlbwLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwNext1
	dbra	d2,VlbwLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until window
VlbwLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlbwLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update background pointers
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwBgNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwBgExit
	move.l	a3,VidBgTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window after background
	move.w	VidBgWiCount3(a5),d0
	bmi	VlbwExit		;window not visible?

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	VidWiWxMask(a5),d2
	and.w	d2,-(a2)

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to first pattern

	move.w	(a0)+,d1
	or.w	d1,(a2)+		;store first pattern
	bra.s	VlbwNext3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window only, but first update background pointers
VlbwWi
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwWiBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwWiBgNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwWiBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwWiBgExit
	move.l	a3,VidBgTilePtr(a5)

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount3(a5),d0
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlbwLoop3
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwNext3
	dbra	d0,VlbwLoop3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update picture and window pointers
VlbwExit
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidWiTileMax(a5),a3
	bcs.s	VlbwWiExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	add.w	#MAP_WRAP*2,VidWiMapOffset(a5)

VlbwWiExit
	move.l	a3,VidWiTilePtr(a5)

	move.l	(sp)+,a1
	move.l	(sp)+,a0
	bra	VidLineQuit

;--------------------------------------------------------------------------
;void VidLineBgChOb(void)
;------------------------
;	One line of background and objects, with checking for window.
VidLineBgChOb
	move.w	LCD_I_WY(a4),d0
	dbmi	d0,VlbcoBg		;window not reached?
	move.w	d0,LCD_I_WY(a4)

	lea	VidLineBgWiOb(pc),a2
	move.l	a2,V_LCD_LINE+2(a4)
	jmp	(a2)			;now render with window

VlbcoBg
	move.w	d0,LCD_I_WY(a4)

	;falls through to VidLineBgOb

;--------------------------------------------------------------------------
;void VidLineBgOb(void)
;----------------------
;	One line of background and objects.
VidLineBgOb
vlboRegs	reg	a0/a1/a4
;a0	pointer to left tile pattern / palm display incl. offset
;a1	pointer to right tile pattern
;a2	pointer into palm display
;a3	pointer pixel line base / to pixel line in object
;a4	pointer to mask
;a6	pointer to actual tile number in background / VidObFirst

	movem.l	vlboRegs,-(sp)

	move.l	VidDisplayPtr(a5),a2
	move.l	VidBgTilePtr(a5),a3
	move.l	VidBgMapStart(a5),a6
	add.w	VidBgMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.l	VidObFirst(a5),a1
	move.w	VidObYOffset(a5),d1
	move.w	(a1,d1.w),d1
	bpl	VlboMasking		;objects behind background/window?

	move.w	VidBgCount2(a5),d0
	move.w	VidBgCount1(a5),d2
	bmi.s	VlboLoop2		;no line wrap?
	bra.s	VlboNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlboLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlboNext1
	dbra	d2,VlboLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlboLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlboLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update pointers
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlboExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlboNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlboNewline
	move.w	d0,VidBgMapOffset(a5)

VlboExit
	move.l	a3,VidBgTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	prepare for objects
	lea	-vidWidthInBytes(a2),a2

	move.l	VidObFirst(a5),a6
	bra	VlboForeground

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render and save mask for this line
VlboMasking
	lea	VidBgWiMask(a5),a4

	move.w	VidBgCount2(a5),d0
	move.w	VidBgCount1(a5),d2
	bmi.s	VlboMLoop2		;no line wrap?
	bra.s	VlboMNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlboMLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	8*$100*4(a1),d1
	or.w	8*$100*4(a0),d1
	move.w	d1,(a4)+		;store mask

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlboMNext1
	dbra	d2,VlboMLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlboMLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	8*$100*4(a1),d1
	or.w	8*$100*4(a0),d1
	move.w	d1,(a4)+		;store mask

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlboMLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update pointers
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlboMExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlboMNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlboMNewline
	move.w	d0,VidBgMapOffset(a5)

VlboMExit
	move.l	a3,VidBgTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects behind background/window
	moveq	#vidWidthInBytes,d1
	sub.w	d1,a2
	sub.w	d1,a4

	move.l	VidObFirst(a5),a6
	move.w	VidObYOffset(a5),d0
	move.w	(a6,d0.w),d1

VlboBgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a4,d1.w),d0
	move.l	(a3)+,d2
	or.l	d0,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	not.l	d0
	and.l	d0,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VlboBgLoop		;more objects?
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects in front of background/window
VlboForeground
	move.w	VidObYOffset(a5),d0
	addq.w	#2,d0
	move.w	(a6,d0.w),d1
	bmi.s	VlboFgQuit		;no objects in front of bg./window?

VlboFgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a3)+,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VlboFgLoop		;more objects?
VlboFgQuit

	addq.w	#2,d0
	move.w	d0,VidObYOffset(a5)

	movem.l	(sp)+,vlboRegs
	bra	VidLineQuit

;--------------------------------------------------------------------------
;void VidLineBgWiOb(void)
;------------------------
VidLineBgWiOb
;	One line of background, window, and objects.
vlbwoRegs	reg	a0/a1/a4
;a0	pointer to left tile pattern / palm display incl. offset
;a1	pointer to right tile pattern
;a2	pointer into palm display
;a3	pointer pixel line base / to pixel line in object
;a4	pointer to mask
;a6	pointer to actual tile number in background / VidObFirst

	movem.l	vlbwoRegs,-(sp)

	move.l	VidDisplayPtr(a5),a2
	move.l	VidBgTilePtr(a5),a3

	move.l	VidObFirst(a5),a1
	move.w	VidObYOffset(a5),d1
	move.w	(a1,d1.w),d1
	bpl	VlbwoMasking		;objects behind backgr./window?

	move.w	VidBgWiCount2(a5),d0
	bmi	VlbwoWi			;only window?

	move.l	VidBgMapStart(a5),a6
	add.w	VidBgMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount1(a5),d2
	bmi.s	VlbwoLoop2		;no wrap before window?
	bra.s	VlbwoNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlbwoLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwoNext1
	dbra	d2,VlbwoLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until window
VlbwoLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlbwoLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update background pointers
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwoBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwoBgNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwoBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwoBgExit
	move.l	a3,VidBgTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window after background
	move.w	VidBgWiCount3(a5),d0
	bmi	VlbwoExit		;window not visible?

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	VidWiWxMask(a5),d2
	and.w	d2,-(a2)

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to first pattern

	move.w	(a0)+,d1
	or.w	d1,(a2)+		;store first pattern
	bra.s	VlbwoNext3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window only, but first update background pointers
VlbwoWi
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwoWiBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwoWiBgNewline	;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwoWiBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwoWiBgExit
	move.l	a3,VidBgTilePtr(a5)

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount3(a5),d0
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlbwoLoop3
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwoNext3
	dbra	d0,VlbwoLoop3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update picture and window pointers
VlbwoExit
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidWiTileMax(a5),a3
	bcs.s	VlbwoWiExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	add.w	#MAP_WRAP*2,VidWiMapOffset(a5)

VlbwoWiExit
	move.l	a3,VidWiTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	prepare for objects
	lea	-vidWidthInBytes(a2),a2

	move.l	VidObFirst(a5),a6
	bra	VlbwoForeground

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render and save mask for this line
VlbwoMasking
	lea	VidBgWiMask(a5),a4

	move.w	VidBgWiCount2(a5),d0
	bmi	VlbwoMWi		;only window?

	move.l	VidBgMapStart(a5),a6
	add.w	VidBgMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount1(a5),d2
	bmi.s	VlbwoMLoop2		;no wrap before window?
	bra.s	VlbwoMNext1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until wrap
VlbwoMLoop1
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	8*$100*4(a1),d1
	or.w	8*$100*4(a0),d1
	move.w	d1,(a4)+		;store mask

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwoMNext1
	dbra	d2,VlbwoMLoop1

	lea	-MAP_WRAP*2(a6),a6	;wrap pointer to tile number
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until window
VlbwoMLoop2
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	8*$100*4(a1),d1
	or.w	8*$100*4(a0),d1
	move.w	d1,(a4)+		;store mask

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns

	dbra	d0,VlbwoMLoop2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update background pointers
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwoMBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwoMBgNewline		;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwoMBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwoMBgExit
	move.l	a3,VidBgTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window after background
	move.w	VidBgWiCount3(a5),d0
	bmi	VlbwoMExit		;window not visible?

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	VidWiWxMask(a5),d2
	and.w	d2,-(a2)
	and.w	d2,-(a4)

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to first pattern

	move.w	8*$100*4(a0),d1
	or.w	d1,(a4)+		;store mask

	move.w	(a0)+,d1
	or.w	d1,(a2)+		;store first pattern
	bra.s	VlbwoMNext3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render window only, but first update background pointers
VlbwoMWi
	lea	$100*4(a3),a3
	cmp.l	VidBgTileMax(a5),a3
	bcs.s	VlbwoMWiBgExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	moveq	#MAP_WRAP*2,d0
	add.w	VidBgMapOffset(a5),d0
	cmp.w	#MAP_WRAP*MAP_WRAP*2,d0
	bcs.s	VlbwoMWiBgNewline	;no vertical wrap?
	sub.w	#MAP_WRAP*MAP_WRAP*2,d0
VlbwoMWiBgNewline
	move.w	d0,VidBgMapOffset(a5)

VlbwoMWiBgExit
	move.l	a3,VidBgTilePtr(a5)

	move.l	VidWiTilePtr(a5),a3
	move.l	VidWiMapStart(a5),a6
	add.w	VidWiMapOffset(a5),a6

	move.w	(a6)+,d1
	lea	2(a3,d1.w),a0		;pointer to "old" right pattern

	move.w	VidBgWiCount3(a5),d0
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	render until finished
VlbwoMLoop3
	move.l	a0,a1			;right becomes left

	move.w	(a6)+,d1
	lea	(a3,d1.w),a0		;pointer to right pattern

	move.w	8*$100*4(a1),d1
	or.w	8*$100*4(a0),d1
	move.w	d1,(a4)+		;store mask

	move.w	(a1)+,d1
	or.w	(a0)+,d1
	move.w	d1,(a2)+		;join patterns
VlbwoMNext3
	dbra	d0,VlbwoMLoop3
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	update picture and window pointers
VlbwoMExit
	move.l	a2,VidDisplayPtr(a5)

	lea	$100*4(a3),a3
	cmp.l	VidWiTileMax(a5),a3
	bcs.s	VlbwoMWiExit		;not last pixel line?

	lea	-8*$100*4(a3),a3

	add.w	#MAP_WRAP*2,VidWiMapOffset(a5)

VlbwoMWiExit
	move.l	a3,VidWiTilePtr(a5)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects behind background/window
	moveq	#vidWidthInBytes,d1
	sub.w	d1,a2
	sub.w	d1,a4

	move.l	VidObFirst(a5),a6
	move.w	VidObYOffset(a5),d0
	move.w	(a6,d0.w),d1

VlbwoBgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a4,d1.w),d0
	move.l	(a3)+,d2
	or.l	d0,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	not.l	d0
	and.l	d0,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VlbwoBgLoop		;more objects?
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	objects in front of background/window
VlbwoForeground
	move.w	VidObYOffset(a5),d0
	addq.w	#2,d0
	move.w	(a6,d0.w),d1
	bmi.s	VlbwoFgQuit		;no objects in front of bg./window?

VlbwoFgLoop
	lea	(a6,d1.w),a3
	move.w	(a3)+,d1
	lea	(a2,d1.w),a0
	move.l	(a3)+,d2
	and.l	d2,(a0)
	move.l	(a3)+,d2
	or.l	d2,(a0)			;insert object

	move.w	(a3),d1
	bpl.s	VlbwoFgLoop		;more objects?
VlbwoFgQuit

	addq.w	#2,d0
	move.w	d0,VidObYOffset(a5)

	movem.l	(sp)+,vlbwoRegs
	bra	VidLineQuit

;--------------------------------------------------------------------------
;UInt16 VidColorTab[]
;--------------------
;	Color conversions for all cases (offsets to functions).
VidColorTab
	dc.w	VidColor0000-VidColorTab-0
	dc.w	VidColor0001-VidColorTab-2
	dc.w	VidColor0010-VidColorTab-4
	dc.w	VidColor0011-VidColorTab-6
	dc.w	VidColor0100-VidColorTab-8
	dc.w	VidColor0101-VidColorTab-10
	dc.w	VidColor0110-VidColorTab-12
	dc.w	VidColor0111-VidColorTab-14
	dc.w	VidColor1000-VidColorTab-16
	dc.w	VidColor1001-VidColorTab-18
	dc.w	VidColor1010-VidColorTab-20
	dc.w	VidColor1011-VidColorTab-22
	dc.w	VidColor1100-VidColorTab-24
	dc.w	VidColor1101-VidColorTab-26
	dc.w	VidColor1110-VidColorTab-28
	dc.w	VidColor1111-VidColorTab-30
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0000
	sub.b	d0,d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0001
	move.b	1(a3),d0
	or.b	(a3),d0
	not.b	d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0010
	move.b	1(a3),d0
	not.b	d0
	and.b	(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0011
	move.b	1(a3),d0
	not.b	d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0100
	move.b	(a3),d0
	not.b	d0
	and.b	1(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0101
	move.b	(a3),d0
	not.b	d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0110
	move.b	1(a3),d0
	move.b	(a3),d1
	eor.b	d1,d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor0111
	move.b	1(a3),d0
	and.b	(a3),d0
	not.b	d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1000
	move.b	1(a3),d0
	and.b	(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1001
	move.b	1(a3),d0
	move.b	(a3),d1
	eor.b	d1,d0
	not.b	d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1010
	move.b	(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1011
	move.b	1(a3),d0
	not.b	d0
	or.b	(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1100
	move.b	1(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1101
	move.b	(a3),d0
	not.b	d0
	or.b	1(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1110
	move.b	1(a3),d0
	or.b	(a3),d0
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
VidColor1111
	move.b	#$ff,d0
	rts

;--------------------------------------------------------------------------
;void VidConvertBgWi(...)
;------------------------
;	Convert background or window tile.
VidConvertBgWi
;d0	(scratch)
;d1	(scratch)
;d2	shift left 0..7 pixel
;d3.w	(counter)
;a0,a1	addresses of color functions
;a2	pointer to destination, will be increased by 4
;a3	pointer to source, will be increased by 8*2
;a4	pointer to GrayCopyTable

	move.w	#8-1,d3
VcbwLoop
	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	swap	d0
	jsr	(a1)
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)
	moveq	#0,d1
	move.b	(a3)+,d1
	or.b	(a3)+,d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	move.l	d1,8*$100*4(a2)

	lea	$100*4(a2),a2

	dbra	d3,VcbwLoop
	lea	4-8*$100*4(a2),a2
	rts

;--------------------------------------------------------------------------
;void VidConvertBgWiAllMarked(...)
;---------------------------------
;	Convert all marked background or window tiles.
VidConvertBgWiAllMarked
	move.w	d0,-(sp)
	move.w	d1,-(sp)
	move.l	a4,-(sp)

	move.l	GrayCopyTable(a5),a4

	move.w	#$80-1,d3
	move.l	VidBgWiTileLow(a5),a3
VcbwamLLoop
	move.b	(a6)+,d0
	bmi.s	VcbwamLNext
	cmp.b	d0,d2
	bne.s	VcbwamLConv		;conversion necessary?
VcbwamLNext
	addq.w	#4,a2
	lea	2*8(a3),a3
	dbra	d3,VcbwamLLoop
	bra.s	VcbwamLReady
VcbwamLConv
	move.b	d2,-1(a6)
	swap	d3
	bsr	VidConvertBgWi
	swap	d3
	dbra	d3,VcbwamLLoop
VcbwamLReady				;convert tiles $00..$7f

	move.w	#$80-1,d3
	move.l	VidBgWiTileHigh(a5),a3
	lea	$800(a3),a3
VcbwamHLoop
	move.b	(a6)+,d0
	bmi.s	VcbwamHNext
	cmp.b	d0,d2
	bne.s	VcbwamHConv		;conversion necessary?
VcbwamHNext
	addq.w	#4,a2
	lea	2*8(a3),a3
	dbra	d3,VcbwamHLoop
	bra.s	VcbwamHReady
VcbwamHConv
	move.b	d2,-1(a6)
	swap	d3
	bsr	VidConvertBgWi
	swap	d3
	dbra	d3,VcbwamHLoop
VcbwamHReady				;convert tiles $80..$ff

	move.l	(sp)+,a4
	move.w	(sp)+,d1
	move.w	(sp)+,d0
	rts

;--------------------------------------------------------------------------
;void VidConvertBgWiAllUsed(...)
;-------------------------------
;	Convert all used background or window tiles.
VidConvertBgWiAllUsed
	move.l	a4,-(sp)

	move.l	GrayCopyTable(a5),a4

	move.w	#$80-1,d3
	move.l	VidBgWiTileLow(a5),a3
VcbwauLLoop
	move.b	(a6)+,d0
	bmi.s	VcbwauLNext
	cmp.b	d0,d2
	beq.s	VcbwauLConv		;conversion necessary?
VcbwauLNext
	addq.w	#4,a2
	lea	2*8(a3),a3
	dbra	d3,VcbwauLLoop
	bra.s	VcbwauLReady
VcbwauLConv
	move.b	d2,-1(a6)
	swap	d3
	bsr	VidConvertBgWi
	swap	d3
	dbra	d3,VcbwauLLoop
VcbwauLReady				;convert tiles $00..$7f

	move.w	#$80-1,d3
	move.l	VidBgWiTileHigh(a5),a3
	lea	$800(a3),a3
VcbwauHLoop
	move.b	(a6)+,d0
	bmi.s	VcbwauHNext
	cmp.b	d0,d2
	beq.s	VcbwauHConv		;conversion necessary?
VcbwauHNext
	addq.w	#4,a2
	lea	2*8(a3),a3
	dbra	d3,VcbwauHLoop
	bra.s	VcbwauHReady
VcbwauHConv
	move.b	d2,-1(a6)
	swap	d3
	bsr	VidConvertBgWi
	swap	d3
	dbra	d3,VcbwauHLoop
VcbwauHReady				;convert tiles $80..$ff

	move.l	(sp)+,a4
	rts

;--------------------------------------------------------------------------
;UInt8 VidReverseTable[0x100]
;----------------------------
;	Table of reversed bytes for object conversion.
VidReverseTable
	dc.b	$00,$80,$40,$c0,$20,$a0,$60,$e0
	dc.b	$10,$90,$50,$d0,$30,$b0,$70,$f0
	dc.b	$08,$88,$48,$c8,$28,$a8,$68,$e8
	dc.b	$18,$98,$58,$d8,$38,$b8,$78,$f8
	dc.b	$04,$84,$44,$c4,$24,$a4,$64,$e4
	dc.b	$14,$94,$54,$d4,$34,$b4,$74,$f4
	dc.b	$0c,$8c,$4c,$cc,$2c,$ac,$6c,$ec
	dc.b	$1c,$9c,$5c,$dc,$3c,$bc,$7c,$fc
	dc.b	$02,$82,$42,$c2,$22,$a2,$62,$e2
	dc.b	$12,$92,$52,$d2,$32,$b2,$72,$f2
	dc.b	$0a,$8a,$4a,$ca,$2a,$aa,$6a,$ea
	dc.b	$1a,$9a,$5a,$da,$3a,$ba,$7a,$fa
	dc.b	$06,$86,$46,$c6,$26,$a6,$66,$e6
	dc.b	$16,$96,$56,$d6,$36,$b6,$76,$f6
	dc.b	$0e,$8e,$4e,$ce,$2e,$ae,$6e,$ee
	dc.b	$1e,$9e,$5e,$de,$3e,$be,$7e,$fe
	dc.b	$01,$81,$41,$c1,$21,$a1,$61,$e1
	dc.b	$11,$91,$51,$d1,$31,$b1,$71,$f1
	dc.b	$09,$89,$49,$c9,$29,$a9,$69,$e9
	dc.b	$19,$99,$59,$d9,$39,$b9,$79,$f9
	dc.b	$05,$85,$45,$c5,$25,$a5,$65,$e5
	dc.b	$15,$95,$55,$d5,$35,$b5,$75,$f5
	dc.b	$0d,$8d,$4d,$cd,$2d,$ad,$6d,$ed
	dc.b	$1d,$9d,$5d,$dd,$3d,$bd,$7d,$fd
	dc.b	$03,$83,$43,$c3,$23,$a3,$63,$e3
	dc.b	$13,$93,$53,$d3,$33,$b3,$73,$f3
	dc.b	$0b,$8b,$4b,$cb,$2b,$ab,$6b,$eb
	dc.b	$1b,$9b,$5b,$db,$3b,$bb,$7b,$fb
	dc.b	$07,$87,$47,$c7,$27,$a7,$67,$e7
	dc.b	$17,$97,$57,$d7,$37,$b7,$77,$f7
	dc.b	$0f,$8f,$4f,$cf,$2f,$af,$6f,$ef
	dc.b	$1f,$9f,$5f,$df,$3f,$bf,$7f,$ff

;--------------------------------------------------------------------------
;void VidConvertOb(...)
;----------------------
;	Convert object tile.
VidConvertOb
;d0	object flags / (shifted pattern)
;d1	(shifted mask)
;d2	(shift left/right 0..7 pixel)
;d3.w	counter 16-1 / 8-1 / 1-1
;d4	object x+8 / (screen offset)
;d5	(mask)
;a0,a1	addresses of color functions
;a2	pointer to destination, may be increased
;a3	pointer to source, may be increased
;a4	pointer to GrayCopyTable
;a6	(pointer to VidReverseTable)

	tst.b	d4
	beq.s	VcoQuit
	cmp.b	#SCR_WIDTH*8+8,d4
	bcc.s	VcoQuit
	moveq	#0,d5
	moveq	#7,d2
	cmp.b	#SCR_WIDTH*8,d4
	bcc	VcoRight
	sub.b	#9,d4
	bcc.s	VcoMid
	not.b	d4
	and.b	d4,d2
	addq.b	#8,d2
	moveq	#0,d4
	bra.s	VcoLeft
VcoQuit
	rts

VcoMid
	eor.b	d2,d4
	and.b	d4,d2
	lsr.b	#3,d4
	ext.w	d4
	add.w	d4,d4

VcoLeft
	add.w	d2,d2

	moveq	#$40,d1
	and.b	d0,d1
	beq.s	VcoLXN
	moveq	#$20,d1
	and.b	d0,d1
	bne	VcoLFF
	bra.s	VcoLNF
VcoLXN
	moveq	#$20,d1
	and.b	d0,d1
	bne	VcoLFN
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting left x normal, y normal
VcoLNNLoop
	move.w	d4,(a2)+

	moveq	#0,d1
	move.b	(a3),d5
	or.b	1(a3),d5
	move.w	d5,d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#2,a3
	addq.w	#4,a2
	dbra	d3,VcoLNNLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting left x normal, y flipped
VcoLNF
	move.w	d3,d0
	addq.w	#1,d0
	add.w	d0,d0
	add.w	d0,a3
VcoLNFLoop
	move.w	d4,(a2)+

	moveq	#0,d1
	move.b	-(a3),d5
	or.b	-(a3),d5
	move.w	d5,d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#4,a2
	dbra	d3,VcoLNFLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting left x flipped, y normal
VcoLFN
	lea	VidReverseTable(pc),a6
VcoLFNLoop
	move.w	d4,(a2)+

	moveq	#0,d1
	move.b	(a3),d5
	or.b	1(a3),d5
	move.w	d5,d1
	move.b	(a6,d1.w),d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#2,a3
	addq.w	#4,a2
	dbra	d3,VcoLFNLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting left x flipped, y flipped
VcoLFF
	move.w	d3,d0
	addq.w	#1,d0
	add.w	d0,d0
	add.w	d0,a3
	lea	VidReverseTable(pc),a6
VcoLFFLoop
	move.w	d4,(a2)+

	moveq	#0,d1
	move.b	-(a3),d5
	or.b	-(a3),d5
	move.w	d5,d1
	move.b	(a6,d1.w),d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#4,a2
	dbra	d3,VcoLFFLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting right
VcoRight
	and.b	d4,d2
	add.w	d2,d2
	moveq	#2*(SCR_WIDTH-2),d4

	moveq	#$40,d1
	and.b	d0,d1
	beq.s	VcoRXN
	moveq	#$20,d1
	and.b	d0,d1
	bne	VcoRFF
	bra.s	VcoRNF
VcoRXN
	moveq	#$20,d1
	and.b	d0,d1
	bne	VcoRFN
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting right x normal, y normal
VcoRNNLoop
	move.w	d4,(a2)+

	move.b	(a3),d5
	or.b	1(a3),d5
	move.w	d5,d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsr.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsr.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#2,a3
	addq.w	#4,a2
	dbra	d3,VcoRNNLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting right x normal, y flipped
VcoRNF
	move.w	d3,d0
	addq.w	#1,d0
	add.w	d0,d0
	add.w	d0,a3
VcoRNFLoop
	move.w	d4,(a2)+

	move.b	-(a3),d5
	or.b	-(a3),d5
	move.w	d5,d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsr.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsr.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#4,a2
	dbra	d3,VcoRNFLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting right x flipped, y normal
VcoRFN
	lea	VidReverseTable(pc),a6
VcoRFNLoop
	move.w	d4,(a2)+

	move.b	(a3),d5
	or.b	1(a3),d5
	move.w	d5,d1
	move.b	(a6,d1.w),d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsr.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsr.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#2,a3
	addq.w	#4,a2
	dbra	d3,VcoRFNLoop
	rts
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert shifting right x flipped, y flipped
VcoRFF
	move.w	d3,d0
	addq.w	#1,d0
	add.w	d0,d0
	add.w	d0,a3
	lea	VidReverseTable(pc),a6
VcoRFFLoop
	move.w	d4,(a2)+

	move.b	-(a3),d5
	or.b	-(a3),d5
	move.w	d5,d1
	move.b	(a6,d1.w),d1
	add.w	d1,d1
	move.w	(a4,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsr.l	d2,d1
	not.l	d1
	move.l	d1,(a2)+

	moveq	#0,d0
	moveq	#0,d1
	jsr	(a0)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	swap	d0
	jsr	(a1)
	and.b	d5,d0
	move.b	(a6,d0.w),d0
	add.l	d0,d0
	move.w	(a4,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a4,d0.w),d1
	lsr.l	d2,d1
	move.l	d1,(a2)+

	addq.w	#4,a2
	dbra	d3,VcoRFFLoop
	rts

;--------------------------------------------------------------------------
;void VidObLink(...)
;void VidObLinkIn(...)
;---------------------
;	Link an object (two entries, the 2nd for initialization).
VidObLink
;d0	offset to object in VidObTiles
;d1	(scratch)
;d2	(scratch)
;d3.w	(counter 16-1 / 8-1)
;d4.w	(scratch)
;a3	pointer into oam
;a6	pointer VidObFirst

	move.w	d0,d4
	add.w	#5*2,d4
	move.w	VidObCount(a5),d3
VolOutLoop
	move.w	(a6,d4.w),d1
	move.w	2(a6,d4.w),d2
	move.w	d1,(a6,d2.w)
	move.w	d2,6*2(a6,d1.w)
	add.w	#7*2,d4
	dbra	d3,VolOutLoop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	check for visibility
VidObLinkIn
	move.w	d0,d4
	move.w	VidObCount(a5),d3

	moveq	#0,d1
	move.b	(a3),d1
	beq.s	VolOff
	sub.w	#16,d1
	bcs.s	VolTopclip
	move.b	1(a3),d2
	beq.s	VolOff
	cmp.b	#SCR_WIDTH*8+8,d2
	bcc.s	VolOff
	move.w	#SCR_HEIGHT*8-1,d2
	sub.w	d1,d2
	bcs.s	VolOff
	sub.w	d3,d2
	bcc.s	VolGoon
	add.w	d2,d3
	not.w	d2
	move.w	d2,-(sp)
	bsr.s	VolGoon
	move.w	(sp)+,d3

VolOff
	add.w	#5*2,d4
	moveq	#-7*2,d2
VolOLoop
	move.w	d2,(a6,d4.w)
	move.w	d4,2(a6,d4.w)
	sub.w	d2,d4
	dbra	d3,VolOLoop
	rts

VolTopclip
	add.w	#5*2,d4
	add.w	d1,d3
	not.w	d1
	moveq	#-7*2,d2
VolTLoop
	move.w	d2,(a6,d4.w)
	move.w	d4,2(a6,d4.w)
	sub.w	d2,d4
	dbra	d1,VolTLoop

	tst.w	d3
	bmi.s	VolQuit
	moveq	#0,d1
	sub.w	#5*2,d4
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	now link lines
VolGoon
	swap	d0

	add.w	d1,d1
	add.w	d1,d1
	moveq	#$80-$100,d2
	and.b	3(a3),d2
	bne.s	VolLLoop1
	addq.w	#2,d1			;offset background / foreground

VolLLoop1
	move.w	d1,d0
VolLLoop2
	move.w	(a6,d0.w),d2
	cmp.w	d2,d4
	bge.s	VolLInsert
	move.w	d2,d0
	add.w	#5*2,d0
	bra.s	VolLLoop2

VolLInsert
	move.w	d4,(a6,d0.w)
	add.w	#5*2,d4
	move.w	d2,(a6,d4.w)
	move.w	d0,2(a6,d4.w)
	add.w	#6*2,d2
	move.w	d4,(a6,d2.w)
	addq.w	#2*2,d4
	addq.w	#2*2,d1
	dbra	d3,VolLLoop1

	swap	d0
VolQuit
	rts

;--------------------------------------------------------------------------
;void VidObTileCheck(...)
;------------------------
;	Check if any object uses this tile and convert.
VidObTileCheck
votcRegs	reg	d1-d3/d5/a0-a4/a6
;d0	(scratch)
;d1.w	offset into video ram
;d2	(scratch)
;d3.w	(scratch)
;a2	base address of video ram
;a3	(scratch)

	move.l	d4,-(sp)
	move.w	d1,d2
	swap	d1

	lsr.w	#4,d2
	move.w	VidObTilemask(a5),d1
	move.l	BASE_RAM_OAM(a4),a3
	lea	$fe00-$10000(a3),a3
	move.w	#40-1,d3
VotcLoop
	move.w	(a3)+,d4
	beq.s	VotcNext
	move.b	(a3),d0
	eor.b	d2,d0
	and.w	d1,d0
	bne.s	VotcNext

	movem.l	votcRegs,-(sp)

	move.w	(a3),d0
	move.w	d1,d2
	ext.w	d2
	lsl.w	#4,d2
	not.w	d2
	swap	d1
	and.w	#$fffe,d1
	and.w	d1,d2
	lea	(a2,d1.w),a3
	move.l	VidObTiles(a5),a2
	moveq	#40-1,d1
	sub.w	d3,d1
	lsl.w	#4+1,d1
	add.w	d1,d2
	move.w	d2,d1
	lsl.w	#3,d2
	sub.w	d1,d2
	add.w	d2,a2
	moveq	#$10,d2
	and.b	d0,d2
	beq.s	VotcPal0
	move.l	VidObColor1L(a5),a0
	move.l	VidObColor1H(a5),a1
	bra.s	VotcConv
VotcPal0
	move.l	VidObColor0L(a5),a0
	move.l	VidObColor0H(a5),a1
VotcConv
	moveq	#1-1,d3
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertOb		;convert one line
	movem.l	(sp)+,votcRegs
VotcNext
	addq.w	#2,a3
	dbra	d3,VotcLoop

	swap	d1
	move.l	(sp)+,d4
	rts

;--------------------------------------------------------------------------
;void VidObLinkConvertAll(...)
;-----------------------------
;	Link and convert from scratch.
VidObLinkConvertAll
;d0-d5/a0-a3/a6	(scratch, not saved)

	move.l	#((-7*2)<<16)+$10000-7*2,d2
	move.l	VidObFirst(a5),a2
	moveq	#SCR_HEIGHT*8*2/4/2-1,d3
VolcaClearLoop
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	dbra	d3,VolcaClearLoop	;clear line offsets

	move.l	BASE_RAM_OAM(a4),a3
	lea	$fe00+40*4-$10000(a3),a3
	move.l	VidObFirst(a5),a6
	move.w	#SCR_HEIGHT*8*2*2+40*16*7*2,d0
	moveq	#40-1,d5
VolcaLinkLoop
	subq.w	#4,a3
	sub.w	#16*7*2,d0
	bsr	VidObLinkIn
	dbra	d5,VolcaLinkLoop	;link in all objects

	move.l	VidObTiles(a5),a2
	move.l	a3,a6
	move.l	BASE_RAM_VID(a4),a3
	lea	$8000-$10000(a3),a3
	move.l	a4,-(sp)
	move.l	GrayCopyTable(a5),a4
	moveq	#40-1,d3
VolcaConvLoop
	move.w	(a6)+,d4
	move.w	VidObTilemask(a5),d1
	and.b	(a6)+,d1
	move.b	(a6)+,d0
	moveq	#$10,d2
	and.b	d0,d2
	beq.s	VolcaPal0
	move.l	VidObColor1L(a5),a0
	move.l	VidObColor1H(a5),a1
	bra.s	VolcaConv
VolcaPal0
	move.l	VidObColor0L(a5),a0
	move.l	VidObColor0H(a5),a1
VolcaConv
	swap	d3
	movem.l	a2/a3/a6,-(sp)
	lsl.w	#3+1,d1
	add.w	d1,a3
	move.w	VidObCount(a5),d3
	bsr	VidConvertOb
	movem.l	(sp)+,a2/a3/a6
	swap	d3
	lea	16*7*2(a2),a2
	dbra	d3,VolcaConvLoop	;convert all objects

	move.l	(sp)+,a4
	rts

;--------------------------------------------------------------------------
;void VidIoWrLcdc(data.b)
;------------------------
;	Write to LCD control.
VidIoWrLcdc
viwlRegs	reg	d3/a0-a1

	pea	(a6)			;save return address

	move.b	(a3),d1
	eor.b	d0,d1
	beq	ViwlNop			;no changes?
	move.b	d0,(a3)

	movem.l	viwlRegs,-(sp)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 7
	moveq	#7,d2			;LCD on: 0/1 = off/on
	btst	d2,d0
	bne.s	Viwl7On			;LCD on?

	btst	d2,d1
	beq	ViwlQuit		;LCD still off?

	lea	MemRamNone(pc),a3
	move.l	a3,V_RAM_WR8087+2(a4)
	move.l	a3,V_RAM_WR888F+2(a4)
	move.l	a3,V_RAM_WR9097+2(a4)
	move.l	a3,V_RAM_WR_OAM+2(a4)
	lea	MemRamMapNone(pc),a3
	move.l	a3,V_RAM_WR989B+2(a4)
	move.l	a3,V_RAM_WR9C9F+2(a4)

	lea	VidLineNone(pc),a6
	move.l	a6,VidLineFunction(a5)

	moveq	#0,d2
	move.b	d2,VidIntMaskVbl(a5)
	move.b	d2,VidIntMaskStat(a5)
	bra	ViwlQuit

Viwl7On
	btst	d2,d1
	beq	Viwl7Quit		;LCD still on?

	or.b	#$01,d1
	and.b	#$fb,d1			;force change bit 0 on / 2 off

	btst	#2,d0
	beq.s	Viwl7Small
	move.w	#16-1,VidObCount(a5)
	move.w	#$fe,VidObTilemask(a5)
	bra.s	Viwl7Goon
Viwl7Small
	move.w	#8-1,VidObCount(a5)
	move.w	#$ff,VidObTilemask(a5)	;set up some variables

Viwl7Goon
	movem.l	d0/d1/d4/d5,-(sp)
	bsr	VidObLinkConvertAll
	movem.l	(sp)+,d0/d1/d4/d5

	lea	MemRamOam(pc),a3
	move.l	a3,V_RAM_WR_OAM+2(a4)

	move.w	#SCR_YMAX-153,LCD_I_LY(a4)
	lea	VidLineGap(pc),a6
	move.l	a6,V_LCD_LINE+2(a4)

	move.b	#01,VidIntMaskVbl(a5)
	move.b	#02,VidIntMaskStat(a5)
Viwl7Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 0
	moveq	#0,d2			;background/window on: 0/1 = off/on
	btst	d2,d0
	bne.s	Viwl0On			;background/window on?

	btst	d2,d1
	beq	Viwl1BgWiOff		;background/window still off?

	lea	MemRamTileOb(pc),a3
	move.l	a3,V_RAM_WR8087+2(a4)
	move.l	a3,V_RAM_WR888F+2(a4)
	lea	MemRamNone(pc),a3
	move.l	a3,V_RAM_WR9097+2(a4)
	lea	MemRamMapNone(pc),a3
	move.l	a3,V_RAM_WR989B+2(a4)
	move.l	a3,V_RAM_WR9C9F+2(a4)

	lea	VidLineNone(pc),a6
	btst	#1,d0
	beq.s	Viwl0ObOff		;objects off?
	move.w	#SCR_YMAX,d2
	sub.w	LCD_I_LY(a4),d2
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,VidObYOffset(a5)
	lea	VidLineOb(pc),a6
Viwl0ObOff
	move.l	a6,VidLineFunction(a5)
	bra	Viwl1BgWiOff

Viwl0On
	btst	d2,d1
	beq.s	Viwl0Quit		;background/window still on?

	or.b	#$38,d1			;force change bits 5..3

	move.w	#$100/4-1,d3
	move.l	VidBgUsed(a5),a6
	moveq	#-1,d2
Viwl0Loop
	move.l	d2,(a6)+
	dbra	d3,Viwl0Loop		;no background tiles used
Viwl0Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 5
	moveq	#5,d2			;window on: 0/1 = off/on
	btst	d2,d0
	bne	Viwl5On			;window on?

	btst	d2,d1
	beq	Viwl3WiOff		;window still off?

	lea	MemRamTileOb(pc),a1
	lea	MemRamTileBgOb(pc),a2
	lea	MemRamTileBg(pc),a3
	btst	#4,d0
	beq.s	Viwl5Signed		;tile type signed?
	move.l	a2,a1
	lea	MemRamNone(pc),a3
Viwl5Signed
	move.l	a1,V_RAM_WR8087+2(a4)
	move.l	a2,V_RAM_WR888F+2(a4)
	move.l	a3,V_RAM_WR9097+2(a4)

	lea	VidLineBg(pc),a6
	btst	#1,d0
	beq.s	Viwl5ObOff		;objects off?
	move.w	#SCR_YMAX,d2
	sub.w	LCD_I_LY(a4),d2
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,VidObYOffset(a5)
	lea	VidLineBgOb(pc),a6
Viwl5ObOff
	move.l	a6,VidLineFunction(a5)

	lea	MemRamMapBg(pc),a2
	lea	MemRamMapNone(pc),a6
	btst	#3,d0
	beq.s	Viwl5Low		;background at lower address?
	exg	a2,a6
Viwl5Low
	move.l	a2,V_RAM_WR989B+2(a4)
	move.l	a6,V_RAM_WR9C9F+2(a4)
	bra	Viwl3WiOff

Viwl5On
	btst	d2,d1
	beq	Viwl5Quit		;window still on?

	or.b	#$40,d1			;force change bit 6

	moveq	#0,d2
	move.b	LCD_WY(a4),d2
	add.w	LCD_I_LY(a4),d2
	sub.w	#SCR_YMAX-1+1,d2
	bpl.s	Viwl5WiCounter
	swap	d2
Viwl5WiCounter
	move.w	d2,LCD_I_WY(a4)		;lines without window

	lea	MemRamTileOb(pc),a1
	lea	MemRamTileBgWiOb(pc),a2
	lea	MemRamTileBgWi(pc),a3
	btst	#4,d0
	beq.s	Viwl5WiSigned		;tile type signed?
	move.l	a2,a1
	lea	MemRamNone(pc),a3
Viwl5WiSigned
	move.l	a1,V_RAM_WR8087+2(a4)
	move.l	a2,V_RAM_WR888F+2(a4)
	move.l	a3,V_RAM_WR9097+2(a4)

	lea	VidLineBgCh(pc),a6
	btst	#1,d0
	beq.s	Viwl5WiObOff		;objects off?
	move.w	#SCR_YMAX,d2
	sub.w	LCD_I_LY(a4),d2
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,VidObYOffset(a5)
	lea	VidLineBgChOb(pc),a6
Viwl5WiObOff
	move.l	a6,VidLineFunction(a5)

	move.w	#$100/4-1,d3
	move.l	VidWiUsed(a5),a6
	moveq	#-1,d2
Viwl5Loop
	move.l	d2,(a6)+
	dbra	d3,Viwl5Loop		;no window tiles used
Viwl5Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 6
	moveq	#6,d2			;window base: 0/1 = $9800/$9c00
	btst	d2,d1
	beq	Viwl6Quit		;window base not changed?

	move.l	BASE_RAM_VID(a4),a3
	move.l	VidMapIndices(a5),a0

	btst	d2,d0
	beq.s	Viwl6Low		;at lower address?

	move.w	#MAP_WRAP*MAP_WRAP,d2

	btst	#3,d0
	beq.s	Viwl6WiHBgL		;background at lower address?

	lea	MemRamMapNone(pc),a2
	lea	MemRamMapBgWi(pc),a6
	bra.s	Viwl6Goon
Viwl6WiHBgL
	lea	MemRamMapBg(pc),a2
	lea	MemRamMapWi(pc),a6
	bra.s	Viwl6Goon

Viwl6Low
	moveq	#0,d2

	btst	#3,d0
	beq.s	Viwl6WiLBgL		;background at lower address?

	lea	MemRamMapWi(pc),a2
	lea	MemRamMapBg(pc),a6
	bra.s	Viwl6Goon
Viwl6WiLBgL
	lea	MemRamMapBgWi(pc),a2
	lea	MemRamMapNone(pc),a6

Viwl6Goon
	move.l	a2,V_RAM_WR989B+2(a4)
	move.l	a6,V_RAM_WR9C9F+2(a4)

	lea	$9800-$10000(a3),a3
	add.w	d2,a3

	add.w	d2,d2
	add.w	d2,a0
	move.l	a0,VidWiMapStart(a5)

	move.w	d0,-(sp)
	move.w	d1,-(sp)

	moveq	#0,d0
	moveq	#$7f,d1
	move.w	VidWiWxPixel(a5),d2
	move.w	#SCR_HEIGHT*MAP_WRAP-1,d3
	move.l	VidWiUsed(a5),a6
Viwl6Loop
	move.b	(a3)+,d0
	cmp.b	(a6,d0.w),d2
	beq.s	Viwl6Next
	move.b	d1,(a6,d0.w)
Viwl6Next
	dbra	d3,Viwl6Loop		;mark which tiles need conversion

	move.w	(sp)+,d1
	move.w	(sp)+,d0
Viwl6Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 3, two entries: window on or off
Viwl3WiOn
	moveq	#3,d2			;background base: 0/1 = $9800/$9c00
	btst	d2,d1
	beq	Viwl3Quit		;background base not changed?

	move.l	BASE_RAM_VID(a4),a3
	move.l	VidMapIndices(a5),a0

	btst	d2,d0
	beq.s	Viwl3WiLow		;at lower address?

	move.w	#MAP_WRAP*MAP_WRAP,d2

	btst	#6,d0
	beq.s	Viwl3WiLBgH		;window at lower address?

	lea	MemRamMapNone(pc),a2
	lea	MemRamMapBgWi(pc),a6
	bra.s	Viwl3Scan
Viwl3WiLBgH
	lea	MemRamMapWi(pc),a2
	lea	MemRamMapBg(pc),a6
	bra.s	Viwl3Scan

Viwl3WiLow
	moveq	#0,d2

	btst	#6,d0
	beq.s	Viwl3WiLBgL		;window at lower address?

	lea	MemRamMapBg(pc),a2
	lea	MemRamMapWi(pc),a6
	bra.s	Viwl3Scan
Viwl3WiLBgL
	lea	MemRamMapBgWi(pc),a2
	lea	MemRamMapNone(pc),a6
	bra.s	Viwl3Scan

Viwl3WiOff
	moveq	#3,d2			;background base: 0/1 = $9800/$9c00
	btst	d2,d1
	beq	Viwl3Quit		;background base not changed?

	move.l	BASE_RAM_VID(a4),a3
	move.l	VidMapIndices(a5),a0

	btst	d2,d0
	beq.s	Viwl3BgLow		;at lower address?

	move.w	#MAP_WRAP*MAP_WRAP,d2

	lea	MemRamMapNone(pc),a2
	lea	MemRamMapBg(pc),a6
	bra.s	Viwl3Scan
Viwl3BgLow
	moveq	#0,d2

	lea	MemRamMapBg(pc),a2
	lea	MemRamMapNone(pc),a6

Viwl3Scan
	move.l	a2,V_RAM_WR989B+2(a4)
	move.l	a6,V_RAM_WR9C9F+2(a4)

	lea	$9800-$10000(a3),a3
	add.w	d2,a3

	add.w	d2,d2
	add.w	d2,a0
	move.l	a0,VidBgMapStart(a5)

	moveq	#SCR_YMAX-$100,d2
	sub.w	LCD_I_LY(a4),d2
	add.b	LCD_SCY(a4),d2
	add.w	d2,d2
	add.w	d2,d2
	move.b	d2,-(sp)
	and.w	#(MAP_WRAP-1)*MAP_WRAP,d2
	add.w	VidBgScxTile(a5),d2
	add.w	d2,d2
	move.w	d2,VidBgMapOffset(a5)
	move.w	(sp)+,d2
	and.w	#7*$100*4,d2
	move.l	VidBgTile(a5),a2
	add.w	d2,a2
	move.l	a2,VidBgTilePtr(a5)	;correct actual pointer

	move.w	d0,-(sp)
	move.w	d1,-(sp)

	moveq	#0,d0
	moveq	#$7f,d1
	move.w	VidBgScxPixel(a5),d2
	move.w	#MAP_WRAP*MAP_WRAP-1,d3
	move.l	VidBgUsed(a5),a6
Viwl3Loop
	move.b	(a3)+,d0
	cmp.b	(a6,d0.w),d2
	beq.s	Viwl3Next
	move.b	d1,(a6,d0.w)
Viwl3Next
	dbra	d3,Viwl3Loop		;mark which tiles need conversion

	move.w	(sp)+,d1
	move.w	(sp)+,d0
Viwl3Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 4
	moveq	#4,d2			;tile type: 0/1 = signed/unsigned
	btst	d2,d1
	beq	Viwl4Quit		;tile type not changed?

	btst	#5,d0
	bne.s	Viwl4WiOn		;window on?

	lea	MemRamTileOb(pc),a1
	lea	MemRamTileBgOb(pc),a2
	lea	MemRamTileBg(pc),a3
	bra.s	Viwl4Goon

Viwl4WiOn
	lea	MemRamTileOb(pc),a1
	lea	MemRamTileBgWiOb(pc),a2
	lea	MemRamTileBgWi(pc),a3

	or.b	#$40,d1			;window tiles need conversion

	moveq	#$7f,d2
	moveq	#$80-1,d3
	move.l	VidWiUsed(a5),a6
Viwl4WiLoop
	tst.b	(a6)+
	bmi.s	Viwl4WiNext
	move.b	d2,-1(a6)
Viwl4WiNext
	dbra	d3,Viwl4WiLoop		;mark which tiles need conversion

Viwl4Goon
	move.l	BASE_RAM_VID(a4),a6
	lea	$8000-$10000(a6),a6
	move.l	a6,VidBgWiTileHigh(a5)

	lea	$1000(a6),a6
	btst	#4,d0
	beq.s	Viwl4Signed		;tile type signed?
	move.l	a2,a1
	lea	MemRamNone(pc),a3
	lea	-$1000(a6),a6
Viwl4Signed
	move.l	a1,V_RAM_WR8087+2(a4)
	move.l	a2,V_RAM_WR888F+2(a4)
	move.l	a3,V_RAM_WR9097+2(a4)
	move.l	a6,VidBgWiTileLow(a5)

	moveq	#$7f,d2
	moveq	#$80-1,d3
	move.l	VidBgUsed(a5),a6
Viwl4BgLoop
	tst.b	(a6)+
	bmi.s	Viwl4BgNext
	move.b	d2,-1(a6)
Viwl4BgNext
	dbra	d3,Viwl4BgLoop		;mark which tiles need conversion

	or.b	#$08,d1			;background tiles need conversion
Viwl4Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert background tiles if necessary
	moveq	#3,d2			;background tiles need conversion
	btst	d2,d1
	beq.s	ViwlBgQuit		;no conversions?

	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.w	VidBgScxPixel(a5),d2
	move.l	VidBgUsed(a5),a6
	move.l	VidBgTile(a5),a2
	bsr	VidConvertBgWiAllMarked
ViwlBgQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	convert window tiles if necessary
	moveq	#6,d2			;window tiles need conversion
	btst	d2,d1
	beq.s	ViwlWiQuit		;no conversions?

	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.w	VidWiWxPixel(a5),d2
	move.l	VidWiUsed(a5),a6
	move.l	VidWiTile(a5),a2
	bsr	VidConvertBgWiAllMarked
ViwlWiQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 1, two entries: background/window on or off
Viwl1BgWiOn
	moveq	#1,d2			;objects on: 0/1 = off/on
	btst	d2,d1
	beq.s	Viwl1Quit		;state not changed?

	btst	#5,d0
	beq.s	Viwl1WiOff		;window off?
	lea	VidLineBgCh(pc),a2
	lea	VidLineBgChOb(pc),a3
	bra.s	Viwl1Goon
Viwl1WiOff
	lea	VidLineBg(pc),a2
	lea	VidLineBgOb(pc),a3
	bra.s	Viwl1Goon

Viwl1BgWiOff
	moveq	#1,d2			;objects on: 0/1 = off/on
	btst	d2,d1
	beq.s	Viwl1Quit		;state not changed?

	lea	VidLineNone(pc),a2
	lea	VidLineOb(pc),a3
Viwl1Goon
	btst	#1,d0
	beq.s	Viwl1ObOff		;objects off?
	move.w	#SCR_YMAX,d2
	sub.w	LCD_I_LY(a4),d2
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,VidObYOffset(a5)
	move.l	a3,a2
Viwl1ObOff
	move.l	a2,VidLineFunction(a5)
Viwl1Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	handle bit 2
	moveq	#2,d2			;object height: 0/1 = 8/16
	btst	d2,d1
	beq.s	Viwl2Quit		;height not changed?

	btst	d2,d0
	beq.s	Viwl2Small
	move.w	#16-1,VidObCount(a5)
	move.w	#$fe,VidObTilemask(a5)
	bra.s	Viwl2Goon
Viwl2Small
	move.w	#8-1,VidObCount(a5)
	move.w	#$ff,VidObTilemask(a5)	;set up some variables

Viwl2Goon
	movem.l	d0/d1/d4/d5,-(sp)
	bsr	VidObLinkConvertAll
	movem.l	(sp)+,d0/d1/d4/d5
Viwl2Quit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ViwlQuit
	movem.l	(sp)+,viwlRegs

	move.l	VidLineFunction(a5),a6
	move.w	LCD_I_LY(a4),d2
	bmi.s	ViwlNop			;not in visible part?
	move.l	a6,V_LCD_LINE+2(a4)

ViwlNop
	rts

;--------------------------------------------------------------------------
;Ptr VidIoRdStat(Ptr p)
;----------------------
;	Read from LCD status and interrupt mode.
VidIoRdStat
	and.b	#$f8,(a3)

	moveq	#SCR_YMAX-1-$100,d0
	sub.w	LCD_I_LY(a4),d0
	cmp.b	#SCR_YMAX+1,d0
	bcs.s	VirsVisible
	or.b	#$01,(a3)
	bra.s	VirsLyc

VirsVisible
	cmp.w	#OPS_PER_LINE-10,d6
	bcs.s	VirsLyc
	or.b	#$02,(a3)
	cmp.w	#OPS_PER_LINE-5,d6
	bcc.s	VirsLyc
	or.b	#$01,(a3)

VirsLyc
	cmp.b	LCD_LYC(a4),d0
	bne.s	VirsQuit
	or.b	#$04,(a3)

VirsQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void VidIoWrStat(data.b)
;------------------------
;	Write to LCD status and interrupt mode.
VidIoWrStat
	move.b	d0,(a3)			;store value

	move.b	#$00,VidIntMaskSV(a5)
	btst	#4,d0
	beq.s	ViwsVblOk
	move.b	#$02,VidIntMaskSV(a5)
ViwsVblOk				;set up for "vblank"

	btst	#6,d0
	beq.s	ViwsQuit
	moveq	#SCR_YMAX-1-$100,d1
	sub.w	LCD_I_LY(a4),d1
	cmp.b	LCD_LYC(a4),d1
	bne.s	ViwsQuit
	move.b	VidIntMaskStat(a5),d0
	or.b	d0,INT_IF(a4)
	pea	(a6)
	jmp	INT_CHECK(a4)		;check for "LYC"

ViwsQuit
	and.b	#$28,d0
	bne.s	ViwsNotImpl
	jmp	(a6)

ViwsNotImpl
	moveq	#FTR_NOT_IMPL,d0
	rts

;--------------------------------------------------------------------------
;void VidIoWrScy(data.b)
;-----------------------
;	Write to scroll y.
VidIoWrScy
	move.b	d0,(a3)			;store value

	sub.w	LCD_I_LY(a4),d0
	add.b	#SCR_YMAX,d0
	add.w	d0,d0
	add.w	d0,d0
	move.b	d0,-(sp)
	and.w	#(MAP_WRAP-1)*MAP_WRAP,d0
	add.w	VidBgScxTile(a5),d0
	add.w	d0,d0
	move.w	d0,VidBgMapOffset(a5)
	move.w	(sp)+,d0
	and.w	#7*$100*4,d0
	move.l	VidBgTile(a5),a3
	add.w	d0,a3
	move.l	a3,VidBgTilePtr(a5)	;correct actual pointer
	jmp	(a6)

;--------------------------------------------------------------------------
;void VidIoWrScx(data.b)
;-----------------------
;	Write to scroll x.
VidIoWrScx
	move.l	d3,-(sp)

	move.b	d0,(a3)			;store value

	move.w	VidBgScxPixel(a5),d1
	moveq	#7,d2
	and.b	d0,d2
	add.w	d2,d2
	move.w	d2,VidBgScxPixel(a5)
	lsr.b	#3,d0
	ext.w	d0
	move.w	d0,d3
	sub.w	VidBgScxTile(a5),d3
	move.w	d0,VidBgScxTile(a5)
	add.w	d3,d3
	add.w	d3,VidBgMapOffset(a5)	;update variables

	moveq	#SCR_WIDTH-1,d3
	sub.w	#MAP_WRAP-SCR_WIDTH,d0
	bmi.s	ViwsxCounterBg
	sub.w	d0,d3
	exg	d0,d3
ViwsxCounterBg
	move.w	d0,VidBgCount1(a5)
	move.w	d3,VidBgCount2(a5)	;set up counters (background)

	move.w	d0,VidBgWiCount1(a5)
	move.w	VidBgWiCount2(a5),d0
	bmi.s	ViwsxCounterExit
	move.w	VidBgWiCount3(a5),d0
	bmi.s	ViwsxCounterBgWi
	sub.w	d0,d3
	bpl.s	ViwsxCounterBgWi
	move.w	d3,VidBgWiCount1(a5)
	moveq	#SCR_WIDTH-1,d3
	sub.w	d0,d3
ViwsxCounterBgWi
	move.w	d3,VidBgWiCount2(a5)	;set up counters (backgr./window)
ViwsxCounterExit

	moveq	#$81-$100,d0
	move.b	LCD_LCDC(a4),d3
	and.b	d0,d3
	cmp.b	d0,d3
	bne	ViwswxQuit		;background off?

	move.l	VidBgUsed(a5),a3
	move.l	VidBgTile(a5),a2
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ViwswxShift
	sub.w	d2,d1
	beq	ViwswxQuit		;no new pixel shift?
	bcc.s	ViwswxRight		;shift right?
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	neg.w	d1
	move.w	#$100-1,d3
ViwswxLLoop
	move.b	(a3)+,d0
	bmi.s	ViwswxLNext
	cmp.b	d0,d2
	bne.s	ViwswxLCorr		;correction necessary?
ViwswxLNext
	addq.w	#4,a2
	dbra	d3,ViwswxLLoop
	bra	ViwswxQuit
ViwswxLCorr
	move.b	d2,-1(a3)

	swap	d3
	move.w	#8-1,d3
ViwswxLIloop
	move.l	(a2),d0
	lsl.l	d1,d0
	move.l	d0,(a2)
	move.l	8*$100*4(a2),d0
	lsl.l	d1,d0
	move.l	d0,8*$100*4(a2)

	lea	$100*4(a2),a2

	dbra	d3,ViwswxLIloop
	swap	d3
	lea	4-8*$100*4(a2),a2

	dbra	d3,ViwswxLLoop
	bra.s	ViwswxQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ViwswxRight
	move.w	#$100-1,d3
ViwswxRLoop
	move.b	(a3)+,d0
	bmi.s	ViwswxRNext
	cmp.b	d0,d2
	bne.s	ViwswxRCorr		;correction necessary?
ViwswxRNext
	addq.w	#4,a2
	dbra	d3,ViwswxRLoop
	bra.s	ViwswxQuit
ViwswxRCorr
	move.b	d2,-1(a3)

	swap	d3
	move.w	#8-1,d3
ViwswxRIloop
	move.l	(a2),d0
	lsr.l	d1,d0
	move.l	d0,(a2)
	move.l	8*$100*4(a2),d0
	lsr.l	d1,d0
	move.l	d0,8*$100*4(a2)

	lea	$100*4(a2),a2

	dbra	d3,ViwswxRIloop
	swap	d3
	lea	4-8*$100*4(a2),a2

	dbra	d3,ViwswxRLoop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ViwswxQuit
	move.l	(sp)+,d3
	jmp	(a6)

;--------------------------------------------------------------------------
;void VidIoWrLyc(data.b)
;-----------------------
;	Write to LCD y compare.
VidIoWrLyc
	move.b	d0,(a3)			;store value

	move.b	d0,VidIntLycReload(a5)
	moveq	#SCR_YMAX-$100,d1
	sub.w	LCD_I_LY(a4),d1
	sub.b	d1,d0
	move.b	d0,VidIntLycCount(a5)	;set up counter

	addq.b	#1,d0
	bne.s	ViwlcQuit
	btst	#6,LCD_STAT(a4)
	beq.s	ViwlcQuit
	move.b	VidIntMaskStat(a5),d0
	or.b	d0,INT_IF(a4)
	pea	(a6)
	jmp	INT_CHECK(a4)		;check for "LYC"

ViwlcQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void VidIoWrDma(data.b)
;-----------------------
;	Write to dma register.
VidIoWrDma
viwdRegs	reg	d3/a0-a1

	movem.l	viwdRegs,-(sp)

	move.l	BASE_RAM_OAM(a4),a1
	lea	$fe00+40*4-$10000(a1),a1
	lea	40*4(a3),a0
	moveq	#40*4-4-$100,d3
ViwdLoop
	move.l	-(a1),d2
	move.l	-(a0),d0
	move.l	d0,(a1)
	move.b	d3,d1
	move.l	a1,a3
	bsr	MemRamOamEntry
	subq.b	#4,d3
	bcc.s	ViwdLoop

	movem.l	(sp)+,viwdRegs
	rts

;--------------------------------------------------------------------------
;void VidIoWrBgp(data.b)
;-----------------------
;	Write to background/window palette.
VidIoWrBgp
viwbRegs	reg	d3/a0-a1

	pea	(a6)			;save return address

	cmp.b	(a3),d0
	beq	ViwbQuit		;no changes?
	move.b	d0,(a3)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	moveq	#0,d1
	moveq	#0,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2			;calculate color indices

	lea	VidColorTab(pc),a3
	add.w	d1,d1
	add.w	(a3,d1.w),d1
	lea	(a3,d1.w),a2
	move.l	a2,VidBgWiColorH(a5)
	add.w	d2,d2
	add.w	(a3,d2.w),d2
	add.w	d2,a3
	move.l	a3,VidBgWiColorL(a5)	;set up color functions
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	moveq	#$81-$100,d0
	move.b	LCD_LCDC(a4),d1
	and.b	d0,d1
	cmp.b	d0,d1
	bne	ViwbQuit		;background/window off?

	movem.l	viwbRegs,-(sp)

	move.l	a3,a0
	move.l	a2,a1

	move.l	VidBgUsed(a5),a6
	move.l	VidBgTile(a5),a2
	move.w	VidBgScxPixel(a5),d2
	bsr	VidConvertBgWiAllUsed

	btst	#5,LCD_LCDC(a4)
	beq.s	ViwbExit		;window off?

	move.l	VidWiUsed(a5),a6
	move.l	VidWiTile(a5),a2
	move.w	VidWiWxPixel(a5),d2
	bsr	VidConvertBgWiAllUsed

ViwbExit
	movem.l	(sp)+,viwbRegs
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ViwbQuit
	rts

;--------------------------------------------------------------------------
;void VidIoWrObp0(data.b)
;------------------------
;	Write to object palette 0.
VidIoWrObp0
viwoRegs	reg	d3-d5/a0-a1/a4

	pea	(a6)			;save return address

	cmp.b	(a3),d0
	beq	Viwo0Quit		;no changes?
	move.b	d0,(a3)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	moveq	#0,d1
	moveq	#0,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2			;calculate color indices

	lea	VidColorTab(pc),a3
	add.w	d1,d1
	add.w	(a3,d1.w),d1
	lea	(a3,d1.w),a2
	move.l	a2,VidObColor0H(a5)
	add.w	d2,d2
	add.w	(a3,d2.w),d2
	add.w	d2,a3
	move.l	a3,VidObColor0L(a5)	;set up color functions
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	btst	#7,LCD_LCDC(a4)
	beq.s	Viwo0Quit		;lcd off?

	movem.l	viwoRegs,-(sp)

	move.l	a3,a0
	move.l	a2,a1
	move.l	VidObTiles(a5),a2
	move.l	BASE_RAM_VID(a4),a3
	lea	$8000-$10000(a3),a3
	move.l	BASE_RAM_OAM(a4),a6
	lea	$fe00-$10000(a6),a6
	move.l	GrayCopyTable(a5),a4

	moveq	#40-1,d3
Viwo0Loop
	move.w	(a6)+,d4
	move.w	VidObTilemask(a5),d1
	and.b	(a6)+,d1
	move.b	(a6)+,d0
	moveq	#$10,d2
	and.b	d0,d2
	bne.s	Viwo0Next		;not this palette?

	swap	d3
	movem.l	a2/a3/a6,-(sp)
	lsl.w	#3+1,d1
	add.w	d1,a3
	move.w	VidObCount(a5),d3
	bsr	VidConvertOb
	movem.l	(sp)+,a2/a3/a6
	swap	d3
Viwo0Next
	lea	16*7*2(a2),a2
	dbra	d3,Viwo0Loop		;convert object with new palette

Viwo0Exit
	movem.l	(sp)+,viwoRegs
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Viwo0Quit
	rts

;--------------------------------------------------------------------------
;void VidIoWrObp1(data.b)
;------------------------
;	Write to object palette 1.
VidIoWrObp1
	pea	(a6)			;save return address

	cmp.b	(a3),d0
	beq	Viwo1Quit		;no changes?
	move.b	d0,(a3)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	moveq	#0,d1
	moveq	#0,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2
	add.b	d0,d0
	addx.b	d1,d1
	add.b	d0,d0
	addx.b	d2,d2			;calculate color indices

	lea	VidColorTab(pc),a3
	add.w	d1,d1
	add.w	(a3,d1.w),d1
	lea	(a3,d1.w),a2
	move.l	a2,VidObColor1H(a5)
	add.w	d2,d2
	add.w	(a3,d2.w),d2
	add.w	d2,a3
	move.l	a3,VidObColor1L(a5)	;set up color functions
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	btst	#7,LCD_LCDC(a4)
	beq.s	Viwo1Quit		;lcd off?

	movem.l	viwoRegs,-(sp)

	move.l	a3,a0
	move.l	a2,a1
	move.l	VidObTiles(a5),a2
	move.l	BASE_RAM_VID(a4),a3
	lea	$8000-$10000(a3),a3
	move.l	BASE_RAM_OAM(a4),a6
	lea	$fe00-$10000(a6),a6
	move.l	GrayCopyTable(a5),a4

	moveq	#40-1,d3
Viwo1Loop
	move.w	(a6)+,d4
	move.w	VidObTilemask(a5),d1
	and.b	(a6)+,d1
	move.b	(a6)+,d0
	moveq	#$10,d2
	and.b	d0,d2
	beq.s	Viwo1Next		;not this palette?

	swap	d3
	movem.l	a2/a3/a6,-(sp)
	lsl.w	#3+1,d1
	add.w	d1,a3
	move.w	VidObCount(a5),d3
	bsr	VidConvertOb
	movem.l	(sp)+,a2/a3/a6
	swap	d3
Viwo1Next
	lea	16*7*2(a2),a2
	dbra	d3,Viwo1Loop		;convert object with new palette

Viwo1Exit
	movem.l	(sp)+,viwoRegs
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Viwo1Quit
	rts

;--------------------------------------------------------------------------
;void VidIoWrWy(data.b)
;----------------------
;	Write to window y coordinate.
VidIoWrWy
	move.b	d0,(a3)			;store value

	moveq	#$a1-$100,d1
	move.b	LCD_LCDC(a4),d2
	and.b	d1,d2
	cmp.b	d1,d2
	bne.s	ViwwyQuit		;window off?

	move.w	LCD_I_LY(a4),d2
	bmi.s	ViwwyQuit		;not in visible part?

	lea	VidLineBgCh(pc),a3
	btst	#1,LCD_LCDC(a4)
	beq.s	ViwwyObOff		;objects off?
	lea	VidLineBgChOb(pc),a3
ViwwyObOff
	move.l	a3,V_LCD_LINE+2(a4)

	and.w	#$ff,d0
	add.w	d2,d0
	sub.w	#SCR_YMAX-1+1,d0
	move.w	d0,LCD_I_WY(a4)
	bmi.s	ViwwyQuit		;window shown yet?

	move.w	#0,VidWiMapOffset(a5)
	move.l	VidWiTile(a5),a3
	move.l	a3,VidWiTilePtr(a5)	;correct actual pointer

ViwwyQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void VidIoWrWx(data.b)
;----------------------
;	Write to window x coordinate.
VidIoWrWx
	move.l	d3,-(sp)

	move.b	d0,(a3)			;store value

	move.w	VidWiWxPixel(a5),d1
	move.b	d0,d2
	not.b	d2
	and.w	#7,d2
	add.w	d2,d2
	move.w	d2,VidWiWxPixel(a5)
	moveq	#-1,d3
	lsl.w	d2,d3
	move.w	d3,VidWiWxMask(a5)	;update variables

	moveq	#SCR_WIDTH-1,d3
	subq.b	#8,d0
	bcs.s	ViwwxFull
	move.w	VidBgCount2(a5),d2
	cmp.b	#SCR_WIDTH*8-1,d0
	bcc.s	ViwwxEmpty
	lsr.b	#3,d0
	ext.w	d0
	sub.w	d0,d3
	move.w	VidBgCount1(a5),d0
	sub.w	d3,d2
	bpl.s	ViwwxCounters
	moveq	#-1,d0
	moveq	#SCR_WIDTH-1,d2
	sub.w	d3,d2
	bra.s	ViwwxCounters
ViwwxFull
	moveq	#-1,d2
	bra.s	ViwwxCounters
ViwwxEmpty
	move.w	VidBgCount1(a5),d0
	moveq	#-1,d3
ViwwxCounters
	move.w	d0,VidBgWiCount1(a5)
	move.w	d2,VidBgWiCount2(a5)
	move.w	d3,VidBgWiCount3(a5)	;set up counters

	moveq	#$a1-$100,d0
	move.b	LCD_LCDC(a4),d3
	and.b	d0,d3
	cmp.b	d0,d3
	bne	ViwswxQuit		;window off?

	move.w	VidWiWxPixel(a5),d2
	move.l	VidWiUsed(a5),a3
	move.l	VidWiTile(a5),a2
	bra	ViwswxShift		;check for tile shifting

;=== The end... ===========================================================
