/**	
	Code to access the database containing the Transaction information

	$Author: chaese $
	$Revision: 1.10 $
	Last updated: $Date: 2000/11/16 02:38:32 $
	
	$Log: TransactionsDB.h,v $
	Revision 1.10  2000/11/16 02:38:32  chaese
	Merged New Transaction and Transaction Details forms
	
	Revision 1.9  2000/11/12 19:10:11  dizzy
	Structures and functions modified to accomodate memorising the cleared atribute as a category
	rather then a member in the Transaction structure.
	
	Revision 1.8  2000/11/12 14:34:12  dizzy
	Initial transaction category support
	
	Revision 1.7  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.6  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.5  2000/10/21 14:35:29  dizzy
	added delete functionality (for accounts and transactions)
	
	Revision 1.4  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.3  2000/09/15 10:10:49  dizzy
	First useful version
	
	Revision 1.2  2000/09/13 11:50:27  dizzy
	The transaction module is fully functional

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#ifndef __TRANSACTIONS_DB__
#define __TRANSACTIONS_DB__

#include <PalmOS.h>
#include <DateTime.h>
#include <FeatureMgr.h>

#include "PalmCoinsRsc.h"
#include "MiscFunc.h"
#include "AccountsDB.h"

/// type of the Transaction database
#define TRANSACTIONSDB_TYPE		'DATA'
/// db name
#define TRANSACTIONSDB_NAME	"PalmCoins-Transactions-DyCo"

//define this variables as extern if the file is not included from TransactionsDB.c
#ifndef __TRANSACTIONS_DB_C__
	/// reference to the database
	extern DmOpenRef		TransactionsDB_ref;
	/// total number of categories
	extern UInt32			TransactionsDB_num_cat;
	/// the currenct category
	extern UInt16			TransactionsDB_current_cat;
	/// the current record (to be displayed/edited/etc)
	extern Int16			TransactionsDB_current_record;
	/// indictates that a new record is being edited
	extern Boolean			TransactionsDB_record_is_new;
	///the strings corresponding to the payment method enum
	extern Char*			TransactionsDB_types[6];
	///the number of types stored in TransactionDB_types
	extern UInt16			TransactionsDB_types_no;
#endif

/// payment method
typedef enum { pm_card, pm_cash, pm_cheque, pm_so, pm_dd, pm_transfer } payment_method;
	
	
///the packed structure of a transaction - see the unpacked struct for explanations
typedef struct {
	UInt32			source_acc;
	UInt32			dest_acc;
	DateType		date;
	Int32			ammount;
	payment_method	method;
	Char			payee[1];
} packedTransactionStructure;

///unpacked structure of an Transaction
typedef struct {
	/// the Transaction where the transaction originates
	UInt32			source_acc;
	/// destination Transaction for the transaction
	UInt32			dest_acc;
	///date on which the transaction was made
	DateType		date;
	/// transaction total
	Int32			ammount;
	///tthe method of payment
	payment_method	method;
	/// person/company 
	Char*			payee;
	/// note on the transaction	
	Char*			note;
} transactionStructure;



// ***** FUNCTIONS

/** 
	sets the category of a given record
	@param index	the index of the record
	@param cat		the record category
*/
void	TransactionsDB_SetRecordCategory(UInt16 index, UInt16 cat);


/// unpacks an Transaction record
void	TransactionsDB_UnpackTransaction (transactionStructure* trans, const packedTransactionStructure* packed_trans);

/// packs an Transaction record
void	TransactionsDB_PackTransaction (transactionStructure* trans, MemHandle db_entry);

/// fetch an unpacked record by passing its unique id
void	TransactionsDB_GetUnpackedRecord(UInt32 id,transactionStructure* ts);


/** 
	get the date of the transaction
	@param id unique id of the record
	@param pDate pointer to allocated date structure that will receive the date
*/
void	TransactionsDB_RecordGetDate(UInt32 id, DateType * pDate);

/** 
	@param id unique id of the record
	@return the ammount involved in a transaction		
*/
Int32	TransactionsDB_RecordGetAmmount(UInt32 id);

/** 
	fetch the destination account informations
	@param id unique id of the record
	@param as pointer to an accountStructure where the information will be saved
 */			
void	TransactionsDB_RecordGetDestAcc(UInt32 id,accountStructure* as);

/** 
	fetch the destination account informations
	@param id unique id of the record
	@param as pointer to an accountStructure where the information will be saved
 */			
void	TransactionsDB_RecordGetSrcAcc(UInt32 id,accountStructure* as);

/**
	Returns the id of a given transaction type string
	@param name the name for the transaction_type
	@return the associated id
*/
payment_method		TransactionsDB_GetTypeIDByName(Char* name);

/// initialize the databases
 Err	TransactionsDB_Init();

/// closes the database
void 	TransactionsDB_Close();

/// gets a locked pointer to the app info block
MemPtr	TransactionsDB_GetLockedAppInfo();

/// main purpose of this function is to initialise the categories from the database
MemPtr TransactionsDB_AppInfoInit(LocalID dbID,UInt16 cardNo);


///init the Transactions DB with some default records
void	TransactionsDB_InitializeTransactions();

/**
	creates a new record and update the balances of the two involved accounts
	@param s pointer to the structure containing the transaction data
	@param cleared boolean which indicates the status of the record. The function will place the record in the right category
	@return id of the newly created record
	
	note: this function is likely to change it's prototype in order to support scheduled transactions
*/
UInt16	TransactionsDB_NewRecord(transactionStructure* s, Boolean cleared);
/**
	deletes a record, given its id
	@param id the id of the record to be deleted
*/
void TransactionsDB_DeleteRecord(UInt16 id);

/**
	Compares two packed transaction structures.
	For parameters see the PalmOS reference
*/
Int16	TransactionsDB_CompareRecords(packedTransactionStructure* rec1,packedTransactionStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info);

#endif
