/*-
 * Copyright (c) 2000 Jason K. Fritcher <jkf@wolfnet.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: palmkey.c,v 1.8 2000/02/28 02:19:16 jkf Exp $
 */

#include <Pilot.h>

#include <sys/types.h>
#include "types.h"
#include "md4/md4.h"
#include "md5/md5.h"
#include "sha1/sha1.h"

#include "palmkey_rcp.h"

extern char *btoe(char *, char *);
static void DoAboutView(void);

struct PKeyPrefs {
    Word	hash;
    Word	format;
    Word	count;
    Char	seed[17];
};

static Word	Hash = 0;
static Word	format = 0;
static char *(*disp)(char *, char *);
static UInt32	count;
static CharPtr  buf;
static struct PKeyPrefs prefs;
static char	*file = __FILE__;
static char	dict = 0;

void FatalErrorIf(int cond, int MsgResource, char *file, int line)
{
    if (cond) {
	DmOpenRef dbp;
	VoidHand msg;
	UInt16   idx;

	/* Open ourselves, find the string resource, and get a handle to it. */
	dbp = DmOpenDatabaseByTypeCreator('appl', 'PKey', dmModeReadOnly);
	idx = DmFindResource(dbp, 'tSTR', MsgResource, NULL);
	msg = DmGetResourceIndex(dbp, idx);

	/* Display error message. */
	ErrDisplayFileLineMsg(file, line, MemHandleLock(msg));
    }
}

void NonFatalError(int AlertID, int MsgResource)
{
    DmOpenRef dbp;
    VoidHand msg;
    UInt16   idx;

    /* Open ourselves, find the string resource, and get a handle to it. */
    dbp = DmOpenDatabaseByTypeCreator('appl', 'PKey', dmModeReadOnly);
    idx = DmFindResource(dbp, 'tSTR', MsgResource, NULL);
    msg = DmGetResourceIndex(dbp, idx);

    /* Show error message. */
    FrmCustomAlert(AlertID, MemHandleLock(msg), " ", " ");

    /* Clean up after ourself. */
    MemHandleUnlock(msg);
    DmCloseDatabase(dbp);
}

static void *GetObjectPtr(FormPtr frm, UInt16 ObjID)
{
	return(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ObjID)));
}

static void lowercase(char *s)
{
    char *p;

    /* Go through string one character at a time and
       convert any capital letters into lowercase. */
    for(p = s; *p; p++)
	if (*p >= 'A' && *p <= 'Z')
	    *p += ('a' - 'A');
}

static void sevenbit(char *s)
{
    char *p;

    /* Go through string one character at a time and strip the MSB off. */
    for(p = s; *p; p++)
	*p &= 0x7F;
}

char *put8(char *out, char *s)
{
    UInt16	i;
    Char	p[3];

    /* Truncate string buffer. */
    out[0] = 0;

    /* Cycle through each byte of binary response. */
    for(i = 0; i < 8; i++) {
	/* Insert space, if needed. */
	if ((i > 0) && !(i % 2)) {
	    int len = StrLen(out);
	    out[len] = ' ';
	    out[len+1] = 0;
	}

	/* Convert high nibble of byte to hexadecimal. */
	p[0] = ((s[i] & 0xF0) >> 4) + '0';
	if (p[0] > '9')
	    p[0] += 7;

	/* Convert low nibble of byte to hexadecimal. */
	p[1] =  (s[i] & 0x0F) + '0';
	if (p[1] > '9')
	    p[1] += 7;

	/* Null terminate converted byte and copy it into the response string. */
	p[2] = 0;
	StrCat(out, p);
    }

    return(out);
}

static void do_md4(void)
{
    UInt32	results[4];
    MD4_CTX	*md;

    /* Allocate md4 context structure. */
    md = MemPtrNew(sizeof(MD4_CTX));
    FatalErrorIf(md == NULL, NullPtrMsg, file, __LINE__);

    /* Initialize context, add our string, and compute the hash. */
    MD4Init(md);
    MD4Update(md, buf, StrLen(buf));
    MD4Final((unsigned char *)results, md);

    /* Fold the 128 bit result into 64 bits. */
    results[0] ^= results[2];
    results[1] ^= results[3];
 
    /* Take previous hash and do everything again, count times. */
    while (count-- != 0) {
	MD4Init(md);
	MD4Update(md, (unsigned char *)results, 8);
	MD4Final((unsigned char *)results, md);

	results[0] ^= results[2];
	results[1] ^= results[3];
    }

    /* Free the md4 context. */
    MemPtrFree(md);

    /* Call our display conversion routine. */
    (*disp)(buf, (char *)results);
}

static void do_md5(void)
{
    UInt32	results[4];
    MD5_CTX	*md;

    /* Allocate md5 context structure. */
    md = MemPtrNew(sizeof(MD5_CTX));
    FatalErrorIf(md == NULL, NullPtrMsg, file, __LINE__);

    /* Initialize context, add our string, and compute the hash. */
    MD5Init(md);
    MD5Update(md, buf, StrLen(buf));
    MD5Final((unsigned char *)results, md);

    /* Fold the 128 bit result into 64 bits. */
    results[0] ^= results[2];
    results[1] ^= results[3];

    /* Take previous hash and do everything again, count times. */
    while (count-- != 0) {
	MD5Init(md);
	MD5Update(md, (unsigned char *)results, 8);
	MD5Final((unsigned char *)results, md);

	results[0] ^= results[2];
	results[1] ^= results[3];
    }

    /* Free the md5 context. */
    MemPtrFree(md);

    /* Call our display converstion routine. */
    (*disp)(buf, (char *)results);
}

static void do_sha1(void)
{
    UInt32	results[5];
    SHA1_CTX	*sha;

    /* Allocate sha1 context structure. */
    sha = MemPtrNew(sizeof(SHA1_CTX));
    FatalErrorIf(sha == NULL, NullPtrMsg, file, __LINE__);

    /* Initialize context, add our string, and compute the hash. */
    SHA1Init(sha);
    SHA1Update(sha, buf, StrLen(buf));
    SHA1Final((unsigned char *)results, sha);

    /* Fold the 160 bit result into 64 bits. */
    results[0] ^= results[2];
    results[1] ^= results[3];
    results[0] ^= results[4]; 

    /* Take previous hash and do everything again, count times. */
    while(count-- != 0) {
	SHA1Init(sha);
	SHA1Update(sha, (unsigned char *)results, 8);
	SHA1Final((unsigned char *)results, sha);

	results[0] ^= results[2];
	results[1] ^= results[3];
	results[0] ^= results[4];
    }

    /* Free the sha1 context. */
    MemPtrFree(sha);

    /* Call our display conversion routine. */
    (*disp)(buf, (char *)results);
}

static void MainViewShowResponse(FormPtr frm)
{
    FieldPtr fld;

    /* Get pointer to response field. */
    fld = GetObjectPtr(frm, ResponseField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Stuff the response into the field and have the field redraw itself. */
    FldSetTextPtr(fld, buf);
    FldDrawField(fld);
}

static void MainViewComputeResponse(void)
{
    FormPtr frm;
    FieldPtr fld;

    /* Get pointer to the currently active form. */
    frm = FrmGetActiveForm();
    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);

    /* Truncate string buffer. */
    buf[0] = 0;

    /* Get pointer to count field. */
    fld = GetObjectPtr(frm, CountField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Check field for content, and process it if anything is there. */
    if (FldGetTextLength(fld) > 0) {
	count = StrAToI(FldGetTextPtr(fld));
	if (count == 0) {
	    NonFatalError(InputAlert, ZeroCount);
	    return;
	}
    } else {
	NonFatalError(InputAlert, BlankCount);
	return;
    }

    /* Get pointer to seed field. */
    fld = GetObjectPtr(frm, SeedField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Does the field contain anything? */
    if (FldGetTextLength(fld) > 0) {
	CharPtr chr, p;

	p = chr = FldGetTextPtr(fld);
	FatalErrorIf(chr == NULL, NullPtrMsg, file, __LINE__);

	/* Cycle through the field and check for invalid characters,
	   displaying an error message if there are any. */
	while(*p) {
	    if (!((*p >= 'a' && *p <= 'z') || (*p >= 'A' && *p <= 'Z') || (*p >= '0' && *p <= '9'))) {
		NonFatalError(InputAlert, InvalidSeed);
		return;
	    }
	    p++;
	}
	/* Copy to our buffer and convert to lowercase. */
	StrCopy(buf, chr);
	lowercase(buf);
    } else {
	NonFatalError(InputAlert, BlankSeed);
	return;
    }

    /* Get pointer to passphrase field. */
    fld = GetObjectPtr(frm, PassphraseField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Retrieve password field and check it for content. Display
       error message if things are not satisfactory. */
    if (FldGetTextLength(fld) > 5) {
	StrCat(buf, FldGetTextPtr(fld));
    } else {
	NonFatalError(InputAlert, ShortPassphrase);
	return;
    }

    /* Strip the MSB from all characters in the buffer. */
    sevenbit(buf);

    /* Determine the currently selected hash and do the work. */
    switch(Hash) {
	case 0:
	    do_md4();
	    MainViewShowResponse(frm);
	    break;
	case 1:
	    do_md5();
	    MainViewShowResponse(frm);
	    break;
	case 2:
	    do_sha1();
	    MainViewShowResponse(frm);
	    break;
	default:
	    break;
    }
}

static void MainViewSaveForm(void)
{
    FormPtr frm;
    FieldPtr fld;
    CharPtr txt;

    /* Get pointer to the main form. */
    frm = FrmGetFormPtr(MainView);
    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);

    /* Save the current hash and format selections. */
    prefs.hash = Hash;
    prefs.format = format;

    /* Get pointer to count field. */
    fld = GetObjectPtr(frm, CountField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Get pointer to field's text and save it. */
    txt = FldGetTextPtr(fld);
    if (txt == NULL) {
	prefs.count = 0;
    } else {
	prefs.count = StrAToI(txt);
    }

    /* Get pointer to seed field. */
    fld = GetObjectPtr(frm, SeedField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);

    /* Get pointer to field's text and save it. */
    txt = FldGetTextPtr(fld);
    if (txt == NULL)
	prefs.seed[0] = 0;
    else
	StrCopy(prefs.seed, txt);
}

static void MainViewClearFields(void)
{
    FormPtr frm;
    FieldPtr fld;

    /* Get pointer to currently active form. */
    frm = FrmGetActiveForm();
    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);

    /* Get pointer to count field and clear it's text. */
    fld = GetObjectPtr(frm, CountField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
    FldDelete(fld, 0, FldGetTextLength(fld));

    /* Get pointer to seed field and clear it's text. */
    fld = GetObjectPtr(frm, SeedField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
    FldDelete(fld, 0, FldGetTextLength(fld));

    /* Get pointer to passphrase field and clear it's text. */
    fld = GetObjectPtr(frm, PassphraseField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
    FldDelete(fld, 0, FldGetTextLength(fld));

    /* Get pointer to response field and clear it's text. */
    fld = GetObjectPtr(frm, ResponseField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
    buf[0] = 0;
    FldSetTextPtr(fld, buf);
    FldDrawField(fld);

    /* Remove any current focus. */
    FrmSetFocus(frm, noFocus);
}

static void MainViewInit(void)
{
    FormPtr frm;
    ListPtr lst;
    ControlType *ctl;
    CharPtr label;
    FieldAttrType fldattr;
    FieldType *fld;
    VoidHand tmp;

    /* Get pointer to main form. */
    frm = FrmGetFormPtr(MainView);
    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);

    /* Initialize list to select an item. */
    lst = GetObjectPtr(frm, HashList);
    FatalErrorIf(lst == NULL, NullPtrMsg, file, __LINE__);
    LstSetSelection(lst, Hash);

    /* Make popup trigger's label match selection made above. */
    ctl = GetObjectPtr(frm, HashListTrigger);
    FatalErrorIf(ctl == NULL, NullPtrMsg, file, __LINE__);
    label = LstGetSelectionText(lst, Hash);
    FatalErrorIf(label == NULL, NullPtrMsg, file, __LINE__);
    CtlSetLabel(ctl, label);

    /* Check format type and set default if the current one is invalid. */
    if ((format != ResponseEngFormat) && (format != ResponseHexFormat))
	format = ResponseEngFormat;
    ctl = GetObjectPtr(frm, format);
    FatalErrorIf(ctl == NULL, NullPtrMsg, file, __LINE__);
    CtlSetValue(ctl, 1);
    if (format == ResponseEngFormat)
	disp = btoe;
    if (format == ResponseHexFormat)
	disp = put8;

    /* Is count greater than zero? If so, stuff it into the count field. */
    if (prefs.count > 0) {
	CharPtr txt;
	tmp = MemHandleNew(5);
	FatalErrorIf(tmp == NULL, NullHandleMsg, file, __LINE__);
	txt = MemHandleLock(tmp);
	FatalErrorIf(txt == NULL, NullPtrMsg, file, __LINE__);
	StrIToA(txt, prefs.count);
	MemHandleUnlock(tmp);
	fld = GetObjectPtr(frm, CountField);
	FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
	FldSetTextHandle(fld, (Handle)tmp);
    }

    /* Does seed contain something? If so, stuff it into the seed field. */
    if (prefs.seed[0] != 0) {
	CharPtr txt;
	tmp = MemHandleNew(StrLen(prefs.seed) + 1);
	FatalErrorIf(tmp == NULL, NullHandleMsg, file, __LINE__);
	txt = MemHandleLock(tmp);
	FatalErrorIf(txt == NULL, NullPtrMsg, file, __LINE__);
	StrCopy(txt, prefs.seed);
	MemHandleUnlock(tmp);
	fld = GetObjectPtr(frm, SeedField);
	FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
	FldSetTextHandle(fld, (Handle)tmp);
    }

    /* Make count field numeric-only. */
    fld = GetObjectPtr(frm, CountField);
    FatalErrorIf(fld == NULL, NullPtrMsg, file, __LINE__);
    FldGetAttributes(fld, &fldattr);
    fldattr.numeric = 1;
    FldSetAttributes(fld, &fldattr);

    /* Remove the english button if the dictionary is not present. */
    if (dict == 0) {
	ControlPtr ctl;
	ctl = GetObjectPtr(frm, ResponseEngFormat);
	CtlHideControl(ctl);
    }
}

static Boolean MainViewHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr frm;

    /* Check out eType to see what kind of event this is. */
    switch(event->eType) {
	case ctlSelectEvent:
	    /* See which control was tapped. */
	    switch (event->data.ctlSelect.controlID) {
		case ComputeButton:
		    MainViewComputeResponse();
		    handled = true;
		    break;
		case ClearButton:
		    MainViewClearFields();
		    handled = true;
		    break;
		case ResponseEngFormat:
		    disp = btoe;
		    format = ResponseEngFormat;
		    handled = false;
		    break;
		case ResponseHexFormat:
		    disp = put8;
		    format = ResponseHexFormat;
		    handled = false;
		    break;
		default:
		    break;
	    }
	    break;
	case popSelectEvent:
	    /* See which popup item was selected. */
	    switch(event->data.popSelect.controlID) {
		case HashListTrigger:
		    Hash = event->data.popSelect.selection;
		    break;
		default:
	    }

	    handled = false;
	    break;
	case menuEvent:
	    /* See which menu the user wants. */
	    switch(event->data.menu.itemID) {
		case AboutMenuID:
		    DoAboutView();
		    handled = true;
		    break;
		default:
		    handled = false;
		    break;
	    }
	    break;
	case frmSaveEvent:
	    MainViewSaveForm();
	    handled = true;
	    break;
	case frmOpenEvent:
	    frm = FrmGetActiveForm();
	    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);
	    MainViewInit();
	    FrmDrawForm(frm);
	    handled = true;
	    break;
	default:
	    break;
    }

    return(handled);
}

static void DoAboutView(void)
{
    FormPtr frm, oldfrm;

    /* Get pointer to the currently active form. */
    oldfrm = FrmGetActiveForm();
    FatalErrorIf(oldfrm == NULL, NullPtrMsg, file, __LINE__);

    /* Get pointer to About form. */
    frm = FrmInitForm(AboutView);
    FatalErrorIf(frm == NULL, NullPtrMsg, file, __LINE__);

    /* Make About form active and wait for it to finish. */
    FrmSetActiveForm(frm);
    FrmDoDialog(frm);

    /* Restore the previously active form. */
    FrmSetActiveForm(oldfrm);

    /* Delete About form. */
    FrmDeleteForm(frm);
}

static Boolean ApplicationHandleEvent(EventPtr event)
{
    FormPtr frm;
    Int     formId;
    Boolean handled = false;

    /* If this is a load form event, load the requested form,
       initialize it, make it active, and set its handler. */
    if (event->eType == frmLoadEvent) {
	formId = event->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	    case MainView:
		FrmSetEventHandler(frm, MainViewHandleEvent);
		break;
	}

	handled = true;
    }

    return(handled);
}

static void EventLoop(void)
{
    EventType event;
    Word      error;

    /* Boiler plate event loop. */
    do {
	EvtGetEvent(&event, evtWaitForever);
	if (! SysHandleEvent(&event))
	    if (! MenuHandleEvent(NULL, &event, &error))
		if (! ApplicationHandleEvent(&event))
		    FrmDispatchEvent(&event);
    } while (event.eType != appStopEvent);
}

static void CheckForDict(void)
{
    DmOpenRef dbp;

    dbp = DmOpenDatabaseByTypeCreator('Dict', 'PKey', dmModeReadOnly);
    if (dbp != 0) {
	dict = 1;
	DmCloseDatabase(dbp);
    } else {
	dict = 0;
    }
}

static void LoadPreferences(void)
{
    Int prefsize = sizeof(prefs);
    Err err;

    /* Load our preferences from the preferences database. */
    err = PrefGetAppPreferences('PKey', 0, &prefs, &prefsize, true);
    if ((err == noPreferenceFound)) {
	err = PrefGetAppPreferences('PKey', 0, &prefs, &prefsize, false);
	if (err == noPreferenceFound) {
	    Hash = 1;
	    format = dict ? ResponseEngFormat : ResponseHexFormat;
	} else {
	    Hash = prefs.hash;
	    format = dict ? prefs.format : ResponseHexFormat;
	}
    } else {
	Hash = prefs.hash;
	format = dict ? prefs.format : ResponseHexFormat;
	PrefSetAppPreferences('PKey', 0, 0, NULL, 0, true);
    }
}

static Err StartApplication(void)
{
    /* Allocate memory for concat'ed passphrase buffer. */
    buf = MemPtrNew(80);
    FatalErrorIf(buf == NULL, NullPtrMsg, file, __LINE__);
    MemSet(buf, 80, 0);

    /* Check for the dictionary. */
    CheckForDict();

    /* Load our preferences. */
    LoadPreferences();

    /* Initiate main form loading. */
    FrmGotoForm(MainView);

    return(0);
}

static void SavePreferences(void)
{
    /* Save our preferences into the preferences database. */
    PrefSetAppPreferences('PKey', 0, 0, &prefs, sizeof(prefs), false);
}

static void StopApplication(void)
{
    /* Have all active forms save their data and then close. */
    FrmSaveAllForms();
    FrmCloseAllForms();

    /* Save our preferences. */
    SavePreferences();

    /* Free any buffer memory we allocated. */
    FatalErrorIf(buf == NULL, NullPtrMsg, file, __LINE__);
    MemPtrFree(buf);
}

static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags)
{
    DWord	romVersion;

    /* Boiler plate rom version checking. */
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion) {
	if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
			== (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
	    FrmAlert(RomAlert);

	    if (romVersion < 0x02000000) {
		AppLaunchWithCommand(sysFileCDefaultApp,
			sysAppLaunchCmdNormalLaunch, NULL);
	    }
	}

	return(sysErrRomIncompatible);
    }

    return(0);
}

DWord PilotMain(Word launchCode, Ptr cmdPBP, Word launchFlags)
{
    Err err;

    err = RomVersionCompatible(0x02003000, launchFlags);
    if (err)
	return(err);

    /* Only respond to the normal startup launch code. */
    if (launchCode == sysAppLaunchCmdNormalLaunch) {
	if ((err = StartApplication()) == 0) {
	    EventLoop();
	    StopApplication();
	}
    }

    return(err);
}
