/*********************************************************************
 * Filename:      flio.c
 * Description:	  file-like i/o for stdio files and dynamic buffers
 * Author:        Thomas Roessler <roessler@guug.de>
 * Modified at:   Mon Jun  1 15:09:52 1998
 * Modified by:   Tage Stabell-Kulo <tage@acm.org>
 *
 * 	Copyright (c) Thomas Roessler
 * 	See the file COPYING for details
 *
 * $Id: flio.c,v 1.22 1999/03/08 19:58:46 perm Exp $
 *
 ********************************************************************/
/* All PalmOS code by Per Harald Myrvang <perm@pasta.cs.uit.no> */


#ifndef __palmos__
# include <assert.h>
# include <stdlib.h>
# include <stdio.h>
# include <errno.h>
# include <string.h>
# ifdef WIN32
#  include <string.h>
# endif
#else	/* not __palmos__ */
# include <unix_stdio.h>
# define APPNETTIMEOUT -1	/* I.e. wait forever */
#endif /* not __palmos__ */

#include "pgplib.h"

#include "flio.h"

#ifdef __palmos__
# define CHUNKSIZE 512
#else	/* __palmos__ */
# define CHUNKSIZE 1024
#endif	/* __palmos__ */

#ifdef __palmos__
extern int pgplib_errno;	/* Located and defined in error.c */
# define errno pgplib_errno	/* A kludge? Indeed. */
#endif	/* __palmos__ */

/* TODO: Rewrite this to work with records and stuff --Perm */
static int 
flio_enlarge_buffer(flio_t *p, u_long chunksize)
{
    u_char *scratch;

    if (!FLIO_IS_BUFF(p->magic)) {
	errno = EINVAL;
	return -1;
    }
    
    scratch = realloc(p->d.buff.b, p->d.buff.size + chunksize);
    if (scratch == NULL) {
	errno = ENOMEM;
	return -1;
    }

    p->d.buff.size += chunksize;
    p->d.buff.b = scratch;

    /* Make sure we clear the "new" area. Otherwise we might end up with 
     * all sort of junk in the buffer.
     * 15Jan1998 Per H. Myrvang <perm@stud.cs.uit.no> 
     */
    memset(p->d.buff.b + p->d.buff.size - chunksize, 0, chunksize);
    
    return 0;
}

void 
flio_close(flio_t *p)
{
    if (p == NULL) 
	return;

    if (FLIO_IS_EXTERN(p->magic)) {
	/* The only thing we should do is to flush.  File descriptors
	 * doesn't need flushing.
	 */
	if (FLIO_IS_FILE(p->magic))
#ifndef __palmos__
	    (void)fflush(p->d.f);
#else	/* __palmos__ */
	ErrDisplay("flio can't be of FILE type under PalmOS!");
#endif	/* __palmos__ */
	return;
    }
    else {
#ifndef __palmos__
	if (FLIO_IS_FILE(p->magic) && p->d.f != NULL) {
	    fclose(p->d.f);		/* fclose fflush'es */
	} 
#else	/* __palmos__ */
	if (FLIO_IS_FILE(p->magic)) {
	    ErrDisplay("flio can't be of FILE type under PalmOS!");
	    return;
	}
#endif	/* __palmos__ */
	else if (FLIO_IS_BUFF(p->magic) && p->d.buff.b != NULL) {
	    memset(p->d.buff.b, 0, p->d.buff.size);
	    free(p->d.buff.b);
	} else
#ifndef __palmos__
	    assert("Impossible" == 0);
#else	/* __palmos__ */
	{
	    ErrDisplay("flio_close(): Impossible file type!");
	    return;
	}
#endif	/* __palmos__ */
    }
    memset(p, 0, sizeof(*p));
    free(p);
}

flio_t *
flio_stdin(void)
{
#ifndef __palmos__
    flio_t *p;
    
    if(!(p = malloc(sizeof(flio_t)))) {
	return NULL;
    }
    p->magic = FLIO_FILE;
    p->d.f = stdin;
    return p;
#else	/* __palmos__ */
    return NULL;
#endif	/* __palmos__ */
}

flio_t *
flio_stdout(void)
{
#ifndef __palmos__
    flio_t *p;

    if (!(p = malloc(sizeof(flio_t)))) {
	return NULL;
    }
    p->magic = FLIO_FILE;
    p->d.f = stdout;
    return p;
#else	/* __palmos__ */
    return NULL;
#endif	/* __palmos__ */
}

flio_t *
flio_stderr(void)
{
#ifndef __palmos__
    flio_t *p;

    if (!(p = malloc(sizeof(flio_t)))) {
	return NULL;
    }
    p->magic = FLIO_FILE;
    p->d.f = stderr;
    return p;
#else	/* __palmos__ */
    return NULL;
#endif	/* __palmos__ */
}

flio_t *
flio_fopen(const char *fname, const char *mode)
{
#ifndef __palmos__
    flio_t *p;
    FILE *f;

    if(!(f = fopen(fname, mode)))
	return NULL;

    if(!(p = malloc(sizeof(flio_t)))) {
	fclose(f);
	errno = ENOMEM;
	return NULL;
    }
    
    p->magic = FLIO_FILE;
    p->d.f = f;
    return p;
#else	/* __palmos__ */
    return NULL;
#endif	/* __palmos__ */
}

/* Attach a flio to a file descriptor */
#ifndef __palmos__
flio_t *
flio_fdreopen(int fd)
{
    flio_t *p;

    p = calloc(1, sizeof(flio_t));
    if (p == NULL) {
	errno = ENOMEM;
	return (NULL);
    }
    p->magic = FLIO_DESC | FLIO_EXTERN;
    p->d.fd = fd;
    return(p);
}
#else	/* __palmos__ */
/* On the pilot this function allows you to attach a socket to a flio. */
flio_t *
flio_fdreopen(Word libRef, NetSocketRef	socket)
{
    flio_t *p;

    p = calloc(1, sizeof(flio_t));
    if (p == NULL) {
	errno = ENOMEM;
	return (NULL);
    }
    p->magic = FLIO_DESC | FLIO_EXTERN;
    p->d.s.libRef = libRef;
    p->d.s.socket = socket;
    return(p);
}
#endif /* __palmos__ */

/* Attach a flio to a FILE(3) */
flio_t *
flio_freopen(FILE *f)
{
#ifndef __palmos__
    flio_t *p;

    if ( ! (p = malloc(sizeof(flio_t)))) {
	errno = ENOMEM;
	return NULL;
    }
    p->magic = FLIO_FILE |  FLIO_EXTERN;
    p->d.f = f;
    return(p);
#else	/* __palmos__ */
    return NULL;
#endif	/* __palmos__ */
} 


flio_t *
flio_newbuff(void)
{
    flio_t *p;

    if(!(p = malloc(sizeof(flio_t)))) {
	errno = ENOMEM;
	return NULL;
    }

    p->magic = FLIO_BUFF;
    p->d.buff.b = NULL;
    p->d.buff.p = 0;
    p->d.buff.size = 0;
    return p;
}


int 
flio_getpos(flio_t *p, flio_pos_t *pos)
{
#ifndef __palmos__
    int	i;
#endif	/* __palmos__ */

    pos->magic = p->magic;
    if (FLIO_IS_BUFF(p->magic)) {
	pos->d.p = p->d.buff.p;
	return 0;
    } 
    else if (FLIO_IS_FILE(p->magic)) {
#ifndef __palmos__
	return fgetpos(p->d.f, &pos->d.f);
#else	/* __palmos__ */
	ErrDisplay("flio can't be of FILE type under PalmOS!");
	return -1;
#endif	/* __palmos__ */
    } else if (FLIO_IS_DESC(p->magic)) {
	/* At the meoment, return -1.  Can someone please
	 * fix this
	 * XXX */
	 /* TODO: Allow for the pilot here. -1 is probably ok.  --Perm */
	return (-1);
    } else {
	errno = EINVAL;
	return -1;
    }
}
    

int 
flio_setpos(flio_t *p, flio_pos_t *pos)
{

    if ( ! FLIO_IS_SAME_TYPE(pos->magic, p->magic))
	return -1;
    if (FLIO_IS_BUFF(p->magic)) {
	if(pos->d.p > (long)p->d.buff.size)
	    return -1;
	p->d.buff.p = pos->d.p;
	return 0;
    } else if (FLIO_IS_FILE(p->magic))
#ifndef __palmos__
	return fsetpos(p->d.f, &pos->d.f);
#else	/* __palmos__ */
    {
	ErrDisplay("flio can't be of FILE type under PalmOS!");
	return -1;
    }
#endif	/* __palmos__ */
    else 
	/* Fixing is needed here as well
	 * XXX */	
	 /* TODO: Allow for the pilot here. -1 is probably ok.  --Perm */
	return -1;
}


int 
flio_getc(flio_t *p, int *i)
{
    int j;


    if (p == NULL || i == NULL)
	return(-1);

    if (FLIO_IS_FILE(p->magic)) {
#ifndef __palmos__
	*i = fgetc(p->d.f);
	if (*i == EOF) {
	    if (feof(p->d.f) != 0)
		return (1);
	    if (ferror(p->d.f) != 0)
		return(-1);
	    assert("Impossile" == 0);
	    /*NOTREACHD*/
	}
#else	/* __palmos__ */
	ErrDisplay("flio can't be of FILE type under PalmOS!");
#endif	/* __palmos__ */
	return 0;
    } else if (FLIO_IS_BUFF(p->magic)) {
	if(p->d.buff.p < p->d.buff.size) {
	    *i = (int) p->d.buff.b[p->d.buff.p++];
	    return 0;
	}
	else {
	    return EOF;
	}
    } else if (FLIO_IS_DESC(p->magic)) {
	*i = 0;
#ifndef __palmos__
	j = read(p->d.fd, i, 1);
#else	/* __palmos__ */
	j = NetLibReceive(p->d.s.libRef, p->d.s.socket, i, 1,
		0,0,0, APPNETTIMEOUT,
		NULL);	/* TODO: Use some other value for errno? */
#endif	/* __palmos__ */
	switch (j) {
	case -1:
	    return -1;
	case 0:
	    return 1;
	default:
	    return 0;
	}
    } else {
	errno = EINVAL;
	return -1;
    }
#ifndef __palmos__
    assert(0);
#else	/* __palmos__ */
    ErrDisplay("0");	/* _Very_ informative. I'm being ironic. Haha. */
#endif	/* __palmos__ */
}

/* XXX - use flio_getc()? */

char *
flio_gets(char *b, u_long s, flio_t *p)
{

    u_char *source; 
    char *dest;
    int i,j;

    if (FLIO_IS_FILE(p->magic))
#ifndef __palmos__
	return fgets(b, s, p->d.f);
#else	/* __palmos__ */
	{
	    ErrDisplay("flio can't be of FILE type under PalmOS!");
	    return NULL;
	}
#endif	/* __palmos__ */
    else if (FLIO_IS_BUFF(p->magic)) {
	dest = b;
	source = p->d.buff.b;
	if(!s-- || !p->d.buff.size || p->d.buff.p == p->d.buff.size)
	    return NULL;
	
	while(s-- && p->d.buff.p < p->d.buff.size) 
	{
	    if((*dest++ = source[p->d.buff.p++]) == '\n')
		break;
	}
	    
	*dest = '\0';
	return b;
    } else if (FLIO_IS_DESC(p->magic)){
	dest = b;
	while (s--) {
#ifndef __palmos__
	    i = read(p->d.fd, &j, 1);
#else	/* __palmos__ */
	    i = NetLibReceive(p->d.s.libRef, p->d.s.socket, &j, 1,
		    0,0,0, APPNETTIMEOUT,
		    NULL);	/* TODO: Use some other value for errno? */
#endif	/* __palmos__ */
	    if (i == 0 || i == -1) {
		b = NULL;
	    }
	    if ((*dest++ = (u_char)j) == '\n')
		break;
	}
	*dest = 0;
	return(b);
    } else
	return NULL;
}

int 
flio_putc(int c, flio_t *p)
{
    unsigned char d = (unsigned char) c;
    int i;


    if (FLIO_IS_FILE(p->magic))
#ifndef __palmos__
	return fputc(c, p->d.f);
#else	/* __palmos__ */
	{
	    ErrDisplay("flio can't be of FILE type under PalmOS!");
	    return NULL;
	}
#endif	/* __palmos__ */

    else if (FLIO_IS_BUFF(p->magic)) {
	if (p->d.buff.p == p->d.buff.size) {
	    if (flio_enlarge_buffer(p, CHUNKSIZE) == -1)
		return EOF;
	}
	
	p->d.buff.b[p->d.buff.p++] = d;
	return (int) d; /* fputc semantics */
    } else if (FLIO_IS_DESC(p->magic)) {
#ifndef __palmos__
	i = write(p->d.fd, &d, 1);
#else	/* __palmos__ */
	i = NetLibSend(p->d.s.libRef, p->d.s.socket, &d, 1,
		0,0,0, APPNETTIMEOUT,
		NULL);	/* TODO: Use some other value for errno? */
#endif	/* __palmos__ */
	if (i == 0 || i == -1)
	    return (EOF);		/* putc semantics */
	return (int)d;
    } else {
	errno = EINVAL;
	return EOF;
    }
}


int 
flio_puts(const char *s, flio_t *p)
{

    if (FLIO_IS_FILE(p->magic)) {
#ifndef __palmos__
	return fputs(s, p->d.f);
#else	/* __palmos__ */
	{
	    ErrDisplay("flio can't be of FILE type under PalmOS!");
	    return NULL;
	}
#endif	/* __palmos__ */
    } else if (FLIO_IS_BUFF(p->magic) || FLIO_IS_DESC(p->magic)) {
	int rv;

	for(rv = 0; *s && (rv = flio_putc(*s, p)) != EOF; s++)
	    ;
	return rv;
    }
    else {
	errno = EINVAL;
	return EOF;
    }
}


int 
flio_write(void *ptr, u_long s, u_long n, flio_t *p)
{
    int i;
    u_char *src;

    if (FLIO_IS_FILE(p->magic)) {
#ifndef __palmos__
	return fwrite(ptr, s, n, p->d.f);
#else	/* __palmos__ */
	{
	    ErrDisplay("flio can't be of FILE type under PalmOS!");
	    return 0;
	}
#endif	/* __palmos__ */
    } else if (FLIO_IS_BUFF(p->magic)) {
	for (i = 0, src = ptr; n; n--, i++) {
	    if (p->d.buff.p + s >= p->d.buff.size) {
		if (flio_enlarge_buffer(p, s) == -1)
		    return i;
	    }
	    memcpy(p->d.buff.b + p->d.buff.p, src, s);
	    p->d.buff.p += s;
	    src += s;
	}
	return i;
    } else if (FLIO_IS_DESC(p->magic)) {
#ifndef __palmos__
	i = write(p->d.fd, ptr, n*s);	/* was ..., n*s, ptr */
#else	/* __palmos__ */
	i = NetLibSend(p->d.s.libRef, p->d.s.socket, ptr, n*s,
		0,0,0, APPNETTIMEOUT,
		NULL);	/* TODO: Use some other value for errno? */
#endif	/* __palmos__ */
	if (i == -1)
	    return (0);
	else
	    return i;
    }
    else {
	errno = EINVAL;
	return 0;
    }
}
