/*********************************************************************
 * Filename:      keybase.c
 * Description:
 * Author:        Tage Stabell-Kulo <tage@acm.org>
 * Modified at:   Sat Feb 28 12:31:30 1998
 * Modified by:   Tage Stabell-Kuloe <tage@acm.org>
 *
 * 	Copyright (c) University of Troms, Norway
 * 	See the file COPYING for details
 *
 * $Id: keybase.c,v 1.16 1998/02/28 12:30:00 tage Exp $
 *
 ********************************************************************/


/*
 * In order to keep track of encryption keys we store them in a set
 * of databases.  The "real" index is the KeyID (64 bit long) and
 * you can use this to find a key based on the KeyID.
 *
 * Then, the string(s) associated with each key ("User ID Packet" in
 * PGP terms) are stored in a separate database for faster lookup.
 */

/* A name must be provided.  From it, we will create two file-names.
 * If these files exists, we will use them, if not, we will create them.
 */

#include <pgplib.h>


#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
/* use DB */
#ifdef HAVE_DB_H
#  include <db.h>
#  define DATA_TYPE DBT 
#  define DATA_PTR(item) ((item).data)
#  define DATA_PTR_TYPE void* 
#  define DATA_SIZE(item) ((item).size)
#  define DATA_SIZE_TYPE size_t
#  define DATA_RELEASE(item)
#  define DATA_ALLOC_COPY_RELEASE(item, buf, len)                  \
              (                                                    \
               (                                                   \
                (len) =                                            \
                (                                                  \
                 (((buf) = malloc(DATA_SIZE(item))) != NULL)       \
                 ?                                                 \
                 (                                                 \
                  memcpy((buf), DATA_PTR(item), DATA_SIZE(item))   \
                  ,                                                \
                  DATA_SIZE(item)                                  \
                 )                                                 \
                 :                                                 \
                 0                                                 \
                )                                                  \
               )                                                   \
               ,                                                   \
               (buf)                                               \
              )

#  define DB_CLOSE(db) (db->close)(db)
#  define DB_SYNC(db) (db->sync)(db, 0)
#  define DB_FETCH(db, key, contents) ((DATA_PTR(contents) = NULL), db->get(db, &key,&contents, 0)) 
#  define DB_STORE(db, key, contents, replace) \
              ((db)->put)((db), &(key), &(contents), (((replace) == REPLACE) ? 0 : R_NOOVERWRITE))
#  define DB_EXISTS(db, key, contents) \
              ((db)->get)((db), &(key), &(contents), 0)
#endif

/* use gdbm */
#ifdef HAVE_GDBM_H
#  include <gdbm.h>
#  define DATA_TYPE datum
#  define DATA_PTR(item) ((item).dptr)
#  define DATA_PTR_TYPE char*
#  define DATA_SIZE(item) ((item).dsize)
#  define DATA_SIZE_TYPE int
#  define DATA_RELEASE(item) free(DATA_PTR(item))
#  define DATA_ALLOC_COPY_RELEASE(item, buf, len) ((len) = DATA_SIZE(item), (buf) = DATA_PTR(item))
#  define DB_CLOSE(db) gdbm_close(db)
#  define DB_SYNC(db) gdbm_sync(db)
#  define DB_FETCH(db, key, contents) ((contents = gdbm_fetch(db, key)), \
                                       ((DATA_PTR(contents) == NULL) ? 1 : 0))
#  define DB_STORE(db, key, contents, replace) \
              gdbm_store((db), (key), (contents), (((replace) == REPLACE) ? GDBM_REPLACE : GDBM_INSERT))
#  define DB_EXISTS(db, key, contents) (gdbm_exists((db), (key)) ? 0 : 1)
#endif
#ifdef WIN32
# define MAXPATHLEN _MAX_PATH
#endif
#include <md5.h>

#if (HAVE_STRERROR == 0)
extern char *  strerror (int);
#endif

#ifndef FALSE
#   define FALSE 0 
#   define TRUE 1 
#endif

#define KEYEXT		".key"
#define STRINGEXT	".string"
#define	REVSTRINGEXT	".revstring"
#define KEYIDEXT	".keyid"
#define KEYHASHEXT	".hash"

#ifdef HAVE_DB_H             
static DB      *keydb;          
static DB      *stringdb;                  
static DB      *revstringdb;   
static DB      *keyiddb; 
static DB      *keyhashdb;
#endif           
#ifdef HAVE_GDBM_H
static GDBM_FILE       keydb;
static GDBM_FILE       stringdb;
static GDBM_FILE       revstringdb;
static GDBM_FILE       keyiddb;
static GDBM_FILE       keyhashdb;
#endif          

int	dbase_open = 0;


void
close_databases()
{
    pgplib_errno = PGPLIB_ERROR_DB_FATAL;
    exit(1);
}

int
keydb_init(char 	*caller_name,
	   openmode_t	caller_mode)
{
    char 	filename_b[MAXPATHLEN];
    char	*filename;
    char 	*p;
    int		mode;
#ifdef HAVE_DB_H
    HASHINFO   *hinfo;
#endif

    if(dbase_open)
      return 0;

    if (caller_name == NULL)
	filename = NULL;
    else
	filename = filename_b;

    switch(caller_mode) {
    case READER:
#ifdef HAVE_DB_H
# ifdef WIN32
	mode = O_RDONLY|O_BINARY;
# else
	mode = O_RDONLY;
# endif
#endif
#ifdef HAVE_GDBM_H
	mode = GDBM_READER;
#endif
	break;
    case WRITER:
#ifdef HAVE_DB_H
# ifdef WIN32
	mode = O_RDWR|O_BINARY;
# else
	mode = O_RDWR;
# endif
#endif
#ifdef HAVE_GDBM_H
	mode = GDBM_WRITER;
#endif
	break;
    case NEWDB:
#ifdef HAVE_DB_H
# ifdef WIN32
	mode = O_CREAT | O_TRUNC | O_RDWR | O_BINARY;
# else
	mode = O_CREAT | O_TRUNC | O_RDWR;
# endif
#endif
#ifdef HAVE_GDBM_H
	mode = GDBM_NEWDB;
#endif
	break;
    default:
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    keydb = NULL; stringdb = NULL; revstringdb = NULL; keyiddb = NULL;

#ifdef HAVE_DB_H
#if 0

    You need to tune these to meet your requirement, in particular,
    how much memory you have available for the system.
    Be careful, you can easily get a 100 MB process if you have a lot of
    keys at hand.

    hinfo = calloc(1, sizeof(*hinfo));
    if ( hinfo == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    hinfo->cachesize = 5*CACHE_SIZE;
    hinfo->bsize     = BSIZE;
    hinfo->nelem     = NELEM;
#else
    hinfo = NULL;
#endif
#endif

    if(filename) {
	filename[0] = '\0';
	strcat(filename, caller_name);
	strcat(filename, KEYEXT);
    }
#ifdef HAVE_DB_H
    keydb = dbopen (filename, mode, 0600, DB_HASH, hinfo);
#endif
#ifdef HAVE_GDBM_H
    keydb = gdbm_open(filename, 0, caller_mode, 0600, close_databases);
#endif
    if ( keydb == NULL )
	goto err;

#ifdef HAVE_DB_H
#if 0
    More memory tuning

    hinfo.cachesize = CACHE_SIZE;
    hinfo.bsize     = BSIZE;
#endif
#endif
    if(filename) {
	p = strrchr(filename, '.');
	*p = '\0';
	strcat(filename, STRINGEXT);
    }
#ifdef HAVE_DB_H
    stringdb = dbopen(filename, mode, 0600, DB_HASH, hinfo);
#endif
#ifdef HAVE_GDBM_H
    stringdb = gdbm_open(filename, 0, caller_mode, 0600, close_databases);
#endif
    if ( stringdb == NULL )
	goto err;

    if(filename) {
	p = strrchr(filename, '.');
	*p = '\0';
	strcat(filename, REVSTRINGEXT);
    }
#ifdef HAVE_DB_H
    revstringdb = dbopen (filename, mode, 0600, DB_HASH, hinfo);
#endif
#ifdef HAVE_GDBM_H
    revstringdb = gdbm_open(filename, 0, caller_mode, 0600, close_databases);
#endif
    if (revstringdb == NULL )
	goto err;

    if(filename) {
	p = strrchr(filename, '.');
	*p = '\0';
	strcat(filename, KEYIDEXT);
    }
#ifdef HAVE_DB_H
    keyiddb = dbopen (filename, mode, 0600, DB_HASH, hinfo);
#endif
#ifdef HAVE_GDBM_H
    keyiddb = gdbm_open(filename, 0, caller_mode, 0600, close_databases);
#endif
    if ( keyiddb == NULL )
	goto err;

    if(filename) {
	p = strrchr(filename, '.');
	strcpy(p, KEYHASHEXT);
    }
#ifdef HAVE_DB_H
    keyhashdb = dbopen(filename, mode, 0600, DB_HASH, hinfo);
#endif
#ifdef HAVE_GDBM_H 
    keyhashdb = gdbm_open(filename, 0, caller_mode, 0600, close_databases);
#endif 

    dbase_open = 1;
    return(0);

err:
    keydb_end();
    pgplib_errno = PGPLIB_ERROR_DB_FATAL;
    return -1;
}


void
keydb_end(void)
{
    if ( keydb != NULL )
	DB_CLOSE(keydb);
    if ( stringdb !=  NULL )
	DB_CLOSE(stringdb);
    if ( revstringdb != NULL )
	DB_CLOSE(revstringdb);
    if ( keyiddb != NULL )
	DB_CLOSE(keyiddb);
    if(keyhashdb != NULL)
	DB_CLOSE(keyhashdb);
    dbase_open = 0;
    keydb = stringdb = revstringdb = keyiddb = keyhashdb =NULL;
}

void
keydb_sync(void)
{
    if ( keydb != NULL )
	DB_SYNC(keydb);
    if ( stringdb !=  NULL )
	DB_SYNC(stringdb);
    if ( revstringdb != NULL )
	DB_SYNC(revstringdb);
    if ( keyiddb != NULL )
	DB_SYNC(keyiddb);
    if(keyhashdb != NULL)
	DB_SYNC(keyhashdb);
}

/* Below is all the routines to store, fetch, alter and delete
 * PGP compatible keys and certificates.
 * All these routines take and return nicely formatted keys and certificates
 * as pointers to struct's.  The internal format (in the databases)
 * are less nice since it has to be marshalled (and, urk, compatible).
 *
 * In the Key database, the keys (together with UserID and signatures)
 * are stored in PGP format.  This way we can print the
 * key fast by simply "armour" the buffer directly.  It takes some
 * effort to convert the buffer to the correct type of C-struct's,
 * but that is the price we have to pay.
 *
 * The strings (UserID) are stored in a separate database.  This
 * way we can locate one _very_ fast
 *
 * The KeyID that pgp presents (only 32 bit) are stored in a third
 * database.
 */

/*
 * Take a string ("0x12345678") and convert it into the corresponding
 * number, and find the key.  No key: return 1 and set keyid to 0.
 * On error return -1;
 */

int
find_shortid(char 	*short_string,
	     PGPKEYID	*keyid)
{

    DATA_TYPE	key, contents;
    u_long	short_id;
    char	*p;

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    if ( short_string == NULL || *short_string == 0 || keyid == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    PGPKEYID_ZERO(*keyid);

    short_id = strtoul(short_string, &p, 0);
    if ( p == NULL || *p != 0 ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    DATA_PTR(key) = (char *)&short_id;
    DATA_SIZE(key) = sizeof(short_id);

    switch (DB_FETCH(keyiddb, key, contents))
    {
    default:
    case -1:
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    case 1:
	return 1;
    case 0:
	if (DATA_SIZE(contents) != sizeof(PGPKEYID))
	{
	    DATA_RELEASE(contents);
	    pgplib_errno = pgplib_errno = PGPLIB_ERROR_IMPOSSIBLE;
	    return(-1);
	}
	memcpy(keyid, DATA_PTR(contents), sizeof(PGPKEYID));
	DATA_RELEASE(contents);
	return(0);
    }	
    return(1);
}

int
store_pubkey(PGPpubkey_t	*pubkey,
	      repl_t		replace)
{
    DATA_TYPE	key, contents;
    PGPKEYID	key_id;
    u_long	short_id;
    int		key_size;
    u_char	*key_buf;
    /* scratch */
    int 	i;

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    if ((replace != INSERT && replace != REPLACE) || pubkey == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }
    if ( PGPKEYID_FromBN(&key_id,pubkey->N) != 0) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    i = pubkey2buf(pubkey,  &key_buf);
    if ( i != 0 ) {
	return(-1);
    }

    key_size = packet_length(key_buf);
    if ( key_size == -1 )
	return(-1);

    i = store_pubkey_rec(key_id, replace, key_buf, key_size);
    free(key_buf);
    switch (i) {
    case -1:
	/*FALLTHROUGH*/
    case 1:
	return(i);
    case 0:
	break;
    default:
	pgplib_errno = PGPLIB_ERROR_IMPOSSIBLE;
	return (-1);
    }

    /* The store went well.  We then proceed by adding the shorter
     * id presented by "pgp -kv", a string representing a 32 bit
     * number.  However, we store the string as the number it is
     * supposed to represent (occupies less than half the space)
     */   

    PGPKEYID_ExtractShort(&key_id, &short_id);

    DATA_PTR(key) = (void *)&short_id;
    DATA_SIZE(key) = sizeof(short_id);

    DATA_PTR(contents) = (void *) &key_id;
    DATA_SIZE(contents)= sizeof(key_id);

    i = DB_STORE(keyiddb, key, contents, INSERT);
    if ( i == -1 ) {
	pgplib_errno = PGPLIB_ERROR_DB_FATAL;
	return(-1);
    }
    else
	if (i == 1)
	    /* R_NOOVERWRITE flag was set and the key already exists.
	     * This case is identified for further modification, if
	     * required.
	     */

           ;

    return(0);
}

int
get_pubkey_rec(PGPKEYID	key_id,
	       int	*len,
	       u_char 	**buf)
{

    DATA_TYPE	key, contents;


    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    if ( PGPKEYID_IsZero(key_id) || len == NULL || buf == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    DATA_PTR(key) = (void *)&key_id;
    DATA_SIZE(key) = sizeof(key_id);

    switch (DB_FETCH(keydb, key, contents))
    {
    default:
    case -1:
        pgplib_errno = PGPLIB_ERROR_SYSERR;
	return -1;
    case 1:
	return(1);
    case 0:
	if (DATA_ALLOC_COPY_RELEASE(contents, *buf, *len) == NULL) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return -1 ;
	}
	return 0;
    }
    /*NOTREACHED*/
}

int
store_pubkey_rec(PGPKEYID	key_id,
		 repl_t 	user_replace,
		 u_char		*buf,
		 int 		size)
{
    DATA_TYPE 	key, contents;
    int 	i;
    u_char	digest[16];

    if ( ! dbase_open) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    if ((user_replace != REPLACE  && user_replace != INSERT) || buf == NULL
	|| size < 1)
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    DATA_PTR(key) = (void *)&key_id;
    DATA_SIZE(key) = sizeof(key_id); 

    DATA_PTR(contents) = (void *)buf;
    DATA_SIZE(contents) = size;

    assert(DATA_SIZE(contents) > 0);
    assert(DATA_SIZE(key) > 0);
    assert(DATA_PTR(contents) != NULL);
    assert(DATA_PTR(key) != NULL);

    i = DB_STORE(keydb, key, contents, user_replace);
    if (i == 0)
    {
	MD5(buf, size, digest);
	DATA_PTR(key) = (char*)&key_id;
	DATA_SIZE(key) = sizeof(key_id);
	DATA_PTR(contents) = digest;
	DATA_SIZE(contents) = sizeof(digest);  
	DB_STORE(keyhashdb, key, contents, user_replace);
    }

    return( i );
}

/*
 * If the key exists, return 1.  If not, return 0.  Error: -1
 */

int
key_exists(PGPKEYID	key_id)
{
#ifdef HAVE_DB_H
    DATA_TYPE	contents;
#endif
    DATA_TYPE	key;    
    int 	i;

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    DATA_PTR(key) = (void *)&key_id;
    DATA_SIZE(key) = sizeof(key_id);

    i = DB_EXISTS(keydb, key, contents);
    if ( i == -1 ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }

    return ( 1 - i );
}

/*
 * Storing a UID consists of two actions.  First, store the KeyID
 * indexed by the UID.  Then, store the UID indexed by the KeyID.
 */

int
store_uid(PGPuserid_t	*user_id,
	  PGPKEYID 	key_id)
{
    DATA_TYPE	key, contents;
    res_t	what;
    int		used, how_far, found, newlen;

    PGPuserid_t	*uid;
    u_char	*buf;
    /* scratch */
    int 	i;
    u_char	*p;
    void	*tmpptr;

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    if ( user_id == NULL || user_id->name == NULL 
	 || PGPKEYID_IsZero(key_id)) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    DATA_PTR(key) = (DATA_PTR_TYPE)user_id->name;
    DATA_SIZE(key) = strlen(user_id->name)+1;

    found = 0;

    if (DB_FETCH(stringdb, key, contents) == 0)
    {
	void *buf;
	PGPKEYID *tmp;
	int j;
	for (j = 0; 
	     !found && (DATA_SIZE_TYPE)(j * sizeof(PGPKEYID)) < DATA_SIZE(contents); 
	     j++) {
	    /* Tage
	    tmp =  DATA_PTR(contents) + (j * sizeof(PGPKEYID));
	    if (PGPKEYID_EQUAL(tmp, key_id)) {
	    */
	    tmp =  (PGPKEYID *)(DATA_PTR(contents) + (j * sizeof(PGPKEYID)));
            if (PGPKEYID_EQUAL(*tmp, key_id)) {
		found = 1;
	    }
	}

	if (!found)
	{
	    if (!(buf = malloc(DATA_SIZE(contents) + sizeof(key_id))))
	    {
		DATA_RELEASE(contents); 
		pgplib_errno = PGPLIB_ERROR_SYSERR;
		return -1;
	    }

	    memcpy(buf, DATA_PTR(contents), DATA_SIZE(contents));
	    memcpy((char *)buf + DATA_SIZE(contents), &key_id, sizeof(key_id));
	    DATA_RELEASE(contents); 
	    DATA_SIZE(contents) += sizeof(key_id);
	    tmpptr = DATA_PTR(contents) = buf;
	}
    }
    else
    {
	if (!(tmpptr = DATA_PTR(contents) = malloc(sizeof(key_id))))

	{
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return -1;
	}
	memcpy(DATA_PTR(contents), (void *) &key_id, sizeof(key_id));
	DATA_SIZE(contents) = sizeof(key_id);

    }
    if ( ! found)
    {  
	i = DB_STORE(stringdb, key, contents, REPLACE);
	free(tmpptr);

	if ( i == -1 ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	}

    }

    DATA_PTR(key)= (DATA_PTR_TYPE) &key_id;
    DATA_SIZE(key) = sizeof(key_id);

    switch(DB_FETCH(revstringdb, key, contents))
    {
    default:
	assert(0);
    case -1:
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    case 1:
	/* This is the first UID for this key */
	i = uid2buf(user_id, (u_char **) &tmpptr);
	DATA_PTR(contents) = tmpptr;
	if ( i != 0 )
	    return(-1);
	DATA_SIZE(contents) = packet_length((u_char*)tmpptr);
	if ( DATA_SIZE(contents) == -1 )
	{
	    free(tmpptr);
	    return(-1);
	}
	i = DB_STORE(revstringdb, key, contents, REPLACE);
	free(tmpptr);
	if ( i != 0 ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	}
	else
	    return(0);
	/*NOTREACHED*/
	assert(0);
    case 0:
	/* There was already (at least) one UID associated with
	 * this key.  We have to walk through them to check if
	 * the one we have already is here.
	 * Note that we (should) have a PGP compatile buffer at hand.
	 */
	how_far = 0;
	found = 0;
	while( how_far < DATA_SIZE(contents)  && found == 0) {
	    i = get_keyentry_from_buf( (u_char *)DATA_PTR(contents) + how_far,
				       DATA_SIZE(contents) - how_far,
				       &what, NULL, NULL, &uid, NULL, NULL,
				       &used);
	    if ( i != 0 ) {
		DATA_RELEASE(contents);
		return(-1);
	    }

	    /*
	     * The string DB should not contain anything but UID's.
	     */
	    assert( what == UID);

	    how_far += used;

	    if ( strcmp(user_id->name, uid->name) == 0 ) {
		found = 1;
	    }
	    free_uid(uid);
	}
	if ( found == 1 )
	    return(1);

	if ( uid2buf(user_id, &buf) != 0 )
	    return(-1);

	if ( (newlen = packet_length(buf)) == -1 )
	    return -1;

	if ( (p = malloc(DATA_SIZE(contents) + newlen)) == NULL ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	}
	memcpy(p, DATA_PTR(contents), DATA_SIZE(contents));
	memcpy(p+DATA_SIZE(contents), buf, newlen);
	free(buf);
	DATA_PTR(contents) = p;
	DATA_SIZE(contents) += newlen;

	i = DB_STORE(revstringdb, key, contents, REPLACE);
	free(p);
	if ( i != 0 ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	}
	return(0);
    }
    assert(0);
    return(0);
    /*NOTREACHED*/
}


int
keys_from_function(int (*func)(const void *, const char *uid),
		   void *info,
		   int *num_keys,
		   PGPKEYID **keys)
{
    DATA_TYPE key, contents;
#ifdef HAVE_GDBM_H
    DATA_TYPE nextkey;
#endif
    char *n = NULL;
    char *t;
    int nl = 0;
#ifdef HAVE_DB_H
    int rv;
#endif
    int alloced = 0;
    PGPKEYID *tmp;

    *num_keys = 0;
    *keys = NULL;

#ifdef HAVE_DB_H
    rv = stringdb->seq(stringdb, &key, &contents, R_FIRST);
    while(rv == 0) {
#endif
#ifdef HAVE_GDBM_H
    key = gdbm_firstkey(stringdb);
    while (DATA_PTR(key) != NULL) {
	nextkey = gdbm_nextkey(stringdb, key);
#endif
	if(DATA_SIZE(key) + 1 >= nl)

	{
	    if((t = realloc(n, (nl = DATA_SIZE(key) + 1))) == NULL)
	    {
		free(n);
		pgplib_errno = PGPLIB_ERROR_SYSERR;
		*keys = NULL;
		*num_keys = 0;
#ifdef HAVE_GDBM_H
		free(DATA_PTR(key));
		if (DATA_PTR(nextkey)) 
		    free(DATA_PTR(nextkey));
#endif
		return -1;
	    }
	    n = t;
	}

	memcpy(n, DATA_PTR(key), DATA_SIZE(key));
	n[DATA_SIZE(key)] = '\0';
	
	if(func(info, n))
	{
#ifdef WIN32
	    unsigned int j;
#else
	    int j;
#endif
#ifdef HAVE_GDBM_H
      contents = gdbm_fetch(stringdb, key);
#endif
           for(j = 0; (DATA_SIZE_TYPE)(j * sizeof(PGPKEYID)) < DATA_SIZE(contents); j++)
	   {
		if(*num_keys >= alloced)
		{
		    if((tmp = realloc(*keys, (alloced += 128) * sizeof(PGPKEYID))) == NULL)
		    {
			pgplib_errno = PGPLIB_ERROR_SYSERR;
			if(*keys) free(*keys);
			*keys = NULL;
#ifdef HAVE_GDBM_H
			free(DATA_PTR(key));
			if (DATA_PTR(nextkey)) free(DATA_PTR(nextkey));
			free(DATA_PTR(contents));
#endif
			return -1;
		    }
		    *keys = tmp;
		}
		memcpy(*keys + (*num_keys)++,(PGPKEYID *) DATA_PTR(contents) 
		       + j, sizeof(PGPKEYID));
	    }
#ifdef HAVE_GDBM_H
	   free(DATA_PTR(contents));
#endif
	}

#ifdef HAVE_DB_H
        rv = stringdb->seq(stringdb, &key, &contents, R_NEXT);
#endif
#ifdef HAVE_GDBM_H
	free(DATA_PTR(key));
	key = nextkey;
#endif
    }
    
    if(n) free(n);
    return 0;
}

struct search_info {
    int save_case;
    char *s;
};

static int keycompare(const void *info, const char *id)
{
    struct search_info *_info = (struct search_info *)info;
    char *s = strdup(id);
    char *u;
    int rv = 0;

    if(!s)
	return 0;
    
    if(!_info->save_case)
	for(u = s; *u; u++)
	    *u = toupper(*u);
    
    if(!*_info->s || strstr(s, _info->s))
	rv = 1;
    
    free(s);

    return rv;
}

/*
 * Find a string in the database.  We start by checking if we can
 * locate the exact string.  If not, we scan the database looking
 * for matches.
 * If save_case is TRUE we preserve case, if not, everything
 * is converted to UPPER.
 * Return an the number of KeyID's we found with a match (in num_keys),
 * and an array (at least) that big of the numbers.
 */

int
keys_from_string(char		*string,
		 int		save_case,
		 int		*num_keys,
		 PGPKEYID	**keys)
{
    DATA_TYPE	key, contents;
    int		ret_value;
    struct 	search_info si;
    char	*tmp;

    if ( string == NULL || num_keys == NULL || keys == 0) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    *num_keys = 0;
    *keys = NULL;


    if (*string) {
	DATA_PTR(key) = (DATA_PTR_TYPE)string;
	DATA_SIZE(key) = strlen(string) + 1;

	switch (DB_FETCH(stringdb, key, contents))
	{
	default:
	case -1:
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	case 0:
	    if (DATA_ALLOC_COPY_RELEASE(contents, (char*)*keys, *num_keys) 
		== NULL)
	    {
		pgplib_errno = PGPLIB_ERROR_SYSERR;
		return(-1);
	    }
	    *num_keys /= sizeof(PGPKEYID);
	    return 0;
	case 1:
	    break;
	}
    }

    si.save_case = save_case;
    if(!(si.s = strdup(string)))
    {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return -1;
    }

    if(!save_case)
    {
	for(tmp = si.s; *tmp; tmp++)
	    *tmp = toupper(*tmp);
    }

    ret_value = keys_from_function(keycompare, (void *) &si, num_keys, keys);
    
    free(si.s);
    return ret_value;
    
}

/*
 * Find the UIDs attached to this key.  Return of 1 means no UIDs
 * (or possibly unknown key ?), -1 is error, 0 is OK.
 * The array of pointers to UID's is NULL terminated.
 */

int
uids_from_keyid(PGPKEYID	key_id,
		PGPuserid_t	***uidpp)
{
    DATA_TYPE	key, contents;
    int		offset;
    res_t	what;
    int		ent;
    int		used;
    PGPuserid_t	**arr, **tmp;
    /* scratch */
    int		i,j;

    if ( PGPKEYID_IsZero(key_id) || uidpp == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    if ( ! dbase_open ) {
	pgplib_errno = PGPLIB_ERROR_DB_NOT_OPEN;
	return(-1);
    }

    DATA_PTR(key) = (void *)&key_id;
    DATA_SIZE(key) = sizeof(key_id);

    memset(&contents, 0 , sizeof(contents));

    switch (DB_FETCH(revstringdb, key, contents))
    { 
    default:
    case -1:
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return -1;
    case 1: 
	return 1;
    case 0:
	break;
    }    

    /* Ensure our array is NULL terminated */
    if ( (arr = calloc(1, sizeof(PGPuserid_t *))) == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	DATA_RELEASE(contents);
	return(-1);
    }

    ent = 1;
    offset = 0;
    do {
	tmp = realloc(arr, (ent+1) * sizeof(PGPuserid_t *));
	if ( tmp == NULL ) {
	    free(arr);
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    DATA_RELEASE(contents);
	    return(-1);
	}
	arr = tmp;
	arr[ent] = NULL;
	i = get_keyentry_from_buf((u_char*)DATA_PTR(contents) + offset,
				  DATA_SIZE(contents) - offset,
				  &what,
				  NULL, NULL, &arr[ent-1], NULL, NULL,
				  &used);
	switch ( i ) {
	case -1:
	    for(j=0; j<ent-2; j++)
		free(arr[j]);
	    free(arr);
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    DATA_RELEASE(contents);
	    return(-1);
	case 1:
	    *uidpp = arr;
	    DATA_RELEASE(contents);
	    return(0);
	case 0:
	    /* the revstring database is suppoed to only hold UID's */
	    assert (what == UID);
	    offset += used;
	    ent += 1;
	    break;
	default:
	    assert(0);
	    /*NOTREACHED*/
	}
    } while (offset < DATA_SIZE(contents));
    *uidpp = arr;
    DATA_RELEASE(contents);
    return(0);
}

/*************************************************************
 * Now comes the higher level routines to access the databases.
 * They were moved here from keymgmnt.c after a suggestion
 * made by tzeruch@ceddec.com.
 *************************************************************
 */

/*
 * We find the key, and add the UserID, unless it is already present;
 * both the length and the characters themselves have to match (case
 * sensitive).
 * We add the UserID to the UserID database with the KeyID as
 * data.
 * We add the UserID to the reversUID database with the keyID as key
 * and UID as data.
 */

int
add_uid(PGPKEYID	key_id,
	PGPuserid_t	*user_id)
{
    u_char 	*rec;
    int		uid_len;
    res_t	what;
    int 	found;
    int 	sofar, buffer_len, used;
    int		found_pubkey;
    PGPuserid_t	*tmpuid;
    /* scratch */
    int		i;
    u_char	*p, *q;

    if ( user_id == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    i = get_pubkey_rec(key_id, &buffer_len, &rec);
    if ( i != 0 ) {
	return(i);
    }

    /* We parse the buffer, skipping things we don't care about,
     * looking for a matching UserID.
     * if we don't find a match, we add this at the end.
     */
    found = 0;
    found_pubkey = 0;
    sofar = 0;
    while ((sofar < buffer_len) && (found == 0))
    {
	i = get_keyentry_from_buf(rec + sofar, buffer_len - sofar, &what,
				  NULL, NULL, &tmpuid, NULL, NULL, &used);

	if (i == 1 || i == -1)
	    return -1;
	sofar += used;

	if (what == UID)
	{
	    if(!strcmp(user_id->name, tmpuid->name))
		found = 1;
	    free_uid(tmpuid);
	    free(tmpuid);
	}
    } /* While */

    if ( found == 0 ) {

	/* If we get here, we have a new UserID.  Let's add it. */
	i  = uid2buf(user_id, &p);
	if ( i  != 0 )
	    return(-1);
	uid_len = packet_length(p);
	if ( uid_len == -1 ) {
	    free(p);
	    free(rec);
	    return(-1);
	}
	q = realloc(rec, buffer_len + uid_len);
	if ( q == NULL ) {
	    free(p);
	    free(rec);
	    return(-1);
	}
	rec = q;
	memcpy(rec + buffer_len, p, uid_len);
	free(p);
	i = store_pubkey_rec(key_id, REPLACE, rec,
			     buffer_len + uid_len);
	if ( i != 0 ) {
	    return(-1);
	}
    }
    free(rec);

    /* Then, store the string in the string_db with the KeyID as data.
     * this ensures fast lookup.  We try even if FOUND==TRUE, since we
     * might update the database.
     */
    i = store_uid(user_id, key_id);
    if ( i == 1 && found != 1)
	return(i);
    return(0);
}

/* Add a trust packet to a public key.
 * The type of the trust packet depends on the
 * uid parameter: uid == NULL implies that we have
 * a key trust packet, while uid != NULL means
 * that we are talking about the trust in a 
 * specific key-user id binding.
 */

int
add_trust(PGPKEYID	key_id,
	  PGPuserid_t	*uid,
	  PGPtrust_t	*trust)
{
    u_char *buf;    
    u_char *p;
    u_char *tbuf;
    int p_len;
    int buffer_len;
    int i;
    int sofar;
    res_t what;
    int used;
    PGPuserid_t *tmpuid;
    int found;

    if( PGPKEYID_IsZero(key_id) || trust == 0 )
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

 retry:
    
    i = get_pubkey_rec(key_id, &buffer_len, &buf);
    if ( i != 0 ) {
	return(i);
    }

    sofar = 0;

    if(uid) {
	found = 0;
	while(sofar < buffer_len && !found)
	{
	    i = get_keyentry_from_buf(buf + sofar, buffer_len - sofar, &what,
				      NULL, NULL, &tmpuid, NULL, NULL, &used);
	    
	    if(i == -1)
		return -1;
	    else if(i)
		break;

	    sofar += used;

	    if(what == UID)
	    {
		if(!strcmp(tmpuid->name, uid->name))
		    found = 1;
		free_uid(tmpuid);
	    } 
	}

	if(!found) { 		 /* add this user-ID */
	    i = add_uid(key_id, uid);
	    free(buf);
	    if(i) return i;
	    goto retry;
	}

    } else {
	sofar = packet_length(buf);
    }
    
    if(sofar < buffer_len)
    {
	if(is_ctb(buf[sofar]) && is_ctb_type(buf[sofar], CTB_KEYCTRL_TYPE))
	{
	    p_len = packet_length(buf + sofar);
	    memmove(buf + sofar, buf + sofar + p_len,
		    buffer_len - p_len - sofar);
	    buffer_len -= p_len;
	}
    }

    trust2buf(trust, &tbuf);
    
    if((p = realloc(buf, buffer_len + (p_len = packet_length(tbuf)))) == NULL)
    {
	int rn = errno;
	free(buf);
	free(tbuf);
	errno = rn;
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return -1;
    }

    buf = p;

    memmove(buf + sofar + p_len, buf + sofar, buffer_len - sofar);
    memcpy (buf + sofar, tbuf, p_len);

    i = store_pubkey_rec(key_id, REPLACE, buf, buffer_len+p_len);
    free(buf);
    free(tbuf);

    return i;
}


/*
 * Adding a signature to a key, verifying that it isn't already
 * there.
 * If we added, return 0.  No adding returns 1, error gives -1.
 */

int
add_sig(PGPKEYID	key_id,
	PGPuserid_t 	*uid,
	PGPsig_t 	*sig)
{

    u_char 	*buf;
    int 	found;
    int		buffer_len, sofar, used;
    res_t	what;
    u_char	*sig_buf;
    int		sig_len;
    PGPsig_t	*tmpsig;
    PGPuserid_t *tmpuid;

    /* scratch */
    u_char	*p;
    int 	i, j;

    if(!sig) 
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    if((sig->classification >= K0_SIGNATURE_BYTE 
	&& sig->classification <= K3_SIGNATURE_BYTE)
       || sig->classification == KR_SIGNATURE_BYTE)
    {
	if(!uid)
	{
	    pgplib_errno = PGPLIB_ERROR_EINVAL;
	    return -1;
	}
    } 
    else 
	uid = NULL;

 retry:
    
    i = get_pubkey_rec(key_id, &buffer_len, &buf);
    if ( i != 0 ) {
	return(i);
    }

    sofar = 0;

    if(uid) {
	for(found = 0; sofar < buffer_len && !found; ) {
	    i = get_keyentry_from_buf(buf + sofar, buffer_len - sofar, 
				      &what, NULL, NULL, &tmpuid, 
				      NULL, NULL, &used);
	    
	    sofar += used;
	    if(i == -1)
		return -1;
	    else if(i)
		break;

	    
	    if(what == UID) {
		if(!strcmp(uid->name, tmpuid->name))
		    found = 1;
		free_uid(tmpuid);
	    }
	}

	if(!found) { 		 /* add this user-ID */
	    i = add_uid(key_id, uid);
	    free(buf);
	    if(i) return i;
	    goto retry;
	}
    }

    found = 0;

    while ( sofar < buffer_len && found == 0 ) {
	i = get_keyentry_from_buf( buf + sofar, buffer_len - sofar, &what, 
				   &tmpsig, NULL, NULL, NULL, NULL, &used);
	if ( i == 1 || i == -1 )
	    return(-1);

	if (what == UID)
	    break;
	else
	    sofar += used;
	
	if (what != SIG)
	    continue;

	/* We must verify that the signature at hand is different from
	 * the one already on the key.  We first check if it is made
	 * in the same date, then if it is made with the same key
	 * (long comparison is faster than quad).  If both matches,
	 * we dig out the actual signatures and compare them.
	 */
	
	/* assert(tmpsig != NULL);*/
	
	if (tmpsig->timestamp == sig->timestamp 
	    && PGPKEYID_EQUAL(tmpsig->key_id, sig->key_id)
	    && tmpsig->classification == sig->classification)
	{
	    j = BN_cmp(tmpsig->I, sig->I);
	    if ( j == 0 ) {
		/* A copy */
		found = 1;
	    }
	    
	}
	free_signature(tmpsig);
    }
    
    if ( found == 1 ) {
	/* It was already there */
	free(buf);
	return(1);
    }

    /* This signature wasn't on the key.  Let us add it. */
    i = signature2buf(sig, &sig_buf);
    if ( i != 0 ) {
	free(buf);
	return(-1);
    }

    if((sig_len = packet_length(sig_buf)) == -1) {
	free(buf);
	return(-1);
    }

    if((p = realloc(buf, buffer_len + sig_len)) == NULL)
    {
	int rn = errno;
	if(buf) free(buf);
	errno = rn;
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    buf = p;

    memmove(buf + sofar + sig_len, buf + sofar, buffer_len - sofar);
    memcpy(buf + sofar, sig_buf, sig_len);
    
    free(sig_buf);
    j = store_pubkey_rec(key_id, REPLACE, buf, buffer_len+sig_len);
    free(buf);
    return j;
}



/*
 * In order to fetch just a public key, we fetch everything
 * stored on this keyID,  Then copy out the key proper and
 * free'ing the rest.
 * Return -1 on error, 1 on no-such-key and 0 on OK.
 */

int
get_only_pubkey(PGPKEYID 	key_id,
		PGPpubkey_t 	**key)
{
    keyent_t	**recp;
    PGPpubkey_t	*tmpkey;
    int		index;
    /* scratch */
    int 	i;

    if ( key == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    i = get_keycontents(key_id, &recp);
    if ( i == -1 || i == 1 )
	return(i);

    tmpkey = calloc(1, sizeof(PGPpubkey_t));
    if ( tmpkey == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    for ( index = 0; recp[index] != NULL; index++) {
	switch (recp[index]->what) {
	case TRUSTBYTE:
	case SIG:
	case UID:
	    continue;
	case PUBKEY:
	    i = copy_pubkey(recp[index]->u.key, tmpkey);
	    if ( i == -1 ) {
		free_keyrecords(recp);
		return(-1);
	    }
	    continue;
	default:
	    assert(0);
	    /*NOTREACHED*/
	}
    }
    free_keyrecords(recp);
    *key = tmpkey;

    return(0);
}


/*
 * Get a public key, and parse out the different packages that
 * we find there.  Return a pointer to an array of pointers to
 * structures.  Everything is malloc'ed and must be freed by the
 * caller; call free_keyentry (see below).
 * A return of 1 means no such key, -1 error, 0=OK.
 */

int
keybuf_to_records(u_char *keybuf, int buf_len, keyent_t ***keyent)
{
    int		used;
    keyent_t	**arr;
    int		ent;
    int		offset;
    res_t	what;
    PGPsig_t	*sig;
    PGPpubkey_t	*key;
    PGPuserid_t	*uid;
    PGPtrust_t	*trust;
    int		ret;
    char	*p;


    if(!keybuf || !buf_len || !keyent)
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    
    /* Ensure our array is NULL terminated */
    arr = calloc(1, sizeof(keyent_t **));
    if ( arr == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }

    ent = 0;
    offset = 0;
    used = 0;

    while ( offset < buf_len ) {
	ret = get_keyentry_from_buf(keybuf+offset,
				    buf_len-offset,
				    &what, &sig, &key,
				    &uid, NULL, &trust, &used);
	if ( ret != 0 ) {
	    return(-1);
	}

	offset += used;

	p = realloc(arr, (ent+2) * sizeof(keyent_t *));
	if ( p == NULL ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    free(arr);
	    return(-1);
	}
	arr = (keyent_t **)p;
	arr[ent+1] = NULL;

	arr[ent] = malloc(sizeof(keyent_t));
	if ( arr[ent] == NULL ) {
	    pgplib_errno = PGPLIB_ERROR_SYSERR;
	    return(-1);
	}
	arr[ent]->what = what;

	switch (what) {
	default:
	case NONE:
	    assert(0);
	case PUBKEY:
	    arr[ent]->u.key = key;
	    break;
	case SIG:
	    arr[ent]->u.sig = sig;
	    break;
	case UID:
	    arr[ent]->u.uid = uid;
	    break;
	case TRUSTBYTE:
	    arr[ent]->u.trust = trust;
	    break;
	}
	ent += 1;
    }

    *keyent = arr;
    return(0);
}

int
get_keycontents_seq(keyent_t ***keyent, int where)
{
#if HAVE_DB_H
    DATA_TYPE key, data;
#endif
#if HAVE_GDBM_H
    static    DATA_TYPE key; /* must retain between invocations */
    DATA_TYPE data, nextkey;
#endif
    int rv;

#if HAVE_DB_H
    if((rv = keydb->seq(keydb, &key, &data, where)) != 0)
	return rv;

    return keybuf_to_records(data.data, data.size, keyent);
#endif
#if HAVE_GDBM_H
    if (where == 3 /* R_FIRST */)
    {    
	key = gdbm_firstkey(keydb);
    } 
    else if (where == 7 /* R_NEXT */)
    { 
	nextkey = gdbm_nextkey(keydb, key);
	free(DATA_PTR(key));
	key = nextkey;
    }
    else 
	return -1;
    if (DATA_PTR(key) == NULL)
	return -1;
    data = gdbm_fetch(keydb, key);
    if (DATA_PTR(data) == NULL)
    {    
	free(DATA_PTR(key));
	return -1;
    }                              
    
    rv = keybuf_to_records(DATA_PTR(data), DATA_SIZE(data), keyent);
    free(DATA_PTR(data));
    
    return rv;              
#endif

}
    

int
get_keycontents(PGPKEYID	key_id,
		keyent_t	***keyent)
{
    int		buf_len;
    u_char	*keybuf;
    int		i;

    if ( keyent == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    i = get_pubkey_rec(key_id, &buf_len, &keybuf);
    if ( i != 0 )
	return(i);

    i = keybuf_to_records(keybuf, buf_len, keyent);
    free(keybuf);
    return i;
}


/* we expect a public key record. */

int
add_pubkey_rec(u_char 	*buf, 
	       int 	size)
{	
    u_char 	digest[16];
    PGPpubkey_t pk;
    int 	new_key;
    DATA_TYPE 	key, data;
    PGPKEYID 	key_id;
    int 	i;
    int 	used;
    PGPsig_t 	*sig, *tmpsig = NULL;
    PGPpubkey_t *pubkey, *tmppubkey = NULL;
    PGPuserid_t *uid, *tmpuid = NULL;
    PGPtrust_t *trust;
    res_t 	what;


    if((i = buf2pubkey(buf, size, &pk)))
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return 1;
    }

    MD5(buf, size, digest);

    if ( PGPKEYID_FromBN(&key_id, pk.N) != 0) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
       	return 1;
    }

    free_pubkey(&pk);

    DATA_PTR(key) = (DATA_PTR_TYPE)&key_id;
    DATA_SIZE(key) = sizeof(key_id);

    /* anything to do? */

    if ((new_key = DB_FETCH(keyhashdb, key, data)) == 0)
    {
        if(DATA_SIZE(data) == sizeof(digest) &&
	   ! memcmp(digest, DATA_PTR(data), DATA_SIZE(data)))
	{ 
	    DATA_RELEASE(data);
	    return 0;
	}
	DATA_RELEASE(data);
    } 

    if(new_key)
    {
	if((i = store_pubkey_rec(key_id, REPLACE, buf, size)))
	    return i;
    }

    used = 0;
    i = 0;
    while(used <  size &&
	  get_keyentry_from_buf(buf + i, size - i,
				&what, &sig, &pubkey, &uid, NULL,
				&trust, &used) == 0)
    {
	i += used;
	switch(what)
	{
	case PUBKEY:
	    if(tmpsig) 
	    {
		free_signature(tmpsig);
		free(tmpsig);
		tmpsig = NULL;
	    }
	    
	    if(tmpuid)
	    {
		free_uid(tmpuid);
		free(tmpuid);
		tmpuid = NULL;
	    }
      
	    if(tmppubkey)
	    {
		free_pubkey(tmppubkey);
		free(tmppubkey);
	    }
	    
	    tmppubkey = pubkey;

	    break;

	case UID:
	    if(tmpsig)
	    {
		free_signature(tmpsig);
		free(tmpsig);
		tmpsig = NULL;
	    }
      
	    if(tmpuid)
	    {
		free_uid(tmpuid);
		free(tmpuid);
	    }
	    tmpuid = uid;
	    
	    if(tmppubkey && !new_key) 
		add_uid(key_id, tmpuid);
	    else if(tmppubkey)
		store_uid(tmpuid, key_id);

	    break;

	case SIG:
	    if(tmpsig)
	    {
		free_signature(tmpsig);
		free(tmpsig);
		tmpsig = NULL;
	    }
	    tmpsig = sig;
	    if(tmppubkey && !new_key)
		add_sig(key_id, tmpuid, tmpsig);
	    
	    break;

	case TRUSTBYTE:
	    if(!tmpsig && tmppubkey && tmpuid && !new_key)
		add_trust(key_id, tmpuid, trust);

	    break;
	    
	default:
	    /* shut up the compiler */
	    break;
	    
	}
    }

    if(tmpuid) {
	free_uid(tmpuid);
	free(tmpuid);
    }

    if(tmppubkey) {
	free_pubkey(tmppubkey);
	free(tmppubkey);
    }

    if(tmpsig) {
	free_signature(tmpsig);
	free(tmpsig);
    }

    return 0;
}

static u_char *enlarge(u_char *buff, int needed, int *now)
{
    u_char *scratch;

    if(needed <= *now)
	return buff;

    if((scratch = realloc(buff, needed)) == NULL)
    {
	free(buff);
	return NULL;
    }

    *now = needed;

    return scratch;
}


/* scan the public key ring, avoiding mallocs and reallocs wherever
 * possible. */

int 
scan_pubring(FILE *fp)
{
    u_char *packet = NULL;
    u_char *pkr = NULL;
    u_char last_ctb = 0;
    int p_len;
    int i;
    int pl = 0;
    int kl = 0;
    int ll;

    int pu = 0;
    int ku = 0;

    packet = malloc(1);
    pl = 1;

    while(1)
    {
	if((pu = fread(packet, 1, 1, fp)) < 1 || !is_ctb(packet[0]))
	    break;

	if((ll = ctb_llength(packet[0])) < 0)
	    break;

	if(!(packet = enlarge(packet, pu + ll, &pl)))
	    break;

	if((i = fread(packet + pu, 1, ll, fp)) < ll || 
	    (p_len = packet_length(packet)) < 0)
	    break;
	
	pu += i;

	if(!(packet = enlarge(packet, p_len, &pl)))
	    break;

	i = p_len - pu;

#ifdef WIN32
	if (fread(packet+pu, 1, i, fp) < (size_t)i)
#else  
	if (fread(packet+pu, 1, i, fp) < i)
#endif
	    break;

	pu += i;

	if(is_ctb_type(packet[0], CTB_KEYCTRL_TYPE))
	{
	    if(last_ctb && is_ctb_type(last_ctb, CTB_USERID_TYPE))
	    {
		if((pkr = enlarge(pkr, ku + p_len, &kl)) == NULL)
		    break;
	
		memcpy(pkr + ku, packet, p_len);
		ku += p_len;
	    }
	}
	else 
	{
	    if(pkr && ku && is_ctb_type(packet[0], CTB_CERT_PUBKEY_TYPE))
	    {
		add_pubkey_rec(pkr, ku);
		ku = 0;
	    } 
	    
	    if((pkr = enlarge(pkr, ku + p_len, &kl)) == NULL)
		break;
	    
	    memcpy(pkr + ku, packet, p_len);
	    ku += p_len;
	}
	last_ctb = packet[0];
    }

    if(pkr && ku)
	add_pubkey_rec(pkr, ku);

    if(pkr) free(pkr);
    if(packet) free(packet);

    if(!feof(fp))
	return -1;
    else
	return 0;
}
