/*********************************************************************
 * Filename:      keybase.c
 * Description:   PalmOS key management routines
 * Author:        Per Harald Myrvang <perm@pasta.cs.uit.no>
 * Modified at:   Sun Jun 07 20:39:27 1998
 *
 * 	Copyright (c) University of Troms, Norway
 * 	See the file COPYING for details
 ********************************************************************/

/* Trap compilation for non-PalmOS targets */
#ifndef __palmos__
# error "PalmOS specific source file (compiling for wrong target?)!"
#endif	/* __palmos__ */

/*
 * We store the keys as PGP packets in records, much like PGP does itself.
 * There's two databases, one for public keys, and one for secret keys. Record
 * #0 (which also has an unique ID of zero) in each database is the index
 * record, where a keyid is mapped to a record's unique identifier for that
 * database.  Each key record have a tiny header, immediately followed by a PGP
 * key package, with trust, uids, and signatures.  This allows for a compact
 * storage format.
 * --Perm
*/

/*
 * Oops, this is buggy. Or at least the usage of keyids as unique record IDs
 * are a bad idea, since a backup-restore cycle will really screw up this
 * mapping. We should, when we e.g. can't find a the index record (or perhaps
 * every time we open the database?), go through every record and make sure
 * that this mapping is correct, or at least sane.
 * --Perm
*/

#include <pgplib.h>
#include <md5.h>

#ifndef offsetof
# define offsetof(s,m) ((long)&(((s *)0))->m)
#endif /* offsetof */

/* Key record entry header */
typedef struct {
    unsigned short	flags;		/* Various flags and metainfo */
    PGPKEYID		keyid;		/* Key id */
    unsigned short	uid_num;	/* # of user ids in this record */
    unsigned short	sig_num;	/* # of signatures in this record */
    u_char		data[0];	/* rest of data */
} keyrecordhead_t;

#if 0
enum keyrecheadflags {
    KEY_PUBLIC, KEY_SECRET,
    KEY_MAX
};
#endif

/* Database Key index record (record #0) entry */
typedef struct {
    unsigned short	flags;
    unsigned long	short_keyid;	/* Short keyid */
    PGPKEYID		keyid;		/* Key id */
    unsigned short	size;		/* Key size (in bits) */
    unsigned long	unique_id;	/* Unique id of this key's record */
} index_entry_t;

/* Database Key index record (record #0) header entry */
typedef struct {
    unsigned short	flags;
    unsigned long	entries;	/* Number of stored entries */
    unsigned long	free;		/* Number of free entries */
    index_entry_t	data[0];	/* Entries begins here */
} index_head_t;

/* INDEX_PAGESIZE is the number of index entries to allocate room for when we
 * grow the index record. */
#define INDEX_PAGESIZE	4


static char default_pubring[] = "PGP PubRing DB";
static char default_secring[] = "PGP SecRing DB";

#define PGPKEYID_CMP2(X,Y) (0 == memcmp((X), (Y), sizeof(PGPKEYID)))
#define PGPKEYID_CPY2(DST,SRC) (memcpy((DST), (SRC), sizeof(PGPKEYID)))

/*
 * Opens a keyring database given the name, and the mode (se DmOpenDatabase).
 * Returns NULL on failure, or a DmOpenRef on success
 */
DmOpenRef
open_keyring(char *dbname,
	unsigned int mode)
{
    DmOpenRef	dref;
    LocalID	lid;

    dref = NULL;
    if (NULL != dbname) {
	lid = DmFindDatabase(0, dbname);
	if (0 != lid) {
	    dref = DmOpenDatabase(0, lid, mode);
	}
    }

    return dref;
}

/* Closes the given keyring database, after unlocking all records.
 * Returns 0 on success, -1 on failure. */
int
close_keyring(DmOpenRef dref)
{
    Err err;
    int ret = -1;;

    if (NULL != dref) {
	//DmResetRecordStates(dref);
	err = DmCloseDatabase(dref);
	if (0 == err)
	    ret = 0;
    }
    return ret;
}

/* Creates a new keyring with a given name and creator id, _including_ the
 * index record (record #0).
 * Returns 0 on success, -1 on failure
 */
int
create_keyring(char *dbname,
	unsigned long creator,
	unsigned short flags)
{
    DmOpenRef	dref;
    LocalID	lid;
    VoidHand	recH;
    VoidPtr	recP;
    UInt	pos;
    ULong	size, ul;
    Err		err;
    index_head_t indexhead;
    int		ret = -1;

    if (NULL != dbname) {
	/* See if database already exists */
	lid = DmFindDatabase(0, dbname);
	if (0 == lid) {
	    /* Try to create database */
	    err = DmCreateDatabase(0, dbname, creator, 'data', false);
	    if (0 == err) {
		/* Try to open the newly created database */
		dref = open_keyring(dbname, dmModeReadWrite);
		if (NULL != dref) {
		    /* Allocate index record */
		    pos = 0;
		    size = sizeof(index_head_t);
		    size += sizeof(index_entry_t) * INDEX_PAGESIZE;
		    recH = DmNewRecord(dref, &pos, size);
		    if (NULL != recH) {
			/* Initialize index record */
			recP = MemHandleLock(recH);
			DmSet(recP, 0, size, 0);
			indexhead.flags = flags;
			indexhead.free = INDEX_PAGESIZE;
			indexhead.entries = 0;
			ul = INDEX_PAGESIZE;
			/* Set recP->free to INDEX_PAGESIZE. */
			DmWrite(recP, 0, &indexhead, sizeof(indexhead));
			MemHandleUnlock(recH);
			ul = 0L;
			/* Set index record's unique id to zero */
			err = DmSetRecordInfo(dref, 0, NULL, &ul);
			if (0 == err)
			    ret = 0;
			DmReleaseRecord(dref, 0, true);
		    }
		    close_keyring(dref);
		}
	    }
	}
    }
    
   return ret; 
}

/* Adds a PGP key id to the index record, given the database, the keyid, 
 * the unique id of the record the key resides, and flags.
 * Returns 0 on success, -1 on failure
 */
static int
add_keyid2index(DmOpenRef dref,
	PGPKEYID *keyid,
	ULong uniqueid,
	unsigned short size,
	short flags)
{
    ULong	ul;
    UInt	index;
    VoidHand	indexH;
    index_head_t *indexP;
    index_entry_t entry;
    Err		err;
    int		i;

    if (NULL == dref)
	return -1;

    /* get index record */
    ul = 0;
    err =  DmFindRecordByID(dref, ul, &index);
    ErrFatalDisplayIf(0 != err, "Index record not found!");
    indexH = DmGetRecord(dref, index);
    ErrFatalDisplayIf(NULL == indexH, "Argh!");
    indexP = (index_head_t *)MemHandleLock(indexH);
    ErrFatalDisplayIf(NULL == indexP, "Eeek!");
    if (flags != indexP->flags) {
	/* We got a problem! The flags in the index record and the flags
	 * provided by the caller doesn't match! Thus we're adding a key to the
	 * wrong database. */
	MemHandleUnlock(indexH);
	DmReleaseRecord(dref, index, false);
	return -1;
    }

    /* Grow index record if needed */
    if (0 == indexP->free) {
	MemHandleUnlock(indexH);
	ul = MemHandleSize(indexH);
	ul += sizeof(index_entry_t) * INDEX_PAGESIZE;
	indexH = DmResizeRecord(dref, index, ul);
	ErrFatalDisplayIf(NULL == indexH, "Can't resize index!");
	indexP = (index_head_t *)MemHandleLock(indexH);
	ErrFatalDisplayIf(NULL == indexP, "Eeek!");
	/* Set recP->free to INDEX_PAGESIZE. */
	ul = INDEX_PAGESIZE;
	i = DmWrite(indexP, (ULong)offsetof(index_head_t, free),
		&ul, sizeof(ul));
	ErrFatalDisplayIf(0 != i, "DmWrite failed!");
    }

    /* Increase the number-of-entries-in-index-record field */
    ul = indexP->entries + 1;
    i = DmWrite(indexP, (ULong)offsetof(index_head_t, entries),
	    &ul, sizeof(ul));
    ErrFatalDisplayIf(0 != i, "DmWrite failed!");

    /* Decrease the number-of-free-entries-in-index-record field */
    ul = indexP->free - 1;
    i = DmWrite(indexP, (ULong)offsetof(index_head_t, free),
	    &ul, sizeof(ul));
    ErrFatalDisplayIf(0 != i, "DmWrite failed!");

    /* Add keyid-to-unique-id mapping to index */
    entry.flags = flags;
    PGPKEYID_ExtractShort(keyid, &entry.short_keyid);
    PGPKEYID_CPY2(&entry.keyid, keyid);
    entry.unique_id = uniqueid;
    entry.size = size;
    /* Offset into record */
    ul = (void *)(&indexP->data[indexP->entries - 1]) - (void *)indexP;
    i = DmWrite(indexP, ul, &entry, sizeof(entry));
    ErrFatalDisplayIf(0 != i, "DmWrite failed!");
    
    DmReleaseRecord(dref, index, true);
    return 0;
}

/* Make a lookup in a database's index record, looking for a keyid, and
 * returning the unique id of the record in this database inwhich the key with
 * the id resides, and the short id of the key. Both the pointer to the returned
 * unique id and the short id may be NULL. If <shortid> is non-null, search for
 * the short keyid in <shortkeyid> instead of <keyid>, returning the long keyid
 * in <keyid>, if <keyid> isn't NULL. If <size> isn't NULL, the number of bits
 * in the key's N is returned there.
 * Returns 0 on success (i.e. we found the keyid), 1 if not found, and -1 on
 * failure.
 */
int
lookup_keyid(DmOpenRef dref,
	PGPKEYID *keyid,
	unsigned long *uniqueid,
	unsigned long *shortkeyid,
	short shortid,
	unsigned short *size)
{
    ULong	ul;
    UInt	index;
    VoidHand	indexH;
    index_head_t *indexP;
    Err		err;
    int		found, ret = -1;	/* Assume failure */

    ul = 0;
    if (shortid) {
	if (NULL == shortkeyid)
	    return -1;
    }
    err =  DmFindRecordByID(dref, ul, &index);
    ErrFatalDisplayIf(0 != err, "Index record not found!");
    if (0 == err) {
	indexH = DmQueryRecord(dref, index);
	if (NULL != indexH) {
	    indexP = (index_head_t *)MemHandleLock(indexH);
	    if (NULL != indexP) {
		ret = 1;	/* Assume not found */
		found = 0;
		for (ul = 0; ul < indexP->entries; ul++) {
		    if (shortid) {
			if (*shortkeyid == indexP->data[ul].short_keyid)
			    found = 1;
		    } else {
		    	if (PGPKEYID_CMP2(keyid, &indexP->data[ul].keyid))
			    found = 1;
		    }

		    if (found) {
			if (NULL != size)
			    *size = indexP->data[ul].size;
			if (NULL != uniqueid)
			    *uniqueid = indexP->data[ul].unique_id;
			if (NULL != shortkeyid && !shortid)
			    *shortkeyid = indexP->data[ul].short_keyid;
			if (shortid)
			    if (NULL != keyid)
				PGPKEYID_CPY2(keyid, &indexP->data[ul].keyid);
			ret = 0;
			break;
		    }
		}
		MemHandleUnlock(indexH);
	    }
	}
    }
    return ret;
}

/* Do a DmWrite on a given record (saves us the hassel of finding and locking
 * the record). If <keyid> is given, it has priority over <uniqueid>. Returns 0
 * on success, -1 on some error/failure */
static int
write_into_record(DmOpenRef dref,
	PGPKEYID *keyid,
	unsigned long uniqueid,
	long offset,
	void *src,
	long len)
{
    ULong	ul;
    UInt	index;
    VoidHand	indexH;
    index_head_t *indexP;
    Err		err;
    int		i, ret = -1;	/* Assume failure */

    if (NULL == dref)
	return -1;

    ul = uniqueid;

    if (NULL != keyid) {
	i = lookup_keyid(dref, keyid, &ul, NULL, 0, NULL);
	if (0 != i)
	    return -1;
    }

    /* get index record */
    err =  DmFindRecordByID(dref, ul, &index);
    if (0 == err) {
	indexH = DmGetRecord(dref, index);
	if (NULL != indexH) {
	    indexP = (index_head_t *)MemHandleLock(indexH);
	    if (NULL != indexP) {
		err = DmWrite(indexP, offset, src, len);
		ErrFatalDisplayIf(0 != err, "DmWrite failed!");
		if (0 == err)
		    ret = 0;
		MemHandleUnlock(indexH);
	    }
	    DmReleaseRecord(dref, index, true);
	}
    }
    return ret;
}

/* Read a bunch of bytes from a given record (saves us the hassel of finding
 * and locking the record). If <keyid> is given, it has priority over
 * <uniqueid>. Allows you to shoot yourself in the foot quite severly. Returns
 * 0 on success, -1 on some error/failure */
static int
read_from_record(DmOpenRef dref,
	PGPKEYID *keyid,
	unsigned long uniqueid,
	long offset,
	void *dest,
	long len)
{
    ULong	ul;
    UInt	index;
    VoidHand	indexH;
    VoidPtr	indexP;
    Err		err;
    int		i, ret = -1;	/* Assume failure */

    if (NULL == dref)
	return -1;

    ul = uniqueid;

    if (NULL != keyid) {
	i = lookup_keyid(dref, keyid, &ul, NULL, 0, NULL);
	if (0 != i)
	    return -1;
    }

    /* get index record */
    err =  DmFindRecordByID(dref, ul, &index);
    if (0 == err) {
	indexH = DmQueryRecord(dref, index);
	if (NULL != indexH) {
	    indexP = (index_head_t *)MemHandleLock(indexH);
	    if (NULL != indexP) {
		err = MemMove(dest, (indexP+offset), len);
		ret = 0;
		MemHandleUnlock(indexH);
	    }
	}
    }
    return ret;
}


/* Creates a new key record from the key data (in PGP buffer format) and the
 * key id. Note that any userids, trust and signatures must be added
 * afterwards, with the appropriate functions.
 * Returns 0 on success, -1 on failure.
 */
static int
new_keyentry(DmOpenRef dref,
	u_char *keybuf,
	int keybuflen,
	PGPKEYID *keyid,
	unsigned short keysize,
	int flags)
{
    ULong	ul;
    UInt	index;
    VoidHand	newH;
    keyrecordhead_t *newP;
    keyrecordhead_t keyhead;
    Err		err;
    int		size;
    int		i, ret = -1;
    unsigned short ush;

    if (NULL == dref || NULL == keybuf || 0 == keybuflen) {
	return -1;
    }

    /* Read flags from index record */
    read_from_record(dref, NULL, 0, offsetof(index_head_t, flags),
	    &ush, sizeof(ush));

    /* If <flags> don't match with what's stored in the index record, we're
     * probably trying to store a key in the wrong database. This is fatal! */
    if (flags != ush) {
	ErrDisplay("Attempting to store key in wrong database!!");
	return -1;
    }

    keyhead.flags = flags;
    PGPKEYID_CPY2(&keyhead.keyid, keyid);
    keyhead.uid_num = 0;
    keyhead.sig_num = 0;

    /* Allocate record */
    index = -1;
    size = sizeof(keyrecordhead_t) + keybuflen;
    newH = DmNewRecord(dref, &index, size);
    if (NULL != newH) {
	/* fill in data */
	newP = (keyrecordhead_t *)MemHandleLock(newH);
	if (NULL != newP) {
	    err = DmWrite(newP, 0, &keyhead, sizeof(keyhead));
	    ErrFatalDisplayIf(0 != err, "DmWrite failed!");
	    if (0 == err) {
		err = DmWrite(newP, sizeof(keyhead), keybuf, keybuflen);
		ErrFatalDisplayIf(0 != err, "DmWrite failed!");
		if (0 == err) {
		    /* Get unique id of record, and to index */
		    err = DmRecordInfo(dref, index, NULL, &ul, NULL);
		    if (0 == err)  {
			i = add_keyid2index(dref, keyid, ul, keysize, flags);
			if (0 == i)
			    ret = 0;
		    }
		}
	    }
	    MemHandleUnlock(newH);
	}
	DmReleaseRecord(dref, index, true);
    }
    
    return ret;
}

/* Appends some data to a key.
 * Return 0 on success, -1 on failure
 */
static int
append2key(DmOpenRef dref,
	PGPKEYID *keyid,
	u_char *append_buf,
	int buflen)
{
    UInt	index;
    ULong	ol, ul;
    VoidHand	recH;
    VoidPtr	recP;
    Err		err;
    int		i, ret = -1;

    if (NULL == dref || NULL == keyid || NULL == append_buf || buflen < 1)
	return -1;

    i = lookup_keyid(dref, keyid, &ul, NULL, 0, NULL);
    if (0 == i) {
	err =  DmFindRecordByID(dref, ul, &index);
	recH = DmGetRecord(dref, index);
	if (NULL != recH) {
	    ol = MemHandleSize(recH);
	    ul = ol + buflen;
	    recH = DmResizeRecord(dref, index, ul);
	    if (NULL != recH) {
		recP = MemHandleLock(recH);
		if (NULL != recP) {
		    err = DmWrite(recP, ol, append_buf, buflen);
		    ErrFatalDisplayIf(0 != err, "DmWrite failed!");
		    if (0 == err)
			ret = 0;
		    MemHandleUnlock(recH);
		}
	    }
	    DmReleaseRecord(dref, index, true);
	}
    }
    return ret;
}


/*
 *
 * Higher level routines
 *
 */

/* Open a public key database, or the default public key database. Will create
 * the database if it doesn't already exists.
 * Returns 0 on success, non-zero on failure
 */
DmOpenRef
open_pubkey_db(char *dbname)
{
    DmOpenRef dbR;
    char *name;

    name = (NULL == dbname) ? default_pubring : dbname;
    dbR = open_keyring(name, dmModeReadWrite);
    if (NULL == dbR) {
	if (0 == create_keyring(name, 'PGPL', KEY_PUBLIC)) {
	    dbR = open_keyring(name, dmModeReadWrite);
	}
    }
    return dbR;
}

/* Open a secret key database, or the default secret key database. Will create
 * the database if it doesn't already exists.
 * Returns 0 on success, non-zero on failure
 */
DmOpenRef
open_seckey_db(char *dbname)
{
    DmOpenRef dbR;
    char *name;

    name = (NULL == dbname) ? default_secring : dbname;
    dbR = open_keyring(name, dmModeReadWrite);
    if (NULL == dbR) {
	if (0 == create_keyring(name, 'PGPL', KEY_SECRET)) {
	    dbR = open_keyring(name, dmModeReadWrite);
	}
    }
    return dbR;
}

/* Closes a database.
 * Returns 0 on success, non-zero on failure
 */
int
close_db(DmOpenRef dbR)
{
    return close_keyring(dbR);
}

/* Adds a public key with a trust parameter to a database. If the trust
 * parameter is NULL, then an unknown trust packet will be added instead.
 * Returns 0 on success, 1 if key already exists, -1 on failure
 */
int
add_pubkey(DmOpenRef dbR,
	PGPpubkey_t *key,
	PGPtrust_t *trust)
{
    PGPKEYID keyid;
    u_char	*key_buf;
    PGPtrust_t	*trst, mytrust = { 0 };
    u_short	size;
    int	key_size;
    int i;
    int ret = -1;

    if (NULL == dbR || NULL == key)
	return -1;
    PGPKEYID_FromBN(&keyid, key->N);			// Extract key id
    size = BN_num_bits(key->N);				// Get key size 
    if (0 == lookup_keyid(dbR, &keyid, NULL, NULL, 0, NULL)) 	// try to look it up
	return 1;
    // otherwise convert it to a buffer and add it to the database.
    i = pubkey2buf(key,  &key_buf);
    if (0 == i) {
	key_size = packet_length(key_buf);
	if (-1 != key_size) {
	    i = new_keyentry(dbR, key_buf, key_size, &keyid, size, KEY_PUBLIC);
	    free(key_buf);
	    key_buf = NULL;
	    if (0 == i) {
		trst = (NULL == trust) ? &mytrust : trust;
		i = trust2buf(trst, &key_buf);
		if (0 == i) {
		    key_size = packet_length(key_buf);
		    if (-1 != key_size) {
			// Now append trust to record
			i = append2key(dbR, &keyid, key_buf, key_size);
			if (0 == i)
			    ret = 0;
		    }
		}
	    }
	}
	if (NULL != key_buf)
	    free(key_buf);
    }
    return ret;
}

/* Adds a secret key with a trust parameter to a database. If the trust
 * parameter is NULL, then an unknown trust packet will be added instead.
 * Returns 0 on success, non-zero on failure
 */
int
add_seckey(DmOpenRef dbR,
	PGPseckey_t *key,
	PGPtrust_t *trust)
{
    PGPKEYID keyid;
    u_char	*key_buf;
    PGPtrust_t	*trst, mytrust = { 0 };
    int	key_size;
    u_short size;
    int i;
    int ret = -1;

    if (NULL == dbR || NULL == key)
	return -1;
    PGPKEYID_FromBN(&keyid, key->N);			// Extract key id
    size = BN_num_bits(key->N);				// Get key size 
    if (0 == lookup_keyid(dbR, &keyid, NULL, NULL, 0, NULL)) 	// try to look it up
	return 1;
    // otherwise convert it to a buffer and add it to the database.
    i = seckey2buf(key,  &key_buf);
    if (0 == i) {
	key_size = packet_length(key_buf);
	if (-1 != key_size) {
	    i = new_keyentry(dbR, key_buf, key_size, &keyid, size, KEY_SECRET);
	    free(key_buf);
	    key_buf = NULL;
	    if (0 == i) {
		trst = (NULL == trust) ? &mytrust : trust;
		i = trust2buf(trst, &key_buf);
		if (0 == i) {
		    key_size = packet_length(key_buf);
		    if (-1 != key_size) {
			// Now append trust to record
			i = append2key(dbR, &keyid, key_buf, key_size);
			if (0 == i)
			    ret = 0;
		    }
		}
	    }
	}
	if (NULL == key_buf)
	    free(key_buf);
    }
    return ret;
}

/* Removes a key from a database.
 * Returns 0 on success, non-zero on failure
 */
int
remove_key(DmOpenRef dbR, PGPKEYID *key)
{
    int ret = -1;
    ErrDisplay("remove_key() not intended for use (yet)!");
    return ret;
}


/* Find a key in a database, then convert it to a public or a secret key
 * structure, depending on the provided pointer, filling in the caller-provided
 * (i.e. allocated) structure pointed to.
 * Return 0 on success, or -1 on failure */
static int
get_key(DmOpenRef dbR, PGPKEYID *keyid,
	PGPpubkey_t *pubkey, PGPseckey_t *seckey)
{
    ULong	ul;
    Long	l;
    UInt	index;
    VoidHand	keyH;
    keyrecordhead_t *keyP;
    Err		err;
    int		i, ret = -1;
    
    if (NULL == keyid || (NULL == pubkey && NULL == seckey))
	return -1;
    
    if (0 != lookup_keyid(dbR, keyid, &ul, NULL, 0, NULL))
	return -1;

    err =  DmFindRecordByID(dbR, ul, &index);
    if (0 == err) {
	keyH = DmQueryRecord(dbR, index);
	if (NULL != keyH) {
	    l = MemHandleSize(keyH) - sizeof(keyrecordhead_t);	/* buf size */
	    keyP = (keyrecordhead_t *)MemHandleLock(keyH);
	    if (NULL != keyP) {
		if (KEY_PUBLIC == keyP->flags) {
		    if (NULL != pubkey) {
			i = buf2pubkey(keyP->data, l, pubkey);
			if (0 == i)
			    ret = 0;
		    }
		}
		else if (KEY_SECRET == keyP->flags) {
		    if (NULL != seckey) {
			i = buf2seckey(keyP->data, l, seckey);
			if (0 == i)
			    ret = 0;
		    }
		}
		else
		    ErrDisplay("Unknown record type!");
		
		MemHandleUnlock(keyH);
	    }
	}
    }

    // lookup keyid, call buf2pubkey/buf2seckey on the found record, depending
    // which one of pubkey/seckey were ==NULL.

    return ret;
}

/* Syntactic sugar for get_key(). <key> must be an already allocated
 * PGPpubkey_t structure.
 * Returns 0 on success, -1 on failure.
 */
int
get_pubkey(DmOpenRef dbR, PGPKEYID *keyid, PGPpubkey_t *key)
{
    return get_key(dbR, keyid, key, NULL);
}

/* Syntactic sugar for get_key(). <key> must be an already allocated
 * PGPseckey_t structure.
 * Returns 0 on success, -1 on failure.
 */
int
get_seckey(DmOpenRef dbR, PGPKEYID *keyid, PGPseckey_t *key)
{
    return get_key(dbR, keyid, NULL, key);
}

/* Adds a user id with a trust parameter to a key. If the trust parameter is
 * NULL, an unknown trust packet will be added instead. 
 * Returns 0 on success, non-zero on failure
 */
int
add_uid2key(DmOpenRef dbR,
	PGPKEYID *keyid,
	PGPuserid_t *userid,
	PGPtrust_t *trust)
{
    ULong	ul;
    u_char	*buf;
    PGPtrust_t	*trst, mytrust = { 0 };
    short	sh;
    int	size;
    int i;
    int ret = -1;

    if (NULL == dbR || NULL == keyid || NULL == userid)
	return -1;
    /* See if key exists, fail if it doesn't */ 
    if (0 != lookup_keyid(dbR, keyid, &ul, NULL, 0, NULL)) 
	return 1;

    // convert it to a buffer and add it to the record
    i = uid2buf(userid, &buf);
    if (0 == i) {
	size = packet_length(buf);
	if (-1 != size) {
	    i = append2key(dbR, keyid, buf, size);
	    free(buf);
	    buf = NULL;
	    if (0 == i) {
		read_from_record(dbR, keyid, 0,
			offsetof(keyrecordhead_t, uid_num),
			&sh, sizeof(sh));
		sh++;
		write_into_record(dbR, keyid, 0,
			offsetof(keyrecordhead_t, uid_num),
			&sh, sizeof(sh));
		trst = (NULL == trust) ? &mytrust : trust;
		i = trust2buf(trst, &buf);
		if (0 == i) {
		    size = packet_length(buf);
		    if (-1 != size) {
			// Now append trust to record
			i = append2key(dbR, keyid, buf, size);
			if (0 == i)
			    ret = 0;
		    }
		}
	    }
	}
	if (NULL != buf)
	    free(buf);
    }

    return ret;
}

/* Adds a signature with a trust parameter to a key. If the trust parameter is
 * NULL, an unknown trust packet will be added instead. 
 * Returns 0 on success, non-zero on failure
 */
int
add_sig2key(DmOpenRef dbR,
	PGPKEYID *keyid,
	PGPsig_t *sig,
	PGPtrust_t *trust)
{
    ULong	ul;
    u_char	*buf;
    PGPtrust_t	*trst, mytrust = { 0 };
    short sh;
    int	size;
    int i;
    int ret = -1;

    if (NULL == dbR || NULL == keyid || NULL == sig)
	return -1;
    /* See if key exists, fail if it doesn't */ 
    if (0 != lookup_keyid(dbR, keyid, &ul, NULL, 0, NULL))
	return 1;

    // convert it to a buffer and add it to the record
    i = signature2buf(sig, &buf);
    if (0 == i) {
	size = packet_length(buf);
	if (-1 != size) {
	    i = append2key(dbR, keyid, buf, size);
	    free(buf);
	    buf = NULL;
	    if (0 == i) {
		read_from_record(dbR, keyid, 0,
			offsetof(keyrecordhead_t, sig_num),
			&sh, sizeof(sh));
		sh++;
		write_into_record(dbR, keyid, 0,
			offsetof(keyrecordhead_t, sig_num),
			&sh, sizeof(sh));
		trst = (NULL == trust) ? &mytrust : trust;
		i = trust2buf(trst, &buf);
		if (0 == i) {
		    size = packet_length(buf);
		    if (-1 != size) {
			// Now append trust to record
			i = append2key(dbR, keyid, buf, size);
			if (0 == i)
			    ret = 0;
		    }
		}
	    }
	}
	if (NULL != buf)
	    free(buf);
    }
    return ret;
}

/* Generate a list of keysids in an array of keyids.
 * Call free()/MemPtrFree on the list returned.
 * Returns 0 on success, non-zero on failure
 */
int
get_keyidlist(DmOpenRef dbR, PGPKEYID **keyidlist, int *entries)
{
    ULong	ul;
    Long	l;
    UInt	index;
    VoidHand	indexH;
    index_head_t *indexP;
    PGPKEYID	*list;
    Err		err;
    int		i, ret = -1;	/* Assume failure */

    if (NULL == dbR || NULL == keyidlist || NULL == entries)
	return -1;
    ul = 0;
    *entries = 0;
    *keyidlist = NULL;
    err =  DmFindRecordByID(dbR, ul, &index);
    ErrFatalDisplayIf(0 != err, "Index record not found!");
    if (0 != err) {
	indexH = DmQueryRecord(dbR, index);
	if (NULL != indexH) {
	    indexP = (index_head_t *)MemHandleLock(indexH);
	    if (NULL != indexP) {
		if (0 < indexP->entries) {
		    l = indexP->entries * sizeof(PGPKEYID);
		    list = (PGPKEYID *)MemPtrNew(l);
		    if (NULL != list) {
			for (i = 0; i < indexP->entries; i++) {
			    PGPKEYID_CPY2(&list[i], &indexP->data[i].keyid);
			}
			*entries = indexP->entries;
			*keyidlist = list;
			ret = 0;
		    }
		} else 
		    ret = 1;
		MemHandleUnlock(indexH);
	    }
	}
    }
    return ret;
}

/* Generate a list of short keysids in an array of keyids.
 * Call free()/MemPtrFree on the list returned.
 * Returns 0 on success, 1 if list would be empty, non-zero on failure
 */
int
get_shortkeyidlist(DmOpenRef dbR, unsigned long **keyidlist, int *entries)
{
    ULong	ul;
    Long	l;
    UInt	index;
    VoidHand	indexH;
    index_head_t *indexP;
    unsigned long	*list;
    Err		err;
    int		i, ret = -1;	/* Assume failure */

    if (NULL == dbR || NULL == keyidlist || NULL == entries)
	return -1;
    ul = 0;
    *entries = 0;
    *keyidlist = NULL;
    err =  DmFindRecordByID(dbR, ul, &index);
    ErrFatalDisplayIf(0 != err, "Index record not found!");
    if (0 == err) {
	indexH = DmQueryRecord(dbR, index);
	if (NULL != indexH) {
	    indexP = (index_head_t *)MemHandleLock(indexH);
	    if (NULL != indexP) {
		if (0 < indexP->entries) {
		    l = indexP->entries * sizeof(unsigned long);
		    list = (unsigned long *)MemPtrNew(l);
		    if (NULL != list) {
			for (i = 0; i < indexP->entries; i++) {
			    list[i] = indexP->data[i].short_keyid;
			}
			*entries = indexP->entries;
			*keyidlist = list;
			ret = 0;
		    }
		} else 
		    ret = 1;
		MemHandleUnlock(indexH);
	    }
	}
    }
    return ret;
}

/* Make a list of the user ids associated with a key.
 * Call free_uid_list() on the returned list.
 * Returns 0 on success, 1 if key wasn't found, 1 if only a partial list could
 * be created, or -1 on failure.
 */
int
list_uids(DmOpenRef dbR, PGPKEYID *keyid, char ***uidlist, int *entries)
{
    ULong	ul;
    Long	l;
    UInt	index;
    VoidHand	keyH;
    keyrecordhead_t *keyP;
    char 	**list;
    u_char	*buf;
    int		buflen;
    Err		err;
    res_t	what;
    PGPuserid_t	*uid;
    int	used, tot_used;
    int	i, j, ret = -1;	/* Assume failure */

    if (NULL == dbR || NULL == uidlist || NULL == entries)
	return -1;
    if (0 != lookup_keyid(dbR, keyid, &ul, NULL, 0, NULL))
	return -1;

    *entries = 0;
    *uidlist = NULL;
    list = NULL;

    err =  DmFindRecordByID(dbR, ul, &index);
    if (0 == err) {
	keyH = DmQueryRecord(dbR, index);
	if (NULL != keyH) {
	    keyP = (keyrecordhead_t *)MemHandleLock(keyH);
	    if (NULL != keyP) {
		list = (char **)calloc((keyP->uid_num + 1),
			sizeof(char *));
		if (NULL != list) {
		    l = MemHandleSize(keyH);
		    l -= sizeof(keyrecordhead_t);	// size of data chunk
		    used = 0;
		    buflen = l;
		    buf = &keyP->data[0];
		    i = 0;

		    ret = 0;	/* Assume success */

		    while(0 < buflen) {
			j = get_keyentry_from_buf(buf, buflen, &what,
				NULL, NULL, &uid, NULL, NULL, &used);

			if (-1 == j) {		/* Failure */
			    ret = -1;
			    break;
			}
			if (1 == j) {		/* Buffer exhaused */
			    ret = 1;
			    break;
			}

			buflen -= used;
			buf += used;
			tot_used += used;

			if (UID != what)
			    continue;

			/* Copy string */
			if (NULL != uid) {
			    list[i] = (char *)calloc(1, StrLen(uid->name) + 1);
			    if (NULL == list[i]) {
				free_uid(uid);
				ret = 1;
				break;
			    }
			    StrCopy(list[i], uid->name);
			    free_uid(uid);
			    i++;
			} else {
			    ret = 1;
			    break;
			}

		    }
		    *entries = i;
		    *uidlist = list;
		}
		MemHandleUnlock(keyH);
	    }
	}
    }

    return ret;
}

/* Free list returned by list_uids(). Returns 0 on success, -1 on failure. */
int
free_uid_list(char **list)
{
    int i, ret = -1;
    if (NULL != list) {
	for (i = 0; NULL != list[i]; i++) {
	    MemSet(list[i], StrLen(list[i]), 0);	/* destroy uids */
	    MemPtrFree(list[i]);
	}
	MemPtrFree(list);
    }
    return ret;
}


/* TODO: Below here */

/* Find a key whose userid[s] match a given search parameter <search>, starting
 * the search at the record with index <startindex>, and, if found, returns the
 * keyid in <keyid>, and the index of the found record in <startindex> (handy
 * for restarts/search again.
 * Returns 0 on success (i.e. search hit), 1 if not found, -1 on failure 
 */
int
search4uid(DmOpenRef dbR, char *search, PGPKEYID *keyid, long *startindex)
{
    int ret = -1;

    return ret;
}

/* Generate a list of signatures */
int
list_signatures(DmOpenRef dbR, PGPKEYID *keyid)
{
    int ret = -1;

    return ret;
}

int
free_signature_list(void)
{
    return -1;
}
