/*********************************************************************
 * Filename:      keymgmnt.c
 * Description:
 * Author:        Tage Stabell-Kulo <tage@acm.org>
 * Modified at:   Mon Apr 20 09:24:41 1998
 * Modified by:   Tage Stabell-Kulo <tage@acm.org>
 *
 * 	Copyright (c) University of Troms, Norway
 * 	See the file COPYING for details
 *
 * $Id: keymgmnt.c,v 1.19 1999/03/08 19:58:46 perm Exp $
 *
 ********************************************************************/
/* All PalmOS code by Per Harald Myrvang <perm@pasta.cs.uit.no> */


#include <pgplib.h>

#include <md5.h>

/* The {rsa,bn,rand}.h are required for key generation */
#include <rsa.h>
#include <bn.h>
#include <rand.h>

#ifndef __palmos__
# include <assert.h>
# include <string.h>
# include <stdlib.h>
# ifdef WIN32
#  include <winsock2.h>
# else  
#  include <netinet/in.h> 
# endif
#else	/* !__palmos__ */
#include <System/Unix/netinet_in.h>
#include <System/SysEvtMgr.h>
#endif	/* __palmos__ */

/* SSLeay */
#include <idea.h>

#ifdef __palmos__
/* The pilot's stupid sprintf cannot comprehend formats such as %02X */
const static char hexed[] = "0123456789ABCDEF";
#endif /* __palmos__ */

/*
 * OK : return 0;
 * error: return -1
 */
int
copy_pubkey(PGPpubkey_t *in, PGPpubkey_t *out)
{

    if ( in == NULL || out == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    /* We copy everyting, then make new BIGNUMs */
    memcpy(out, in, sizeof(PGPpubkey_t));
    out->N = BN_dup(in->N);
    if ( out->N == NULL ) {
	out->E = NULL;
	goto bad;
    }
    out->E = BN_dup(in->E);
    if (out->E == NULL )
	goto bad;

    return(0);

bad:
    BN_free(out->N);
    BN_free(out->E);
    memset(out, 0, sizeof(*out));
    pgplib_errno = PGPLIB_ERROR_SYSERR;
    return(-1);
}

void
free_pubkey(PGPpubkey_t	*key)
{
    if ( key == NULL )
	return;
    if ( key->N != NULL )
	BN_free(key->N);
    if ( key->E != NULL )
	BN_free(key->E);
    memset(key, 0, sizeof(*key));

    return;
}


void
free_keyrecords(keyent_t **records)
{
    int		index;

    if ( records == NULL || records[0] == NULL )
	return;
    for(index=0; records[index] != NULL; index += 1) {
	switch (records[index]->what) {
	case PUBKEY :
	    free_pubkey(records[index]->u.key);
	    free(records[index]->u.key);
	    break;
	case SIG:
	    free_signature(records[index]->u.sig);
	    free(records[index]->u.sig);
	    break;
	case UID:
	    free_uid(records[index]->u.uid);
	    free(records[index]->u.uid);
	    break;
	case TRUSTBYTE:
	    free(records[index]->u.trust);
	    break;
	default:
#ifdef __palmos__
	    /* Helpful error message of the year, eh? --Perm  */
	    ErrFatalDisplayIf(1, "bad what!");
#else	/* __palmos__ */
	    assert(0);
#endif	/* __palmos__ */
	    /*NOTREACHED*/
	}
    }
    free(records);

    return;
}

/*
 * Calculate the fingerprint of a public key and write the
 * resulting _string_ into the memory pointed to by str.
 * The string occupies 50 chars, including the terminating 0.
 * Return 0 on success, -1 on error.
 */
 /* The string occupies much more, namely 58 bytes, including the terminating
  * 0. --Perm */
int
pub_fingerprint(PGPpubkey_t	*key,
		char		str[50])
{
    MD5_CTX	context;
    u_char	digest[16];
    int		len;
    u_char	*p;
    char	*s;
    /* scratch */
    int		i;

    if ( key == NULL || key->N == NULL || key->E == NULL || str == NULL ){
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    MD5_Init(&context);

    len = BN_num_bytes(key->N);
    p = malloc(len);
    if ( p == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }
    BN_bn2bin(key->N, p);

    MD5_Update(&context, p, len);
    free(p);

    len = BN_num_bytes(key->E);
    p = malloc(len);
    if ( p == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }
    BN_bn2bin(key->E, p);

    MD5_Update(&context, p, len);
    free(p);

    MD5_Final(digest, &context);

#if 0
    for(i=0; i < 8; i++, str++) {
	sprintf(str, "%02X ", digest[i]);
    str += 1;
    *str = ' ';
    for(i=8; i < 16; i++, str++)
	sprintf(str, "%02X", digest[i]);
    str += 1;
    *str = 0;
#else
    s = str;
    for(i = 0; i < 16; i++) {	/* 16 was 8 */
#ifdef __palmos__
	/* The pilot's stupid sprintf cannot comprehend formats such as %02X */
	*s++ = hexed[((digest[i] >> 4) & 0xf)];
	*s++ = hexed[(digest[i] & 0xf)];;
	*s++ = ' ';
	if (7 == i)
	    *s++ = ' ';
	/*StrPrintF(str, "%c%c ", hexed[((digest[i] >> 4) & 0xf)],
		hexed[(digest[i] & 0xf)]);*/
#else	/* __palmos__ */
	sprintf(str, "%02X ", digest[i]);
#endif	/* __palmos__ */
    }
    *s = '\0';
#endif /* 0 */

    return(0);
}

/* print_keyrec() deals with FILEs, which doesn't exist under PalmOS (this is,
 * as the observant source code reader now will know, a lie) --Perm */
#ifndef __palmos__
/*
 * This routine should not be present in a library such as this, since
 * the library has been written in ordert to make it possible to write
 * routines such as this one.
 * Regard it as left-overs from the developing phase.  At best.
 */

int
print_keyrec(u_char	*buffer,
	     int 	buf_len,
	     FILE 	*fd)
{
    u_char	*packet;
    int		p_len;
    int		how_much;
    PGPpubkey_t	*key;
    PGPsig_t	*sig;
    PGPuserid_t	*uid;
    int		have_key;
    u_char	CTB;
    /* scratch */
    int		i;
    PGPKEYID	key_id;
    
    packet = NULL;
    how_much = 0;

    while ( how_much < buf_len ) {

	free(packet);
	packet = NULL;

	i = find_packet_in_buffer(buffer+how_much, buf_len-how_much,
				  &packet);
	p_len = packet_length(packet);
	switch( i ) {
	case -1:
	    free(packet);
	    fprintf(stderr, "Some error in print_keyrecords\n");
	    return(-1);
	case 1:
	    free(packet);
	    if ( buf_len != how_much)
		fprintf(stderr, "Syntax error in buffer\n");
	    return(1);
	case 0:
	    /* We found a package.  We consumed this many bytes */
	    how_much += p_len;
	    break;
	default:
	    assert(0);
	}

	CTB = (packet[0] & CTB_TYPE_MASK) >> 2;
	switch ( CTB ) {
	case CTB_COMMENT_TYPE:
	case CTB_CKE_TYPE:
	case CTB_COMPRESSED_TYPE:
	    fprintf(stderr, "CTB = %x\n", packet[0]);
	    assert(0);
	    /*NOTREACHED*/
	case CTB_CERT_SECKEY_TYPE: {
	    u_short	us;

	    memcpy(&us, packet+11, 2);
	    us = ntohs(us);
	    fprintf(fd, "\tSecret key (length=%d)!! \n", us);
	    continue;
	}
 	case CTB_KEYCTRL_TYPE:
	    fprintf(fd, "\tKey Control Packet\n");
	    continue;
	case CTB_CERT_PUBKEY_TYPE:
	    have_key = 1;
	    key = malloc(sizeof(PGPpubkey_t));
	    if ( key == NULL ) {
		fprintf(stderr, "Out of memory\n");
		return(-1);
	    }
	    i = buf2pubkey(packet, p_len, key);
	    PGPKEYID_FromBN(&key_id,key->N);
	    fprintf(fd, "\tPublic Key: %s\n", PGPKEYID_ToStr(&key_id));

	    free_pubkey(key);
	    free(key);
	    continue;
 	case CTB_USERID_TYPE:
	    uid = malloc(sizeof(PGPuserid_t));
	    if ( uid == NULL ) {
		fprintf(stderr, "Out of memory\n");
		return(-1);
	    }
	    i = buf2uid(packet, p_len, uid);
	    fprintf(fd, "\tUserID: %s\n", uid->name);
	    free_uid(uid);
	    free(uid);
	    continue;

	case CTB_SKE_TYPE: {
	    /* a local variable */
	    PGPuserid_t	**uidp;

	    sig = malloc(sizeof(PGPsig_t));
	    if ( sig == NULL ) {
		fprintf(stderr, "Out of memory\n");
		return(-1);
	    }
	    i = buf2signature(packet, p_len, sig);
	    fprintf(fd, "\tSignature with KeyID: %s\n",
		    PGPKEYID_ToStr(&sig->key_id));

	    i = uids_from_keyid(sig->key_id, &uidp);
	    if ( i == 1 || i == -1)
		fprintf(fd, "\t\tUnknown\n");
	    else {
		for(i = 0; uidp[i] != NULL; i++) {
		    fprintf(fd, "\t\t\t%s\n", uidp[i]->name);
		    free_uid(uidp[i]);
		}
		free(uidp);
	    }
	    free_signature(sig);
	    free(sig);
	    continue;
	}
	default:
	    fprintf(stderr, "Unknown CTB=%d\n", CTB);
	    assert(0);
	    /*NOTREACHED*/
	}
    } /* while */
    return(0);
}
#endif	/* __palmos__ */

/* Turns a bignum into binary (mpi). It's here to save some space on the Pilot.
 * Returns 0 on success, -1 on failure. Frees the buffer <buf> on failure.
 * --Perm */
static int
BN2bin(register BIGNUM *N,
	register u_char *buf,
	register int *howfar)
{
    short hs;
    register int i;

    /* prefix to N */
#ifdef WIN32
    hs = htons((u_short)BN_num_bits(N));
#else	/* WIN32 */
    hs = htons(BN_num_bits(N));
#endif	/* WIN32 */
    memcpy(buf + *howfar, &hs, 2);
    *howfar += 2;

    /* N */
    i = BN_bn2bin(N, buf + *howfar);
    if (0 == i) {
	free(buf);
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return -1;
    }
    *howfar += i;
    return 0;
}






/* Find the length of a public key.  Return the length,
 * or 0 upon error
 */

int
pubkey_length(PGPpubkey_t *key)
{
    if (key == NULL)
	return(0);
    return (1 	/* CTB */
	    +2 	/*length*/
	    +1	/*Version*/
	    +4	/*Timestamp*/
	    +2 	/*Valid*/
	    +1	/*Algorithm byte*/
	    +BN_num_bytes(key->N)
	    +BN_num_bytes(key->E));
}

/*
 * Return 0 if all is OK, -1 upon error.
 */

int
pubkey2buf(PGPpubkey_t 	*key,
	   u_char 	**bufptr)
{
    u_char	CTB;
    u_char 	*buf;
    short 	hs;
    unsigned long hl;
    int		howfar;
    int 	i;

    if ( key == NULL || bufptr == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    CTB = CTB_CERT_PUBKEY;
    
    buf = calloc(1, key->length);
    if ( buf == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    howfar = 0;

    /* CTB */
    buf[howfar] = CTB;
    howfar += 1;

    /* length */
    hs = htons( (u_short) (key->length - 1 - ctb_llength(CTB)));
    memcpy(buf+howfar, &hs, 2);
#ifdef __palmos__
    ErrFatalDisplayIf((ctb_llength(CTB) != 2), "ctb_llength(CTB) != 2");
#else	/* __palmos__ */
    assert ( ctb_llength(CTB) == 2);
#endif	/* __palmos__ */
    howfar += 2;

    /* version */
    buf[howfar] = key->version;
    howfar += 1;

    /* timestamp */
    hl = htonl(key->created);
    memcpy(buf+howfar, &hl, SIZEOF_TIMESTAMP);
    howfar += SIZEOF_TIMESTAMP;

    /* Validity */
    hs = htons( (u_short)key->valid);
    memcpy(buf+howfar, &hs, 2);
    howfar += 2;

    /* algorithm */
    buf[howfar] = key->algorithm;
    howfar += 1;

    /* Emit N */
    if (0 == BN2bin(key->N, buf, &howfar)) {
	/* Emit E */
	if (0 == BN2bin(key->E, buf, &howfar)) {
	    if ( key->length != howfar) {
		free(buf);
		pgplib_errno = PGPLIB_ERROR_IMPOSSIBLE;
		return(-1);
	    }
	    *bufptr = buf;
	    return 0;
	}
    }

    return -1;
}


/*
 * Take a buffer and convert the thing into a struct.  It must be
 * assumed that this indeed is the binary representation of a
 * key.  The only thing we can check is the CTB and look for a
 * reasonable size of N.
 * Return 0 on OK, -1 on error, 1 on syntax error (which is unsure).
 */

int
buf2pubkey(u_char	*buf,
	   int 		len,
	   PGPpubkey_t  *key)
{
    int		sofar, ret_value = -1;
    /* scratch */
    u_long 	l;
    u_short	us;
    int		i;

    /* 15 is a minimum number, in that it inluces 2 bytes for MPI for
     * both numbers, but not the numbers themselves.
     */
    if ( buf == NULL || len < 15 || key == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    sofar = 0;
    memset(key, 0, sizeof(PGPpubkey_t));

    if ( ! is_ctb_type(buf[sofar], CTB_CERT_PUBKEY_TYPE )) {
	pgplib_errno = PGPLIB_ERROR_BAD_CTB;
	return(1);
    }
    sofar++;

    /* length of the _entire_ packet */
    key->length = packet_length(buf);
    if ( key->length == -1 )
	return (-1);

    if ( key->length > len ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(1);
    }

    sofar += ctb_llength(*buf);

    /*version */
    memcpy( &key->version, buf+sofar, 1);
    sofar++;

    /* time */
    l = 0;
    memcpy( &l, buf+sofar, SIZEOF_TIMESTAMP);
    key->created = ntohl(l);
    /* XXX Sanity check on timestamp ? */
    sofar += SIZEOF_TIMESTAMP;

    /* validity */
    us = 0;
    memcpy( &us, buf+sofar, 2);
    key->valid = ntohs(us);
    /* XXX Sanity check on timestamp ? */
    sofar += 2;

    /* Read algotithm */
    memcpy( &key->algorithm, buf+sofar, 1);
    if (key->algorithm != RSA_ALGORITHM_BYTE ) {
	pgplib_errno = PGPLIB_ERROR_BAD_ALGORITHM;
	goto bad;
    }
    sofar++;

    ret_value = mpi2bignum(buf+sofar, len-sofar, &i, &key->N);
    if ( ret_value != 0 ) {
	ret_value = -1;
	goto bad;
    }
    sofar += i;

    ret_value = mpi2bignum(buf+sofar, len-sofar, &i, &key->E);
    if ( ret_value != 0 ) {
	ret_value = -1;
	goto bad;
    }

    return(0);

    /* We only get here if something is wrong.  Notice that
     * free(NULL) is not an error, thus we have to memset at the top.
     */
bad:
    BN_free(key->N);
    BN_free(key->E);
    return(ret_value);
}


/*
 * Routines to manipulate secret keys.
 */

void
free_seckey(PGPseckey_t	*key)
{
    if ( key == NULL)
	return;
    if ( key->N != NULL )
	BN_free(key->N);
    if ( key->E != NULL )
	BN_free(key->E);
    if ( key->D != NULL )
	BN_clear_free(key->D);
    if ( key->P != NULL )
	BN_clear_free(key->P);
    if ( key->Q != NULL )
	BN_clear_free(key->Q);
    if ( key->U != NULL )
	BN_clear_free(key->U);
    if (NULL !=  key->dmp1)
	BN_clear_free(key->dmp1);
    if (NULL !=  key->dmq1)
	BN_clear_free(key->dmq1);
    if (NULL !=  key->iqmp)
	BN_clear_free(key->iqmp);
    memset(key, 0, sizeof(PGPseckey_t));
}

/*
 * Turn a buffer into a secret key.
 * We must assume that the buffer contains a key, but try
 * to do verification as we go.
 * Return -1 on error, 0 if all is well.
 */

int
buf2seckey(u_char	*buf,
	   int		buflen,
	   PGPseckey_t	*key)
{
    int		sofar, ret_value;
    u_long	ul;
    u_short	us;
    /* scratch */
    int		i;

    /* 34 is the lowest possible number.  Two bytes are assumed per MPI,
     * but no number itself (All MPI's == 0 )
     */
    if ( buf == NULL || buflen < 34 || key == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    sofar = 0;
    memset(key, 0, sizeof(*key));

    if ( ! is_ctb_type(buf[sofar], CTB_CERT_SECKEY_TYPE )) {
	pgplib_errno = PGPLIB_ERROR_BAD_CTB;
	return(-1);
    }
    sofar++;

    /* length of the _entire_ packet */
    key->length = packet_length(buf);
    if ( key->length > buflen || key->length == -1) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }
    sofar += ctb_llength(*buf);

    /* version*/
    key->version = buf[sofar];
    if (key->version != VERSION_BYTE_NEW
	&& key->version != VERSION_BYTE_OLD)
    {
	pgplib_errno = PGPLIB_ERROR_BAD_ALGORITHM; /* well.... */
	return(-1);
    }
    sofar += 1;
    /* timestamp */
    memcpy(&ul, buf+sofar, 4);
    key->created = ntohl(ul);
    /* XXX verification of timestamp ? */
    sofar += 4;

    /* Validity */
    memcpy(&us, buf+sofar, 2);
    key->valid = ntohs(us);
    sofar += 2;

    /* RSA algorithm byte */
    key->algorithm = buf[sofar];
    if ( key->algorithm != RSA_ALGORITHM_BYTE) {
	pgplib_errno = PGPLIB_ERROR_BAD_ALGORITHM;
	return(-1);
    }
    sofar += 1;

    /* Public N */
    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->N);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    /* Public E */
    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->E);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    /* Type of chipher used */
    key->algorithm = buf[sofar];
    sofar += 1;
    /* The key is only followed by IV when it exists.  Pointed out by
     * Andy Dustman <andy@CCMSD.chem.uga.edu>
     */
    if ( key->algorithm != 0 ) {
	/* Eight bytes of iv */
	memcpy(key->iv, buf+sofar, 8);
	sofar += 8;
    }

    /* Then the four private numbers */
    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->D);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->P);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->Q);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    ret_value = mpi2bignum(buf+sofar, buflen-sofar, &i, &key->U);
    if ( ret_value != 0 )
	goto bad;
    sofar += i;

    /* And, at the end, the checksum */
    if ( sofar+2 > buflen ) {
	ret_value = 1;
	/* XXX EINVAL shoud bee syntax error or something */
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	goto bad;
    }

    memcpy(&key->cksum, buf+sofar, 2);

    return(0);

bad:
    free_seckey(key);

    if ( ret_value == 1 || ret_value == -1 )
	return(ret_value);
    else
	return(-1);
}

/*
 * Turn a secret key into a buffer. This one does virtually no verification of
 * the key, so shooting yourself in the foot IS ALLOWED. Needed on the pilot
 * for the same reasons encrypt_seckey() is needed.		--Perm
 * Return -1 on error, 0 if all is well.
 */
int
seckey2buf(PGPseckey_t	*key,
	   u_char	**bufptr)
{
    register u_char	CTB;
    register u_char 	*buf;
    short 	hs;
    u_short	cs;	/* checksum */
    unsigned long	hl;
    int	howfar;

    if ( key == NULL || bufptr == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    *bufptr = NULL;

    CTB = CTB_CERT_SECKEY;

    buf = calloc(1, key->length);
    if ( buf == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    howfar = 0;

    /* CTB */
    buf[howfar] = CTB;
    howfar += 1;

    /* length */
    hs = htons( (u_short) (key->length - 1 - ctb_llength(CTB)));
    memcpy(buf+howfar, &hs, 2);
    howfar += 2;

    /* version */
    buf[howfar] = key->version;
    howfar += 1;

    /* timestamp */
    hl = htonl(key->created);
    memcpy(buf+howfar, &hl, SIZEOF_TIMESTAMP);
    howfar += SIZEOF_TIMESTAMP;

    /* Validity */
    hs = htons( (u_short)key->valid);
    memcpy(buf+howfar, &hs, 2);
    howfar += 2;

    /* algorithm */
    buf[howfar] = key->algorithm;
    howfar += 1;

    /* Emit N */
    if (BN2bin(key->N, buf, &howfar))
	goto err;
    
    /* Emit E */
    if (BN2bin(key->E, buf, &howfar))
	goto err;

    /* type of chipher used  */
    buf[howfar] = key->algorithm;
    howfar += 1;
    /* The key is only followed by IV when it exists. */
    if ( key->algorithm != 0 ) {
	/* Eight bytes of iv */
	memcpy(buf+howfar, key->iv, 8);
	howfar += 8;
    }

    /* The four private numbers */ 

    /* Emit D */
    if (BN2bin(key->D, buf, &howfar))
	goto err;
    
    /* Emit P */
    if (BN2bin(key->P, buf, &howfar))
	goto err;

    /* Emit Q */
    if (BN2bin(key->Q, buf, &howfar))
	goto err;
    
    /* Emit U */
    if (BN2bin(key->U, buf, &howfar))
	goto err;

    /* And, at the end, the checksum */
    cs = checksum(buf, howfar);		/* calculate checksum of D, P, Q, U */
    cs = htons(cs);			/* network format (bigendian */
    memcpy(buf+howfar, &cs, 2);
    howfar += 2;
    
    if ( key->length != howfar) {
	free(buf);
	pgplib_errno = PGPLIB_ERROR_IMPOSSIBLE;
	return(-1);
    }
    *bufptr = buf;
    return 0;

err:
    return -1;
}


/* Remove the secrets stored in the key. */

void
clear_seckey(PGPseckey_t	*key)
{
    BN_clear_free(key->D); key->D = NULL;
    BN_clear_free(key->P); key->P = NULL;
    BN_clear_free(key->Q); key->Q = NULL;
    BN_clear_free(key->U); key->U = NULL;

    /* Oh, and these must also be clear-free'd. --Perm */
    BN_clear_free(key->dmp1);
    BN_clear_free(key->dmq1);
    BN_clear_free(key->iqmp);
    key->dmp1 = key->dmq1 = key->iqmp = NULL;
}


/* Encrypt/decrypt (same shit) a bignum. Saves some space on the PILOT by
 * having an own routine doing it.  --Perm */
static int
encrypt_BN(register BIGNUM *old,
	    register BIGNUM **new,
	    register u_char *iv,
	    register int *useiv,
	    register IDEA_KEY_SCHEDULE *ks,
	    int enc)
{
    register u_char	*buf;
    register int	len1, len2, ret = -1;

    /* Prepare */
    len1 = BN_num_bytes(old);
    buf = malloc(len1);
    if ( buf != NULL ) {
	/* Make binary for encryption/decryption */
	len2 = BN_bn2bin(old, buf);
	idea_cfb64_encrypt(buf, buf, len1, ks, iv, useiv, enc);
	*new = BN_bin2bn(buf, len1, NULL);
	memset(buf, 0, len1);
	free(buf);
	if (*new != NULL ) {
	   ret = 0;
	}
    }
    return ret;
}

/*
 * In order to actually use a secret key, the secret parts must
 * be decrypted.  Return -1 on error, 1 on incorrect decryprion
 * and 0 on sucess.
 *
 * An armed (decrypted) key can be recognised by the fact that
 * the field algorithm is set to 0.
 *
 * This is a slightly modified version - the original is found somewhere below,
 * inside an #if 0/#endif conditional.   --Perm
 */

int
decrypt_seckey(PGPseckey_t 	*old,
	       u_char 		digest[16],
	       PGPseckey_t	*new)
{
    IDEA_KEY_SCHEDULE ks;
    u_char	iv[8];
    int		useiv;	/* Keep track of IV usage */
    BIGNUM	*tmp;
    /* scratch */
    int		i;
    BN_CTX	*ctx;

    if ( old == NULL || digest == NULL || new == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    /* Set things up */
    memcpy(new, old, sizeof(*new));
    new->D = new->P = new->Q = new->U = new->dmp1 = new->dmq1 = new->iqmp = NULL;
    new->algorithm = 0;

    new->N = BN_dup(old->N);
    new->E = BN_dup(old->E);

    idea_set_encrypt_key(digest, &ks);
    memcpy(iv, old->iv, 8);
    useiv = 0;

    pgplib_errno = PGPLIB_ERROR_MAX;

    /* Do the number D */
    if (encrypt_BN(old->D, &new->D, iv, &useiv, &ks, IDEA_DECRYPT))
	goto bad;

    /* The number P */
    if (encrypt_BN(old->P, &new->P, iv, &useiv, &ks, IDEA_DECRYPT))
	goto bad;

    /* The number Q */
    if (encrypt_BN(old->Q, &new->Q, iv, &useiv, &ks, IDEA_DECRYPT))
	goto bad;

    /*
     * There are two ways for us to check whether the password was correct
     * or not: Either verify that P*Q = N, or using the PGP checksum
     * provided with the key.   The latter is calculated over the MPI's
     * (that is, on the numbers, after decryption, but including the
     * prefixed bitcount).  Since we store numbers as BIGNUMs we use the
     * former method.
     * As usual, I solicit your comments.
     */

    tmp = BN_new();
    i = BN_mul(tmp, new->P, new->Q);
    if ( i == 0 )
	goto bad;
    i = BN_cmp(tmp, new->N);
    BN_free(tmp);
    if ( i != 0 ) {
	clear_seckey(new);
	pgplib_errno = PGPLIB_ERROR_WRONG_PASS;
	return(1);
    }

    /* The number U */
    if (encrypt_BN(old->U, &new->U, iv, &useiv, &ks, IDEA_DECRYPT))
	goto bad;

    /* Here we initialize the additional fields to utilitize the
     * RSA_eay_mod_exp function. If something goes wrong, we fall back to the
     * default slow method. There are some severe weirdness regarding the
     * mixing of p and q and whatnot, mainly from the fact that PGP
     * (Zimmermann's code) calculates U from inv(p mod q), which is unusable to
     * the SSLeay code, which requires inv(q mod p). We cannot, AFAIK, easily
     * transform inv(p mod q) into inv(q mod p). See comments in secencrypt.c
     * for details.
     *
     * Thus we have the following options:
     * A) at decryption-time recompute U in the armed key so it's usable by
     *    SSLeay, and make sure we store the PGP-correct U in the secret-key
     *    before disarming it when we generate keypairs, or
     * B) leave U as it is, and instead store inv(q mod p) elsewhere in the
     *    PGPseckey_t structure, along with dm[pq]1.
     *
     * Now, option B has the advantage that it will correctly survive weird
     * operations (not that I can currently think of why anyone would want to
     * do this) seckey2buf(encrypt_seckey(decrypt_seckey(buf2seckey()))), at
     * the cost of keeping U around, doing nothing.
     * Option A, on the other hand, won't correcty survive the aforementioned
     * operation, but it will use less memory. However, the memory saved is
     * negligably small, unless the key is ridiculously large.
     *
     * Since PGPlib isn't [currently] targeted for smartcards and other wimpy
     * devices, I decide to risk wasting 512 bytes in exchange for the
     * banishment of unsurvivability in the face of weird operations, and
     * thusly choose option B.
     *
     * Anyone disagreeing can come to alt.flame and fight it out.
     *
     * --Perm
     */
    ctx = BN_CTX_new();
    if (NULL != ctx) {
	BIGNUM *r1 = NULL, *r2 = NULL;	/* scratch */

	i = 0;	/* to distinguish failure from success at end. */
	new->dmp1 = new->dmq1 = NULL;
	r1 = ctx->bn[0];
	r2 = ctx->bn[1];
	ctx->tos += 2;

	/* calculate d mod (p-1) */
	new->dmp1 = BN_new();
	if (new->dmp1 == NULL) goto permerr;
	if (!BN_sub(r1, new->P, BN_value_one())) goto permerr; /* p-1 */
	if (!BN_mod(new->dmp1, new->D, r1, ctx)) goto permerr;

	/* calculate d mod (q-1) */
	new->dmq1 = BN_new();
	if (new->dmq1 == NULL) goto permerr;
	if (!BN_sub(r2, new->Q, BN_value_one())) goto permerr; /* q-1 */
	if (!BN_mod(new->dmq1, new->D, r2, ctx)) goto permerr;

	/* calculate inv(q mod p) */
	/* XXX This _acts_ right, but definitly _looks_ wrong, since this is
	 * exactly how we compute U in generate_keypair() below. Maybe we
	 * should reverse P and Q in generate_keypair? Or just scrap iqmp and
	 * use U (we did that at first, but that was messy!) */
	new->iqmp = BN_mod_inverse(new->P, new->Q, ctx);
	if (NULL == new->iqmp) goto permerr;

	i = 1;	/* so far, so good. */
permerr:
	if (0 == i) {
	    BN_clear_free(new->dmp1);
	    BN_clear_free(new->dmq1);
	    BN_clear_free(new->iqmp);
	    new->dmp1 = new->dmq1 = new->iqmp = NULL;
	}
	BN_CTX_free(ctx);
	ctx = NULL;
    }

    return(0);

bad:

    clear_seckey(new);
    if ( pgplib_errno == PGPLIB_ERROR_MAX )
	pgplib_errno = PGPLIB_ERROR_SYSERR;
    return -1;
}

/* On the Pilot, it is sometimes necessary to create keypairs, even if it takes
 * forever. It might also be necessary to change the passphrase (i.e. if we use
 * some other method than text e.g. pass-symbols). Therefore we must be able to
 * encrypt a secret key, i.e. disarming it. This function will *MODIFY* the
 * input key. If you want to keep the disarmed key, duplicate it first.  --Perm
 */
int
encrypt_seckey(PGPseckey_t 	*key,
	       u_char 		digest[16])
{
    IDEA_KEY_SCHEDULE ks;
    u_char	iv[8];
    int		useiv;	/* Keep track of IV usage */
    /* scratch */
    BIGNUM *D, *P, *Q, *U;

    if ( key == NULL || digest == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    /* Set things up */
    D = P = Q = U = NULL;

    /* Initialize IV */
    idea_set_encrypt_key(digest, &ks);
    memset(iv, 0, 8);
    useiv = 0;

    pgplib_errno = PGPLIB_ERROR_MAX;

    /* Do the number D */
    if (encrypt_BN(key->D, &D, iv, &useiv, &ks, IDEA_ENCRYPT))
	goto bad;
    BN_clear_free(key->D); /* Clear and free old secret-key components */
    key->D = D;

    /* The number P */
    if (encrypt_BN(key->P, &P, iv, &useiv, &ks, IDEA_ENCRYPT))
	goto bad;
    BN_clear_free(key->P); /* Clear and free old secret-key components */
    key->P = P;

    /* The number Q */
    if (encrypt_BN(key->Q, &Q, iv, &useiv, &ks, IDEA_ENCRYPT))
	goto bad;
    BN_clear_free(key->Q); /* Clear and free old secret-key components */
    key->Q = Q;

    /* The number U */
    if (encrypt_BN(key->U, &U, iv, &useiv, &ks, IDEA_ENCRYPT))
	goto bad;
    BN_clear_free(key->U); /* Clear and free old secret-key components */
    key->U = U;

    /* Clear dmp1 and dmq1 */
    if (NULL != key->dmp1) BN_clear_free(key->dmp1);
    if (NULL != key->dmq1) BN_clear_free(key->dmq1);
    if (NULL != key->iqmp) BN_clear_free(key->iqmp);

#if 0
    /* Copy new iv into key */
    memcpy(key->iv, iv, 8);
#endif
    /* We must _clear_ the IV! If we don't then the first 64 bits will be out
     * of sync. We can _set_ it to something other than zero, but it _must_ be
     * set to the same as the initial setting (i.e. before encrypting) --Perm
     */
    memset(key->iv, 0, 8);

    D = P = Q = U = NULL;

    return 0;

bad:
    if (NULL != D) BN_clear_free(D);
    if (NULL != P) BN_clear_free(P);
    if (NULL != Q) BN_clear_free(Q);
    if (NULL != U) BN_clear_free(U);

    if ( pgplib_errno == PGPLIB_ERROR_MAX )
	pgplib_errno = PGPLIB_ERROR_SYSERR;
    return -1;
}

/* This is the original decrypt_seckey()    -- Perm */
#if 0
/*
 * In order to actually use a secret key, the secret parts must
 * be decrypted.  Return -1 on error, 1 on incorrect decryprion
 * and 0 on sucess.
 *
 * An armed (decrypted) key can be recognised by the fact that
 * the field algorithm is set to 0.
 *
 */

int
decrypt_seckey(PGPseckey_t 	*old,
	       u_char 		digest[16],
	       PGPseckey_t	*newkey)
{
    u_char	*buf;
    IDEA_KEY_SCHEDULE ks;
    int		len1, len2;
    u_char	iv[8];
    int		useiv;	/* Keep track of IV usage */
    BIGNUM	*tmp;
    /* scratch */
    int		i;

    if ( old == NULL || digest == NULL || newkey == NULL) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }

    /* Set things up */
    memcpy(newkey, old, sizeof(*newkey));
    newkey->D = newkey->P = newkey->Q = newkey->U = NULL;
    newkey->algorithm = 0;

    newkey->N = BN_dup(old->N);
    newkey->E = BN_dup(old->E);

    idea_set_encrypt_key(digest, &ks);
    memcpy(iv, old->iv, 8);
    useiv = 0;

    pgplib_errno = PGPLIB_ERROR_MAX;

    /* Do the number D */
    len1 = BN_num_bytes(old->D);
    buf = malloc(len1);
    if ( buf == NULL )
	goto bad;
    /* Make binary for decryption */
    len2 = BN_bn2bin(old->D, buf);
    idea_cfb64_encrypt(buf, buf, len1, &ks, iv, &useiv, IDEA_DECRYPT);
    newkey->D = BN_bin2bn(buf, len1, NULL);
    memset(buf, 0, len1);
    free(buf);
    if ( newkey->D == NULL )
	goto bad;

    /* The number P */
    len1 = BN_num_bytes(old->P);
    buf = malloc(len1);
    if ( buf == NULL )
	goto bad;
    len2 = BN_bn2bin(old->P, buf);
    idea_cfb64_encrypt(buf, buf, len1, &ks, iv, &useiv, IDEA_DECRYPT);
    newkey->P = BN_bin2bn(buf, len1, NULL);
    memset(buf, 0, len1);
    free(buf);
    if ( newkey->P == NULL )
	goto bad;

    /* The number Q */
    len1 = BN_num_bytes(old->Q);
    buf = malloc(len1);
    if ( buf == NULL )
	goto bad;
    len2 = BN_bn2bin(old->Q, buf);
    idea_cfb64_encrypt(buf, buf, len1, &ks, iv, &useiv, IDEA_DECRYPT);
    newkey->Q = BN_bin2bn(buf, len1, NULL);
    memset(buf, 0, len1);
    free(buf);
    if ( newkey->Q == NULL )
	goto bad;

    /*
     * There are two ways for us to check whether the password was correct
     * or not: Either verify that P*Q = N, or using the PGP checksum
     * provided with the key.   The latter is calculated over the MPI's
     * (that is, on the numbers, after decryption, but including the
     * prefixed bitcount).  Since we store numbers as BIGNUMs we use the
     * former method.
     * As usual, I solicit your comments.
     */

    tmp = BN_new();
    i = BN_mul(tmp, newkey->P, newkey->Q);
    if ( i == 0 )
	goto bad;
    i = BN_cmp(tmp, old->N);
    BN_free(tmp);
    if ( i != 0 ) {
	clear_seckey(newkey);
	pgplib_errno = PGPLIB_ERROR_WRONG_PASS;
	return(1);
    }

    /* The number U */
    len1 = BN_num_bytes(old->U);
    buf = malloc(len1);
    if ( buf == NULL )
	goto bad;
    len2 = BN_bn2bin(old->U, buf);
    idea_cfb64_encrypt(buf, buf, len1, &ks, iv, &useiv, IDEA_DECRYPT);
    newkey->U = BN_bin2bn(buf, len1, NULL);
    memset(buf, 0, len1);
    free(buf);
    if ( newkey->U == NULL )
	goto bad;

    return(0);

bad:

    clear_seckey(newkey);
    if ( pgplib_errno == PGPLIB_ERROR_MAX )
	pgplib_errno = PGPLIB_ERROR_SYSERR;
    return -1;
}
#endif /* 0 */

/* Default callback function - to show progress. One big hack. *shudder*
 * Doesn't actually do anything unless the target platform is PalmOS. Fix? Add
 * nice fprintf(stderr, "+") to show progress?
 * Anyway, here are the meanings of parameters' values :
 *    level:	0: BN_generate_prime found prime candidate number
 *		1: BN_generate_prime canditate survived a primality test
 *		2: RSA_generate_key discovered bogus prime (GCD != 1), or
 *		   BN_generate_prime found a strong prime (this is unlikely,
 *		   since RSA_generate_key doesn't ask for a strong prime)
 *		3: RSA_generate_key found a sound prime (i.e usable).
 *    round:	Round of the current level (i.e. 3,0 = got P, 3,1 = got Q).
 * --Perm */
void
rsa_key_generate_callback(int level, int round)
{
#ifdef __palmos__
    register WinHandle wh;
    register int i, state, y;
    unsigned char c;

    /* We're now about to do some magic - since global data seems to be
     * unavailable in the callback function (even with Ian's hack), we have
     * to get a static storage elsewhere. Here we exploit the currently active
     * draw window's (which we have made in rsa_key_generate_callback_control()
     * below) graphicstate's custom pattern. :)  --Perm */
    wh = WinGetDrawWindow();
    i = wh->gstate->customPtn[0];
    state = wh->gstate->customPtn[1];
    y = wh->gstate->customPtn[2];

    if (state & (1 << 1)) {
	/* We must do this in assembler, since both statics and stack
	 * initialization could overflow the stack, but most of all I just
	 * love M68K assembly :) --Perm
	 */
	asm ("
	move.w	#16, -(%%sp)		| snd.param3 (volume)
	move.w	#40, -(%%sp)		| snd.param2 (Duration in milliseconds)
	moveq.l	#0,%%d0			| Vary Hz by <level>*4
	move.w	%0, %%d0
	lsl.w	#4, %%d0
	add.w	#440, %%d0
	move.l	%%d0, -(%%sp)		| snd.param1 (Hz)
	move.w	#0x0100, -(%%sp)	| snd.cmd 
	move.l	%%sp, %%d0		| SndCommandType is now on the stack
	clr.w	-(%%sp)
	move.l	%%d0, -(%%sp)
	clr.l	-(%%sp)
	trap	#15
	dc.w	0xa233	| SndDoCmd
	adda.w	#10+10, %%sp		| Restore sp (SndCommandType + args)
	" : : "g" (level) );
    }
    if (state & 1) {		/* Should we clear anyting? */
	WinEraseWindow();
	i = -6;
	state &= ~1;
    }
    if (0 == level) c = '\x95';	/* Bullet */
    if (1 == level) c = '+';
    if (2 == level) c = '*';
    if (3 == level) {
	c = '\x86';		/* dagger */
	state |= 1;
    }
    if (0x5555 == level) {	/* special reset magic */
	i = -6;
	state &= ~1;
	y = 1;
    } else {
	i += 6;
	if (154 < i) {
	    i = 0;
	    y += 4;
	}
	WinDrawChars(&c, 1, i + 3, y);
	EvtResetAutoOffTimer();
    }
    wh->gstate->customPtn[0] = i;
    wh->gstate->customPtn[1] = state;
    wh->gstate->customPtn[2] = y;
#else	/* not __palmos__ */
    static 
    char c;
    switch (level) {
	case 0: c = '.'; break;
	case 1: c = '+'; break;
	case 2: c = '*'; break;
	case 3: c = '#'; break;
	default: c = '?'; break;
    }
    fprintf(stdout, "%c", c);
    fflush(stdout);

    return;
#endif	/* __palmos__ */
}

/* Initalize key generation feedback... thingie. If <control> is nonzero, we're
 * initializing, if it's zero, we're terminating. <flags> are the flags passed
 * to generate_keypair().
 * Doesn't actually do anything unless the target platform is PalmOS. Fix? Add
 * nice control to above mentioned fprintf()?
 * --Perm
 */
static void
rsa_key_generate_callback_control(register int control, register u_short flags)
{
#ifdef __palmos__
    static WinHandle wh = NULL, oldwin = NULL, savewin = NULL;
    static RectangleType framerec = { {0,120}, {156, 37} };
    static RectangleType winrec = { {2,133}, {156, 23} };
    static RectangleType saverec = { {0,118}, {159, 41} };
    static SndCommandType done_beep = { 1, 1400, 100, sndMaxAmp };
    static char info[] = " Generating keypair... ";
    register FontID fnt;
    register int i, l;
    Word w;

    if (control) {
	if (!(GENKEY_NO_PROGRESSWINDOW & flags)) {
	    /* We're initializing */
	    oldwin = WinGetDrawWindow();

	    /* Save the portion of the old window we'll be overwriting */
	    savewin = WinSaveBits(&saverec, &w);
	    if (NULL == savewin) goto argh;
	    WinEraseRectangle(&saverec, 0);
	    WinDrawRectangleFrame(dialogFrame, &framerec);

	    fnt = FntSetFont(1);
	    l = StrLen(info);
	    i = FntLineWidth(info, l);
	    WinDrawInvertedChars(info, l, (80 - (i / 2)), saverec.topLeft.y + 2);
	    FntSetFont(fnt);
	    WinDrawLine(1, saverec.topLeft.y + 13,
		    158, saverec.topLeft.y + 13);

	    /* Create, initialize and draw progress window */
	    wh = WinCreateWindow(&winrec, noFrame, 1, 0, &w);
	    if (NULL == wh) goto argh;	/* bail out */

	    /* We're now about to do some magic - since global data seems to be
	     * unavailable in the callback function (even with Ian's hack), we
	     * have to get a static storage elsewhere.  Here we exploit the
	     * currently active draw window's (which we have made ourselves)
	     * graphicstate's custom pattern. And since we're the creators, we
	     * don't care about saving what's there. :)  --Perm */

	    WinInitializeWindow(wh);
	    WinSetDrawWindow(wh);
	    WinDrawRectangleFrame(simpleFrame, &winrec);

	    /* Set state to pass to callback */
	    wh->gstate->customPtn[1] = (GENKEY_BEEP_UPDATE & flags) ? 1<<1 : 0;
	    rsa_key_generate_callback(0x5555, 0); /* Reset callback */
	}
    } else {
	/* We're terminating */
	if (NULL != wh)
	    WinDeleteWindow(wh, 1);
	if (NULL != oldwin)
	    WinSetDrawWindow(oldwin);
	if (NULL != savewin)
	    WinRestoreBits(savewin, saverec.topLeft.x, saverec.topLeft.y);
	savewin = wh = oldwin = NULL;

	if (GENKEY_BEEP_WHEN_DONE & flags)
	    SndDoCmd(NULL, &done_beep, 0);
    }
argh:
	return;
#else	/* not __palmos__ */
	if (0 == control) {
	    fprintf(stdout, "\n");
	    fflush(stdout);
	}

	return;
#endif	/* __palmos__ */
}


/* generate a RSA keypair, given the keysize, some value for RSA's e (e.g. 3,
 * 11 or 0x1001), a passphrase, a couple of pointers to some keystructs, and
 * some random-source, not to forget the number of days the keypair shall
 * remain valid (0 means forever). The random-source may be NULL, if you trust
 * SSLEAY's rsa-key generator to do a good job initializing the SSLEAY RNG.
 * Returns -1 on failure, 0 on success.
 * Note that this function has been tuned for the PalmPilot, so it may seem
 * weird at places (but that's tradeoffs for ya - in _my_ days...).
 * --Perm 
 */
int
generate_keypair(PGPseckey_t *sec,	/* secret key (empty struct) */
		 PGPpubkey_t *pub,	/* public key (ditto) */
		 int keysize,		/* keysize in bits (>=384) */
		 u_long e,		/* >2, e.g. 3, 11, some small prime */
		 u_short valid,		/* # of days the keypair is valid */
		 u_char digest[16],	/* passphrase */
		 u_char *rndsrc,	/* random stir source (or NULL) */
		 u_long rndsrclen,	/* # of bytes in rndsrc */
		 u_short flags)		/* Various flags. Pass 0. */
{
    register RSA *rsa;
    register BN_CTX *ctx;
    register long sl, pl;
    int i;
    int ret = -1;	/* assume failure */
    
    /* Pointer sanity checks */
    if (NULL == sec || NULL == pub || NULL == digest) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    /* We allow only keysizes between 64 and 4096, inclusive. e must be >2,
     * and there must be something in the passphrase. Yawn. */
    if (64 > keysize || 4096 < keysize || 3 > e) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return -1;
    }

    /* If there's a rndsrc, use it to stir the rng */
    if (NULL != rndsrc) {
	if (0 < rndsrclen) {
	    RAND_seed(rndsrc, rndsrclen);	/* Add seed data to the RNG state */
	}
    }

    /* Clear key-structs */
    memset(sec, 0, sizeof(*sec));
    memset(pub, 0, sizeof(*pub));

    if (flags & GENKEY_NO_CALLBACK)		/* NO_CALLBACK implies: */
	flags |= GENKEY_NO_PROGRESSWINDOW;	/* no progress windows */

    /* Initialize callback */
    rsa_key_generate_callback_control(1, flags);

    rsa = RSA_generate_key(keysize, e,
	    (flags & GENKEY_NO_CALLBACK) ? NULL : rsa_key_generate_callback);

    if (NULL != rsa) {
	/* Accoring to Andy Dustman <andy@CCMSD.chem.uga.edu> and Derek Atkins
	 * <warlord@MIT.EDU> the number U in PGP is inv(p mod q) while the
	 * SSLeay number iqmp is inv(q mod p).  In order to use PGP and SSLeay
	 * together we have to switch the numbers before we use them.  */

	/* We do U first, because BN_CTX_new() eats some of the memory that
	 * we'd use on the BN_dup()s, so we free it before we do any BN_dup()s.
	 * --Perm */
	if (NULL == (ctx = BN_CTX_new())) goto argh;
	/* PGP 2.6.x wants U as inv(p mod q), if p<q. Later, in
	 * decrypt_seckey(), we use P and Q to compute inv(p mod q), which is
	 * what Eric Young's RSA implementation want. It's a drag, but...
	 * --Perm */
	/* TODO: See if PGP 2.6 understands this U */
	i =  BN_cmp(rsa->p, rsa->q);
	if (-1 == i) {
	    sec->U = BN_mod_inverse(rsa->p, rsa->q, ctx);
	} else if (1 == i) {
	    sec->U = BN_mod_inverse(rsa->q, rsa->p, ctx);
	} else {
	    /* If P == Q, make merry hell */
#ifdef __palmos__
	    ErrDisplay("ARGH! P == Q (very bad)!");
#else	/* not __palmos__ */
	    assert(0); /* If P == Q, make merry hell */
#endif	/* __palmos__ */
	}
	
	BN_CTX_free(ctx);
	if (NULL == sec->U) goto argh;
	if (NULL == (sec->N = BN_dup(rsa->n))) goto argh;
	BN_clear_free(rsa->n); rsa->n = NULL;
	if (NULL == (sec->E = BN_dup(rsa->e))) goto argh;
	BN_clear_free(rsa->e); rsa->e = NULL;
	if (NULL == (sec->D = BN_dup(rsa->d))) goto argh;
	BN_clear_free(rsa->d); rsa->d = NULL;
	if (NULL == (sec->P = BN_dup(rsa->p))) goto argh;
	BN_clear_free(rsa->p); rsa->p = NULL;
	if (NULL == (sec->Q = BN_dup(rsa->q))) goto argh;
	/* We free the rsa-struct as soon as possible, not wanting to hog
	 * memory in the dynamic heap longer than strictly necessary. We pull
	 * pub->[NE] from sec's instead. -- Perm */
	RSA_free(rsa);
	rsa = NULL;
	/* That mess aside, we can fill in the rest of the seckey structure */
#ifdef __palmos__
	sec->created = htonl(TimGetSeconds() - POSIX_EPOCH_KLUDGE);
#else	/* __palmos__ */
	sec->created = htonl(time(NULL));
#endif	/* __palmos__ */
	sec->version = VERSION_BYTE_NEW;
	sec->valid = htons(valid);
	sec->algorithm = RSA_ALGORITHM_BYTE;
	/* Calculate length */
	sl =  BN_num_bytes(sec->N);
	sl += BN_num_bytes(sec->E);
	pl = sl;
	sl += BN_num_bytes(sec->D);
	sl += BN_num_bytes(sec->P);
	sl += BN_num_bytes(sec->Q);
	sl += BN_num_bytes(sec->U);

#ifdef DEBUG
	{
	    /* This block for debugging only! */
	    char *s;
	    s = BN_bn2ascii(sec->N);
	    fprintf(stderr, "\nRSA_generate_key(): N = %s\n", s); free(s);
	    s = BN_bn2ascii(sec->E);
	    fprintf(stderr, "RSA_generate_key(): E = %s\n", s); free(s);
	    s = BN_bn2ascii(sec->P);
	    fprintf(stderr, "RSA_generate_key(): P = %s\n", s); free(s);
	    s = BN_bn2ascii(sec->Q);
	    fprintf(stderr, "RSA_generate_key(): Q = %s\n", s); free(s);
	    s = BN_bn2ascii(sec->D);
	    fprintf(stderr, "RSA_generate_key(): D = %s\n", s); free(s);
	    s = BN_bn2ascii(sec->U);
	    fprintf(stderr, "RSA_generate_key(): U = %s\n", s); free(s);
	}
#endif

	/* Encrypt sensitive key components */
	if (encrypt_seckey(sec, digest)) goto argh;
	
	/* This length calculation is drawn from PGP's keymgmt.c  --Perm */
	sec->length =	3 + 	/* CTB + length */
			1 + 	/* Version */
	    		4 +	/* timestamp */
			2 +	/* valid */
			1 +	/* algorithm byte */
			2 +	/* N prefix */
			2 +	/* E prefix */
			1 +	/* IDEA algorithm byte */

			/*
			 * TODO: check to see if it is neccessary to add IV
			 * vector length - invariant: allways do it, since we
			 * always encrypt the key!
			 */
			8 +	/* IV vector length XXX: really? we *MUST* arm
				   the secret key, and add the iv if this is
				   added! */
			2 +	/* D prefix */
			2 +	/* P prefix */
			2 +	/* Q prefix */
			2 +	/* U prefix */
			sl +	/* sum of number of bytes in (N,E,P,D,Q,U) */
			2;	/* checksum */
	/* The actual calculation of sec->cksum cannot be done on non-mpi'ed
	 * values (and be right), so we have to postpone it until we convert
	 * the PGPseckey-structure into a buffer in seckey2buf().  --Perm */
	/* XXX Actually, that's b.s., since we have to calc the checksum before
	 * we encrypt. We don't do that. So Phil's PGP will barf on any keys
	 * generated by pgplib. We'd better fix it. --Perm */
	sec->cksum = 0;

	/* The pubkey is quite simple */
	if (NULL == (pub->N = BN_dup(sec->N))) goto argh;
	if (NULL == (pub->E = BN_dup(sec->E))) goto argh;
	pub->created = sec->created;
	pub->valid = sec->valid;
	pub->version= sec->version;
	pub->algorithm = sec->algorithm;
	/* This length calculation is drawn from PGP's keymgmt.c  --Perm */
	pub->length =	3 + 	/* CTB + length */
			1 + 	/* Version */
	    		4 +	/* timestamp */
			2 +	/* valid */
			1 +	/* algorithm byte */
			2 +	/* N prefix */
			2 +	/* E prefix */
			pl;	/* sum of number of bytes in (N, E) */

	ret = 0;
argh:
	if (NULL != rsa)
	    RSA_free(rsa);	/* destroy and free - last thing we do */
    }

    /* Terminate callback */
    rsa_key_generate_callback_control(0, flags);

    return ret;
	
}
