/*********************************************************************
 * Filename:      misc.c
 * Author:        Tage Stabell-Kulo <tage@acm.org>
 * Modified at:   Sat Feb 28 12:33:07 1998
 * Modified by:   Tage Stabell-Kuloe <tage@acm.org>
 *
 * 	Copyright (c) University of Troms, Norway
 * 	See the file COPYING for details
 *
 * $Id: misc.c,v 1.16 1999/03/06 19:02:17 perm Exp $
 *
 ********************************************************************/
/* All PalmOS code by Per Harald Myrvang <perm@pasta.cs.uit.no> */

#include <pgplib.h>

#include <ctype.h>
#include <stdlib.h>
#ifdef WIN32
# include <winsock2.h>
#else  
# include <netinet/in.h>
#endif
#include <string.h>

#ifndef HAVE_STRTOUQ
extern u_quad_t
strtouq(const char *, char **, register int);
#endif 
       
/* SSLeay */
#include <bn.h>

/*
 * Added to satisfy Linux.
 */
#ifndef NBBY
#  define NBBY 8
#endif


#ifdef __palmos__
/* Added this one, since PalmOS doens't have such a thing. --Perm */
void *
calloc(u_long nmemb, u_long size)
{
    void *ret;
    ret = MemPtrNew((nmemb * size));
    if (NULL != ret)
	MemSet(ret, (nmemb * size), 0);
    return ret;
}


/* And the realloc stinks too, so here's a better one, based on Ian Goldberg's
 * PalmOS_realloc() in his SSLeay-port. */
/* XXX This function should be rewritten so that it will handle reallocation of
 * storage-chunks properly, i.e. if <ptr> points to a record, etc. etc. */
void *
realloc(void *ptr, u_long size)
{
    Err err;
    unsigned long oldsize, copysize;
    void *newptr;

    /* Remember the special cases */
    if (NULL == ptr)			/* same as malloc(2) */
	return MemPtrNew(size);

    if (0 == size) {			/* same as free(2) */
	MemPtrFree(ptr);
	return NULL;
    }

    newptr = ptr;

    err = MemPtrResize(ptr, size);	/* First try a simple Resize. */
    if (err) {
	/* Next, try to allocate a new block of the requested size, and moving
	 * the old data there. */
	newptr = MemPtrNew(size);
	if (NULL != newptr) {	/* If this fails, ptr will still be valid */
	    oldsize = MemPtrSize(ptr);
	    copysize = (oldsize < size) ? oldsize : size;
	    MemMove(newptr, ptr, copysize);
	    MemPtrFree(ptr);
	}
	/* Here we could have an else-statement, where we'd use databases and
	 * records to copy the old data into, then freeing the old data chunk,
	 * _then_ trying getting a larger new chunk, and if successful, copyin
	 * the data from the database into the new chunk. Of course, if we
	 * couldn't get a new chunk, we'd have to get a chunk the size of the
	 * old data chunk, and copy the data back into it. In both cases the
	 * database/record would need some cleaning-up. --Perm */
    }
    return newptr;
}

/* This strtol code has been mercilessly snarfed from Matt Dillon's dice
 * library code. Perhaps we should ask for permission? --Perm */

long pgplib_strtol(const char *ptr,
	char **tail,
	int base)
{
    long v = 0;
    short ishex = 0;
    short c;
    short neg = 0;

    while (*ptr == ' ' || *ptr == '\t')
	++ptr;
    if (*ptr == '-') {
	neg = 1;
	++ptr;
    } else if (*ptr == '+')
	++ptr;

    if (ptr[0] == '0' && (ptr[1] == 'x' || ptr[1] == 'X'))
	ishex = 1;

    if (base == 0) {
	base = 10;
	if (ptr[0] == '0') {
	    base = 8;
	    if (ishex)
		base = 16;
	}
    }
    if (base == 16 && ishex)
	ptr += 2;
    for (;;) {
	c = *ptr;
	if (c >= '0' && c <= '9')
	    c -= '0';
	else if (c >= 'a' && c <= 'z')
	    c -= ('a' - 10);
	else if (c >= 'A' && c <= 'Z')
	    c -= ('A' - 10);
	else
	    break;
	if (c >= base)
	    break;
	v = v * base + c;
	++ptr;
    }
    if (tail)
	*tail = ptr;
    if (neg)
	v = -v;
    return(v);
}
#endif /* __palmos__ */

/* The length of the length of a PGP packet depends on the length
 * of the data, and the type.  Take a type, a length, and a "force"
 * (which forces the use of 4 bytes length) and a buffer (at least 5
 * bytes long) and write a CTB and the length (in Network Byte Order)
 * into the buffer.
 * Return the number of bytes written into the buffer.
 * The idea was taken from write_ctb_len in PGP/crypto.c
 *
 * Return -1 on error;
 */
int
make_ctb(u_char ctb_type, u_long length, int force, u_char *buf)
{
    int		llength, llenb, ret = 0;
    u_short	us;
    u_long	ul;

    if ( buf == NULL ) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(-1);
    }
    if (force || (length > 0xffffL)) {
	llength = 4;
	llenb = 2;
    }
    else if ( (u_short)length > 0xff) {
	llength = 2;
	llenb = 1;
    }
    else {
	llength = 1;
	llenb = 0;
    }

    buf[0] = CTB_BYTE(ctb_type, llenb);
    /* convert length to external byteorder */
    switch ( llength) {
    case 1:
	buf[1] = (u_char)length;
	ret = 2;
	break;
    case 2:
	us = htons( (u_short)length);
	memcpy(buf+1, &us, 2);
	ret = 3;
	break;
    case 4:
	ul = htonl(length);
	memcpy(buf+1, &ul, 4);
	ret = 5;
	break;
    default:
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	ret = -1;
    }
    return(ret);
}


/*
 * Copy at most length-1 bytes from buffer into line, stopping after
 * the first newline has been found or buflen bytes has been
 * copied.  In any case, a NULL is appended to the string.
 * If some parameter is unreasonable, 0 is returned, if not, the
 * number of bytes consumed from buf is returned.
 * If a one-byte line is passed in, 0 will obviously be returned.
 */

int
bufgets(u_char	*line,
	int 	linelen,
	u_char 	*buf,
	int 	buflen)
{
    int		i;

    if ( line == NULL || linelen < 1 || buf == NULL || buflen < 1)
	return(0);

    i = 0;
    while ( i < linelen && i < buflen) {
	line[i] = buf[i];
	if ( line[i] == '\n' )
	    if ( i+1 == linelen ) {
		line[i] = 0;
		return(i);
	    }
	    else {
		line[i+1] = 0;
		return(i+1);
	    }
	i += 1;
	continue;
    }
    if ( i == linelen ) {
	line[i-1] = 0;
	/* The last char was overwritten, not consumed */
	return(i-1);
    }
    if ( i == buflen ) {
	if ( i == linelen ) {
	    line[i-1] = 0;
	    return(i-1);
	}
	else {
	    line[i] = 0;
	    return(i);
	}
    }
    if ( i == linelen ) {
	line[i-1] = 0;
	return(i-1);
    }
    return(0);
}


/* PGPKEYID_FromBN - convert a BN into a PGPKEYID simply looking at the
 *		     bottom 64 bits of the BN (unlike BN_get_quad)
 */
int 
PGPKEYID_FromBN(PGPKEYID *key_id,BIGNUM *a)
{
#ifdef __palmos__
    int i, j;
#endif	/* __palmos__ */
    PGPKEYID ret;

    if ( key_id == NULL || a == NULL )
	return(-1);

    /* detect zero specially */
    if ( BN_is_zero(a)) {
      	PGPKEYID_ZERO(ret);
	goto done;
    }
    
#ifndef __palmos__
    /* shortcut little BNs */
    if ( a->top == 1 ) {
#ifdef USE_LONGLONG
	ret = a->d[0];
#else	/* USE_LONGLONG */
# ifdef WIN32
	(*(unsigned __int64 *)&ret = a->d[0];
# else	/* WIN32 */
	(*(unsigned long long *)&ret) = a->d[0];
# endif	/* WIN32 */
#endif	/* USE_LONGLONG */
	goto done;
    }

    /* XXX Byte order ? */
#ifdef USE_LONGLONG
    ret = a->d[0];
    ret |= ( (PGPKEYID) a->d[1] << 32);
#else	/* USE_LONGLONG */
# ifdef WIN32
    (*(unsigned __int64 *)&ret) = a->d[0];
    (*(unsigned __int64 *)&ret) |= ( (unsigned long long) a->d[1] << 32);
# else 	/* WIN32 */
    (*(unsigned long long *)&ret) = a->d[0];
    (*(unsigned long long *)&ret) |= ( (unsigned long long) a->d[1] << 32);
# endif	/* WIN32 */
#endif	/* USE_LONGLONG */
#else	/* __palmos__ */
    /* On the Pilot, the BN_ULONG is not necessarily 32 bits - it's 16. Thus we
     * need some other way to obtain 64 bits than the above casting, which will
     * only work when BN_ULONG is 32 bits. Note that this little code replaces
     * all of the above non-pilot code, and provides maximum portability.
     * --Perm */
    PGPKEYID_ZERO(ret);			/* Clear destination */
    j = 64 / BN_BITS2;			/* Determine number of iterations */
    j = min(a->top, j);
    for (i = 0; i < j; i++) {
	((BN_ULONG *)&ret)[j - i - 1] = a->d[i];	/* Copy bits */
    }
#endif	/* PILOT */

done: ;
    memcpy(key_id,&ret,sizeof(PGPKEYID));
    return(0);
}

int 
PGPKEYID_ExtractShort(PGPKEYID 	*key_id,
		      unsigned 	long *short_id)
{
    if ( key_id == NULL || short_id == NULL )
	return(-1);

#ifdef USE_LONGLONG
# ifdef WIN32
    (*short_id)=(unsigned long)((*key_id) & 0xffffffff);
# else
    (*short_id)=(*key_id) & 0xffffffff;
# endif
#else
# ifdef WIN32
    (*short_id)=(*(unsigned __int64 *)key_id) & 0xffffffff;
# else
    (*short_id)=(*(unsigned long long *)key_id) & 0xffffffff;
# endif
#endif 

    return(0);
}

/*
 * Take 8 bytes (in network order) and make a quad
 * was to_quad
 */
int 
PGPKEYID_FromBytes(PGPKEYID *key_id, u_char *buf)
{
    int 	i;
    PGPKEYID	q;

    if ( key_id == NULL || buf == NULL ) 
	return(-1);

    PGPKEYID_ZERO(q);
    for(i=0; i < 8; i++) {
#ifdef USE_LONGLONG
	q |= ((PGPKEYID) buf[i] << 8*(7-i));
#else
# ifdef WIN32
	(*(unsigned __int64 *)&q) |= ((unsigned __int64) buf[i] << 8*(7-i));
# else
	(*(unsigned long long *)&q) |= ((unsigned long long) buf[i] << 8*(7-i));
# endif
#endif
    }

    memcpy(key_id,&q,sizeof(PGPKEYID));
    return(0);
}

/*
 * Copy a quad into a buffer, in network order.
 * was copy_quad
 */
int
PGPKEYID_ToBytes(PGPKEYID 	*q, 
		 u_char        	*buf)
{
    int	i,j;

    for(j=0, i=7; i>=0; i--, j++) {
#ifdef USE_LONGLONG
        buf[j] =  (u_char) ((*q) >> 8*i) & 0xff;
#else
# ifdef WIN32
        buf[j] =  (u_char) ((*(unsigned __int64 *)q) >> 8*i) & 0xff;
# else
        buf[j] =  (u_char) ((*(unsigned long long *)q) >> 8*i) & 0xff;
# endif
#endif
    }
    return(0);
}

/*
 * This is how PGP generates "checksum" of a buffer.  It is only used
 * to add two bytes to a DEK in order to be able to verify that the
 * correct key  has been used.  I've read the sources for PGP to find
 * this out, in particular the file mpiio.c
 */
u_short
checksum(u_char	*buf,
	 int 	count)
{
    u_short cksum;
    int	    i;

    for(i = 0, cksum = 0; count > 0; i++, count--)
	cksum += buf[i];

    return (cksum);
}

/*
 * A buffer, containing a PGP MPI, is converted into a SSLeay
 * BIGNUM. If everything is OK, we return 0.  On error, we return -1.
 */
int
mpi2bignum(u_char	*buf,
	   int		buflen,
	   int		*used,
	   BIGNUM	**N)
{
    u_short	us;
    int		length, numbit;
    int		sofar;
    /* scratch */
    u_char	*p;

    if ( buf == NULL || buflen < 2 || used == NULL || N == NULL )
    {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	return(0);
    }
    sofar = 0;

    /* Length of N */
    memcpy( &us, buf, 2);
    numbit = (int)ntohs(us);
    length = numbit / NBBY;
    if ( numbit % NBBY  != 0 )
	length += 1;
    p = calloc( length, 1);
    if ( p == NULL ) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    sofar += 2;

    if ( buflen - (sofar + length) < 0) {
	free(p);
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    /* Read N */
    memcpy( p, buf+sofar, length);
    sofar += length;
    /* Turn N into a BIGNUM */
    *N = BN_bin2bn(p, length, NULL);
    memset(p, 0, length);
    free(p);
    if (*N == NULL) {
	pgplib_errno = PGPLIB_ERROR_SYSERR;
	return(-1);
    }
    *used = sofar;
    return(0);
}


/*
 * Used to be this macro.
 *   "length of length" field of packet, in bytes (1, 2, 4, 8 bytes):
 *   #define ctb_llength(ctb) ((int) 1 << (int) ((ctb) & CTB_LLEN_MASK))
 * Routine written by Thomas Roessler <roessler@guug.de>
 */

int 
ctb_llength(int ctb)
{
    int n;

    n = ctb & CTB_LLEN_MASK;

    if( ! is_ctb(ctb)) {
	pgplib_errno = PGPLIB_ERROR_BAD_CTB;
	return -1;
    }

    if(n > 2) {
	pgplib_errno = PGPLIB_ERROR_BAD_CTB_LENGTH;
	return -1;
    }

    return 1 << n;
}

#ifndef __palmos__
int 
PGPKEYID_FromAscii(char 	*str,
		   PGPKEYID	*key_id)
{
    char *endptr;
#ifdef USE_LONGLONG
    PGPKEYID tmp;
#else
# ifdef WIN32
    unsigned __int64 tmp;
# else
    unsigned long long tmp;
# endif
#endif

    if ( str == NULL || key_id == NULL )
	return(-1);

    tmp = strtouq(str, &endptr, 0);
    if ( *endptr != '\0' || endptr == str )
	return(-1);
#ifdef USE_LONGLONG
    (*key_id) = tmp;
#else
# ifdef WIN32
    (*(unsigned __int64 *)key_id) = tmp;
# else
    (*(unsigned long long *)key_id) = tmp;
# endif
#endif

  return(0);

}

int 
PGPKEYID_ToAscii(PGPKEYID	*key_id,
		 char 		*buf,
		 int 		maxsize)
{
    int ret;

    if ( key_id == NULL || buf == NULL || maxsize < 1 )
	return(-1);

#ifdef USE_LONGLONG
    ret = sprintf(buf,"%#.16" PRIx64,(*key_id));
#else
    ret = sprintf(buf,"%#.16" PRIx64,(*(unsigned long long *)key_id));
#endif

    return(ret);
}

char *
PGPKEYID_ToStr(PGPKEYID *key_id)
{
    static char buf[128];

    if (PGPKEYID_ToAscii(key_id,buf,sizeof(buf)) > 0)
	return buf;
    else
	return "";
}

#endif /* __palmos__ */



/* ------------------------------------------------------------------------ */
/* stuff that is not used but might be useful to still have handy for later */
/* ------------------------------------------------------------------------ */

#ifdef NOT_USED

/* BN_get_quad - convert a BN into a PGPKEYID and complain if the
 * 		 BN is too large to fit
 */
int 
BN_get_quad(BIGNUM *a, PGPKEYID *key_id)
{
    int i;
    PGPKEYID ret;

    /* detect zero specially */
    if ( BN_is_zero(a)) {
      	PGPKEYID_ZERO(ret);
	goto done;
    }

    i = BN_num_bytes(a);
    if ( i > sizeof(PGPKEYID)) {
	pgplib_errno = PGPLIB_ERROR_EINVAL;
	/*
	return ( ~0 );
	*/
	return (0);
    }

    if ( a->top == 1 ) {
#ifdef USE_LONGLONG
	ret = a->d[0];
#else
# ifdef WIN32
	(*(unsigned __int64 *)&ret) = a->d[0];
# else
	(*(unsigned long long *)&ret) = a->d[0];
# endif
#endif
	goto done;
    }

#ifdef USE_LONGLONG
    ret = a->d[0];
    ret |= ( (PGPKEYID) a->d[1] << 32);
#else
# ifdef WIN32
    (*(unsigned __int64 *)&ret) = a->d[0];
    (*(unsigned __int64 *)&ret) |= ( (unsigned __int64) a->d[1] << 32);
# else
    (*(unsigned long long *)&ret) = a->d[0];
    (*(unsigned long long *)&ret) |= ( (unsigned long long) a->d[1] << 32);
# endif
#endif

done: ;
    memcpy(key_id,&ret,sizeof(PGPKEYID));
    return(1);
}

/* I thought that this might be needed ... but I was wrong --tjh */
int BN_get_shortkeyid(BIGNUM *a,unsigned long *short_id)
{
    unsigned long ret;

    /* detect zero specially */
    if ( BN_is_zero(a)) {
      	ret=0;
	goto done;
    }

    /* XXX Byte order ? */
    ret = a->d[0];

done: ;
    (*short_id)=ret;
    return(1);
}

#endif /* NOT_USED */
