/*********************************************************************
 * Filename:      pubdecrypt.c
 * Description:
 * Author:        Thomas Roessler <roessler@guug.de>
 * Modified at:   Tue May 26 16:06:19 1998
 * Modified by:   Tage Stabell-Kulo <tage@acm.org>
 *
 * 	Copyright (c) Thomas Roessler
 * 	See the file COPYING for details
 *
 * $Id: pubdecrypt.c,v 1.13 1999/03/08 19:58:46 perm Exp $
 *
 ********************************************************************/
/* All PalmOS code by Per Harald Myrvang <perm@pasta.cs.uit.no> */


#include <pgplib.h>

#ifndef __palmos__
# include <assert.h>
# include <stdlib.h>
# include <string.h>
# ifdef WIN32
#  include <winsock2.h>  
# else	/* WIN32 */
#  include <netinet/in.h>
# endif	/* WIN32 */
#else	/* __palmos__ */
# include <System/Unix/netinet_in.h>
#endif	/* __palmos__ */

/* SSLeay */
#include <rsa.h>


/* take a secret key, a public-key encrypted packet,
 * a conventionally encrypted packet, and tries to
 * decrypt this stuff.
 *
 * Return value:
 *
 * 0 = ok, 1 = packet invalid, 2 = wrong secret key, -1 = other failure.
 *
 */

int
sec_decrypt_packet(PGPseckey_t 	*key,
		   u_char 	*PKE_packet,
		   int 		pke_l,
		   u_char 	*conv_packet,
		   int 		conv_l,
		   u_char 	**outbuf,
		   int 		*outlen)
{
    RSA 	*rsa_info;
    PGPpke_t 	pke;
    PGPKEYID 	seckeyid;
    int 	buflen;
    u_char 	*edek_buf;
    u_char 	*dek_buf;
    u_char 	*dek;
    u_short 	cksum;
    int 	i;

    if (key == NULL || PKE_packet == NULL || conv_packet == NULL
	|| outbuf == NULL || outlen == NULL)
	return -1;
    if ( conv_l <= 5 )
	return 1;

    
    *outbuf = NULL;

    i = buf2pke(PKE_packet, pke_l, &pke);
    if ( i != 0 )
	return i;

    if ( PGPKEYID_FromBN(&seckeyid,key->N) != 0 )
	return(-1);

    if(memcmp(&seckeyid, &pke.key_id, sizeof(PGPKEYID))) {
	free_pke(&pke);
	return 2;
    }

    buflen = BN_num_bytes(pke.E);
    edek_buf = calloc(buflen, 1);
    if ( edek_buf == NULL) {
	free_pke(&pke);
	return -1;
    }
    dek_buf = calloc(buflen, 1);
    if ( dek_buf == NULL) {
	free_pke(&pke);
	free(edek_buf);
	return -1;
    }

    BN_bn2bin(pke.E, edek_buf);

    rsa_info = RSA_new();
    rsa_info->n = BN_dup(key->N);
    rsa_info->e = BN_dup(key->E);
    rsa_info->d = BN_dup(key->D);

    /* And these in the hope that they will speed things up (they do, and
     * _very_ noticeably on the 16MHz PalmPilot, if I may say so) - they've
     * been mangled as specified in secencrypt.c.  --Perm */
    rsa_info->q = BN_dup(key->P);
    rsa_info->p = BN_dup(key->Q);
    rsa_info->dmp1 = BN_dup(key->dmq1);
    rsa_info->dmq1 = BN_dup(key->dmp1);
    rsa_info->iqmp = BN_dup(key->iqmp);

    if (pke.algorithm != RSA_ALGORITHM_BYTE)
    {
	pgplib_errno = PGPLIB_ERROR_BAD_ALGORITHM;
	return 1;
    }

    i = 0;
    i = RSA_private_decrypt(buflen, edek_buf, dek_buf, rsa_info,
			    RSA_PKCS1_PADDING);
    RSA_free(rsa_info);
    free(edek_buf);
    free_pke(&pke);

    if (i == -1 || i < 19) {
#ifndef __palmos__
# ifdef DEBUG
	ERR_print_errors_fp(stderr);	/* from SSLeay */
# endif	/* DEBUG */
#endif	/* not __palmos__ */
	free(dek_buf);
	return 1;
    }

    dek = dek_buf + i - 19;

    if(*dek++ != 1) {
	free(dek_buf);
	return 1;
    }

    cksum = checksum(dek, 16);
    /* For some reasons this works on the Pilot, but the original code doesn't,
     * giving a Fatal Exception. Possibly due to odd addresses. Quite a nut to
     * crack, used 4 days to track it down. Bugger it to heck! And Bacon!
     * -- Perm@pasta.cs.uit.no */
    if (((dek[16] << 8) | (dek[17])) != cksum) {
        free(dek_buf);
        return 2;
    }

    i = conventional_key_decrypt(conv_packet, dek, outlen, outbuf);
    free(dek_buf);
    return i;
}
