/*
 * RollingCubes
 *      The RollingCubes game
 *
 * Ken Shirriff         ken.shirriff@eng.sun.com
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Please refer to the file "license.txt"
 * for further important copyright and licensing information.
 */

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "cube.h"
#define AppType    'KSrc'
#define AppDBType   'Data'
#define AppVersion  1

static void displayMoveCount(void);
static void EventLoop(void);
static Boolean cubeEvent(EventPtr event);
static void drawboxes();
static void drawbox(short x0, short y0, short w, short h, short dir);

enum {N=0, S=1, E=2, W=3, TOP=4, BOT=5, EMPTY=6} dir;

// Dimensions of the board: offset of the upper-right corner and size of a box
#define offx 27
#define offy 34
#define box_size 36
// Padded box size including gap
#define box_sizep (box_size+1)

#define ROT_STEPS 10

// sintab and costab of 0 to 90 in steps of ROT_STEPS multiplied by box_size
unsigned char sintab[ROT_STEPS] = {0, 6, 11, 16, 21, 25, 29, 32, 34, 36};
unsigned char costab[ROT_STEPS] = {36, 36, 34, 32, 29, 25, 21, 16, 11, 6};

struct state {
  unsigned int version;
  short moves;			// Number of moves player has made
  char data[3][3];		// State of the board
  short emptyx, emptyy;		// Position of the empty square
};

struct state board = {
  AppVersion,
  0,
  {{TOP, TOP, TOP},
  {TOP, EMPTY, TOP},
  {TOP, TOP, TOP}},
  1, 1
};

// Patterns for drawing the cubes
unsigned short patterns[3][4] = {
  { 0xffff, 0xffff, 0xffff, 0xffff},
  { 0x0000, 0x0000, 0x0000, 0x0000},
  { 0x8800, 0x2200, 0x8800, 0x2200}
};

// The data specifies the direction of the black face of each cube.
// The movetab specifies the new direction after a roll in the
// given direction.
char movetab[4][6] = {
  {BOT,TOP,E,W,N,S}, //Roll N
  {TOP,BOT,E,W,S,N}, //Roll S
  {N,S,BOT,TOP,E,W}, //Roll E
  {N,S,TOP,BOT,W,E} //Roll W
};

// Test if player has won
static short win() {
  int i, j;
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      if (i==1 && j==1) {
        if (board.data[i][j] != EMPTY) {
          return 0;
        }
      } else {
        if (board.data[i][j] != BOT) {
          return false;
        }
      }
    }
  }
  return 1;
}

// Restart the game
static void restart() {
  short i, j;
  board.emptyx = 1;
  board.emptyy = 1;
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      board.data[i][j] = TOP;
    }
  }
  board.data[board.emptyx][board.emptyy]=EMPTY;
  board.moves = 0;
  drawboxes();
}

#define moveLeft 110
#define moveTop 2
#define moveWidth 50
#define moveHeight 11

// Display move count
static void displayMoveCount(void) {
  char buf[20];
  RectangleType r;
  r.topLeft.x = moveLeft;
  r.topLeft.y = moveTop;
  r.extent.x = moveWidth;
  r.extent.y = moveHeight;
  WinEraseRectangle(&r, 0);
  StrCopy(buf, "Moves: ");
  StrIToA(buf+StrLen(buf), board.moves);
  WinDrawChars(buf, StrLen(buf), moveLeft, moveTop);
}

// Draw all the boxes
static void drawboxes()
{
    FormPtr frm;
    char moves[6];
    short i, j;
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	drawbox(offx+i*box_sizep, offy+j*box_sizep, box_size, box_size, board.data[i][j]);
      }
    }
    WinDrawGrayLine(offx-2, offy-2, offx+3*box_sizep-1, offy-2);
    WinDrawGrayLine(offx+3*box_sizep-1, offy-2, offx+3*box_sizep-1,
      offy+3*box_sizep-1);
    WinDrawGrayLine(offx+3*box_sizep-1, offy+3*box_sizep-1, offx-2,
      offy+3*box_sizep-1);
    WinDrawGrayLine(offx-2, offy+3*box_sizep-1, offx-2, offy-2);

    // Display the move counter
    displayMoveCount();
}

// Draw the box at the specified position
static void drawbox(short x0, short y0, short w, short h, short dir) {
  RectangleType r;
  r.topLeft.x = x0;
  r.topLeft.y = y0;
  if (dir == TOP || dir == BOT || dir == EMPTY) {
    WinSetPattern(patterns[dir-TOP]);
    r.extent.x = w-1;
    r.extent.y = h-1;
    WinFillRectangle(&r, 0);
  } else if (dir==N || dir==S) {
    WinSetPattern(patterns[dir==N?0:1]);
    r.extent.x = w-1;
    r.extent.y = h/2;
    WinFillRectangle(&r, 0);
    r.topLeft.y += h/2;
    r.extent.y = h-h/2-1;
    WinSetPattern(patterns[dir==S?0:1]);
    WinFillRectangle(&r, 0);
  } else if (dir==E || dir==W) {
    WinSetPattern(patterns[dir==W?0:1]);
    r.extent.x = w/2;
    r.extent.y = h-1;
    WinFillRectangle(&r, 0);
    WinSetPattern(patterns[dir==E?0:1]);
    r.topLeft.x += w/2;
    r.extent.x = w-w/2-1;
    WinFillRectangle(&r, 0);
  }
  WinDrawGrayLine(x0-1, y0-1, x0-1, y0+h-1);
  WinDrawGrayLine(x0-1, y0+h-1, x0+w-1, y0+h-1);
  WinDrawGrayLine(x0+w-1, y0+h-1, x0+w-1, y0-1);
  WinDrawGrayLine(x0+w-1, y0-1, x0-1, y0-1);
}

// Main loop
DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  if (cmd == sysAppLaunchCmdNormalLaunch)
 {
    FrmGotoForm(formID_cube);

    EventLoop();  // Event loop
  }
  return 0;
}

// Handle a click on a box
static void PenEvent(enum events etype, short nx, short ny)
{
  short i;
  if (etype == penDownEvent) {
    short bx = (nx-offx)/box_sizep;
    short by = (ny-offy)/box_sizep;
    short diffx, diffy;
    short rot_dir, rot_old, rot_new;
    if (nx<offx || bx>=3 || ny<offy || by>=3) {
	return;
    }
    // Figure out which direction we're going
    diffx = board.emptyx - bx;
    diffy = board.emptyy - by;
    if (diffx == 0) {
      if (diffy == -1) {
	rot_dir = N;
      } else if (diffy == 1) {
	rot_dir = S;
      } else {
	return;
      }
    } else if (diffy == 0) {
      if (diffx ==1) {
	rot_dir = E;
      } else if (diffx == -1) {
	rot_dir = W;
      } else {
	return;
      }
    } else {
      return;
    }

    rot_old = board.data[bx][by]; // Old cube position
    rot_new = movetab[rot_dir][rot_old]; // New cube position

    // Do the rotate effect
    for (i=0; i<ROT_STEPS; i++) {
      short x0 = offx+box_sizep*bx;
      short y0 = offy+box_sizep*by;
      short x1 = offx+box_sizep*board.emptyx;
      short y1 = offy+box_sizep*board.emptyy;
      WinSetPattern(patterns[dir-TOP]);
      switch (rot_dir) {
	case N:
	  drawbox(x0,y0+costab[i],box_size,box_size-costab[i], EMPTY);
	  drawbox(x0,y0-sintab[i],box_size,costab[i],rot_old);
	  drawbox(x0,y0-sintab[i]+costab[i],box_size,sintab[i],rot_new);
	  break;
	case S:
	  drawbox(x0,y0,box_size,box_size-costab[i], EMPTY);
	  drawbox(x0,y0+box_size-costab[i]+sintab[i],box_size,costab[i],rot_old);
	  drawbox(x0,y0+box_size-costab[i],box_size,sintab[i],rot_new);
	  break;
	case E:
	  drawbox(x0,y0,box_size-costab[i],box_size, EMPTY);
	  drawbox(x0+box_size-costab[i]+sintab[i],y0,costab[i],box_size,rot_old);
	  drawbox(x0+box_size-costab[i],y0,sintab[i],box_size,rot_new);
	  break;
	case W:
	  drawbox(x0+costab[i],y0,box_size-costab[i],box_size, EMPTY);
	  drawbox(x0-sintab[i],y0,costab[i],box_size,rot_old);
	  drawbox(x0-sintab[i]+costab[i],y0,sintab[i],box_size,rot_new);
	  break;
      }
      SysTaskDelay(5);
    }

    // Update the data

    board.data[board.emptyx][board.emptyy] = rot_new;
    board.emptyx = bx;
    board.emptyy = by;
    board.data[board.emptyx][board.emptyy] = EMPTY;

    board.moves ++;

    // Redraw everything
    drawboxes();
    if (win()) {
      FrmAlert(alertID_win);
      restart();
    }
  }
}

static void EventLoop(void)
{
  short err;
  int formID;
  FormPtr form;
  EventType event;

  do
  {

    EvtGetEvent(&event, evtWaitForever);

    if (SysHandleEvent(&event)) continue;
    if (MenuHandleEvent((void *)0, &event, &err)) continue;

    if (event.eType == frmLoadEvent)
    {
      formID = event.data.frmLoad.formID;
      form = FrmInitForm(formID);
      FrmSetActiveForm(form);
      switch (formID) 
      {
      case formID_cube:
        FrmSetEventHandler(form, (FormEventHandlerPtr) cubeEvent);
        break;
      }
    }

    FrmDispatchEvent(&event);
  } while(event.eType != appStopEvent);
}

static Boolean cubeEvent(EventPtr event)
{
  FormPtr   form;
  int       handled = 0;

  switch (event->eType) 
  {
  case frmOpenEvent:
    form = FrmGetActiveForm();
    FrmDrawForm(form);
    restart();
    handled = 1;
    break;
    
  case menuEvent:
    switch (event->data.menu.itemID)
    {
    case menuitemID_about:
      FrmAlert(alertID_about);
      break;
    case menuitemID_instr:
      FrmAlert(alertID_instr);
      break;
    case menuitemID_restart:
      restart();
      break;
    }
    handled = 1;
    break;

  case penDownEvent:
    PenEvent(event->eType, event->screenX, event->screenY);
    handled = 1;
    break;

  }
  return handled;
}
