#include <PalmOS.h>
#include "SaveToHost.h"


/***********************************************************************
 *   Entry Points
 ***********************************************************************/


/***********************************************************************
 *   Internal Structures
 ***********************************************************************/
typedef struct {
    FontID appFontID;
    Boolean saveBackup;
}StoHPreferenceType;

typedef struct {
    UInt16					renamedCategories;	// bitfield of categories with a different name
    Char 					categoryLabels[dmRecNumCategories][dmCategoryLength];
    UInt8 					categoryUniqIDs[dmRecNumCategories];
    UInt8					lastUniqID;	// Uniq IDs generated by the device are between
    // 0 - 127.  Those from the PC are 128 - 255.
    UInt8					reserved1;	// from the compiler word aligning things
    UInt16					reserved2;
    CountryType 		country;		// Country the database (labels) is formatted for
} StoHAppInfoType;

typedef StoHAppInfoType* StoHAppInfoPtr;


/***********************************************************************
 *   Global variables
 ***********************************************************************/
static Boolean HideSecretRecords;
static DmOpenRef StoHDB;
static FontID AppFontID = stdFont;
static Boolean SaveBackup = true;


/***********************************************************************
 *   Internal Constants
 ***********************************************************************/

#define appFileCreator             'StoH'
#define appVersionNum              0x01
#define appPrefID                  0x00
#define appPrefVersionNum          0x01
#define appDBName                  "StoHDB"
#define appDBType                  'DATA'
// Version Checking define
#define version20                  0x02000000


StoHAppInfoPtr StoHAppInfoGetPtr(DmOpenRef *dbP);
Err	StoHAppInfoInit(DmOpenRef *dbP );
/***********************************************************************
 *   Internal Functions
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  formId - id of the form to display
 *
 * RETURNED:    void *
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void * GetObjectPtr(UInt16 objectID )
{
    FormPtr frmP;


    frmP = FrmGetActiveForm();
    return (FrmGetObjectPtr( frmP, FrmGetObjectIndex( frmP, objectID)));
}


/***********************************************************************
 *
 * FUNCTION: StoHAppInfoGetPtr
 *
 * DESCRIPTION: Return a locked pointer to the AppInfo or NULL
 *
 * PARAMETERS: dbP - open database pointer
 *
 * RETURNS: locked ptr to the AppInfo or NULL
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
StoHAppInfoPtr StoHAppInfoGetPtr(DmOpenRef *dbP )
{
    UInt16 cardNo;
    LocalID dbIDID;
    LocalID appInfoID;


    if (DmOpenDatabaseInfo(dbP, &dbIDID, NULL, NULL, &cardNo, NULL))
        return NULL;
    if (DmDatabaseInfo( cardNo, dbIDID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
        return NULL;
    if (appInfoID == NULL)
        return NULL;
    else
        return MemLocalIDToLockedPtr( appInfoID, cardNo);
}


/***********************************************************************
 *
 *  FUNCTION: StoHAppInfoInit
 *
 *  DESCRIPTION: Create an app info chunk if missing.  Set
 *		the strings to a default.
 *
 *  PARAMETERS: db - open database pointer
 *
 *  RETURNS: 0 if successful, errorcode if not
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
Err	StoHAppInfoInit(DmOpenRef *dbP )
{
    UInt16 cardNo;
    LocalID dbID;
    LocalID appInfoID;
    MemHandle tempH;
    StoHAppInfoPtr appInfoP;
    StoHAppInfoPtr defaultApplicationInfoP;
    Int16 loop;


    appInfoP = StoHAppInfoGetPtr(dbP);
    // If there isn't an ApplicationInfo make space for one
    if (appInfoP == NULL) {
        if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
            return dmErrInvalidParam;

        if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
            return dmErrInvalidParam;

        tempH = DmNewHandle(dbP, sizeof(StoHAppInfoType));
        if (!tempH) 
            return dmErrMemError;

        appInfoID = MemHandleToLocalID( tempH);
        DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL);
        appInfoP = MemHandleLock(tempH);
    }


    // Allocate & Clear the app info
    defaultApplicationInfoP = MemPtrNew(sizeof(StoHAppInfoType));
    if (defaultApplicationInfoP == NULL) {
        ErrDisplay("Unable to init StoHDB");
        return 1;
    }


    MemSet( defaultApplicationInfoP, sizeof(StoHAppInfoType), 0);
    // Init the categories
    for (loop = 0; loop < dmRecNumCategories; loop++) {
        defaultApplicationInfoP->categoryUniqIDs[loop] = loop;
    }


    defaultApplicationInfoP->lastUniqID = dmRecNumCategories - 1;
    // copy in the defaults and free the default app info
    DmWrite(appInfoP, 0, defaultApplicationInfoP,  sizeof(StoHAppInfoType));
    MemPtrFree(defaultApplicationInfoP);
    // Try to use localized app info block strings.
    //StoHLocalizeAppInfo(appInfoP);
    // Localize the field labels to the current country
    //StoHChangeCountry(appInfoP);
    // Unlock
    MemPtrUnlock(appInfoP);
    return 0;
}


/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version meets your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags that indicate if the application 
 *                                UI is initialized.
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 ***********************************************************************/
static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
    UInt32 romVersion;


    // See if we have at least the minimum required version of the ROM or later.
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion) {
        if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
            // TODO: Display an Alert with an alert UI
            // FrmAlert(<RomIncompatibleAlert>);
            // Pilot 1.0 will continuously relaunch this app unless we switch to 
            // another safe one.
            if (romVersion < version20) {


                AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
            }
        }
        return sysErrRomIncompatible;
    }
    return 0;
}

static void FormClose(FormPtr form, Boolean deleteRec) 
{
	FieldPtr	aField;
	MemHandle	tempRec;
	MemPtr	tempPtr;

	UInt16		index=0;
	UInt32		uniqueID;

	aField = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainFormMainField));
	FldSetTextHandle(aField, NULL);
	DmReleaseRecord(StoHDB, index, true);
}

static void UpdateScrollBar(FormPtr form)
{
	UInt16 scrollPos;
	UInt16 textHeight;
	UInt16 fieldHeight;
	Int16 maxValue;
	FieldPtr fld;
	ScrollBarPtr bar;

	fld = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainFormMainField));
	bar = FrmGetObjectPtr(form, FrmGetObjectIndex(form, Scrollbar001));
	
	FldGetScrollValues(fld, &scrollPos, &textHeight,  &fieldHeight);

	if (textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if (scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;

	SclSetScrollBar (bar, scrollPos, 0, maxValue, fieldHeight-1);
}

static void FieldScroll(FormPtr form, Int16 linesToScroll)
{
	UInt16			blankLines;
	Int16			min;
	Int16			max;
	Int16			value;
	Int16			pageSize;
	FieldPtr		fld;
	ScrollBarPtr	bar;
	
	fld = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainFormMainField));

	if (linesToScroll < 0)
		{
		blankLines = FldGetNumberOfBlankLines(fld);
		FldScrollField(fld, -linesToScroll, winUp);
		
		// If there were blank lines visible at the end of the field
		// then we need to update the scroll bar.
		if (blankLines)
			{
			// Update the scroll bar.
			bar = FrmGetObjectPtr(form, FrmGetObjectIndex(form, Scrollbar001));
			SclGetScrollBar (bar, &value, &min, &max, &pageSize);
			if (blankLines > -linesToScroll)
				max += linesToScroll;
			else
				max -= blankLines;
			SclSetScrollBar (bar, value, min, max, pageSize);
			}
		}

	else if (linesToScroll > 0)
		FldScrollField (fld, linesToScroll, winDown);
}

static void PageScroll(FormPtr form, Int16 direction)
{
	Int16 seekDirection;
	UInt16 category;
	UInt16 recordNum;
	Int16 value;
	Int16 min;
	Int16 max;
	UInt32 uniqueID;
	Int16 pageSize;
	UInt16 linesToScroll;
	FieldPtr fld;
	ScrollBarPtr bar;

	fld = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainFormMainField));
	
	if (FldScrollable(fld, direction))
		{
		linesToScroll = FldGetVisibleLines(fld) - 1;
		FldScrollField (fld, linesToScroll, direction);

		// Update the scroll bar.
		bar = FrmGetObjectPtr(form, FrmGetObjectIndex(form, Scrollbar001));
		SclGetScrollBar (bar, &value, &min, &max, &pageSize);

		if (direction == winUp)
			value -= linesToScroll;
		else
			value += linesToScroll;
		
		SclSetScrollBar (bar, value, min, max, pageSize);
	}
}

/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:  frm - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void MainFormInit(FormPtr frmP)
{
	MemHandle	hdlRecOne;
	MemPtr		ptrRecOne;
	UInt16		index=0;
	
	FormPtr		form;
	FieldPtr	aField;
	
	if (DmNumRecords(StoHDB) == 0) {
		hdlRecOne = DmNewRecord(StoHDB, &index, sizeof(Char));
		ptrRecOne = MemHandleLock(hdlRecOne);
		DmWrite(ptrRecOne, 0, "", sizeof(Char));
		MemHandleUnlock(hdlRecOne);
		
		DmReleaseRecord(StoHDB, index, true);
	}

	form = FrmGetFormPtr(MainForm);

	aField = FrmGetObjectPtr (form, FrmGetObjectIndex(form, MainFormMainField));
	aField->attr.hasScrollBar = 1;
	hdlRecOne = DmGetRecord(StoHDB, index);
	FldSetTextHandle(aField, hdlRecOne);

	UpdateScrollBar(form);

	FrmSetFocus(form, FrmGetObjectIndex(form, MainFormMainField));
}

void SaveToHost()
{
	HostFILE*	file;
	MemHandle	hdlRecOne;
	MemPtr		ptrRecOne;
	
	FormPtr		form;
	FieldPtr	aField;
	
	long		temp;

	form = FrmGetFormPtr(MainForm);

	aField = FrmGetObjectPtr (form, FrmGetObjectIndex(form, MainFormMainField));
	hdlRecOne = FldGetTextHandle(aField);
	ptrRecOne = MemHandleLock(hdlRecOne);
	
	file = HostFOpen("paste.txt", "w");
	
	temp = HostFPutS(ptrRecOne, file);
	
	temp = HostFClose(file);
	
	MemHandleUnlock(hdlRecOne);
	
	FrmAlert(alrtSaved);
}

void ReadFromHost()
{
	HostFILE*	file;
	
	MemHandle	hdlRecOne;
	MemPtr		ptrRecOne;
	UInt16		index=0;
	
	MemHandle	hdlBuffer;
	MemPtr		buffer;
	
	FormPtr		form;
	FieldPtr	aField;
	
	long		temp;

	form = FrmGetFormPtr(MainForm);

	aField = FrmGetObjectPtr (form, FrmGetObjectIndex(form, MainFormMainField));
	FldSetTextHandle(aField, NULL);
	DmReleaseRecord(StoHDB, index, true);
	
	buffer = MemPtrNew(aField->maxChars - 1);
	
	file = HostFOpen("paste.txt", "r");
	
	temp = HostFRead(buffer, 1, MemPtrSize(buffer), file);
	((char *) buffer)[temp] = '\0';
	
	temp = HostFClose(file);

	hdlRecOne = DmResizeRecord(StoHDB, index, StrLen(buffer) + 10);
	ptrRecOne = MemHandleLock(hdlRecOne);
	
	DmStrCopy(ptrRecOne, 0, buffer);
	
	MemHandleUnlock(hdlRecOne);
	
	MemPtrFree(buffer);
	
	FldSetTextHandle(aField, hdlRecOne);
	FldDrawField(aField);

	UpdateScrollBar(form);

	FrmSetFocus(form, FrmGetObjectIndex(form, MainFormMainField));
	
	FrmAlert(alrtRead);
}

/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean MainFormDoCommand(UInt16 command)
{
    FieldPtr fldP;
    Boolean handled = true;

	FormPtr		activeForm;
	UInt16		focusIndex;
	UInt16		focusID;

	activeForm = FrmGetActiveForm();
	focusIndex = FrmGetFocus(activeForm);
	if (focusIndex != -1)
		focusID = FrmGetObjectId(activeForm, focusIndex);
	else
		focusID = 0;

    switch (command) {
		case UndoItem:
			switch (focusID) {
				case MainFormMainField:
					FldUndo(FrmGetObjectPtr(activeForm, focusIndex));
					break;

				default:
					handled = false;
					break;
			}
			break;

		case CutItem:
			switch (focusID) {
				case MainFormMainField:
					FldCut(FrmGetObjectPtr(activeForm, focusIndex));
					break;

				default:
					handled = false;
					break;
			}
			break;

		case CopyItem:
			switch (focusID) {
				case MainFormMainField:
					FldCopy(FrmGetObjectPtr(activeForm, focusIndex));
					break;

				default:
					handled = false;
					break;
			}
			break;

		case PasteItem:
			switch (focusID) {
				case MainFormMainField:
					FldPaste(FrmGetObjectPtr(activeForm, focusIndex));
					break;

				default:
					handled = false;
					break;
			}
			break;

		case SelectAllItem:
			switch (focusID) {
				case MainFormMainField:
					fldP = FrmGetObjectPtr(activeForm, focusIndex);
					FldSetSelection(fldP, 0, FldGetTextLength (fldP));
					break;

				default:
					handled = false;
					break;
			}
			break;
			
		case SaveToHostItem:
			SaveToHost();
			handled = true;
			break;
		
		case ReadFromHostItem:
			ReadFromHost();
			handled = true;
			break;
			
		case AboutItem:
			FrmAlert(alrtAbout);
			break;

        default:
            handled = false;
            break;

    }


    return handled;
}

/***********************************************************************
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the
 *              "MainForm" of this application.
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent( EventType * eventP)
{
    Boolean handled = false;
    FormPtr frmP=FrmGetActiveForm();

    switch (eventP->eType) {
        case menuEvent:
            return MainFormDoCommand(eventP->data.menu.itemID);

        case keyDownEvent:
            if (TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)) {
                // SysHandleEvent saw these keys and is now switching apps. 
                // Leave the function.
                return false;
            }

            switch (eventP->data.keyDown.chr) {
                case pageUpChr:
					PageScroll(frmP, winUp);
             		handled = true;
                   break;
                case pageDownChr:
					PageScroll(frmP, winDown);
//            		handled = true;
                    break;
                default:
                    break;
            }
            break;
            
        case frmOpenEvent:
            MainFormInit( frmP);
            FrmDrawForm ( frmP);
            handled = true;
            break;

		case frmSaveEvent:
		case frmCloseEvent:
			FormClose(frmP, false);
			break;

		case fldChangedEvent:
			UpdateScrollBar(frmP);
			handled = true;
			break;

		case sclRepeatEvent:
			FieldScroll(frmP, eventP->data.sclRepeat.newValue - eventP->data.sclRepeat.value);
			break;

        case frmUpdateEvent:
            /*
            MainFormUpdateDisplay( eventP->data.frmUpdate.updateCode);
            handled = true;
            */
            break;

        default:
            break;
    }

    return handled;
}

/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Boolean AppHandleEvent( EventType * eventP)
{
    UInt16 formId;
    FormPtr frmP;


    if (eventP->eType == frmLoadEvent) {
        // Load the form resource.
        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm( formId);
        FrmSetActiveForm( frmP);
        // Set the event handler for the form.  The handler of the currently
        // active form is called by FrmHandleEvent each time is receives an
        // event.
        switch (formId) { //{{PFW_FORM_LOAD
            case MainForm:
                FrmSetEventHandler(frmP, MainFormHandleEvent);
                break;
            default:
                //ErrNonFatalDisplay("Invalid Form Load Event");
                break;

        } //}}PFW_FORM_LOAD

        return true;
    }

    return false;
}

/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppEventLoop(void)
{
    UInt16 error;
    EventType event;


    do {
        EvtGetEvent(&event, evtWaitForever);
        if (! SysHandleEvent(&event))
            if (! MenuHandleEvent(0, &event, &error))
                if (! AppHandleEvent(&event))
                    FrmDispatchEvent(&event);
    } while (event.eType != appStopEvent);
}

/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  This routine opens the application's resource file and
 *               database.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value 0 if nothing went wrong
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static Err AppStart(void)
{
    Err error= 0;
    UInt16 mode;
    SystemPreferencesType sysPrefs;
    StoHPreferenceType prefs;
    UInt16 prefsSize;
    StoHAppInfoPtr appInfoP;


    // Determime if secert record should be shown.
    PrefGetPreferences(&sysPrefs);
    HideSecretRecords = sysPrefs.hideSecretRecords;
    if (sysPrefs.hideSecretRecords)
        mode = dmModeReadWrite;
    else
        mode = dmModeReadWrite | dmModeShowSecret;
    // Find the application's data file.  If it doesn't exist create it.
    StoHDB = DmOpenDatabaseByTypeCreator( appDBType, appFileCreator, mode);
    if (!StoHDB) {
        error = DmCreateDatabase(0, appDBName, appFileCreator, appDBType, false);
        if (error) 
            return error;

        StoHDB = DmOpenDatabaseByTypeCreator(appDBType, appFileCreator, mode);
        if (!StoHDB)
            return 1;

        error = StoHAppInfoInit(StoHDB);
        if (error) 
            return error;
    }

    appInfoP = (StoHAppInfoPtr) StoHAppInfoGetPtr(StoHDB);
    ErrFatalDisplayIf(appInfoP == NULL, "Missing app info block");
    // Update the database to look and behave properly for the given country.
    if (appInfoP->country != sysPrefs.country) {
        // StoHChangeCountry(appInfoP);
        // Read the preferences / saved-state information.  There is only one
        // version of the Applications preferences so don't worry about multiple
        // versions.
        prefsSize = sizeof(StoHPreferenceType);
        if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, true) != 
                noPreferenceFound) {
            AppFontID = prefs.appFontID;
            SaveBackup = prefs.saveBackup;
        }

    }

    MemPtrUnlock(appInfoP);
    return error;
}

/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: This routine close the application's database
 *              and save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static void AppStop(void)
{
	FrmSaveAllForms();
	FrmCloseAllForms();
	
	DmCloseDatabase(StoHDB);
}

/***********************************************************************
 * FUNCTION:    StoHPilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
static UInt32 StoHPilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    Err error;


    error = RomVersionCompatible(version20, launchFlags);
    if (error) 
        return error;

    switch (cmd) {
        case sysAppLaunchCmdNormalLaunch:
            error = AppStart();
            if (error) 
                return error;
            FrmGotoForm(MainForm);
            AppEventLoop();
            AppStop();
            break;
        case sysAppLaunchCmdSaveData:
            FrmSaveAllForms();
            break;
        case sysAppLaunchCmdInitDatabase:
            //AppLaunchCmdDatabaseInit(((SysAppLaunchCmdInitDatabaseType*)cmdPBP)->dbP);
            break;
        default:
            break;
    }
    return 0;
}
/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 ***********************************************************************/
UInt32 PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    return StoHPilotMain( cmd, cmdPBP, launchFlags);
}

