/*
 * TaleLight v1.21
 *
 * TaleLight Communicator Package
 * Copyright 1998 Daniel McCarty
 *
 * (for revisions, see history.txt)
 * (for legal issues, see end of file)
 *
 * Version History
 * 1.21 + removed Start button on Blink Rate form; functionality is
 *        now handled through OK button
 *      + fixed timing bug if the sound was not enabled; code in an
 *        if statement was never being reached
 *      + added list-view (15x9) icon
 *      + fixed OS3.0 bug with no handled = true after FrmReturnToForm
 *        in BREventHandler
 *
 * 1.2  + added feature to send individual characters from the Morse
 *        code table (MCEventHandler, etc...)
 *      + fixed a bug in the Morse code table where numbers were 1 off
 *      * 4/9/98: I think this is the final release!
 *
 * 1.1  + (again) made app OS 1.0 compliant
 *        (--this time it works!  thx, Alan Weiner!)
 *      + some better code here and there (MemMove(), etc)
 *      + "Disable Auto-off" feature
 *      + Morse "Stop" feature
 *
 * 1.0a + (hopefully) made app 1000 and 5000 complaint
 *        (Latest: according to Susan McMullen, still doesn't work)
 *      + fixed tap-and-drag-out bug on TaleLight button (thx, Dan Royea)
 *      + removed "Graffiti" menu item (for 1000/5000 compatibility)
 *        (replaced "Graffiti" menu item with beep for 1000 and 5000's)
 *
 * 1.0  + initial release (see history.txt, readme.txt)
 *
 */

// #pragma pack(2) isn't necessary anymore with GCC 0.5.0
// (structure packing on 2-byte boundaries is automatic now)

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include <SerialMgr.h>        // not included in SysAll.h
#include <SysEvtMgr.h>        // ditto

#define NON_PORTABLE          // for hardware.h
#include <Hardware/Hardware.h>

#include "tl.h"
#include "callback.h"


#define TLAppID         'dmTL'// registered with Palm/3Com
#define TLAppVer        1

#define MAXITEMS        5     // max items available in listID_text

#define MINEVENTTIME    5     // minimum period to poll for events

#define MININTERVAL     1
#define MAXINTERVAL     208

#define DASH_DELAY      5     // multipliers for morseWait[]
#define DOT_DELAY       2


// there are 10 rows in the Morse code table
#define MorseTableRow(x)    (15 + (x % 10) * 12)
#define MorseTableCol(x)    ( 4 + (x / 10) * 36)
#define MorseTableColPos(x) ( 4 +  x * 36)
// The difference between MorseTableCol and MorseTableColPos is that
// Col turns a grid index into a column position, while ColPos turns
// a column index into a column position.


short morseChars[40] =
   {  5,    // A .-         (101)
     24,    // B -     (11000)
     26,    // C --     (11010)
     12,    // D -       (1100)
      2,    // E            (10)
     18,    // F -     (10010)
     14,    // G --       (1110)
     16,    // H      (10000)
      4,    // I          (100)
     23,    // J ---     (10111)
     13,    // K --       (1101)
     20,    // L -     (10100)
      7,    // M --         (111)
      6,    // N -         (110)
     15,    // O ---       (1111)
     22,    // P --     (10110)
     29,    // Q ---     (11101)
     10,    // R -       (1010)
      8,    // S        (1000)
      3,    // T -           (11)
      9,    // U -       (1001)
     17,    // V -     (10001)
     11,    // W --       (1011)
     25,    // X --     (11001)
     27,    // Y ---     (11011)
     28,    // Z --     (11100)
     85,    // . --- (1010101)
     76,    // ? -- (1001100)
     50,    // / --   (110010)
     49,    // = --   (110001)
     63,    // 0 -----   (111111)
     47,    // 1 ----   (101111)
     39,    // 2 ---   (100111)
     35,    // 3 --   (100011)
     33,    // 4 -   (100001)
     32,    // 5    (100000)
     48,    // 6 -   (110000)
     56,    // 7 --   (111000)
     60,    // 8 ---   (111100)
     62     // 9 ----   (111110)
   };
   // 40 characters, encoded in one byte each; the first 1 in the binary
   // value signals the start of the char; after that, a
   //    0 bit is a dot
   //    1 bit is a dash
   // (this scheme wastes at least 40 bits--oh, no!)

short morseWait[5] = { 36, 30, 21, 15, 9 };
   // these "wait values" are arbitrary; there may be a timing
   // table somewhere out there, but we don't know where it is.

char *blinkMetronome[8] =
     { "40-60Largo",
     "60-66Larghetto",
     "66-76Adagio",
     "76-108Andante",
     "108-120Moderato",
     "120-168Allegro",
     "168-200Presto",
     "200-208Prestissimo"
   };
   // the unprintable characters are ASCII 128 (numericSpaceChar)
short blinkIntervalMetronome[8] = { 40, 60, 66, 76, 108, 120, 168, 200 };
   // metronome values for the metronome list

char *blinkSeconds[6] =
     { "second",
     "2 seconds",
     "5 seconds",
     "10 seconds",
     "30 seconds",
     "minute"
   };
short blinkIntervalSeconds[6] = { 1, 2, 5, 10, 30, 60 };
   // seconds values for the "seconds" list

enum {
   statusOff = 0,
   statusOn,
   statusBlinking,
   statusPending,
   statusMorsing
};

typedef struct {
   // application preferences struct
   int     morseText;               // index into items[]
   int     morseSpeed;              // index into morseWait[]
   Boolean morseRepeat;
   Boolean syncSound;
   Boolean syncBacklight;
   char    items[MAXITEMS][32];     // limited by MAXCHARS in tl.rcp
   int     itemCount;               // current number of items in items[]
   // it would save time and space to just always have 5 items, but it
   // looks more professional to start out with a few samples and move to 5

   Boolean blinkSeconds;            // blink in seconds or beats/min
   Boolean blinkItem;               // index into blink-rate popup list
   int     blinkInterval;           // value in blink-rate text field
   Boolean blinkDisableAutoOff;     // Disable Auto-off?

   SWord   selectedMorseChar;       // selected character in Morse table
} TLPrefType;
TLPrefType TLPrefs;


// general utility functions
static VoidPtr GetObjectPtr(int objectID);         // from memopad.c
static Boolean ItemInList(char *item, ListPtr lst);
static void SetFieldText(FieldPtr fld, char *text);

// specific utility functions
static void SetBlinkInterval(void);
static void StopActivity(void);
static void ToggleBlinkStartStop(void);
static void ToggleMorseStartStop(void);
static void ToggleTaleOnOff(void);
static void UpdateIntervalList(char **listChoices, int itemCount);
static void UpdateTextsList(void);

// program functions
static Boolean InitComm(void);
static void CloseComm(void);
static Boolean SendText(void);
static void SendChar(char c);
static void SendCharMain(int index);
static void SendSound(int delay);
static Boolean GetBacklight(void);
static Boolean SetBacklight(Boolean fOn);
static void DrawBlinkRateText(void);

// PalmOS-required functions
static int StartApplication(void);
static void StopApplication(void);
static void EventLoop(void);
static Boolean AppHandleEvent(EventPtr event);
static Boolean TLEventHandler(EventPtr event);
static Boolean BREventHandler(EventPtr event);
static Boolean MCEventHandler(EventPtr event);


// This is a weird place to declare this, since it's only referenced in
// MCEventHandler, but initializing rects out here costs _much_ less
RectangleType  morseTable = {{2, 12}, {141, 123}};

FieldPtr       fldPtrText;

UInt           serialRef;
short          TLStatus;

int            ticksInterval;

DWord          PPBacklight;



DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags) {
   int error;

   if (cmd == sysAppLaunchCmdNormalLaunch) {
      error = StartApplication();
      if (error)
         return error;

      EventLoop();
      StopApplication();
   } else
      return sysErrParamErr;

   return 0;
}


static int StartApplication(void) {
   // ToDo: can we test for a physical serial connection with HwrPluggedIn()?
   int error;

   // might as well try to do this first; if it bombs, we don't waste time
   error = SysLibFind("Serial Library", &serialRef);
   if (error) {
      FrmCustomAlert(alertID_taleLight, "TaleLight could not initialize the serial library.", "", "");
      return error;
   }

   // determine if the Pilot we're on has a backlight or not
   FtrGet(sysFtrCreator, sysFtrNumBacklight, &PPBacklight);

   if (PrefGetAppPreferencesV10(TLAppID, TLAppVer, &TLPrefs, sizeof(TLPrefType)) == NULL) {
      // initialize preferences
      TLPrefs.morseText = 0;                    // "SOS"
      TLPrefs.morseSpeed = 2;                   // "Normal"
      TLPrefs.morseRepeat = false;
      TLPrefs.syncSound = false;                // don't synchronize sound
      TLPrefs.syncBacklight = false;            // don't synchronize backlight
      StrCopy(TLPrefs.items[0], "SOS");         // default items
      StrCopy(TLPrefs.items[1], "HI MOM");
      StrCopy(TLPrefs.items[2], "TESTING 123");
      TLPrefs.itemCount = 3;                    // number of default items

      TLPrefs.blinkSeconds  = true;             // blink in seconds
      TLPrefs.blinkItem     = 1;                // blink once every 2 seconds
      TLPrefs.blinkInterval = 2;                // value in field
      TLPrefs.blinkDisableAutoOff = true;       // don't automatically turn off

      TLPrefs.selectedMorseChar = 0;            // default to A in the table
   }

   SetBlinkInterval();
   FrmGotoForm(formID_tl);

   return 0;
}


static void StopApplication(void) {
   FldSetTextHandle(fldPtrText, NULL);

   if (TLStatus == statusOn)
      CloseComm();

   // write the current preference information
   PrefSetAppPreferencesV10(TLAppID, TLAppVer, &TLPrefs, sizeof(TLPrefType));
}


static void EventLoop(void) {
   Word      error;
   EventType event;

   static ULong ticksStart = 0;

   do {
      // handle the passing of an event from one form to another
      // a better way to do this might be through FrmUpdateForm()
      if (TLStatus == statusPending)
         ToggleBlinkStartStop();

      // we mandate that if this case is true, we process it first
      if ((TLStatus == statusBlinking) && (TimGetTicks() >= ticksStart + ticksInterval)) {
         ticksStart = TimGetTicks();
         SendChar('e');             // one short beep

         if (TLPrefs.blinkDisableAutoOff)
            EvtResetAutoOffTimer(); // reset the auto-off timer at every beat
      }

      // if we're doing nothing, wait forever; otherwise,
      // we need a much faster event loop (5ms)
      EvtGetEvent(&event, (TLStatus ? MINEVENTTIME : evtWaitForever));

      if (!SysHandleEvent(&event))
         if (!MenuHandleEvent(NULL, &event, &error))
            if (!AppHandleEvent(&event))
               FrmDispatchEvent(&event); 
   } while (event.eType != appStopEvent);
}


static Boolean AppHandleEvent(EventPtr event) {
   int     formID;
   FormPtr frm;
   Boolean handled = false;

   if (event->eType == frmLoadEvent) {
      formID = event->data.frmLoad.formID;
      frm = FrmInitForm(formID);
      FrmSetActiveForm(frm);

      switch (formID) {
      case formID_tl:
         FrmSetEventHandler(frm, (FormEventHandlerPtr) TLEventHandler);
         break;
      case formID_blinkRate:
         FrmSetEventHandler(frm, (FormEventHandlerPtr) BREventHandler);
         break;
      case formID_morse:
         FrmSetEventHandler(frm, (FormEventHandlerPtr) MCEventHandler);
         break;
      }
      handled = true;
   }

   return handled;
}


static Boolean TLEventHandler(EventPtr event) {
// TaleLight event handler
   FormPtr    frm;
   ListPtr    lst;
   ControlPtr ctl;

   int        i,
              handled = false;
   CharPtr    ptr;
   DWord      PPVer;

   CALLBACK_PROLOGUE       // see callback.h

   switch (event->eType) {
   case frmOpenEvent:
      frm = FrmGetActiveForm();
      fldPtrText = GetObjectPtr(fieldID_text);

      UpdateTextsList();

      // set the text field to the first entry in the popup list
      lst = GetObjectPtr(listID_text);
      SetFieldText(fldPtrText, LstGetSelectionText(lst, LstGetSelection(lst)));

      // initialize the morse speed list and its popup label
      lst = GetObjectPtr(listID_speed);
      LstSetSelection(lst, TLPrefs.morseSpeed);
      ctl = GetObjectPtr(poptrigID_speed);
      CtlSetLabel(ctl, LstGetSelectionText(lst, LstGetSelection(lst)));

      // initialize the repeat button
      ctl = GetObjectPtr(buttonID_repeat);
      CtlSetValue(ctl, TLPrefs.morseRepeat);

      // set the "Sync Sound" checkbox
      ctl = GetObjectPtr(chkboxID_sound);
      CtlSetValue(ctl, TLPrefs.syncSound);

      FrmDrawForm(frm);
      FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_text));

      // Pilot1000 and 5000's have no backlight; respond accordingly
      // WinDrawChars() must be placed after any FrmDrawForm()
      ctl = GetObjectPtr(chkboxID_backlight);
      // or, if (GHwrMiscFlags & hwrMiscFlagHasBacklight)
      if (PPBacklight & 1)           // test bit0
         CtlSetValue(ctl, TLPrefs.syncBacklight);
      else {
         CtlHideControl(ctl);
         // 100 is a trial-and-error value, 146 is from tl.rcp
         WinDrawChars("Backlight: N/A", 14, 100, 146);
      }

      handled = true;
      break;

   case menuEvent:
      switch (event->data.menu.itemID) {
      case menuitemID_morse:
         // ToDo: implement this on a table instead?
         FrmPopupForm(formID_morse);
         break;
      case menuitemID_help:
         FrmHelp(helpID_tips);
         break;
      case menuitemID_about:
         frm = FrmInitForm(formID_about);
         FrmDoDialog(frm);
         FrmDeleteForm(frm);
         break;
      case menuitemID_undo:
         FldUndo(fldPtrText);
         break;
      case menuitemID_cut:
         FldCut(fldPtrText);
         break;
      case menuitemID_copy:
         FldCopy(fldPtrText);
         break;
      case menuitemID_paste:
         FldPaste(fldPtrText);
         break;
      case menuitemID_keyboard:
         SysKeyboardDialogV10();
         break;
      case menuitemID_graffiti:
         FtrGet(sysFtrCreator, sysFtrNumROMVersion, &PPVer);
         if (PPVer > 2)
            // this only works on OS2.0 or greater
            SysGraffitiReferenceDialog(referenceDefault);
         else
            // the other way to do this is to create two menu resources and
            // use FrmSetMenu() to switch between 1.0 and 2.0 menues; this
            // isn't worth the code and resources it would take
            SndPlaySystemSound(sndError);
         break;
      }
      handled = true;
      break;

   case ctlEnterEvent:
      if (event->data.ctlEnter.controlID == buttonID_taleLight) {
         StopActivity();
         i = InitComm();
         if (TLPrefs.syncSound)
            // one short beep when the pen goes down
            SendSound(morseWait[TLPrefs.morseSpeed] * DOT_DELAY);

         // so the OS will draw the inverted button
         handled = false;
      }
      break;

   case ctlExitEvent:
      if (event->data.ctlEnter.controlID == buttonID_taleLight) {
         CloseComm();
         handled = true;
      }
      break;

   case ctlSelectEvent:
      switch (event->data.ctlEnter.controlID) {

      case buttonID_repeat:
         TLPrefs.morseRepeat = !TLPrefs.morseRepeat;
         break;

      case buttonID_send:
         if (TLStatus) {
            // if toggling the Morse status stopped it, then this was
            // a cancel; break and get out of here
            if (TLStatus == statusMorsing) {
               ToggleMorseStartStop();
               break;
            } else
               StopActivity();
         }

         ToggleMorseStartStop();

         if (SendText()) {
            // only add the text if the user didn't cancel
            // add text to the list of sent items if it isn't already there
            // if it _is_ in the list, move it to the top of the list
            lst = GetObjectPtr(listID_text);
            ptr = FldGetTextPtr(fldPtrText);
            if (ptr) {
               // ToDo: can this code be optimized (combine the for loops?)
               if (!ItemInList((char *) ptr, lst)) {
                  for (i = min(MAXITEMS - 1, TLPrefs.itemCount); i > 0; i--)
                     StrCopy(TLPrefs.items[i], TLPrefs.items[i - 1]);

                  TLPrefs.itemCount = min(MAXITEMS, TLPrefs.itemCount + 1);
               } else {
                  for (i = LstGetSelection(lst); i > 0; i--)
                     StrCopy(TLPrefs.items[i], TLPrefs.items[i - 1]);
               }
               StrCopy(TLPrefs.items[0], ptr);
               UpdateTextsList();
               LstSetSelection(lst, 0);
            }
         }

         if (TLStatus == statusMorsing)
            ToggleMorseStartStop();

         handled = true;
         break;

      case buttonID_blinkrate:
         StopActivity();
         FrmPopupForm(formID_blinkRate);
         handled = true;
         break;

      case buttonID_taleLight:
         CloseComm();
         handled = true;
         break;

      case seltrigID_taleOnOff:
         if (TLStatus && TLStatus != statusOn)
            StopActivity();
         ToggleTaleOnOff();
         handled = true;
         break;

      case seltrigID_blinkStartStop:
         if (TLStatus && TLStatus != statusBlinking)
            StopActivity();
         ToggleBlinkStartStop();
         handled = true;
         break;

      case chkboxID_sound:
         TLPrefs.syncSound = !TLPrefs.syncSound;
         handled = true;
         break;

      case chkboxID_backlight:
         TLPrefs.syncBacklight = !TLPrefs.syncBacklight;
         handled = true;
         break;
      }
      break;

   case popSelectEvent:
      if (event->data.popSelect.listID == listID_text) {
         lst = GetObjectPtr(listID_text);
         SetFieldText(fldPtrText, LstGetSelectionText(lst, LstGetSelection(lst)));

         // don't let the OS complete the redraw
         handled = true;
      }
      else {
         lst = GetObjectPtr(listID_speed);
         TLPrefs.morseSpeed = LstGetSelection(lst);

         handled = false;
      }
      break;

   }

   CALLBACK_EPILOGUE       // see callback.h

   return handled;
}


static Boolean BREventHandler(EventPtr event) {
// Blink Rate event handler
   CharPtr  ptr;
   FormPtr  frm;
   ListPtr  lst;
   ControlPtr ctl;

   int      interval,
            handled = false;
   char     tempInterval[3];

   static TLPrefType TLPrefsTemp;
   static FieldPtr   fieldPtr_interval;

   CALLBACK_PROLOGUE       // see callback.h

   switch (event->eType) {
   case frmOpenEvent:
      // make a local copy of the Prefs struct
      MemMove(&TLPrefsTemp, &TLPrefs, sizeof (TLPrefType));
      frm = FrmGetActiveForm();
      FrmSetControlGroupSelection(frm, groupID_units, (TLPrefs.blinkSeconds ? buttonID_seconds : buttonID_BPM));

      FrmDrawForm(frm);
      DrawBlinkRateText();

      ctl = GetObjectPtr(chkboxID_disableAutoOff);
      CtlSetValue(ctl, TLPrefs.blinkDisableAutoOff);

      // initialize "combo box" field
      fieldPtr_interval = GetObjectPtr(fieldID_interval);
      StrIToA(tempInterval, TLPrefs.blinkInterval);
      SetFieldText(fieldPtr_interval, tempInterval);
      FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_interval));

      handled = true;
      break;

   case ctlSelectEvent:
      switch (event->data.ctlEnter.controlID) {
      case buttonID_seconds:
         TLPrefs.blinkSeconds = true;
         DrawBlinkRateText();
         break;

      case buttonID_BPM:
         TLPrefs.blinkSeconds = false;
         DrawBlinkRateText();
         break;

      case chkboxID_disableAutoOff:
         TLPrefs.blinkDisableAutoOff = !TLPrefs.blinkDisableAutoOff;
         break;

      case buttonID_OK:
         ptr = FldGetTextPtr(fieldPtr_interval);
         if (ptr)
            // set the interval to the most current value in the field
            interval = StrAToI(ptr);

         // ensure user-defined interval is within bounds
         if (interval < MININTERVAL || interval > MAXINTERVAL) {
            FrmAlert(alertID_interval);
            break;
         }
         // only set the status pending after passing the interval test
         TLStatus = statusPending;

         TLPrefs.blinkInterval = interval;
         SetBlinkInterval();
         // fall through again...aaiiieeeeeeeeeee!!!... (isn't this fun!)
      case buttonID_cancel:
         // restore prefs if the user pressed the cancel button
         if (event->data.ctlSelect.controlID == buttonID_cancel)
            MemMove(&TLPrefs, &TLPrefsTemp, sizeof(TLPrefType));

         FldSetTextHandle(fieldPtr_interval, NULL);
         FldEraseField(fieldPtr_interval);
         FrmReturnToForm(formID_tl);

         handled = true;
         break;
      }
      break;

   case popSelectEvent:
      lst = GetObjectPtr(listID_interval);
      TLPrefs.blinkItem = LstGetSelection(lst);
      if (TLPrefs.blinkSeconds)
         TLPrefs.blinkInterval = blinkIntervalSeconds[LstGetSelection(lst)];
      else
         TLPrefs.blinkInterval = blinkIntervalMetronome[LstGetSelection(lst)];
      // sync the interval field
      StrIToA(tempInterval, TLPrefs.blinkInterval);
      SetFieldText(fieldPtr_interval, tempInterval);

      // don't let the OS complete the redraws, etc.
      // (if the OS handles the redraws it draws over the field)
      handled = true;
      break;
   }

   CALLBACK_EPILOGUE       // see callback.h

   return handled;
}


static Boolean MCEventHandler(EventPtr event) {
// Morse Code event handler
   Boolean handled = false;

   int    i = 0,
          x = -1,
          y = -1;
   SWord  selectedChar,
          penX = event->screenX,
          penY = event->screenY;
   static RectangleType morseTableChars[40];

   CALLBACK_PROLOGUE       // see callback.h

   selectedChar = TLPrefs.selectedMorseChar;
   switch (event->eType) {
   case frmOpenEvent:
      // build dimensions of the selection rectangle in the Morse code table
      // this whole pseudo-table is a size-cheap implementation of a table
      for (i = 0; i < 40; i++) {
         morseTableChars[i].topLeft.x = MorseTableCol(i);
         morseTableChars[i].topLeft.y = MorseTableRow(i);
         morseTableChars[i].extent.x = 32;
         morseTableChars[i].extent.y = 13;
      }

      FrmDrawForm(FrmGetActiveForm());
      WinInvertRectangle(&morseTableChars[selectedChar], 3);

      handled = true;
      break;

   case penDownEvent:
      WinDisplayToWindowPt(&penX, &penY);
      if (RctPtInRectangle(penX, penY, &morseTable)) {
         // find x-pos for selection rectangle
         i = 0;
         // -3 = spacing of selection rect
         while (penX > (MorseTableColPos(i) - 3) && i < 4) {
            x++; i++;
         }
         // find y-pos for selection rectangle
         i = 0;
         while (penY > (MorseTableRow(i) - 3) && i < 10) {
            y++; i++;
         }

         // restore the previous selection rectangle
         WinInvertRectangle(&morseTableChars[selectedChar], 3);
         // resolve x-, y-coordinates to grid index
         selectedChar = x * 10 + y;
         WinInvertRectangle(&morseTableChars[selectedChar], 3);

         handled = true;
      }

      break;

   case ctlSelectEvent:
      switch (event->data.ctlEnter.controlID) {
      case buttonID_morseSend:
         SendCharMain(selectedChar);
         handled = true;
         break;

      case buttonID_morseOK:
         FrmReturnToForm(formID_tl);
         handled = true;
         break;
      }
      break;
   }
   TLPrefs.selectedMorseChar = selectedChar;

   CALLBACK_EPILOGUE       // see callback.h

   return handled;
}


static Boolean InitComm(void) {
   // thanks to maus.c (Iain Barclay) for some of the serial code
   Err error;

   // toggle the backlight
   SetBacklight(!GetBacklight());
   // 0 is the only serial port; the 19200 for Baud is arbitrary
   // (300 caused a "flicker" every time the TaleLight was turned on)
   error = SerOpen(serialRef, 0, 19200L);
   if (error) {
      FrmCustomAlert(alertID_taleLight, "TaleLight could not activate the serial port.", "", "");
      return false;
   }

   return true;
}


static void CloseComm(void) {
   // toggle the backlight
   SetBacklight(!GetBacklight());
   SerClose(serialRef);

   return;
}


static Boolean GetBacklight(void) {
   if (PPBacklight & 1)     // test bit0
      return HwrBacklight(false, false);
   else
      return false;
}


static Boolean SetBacklight(Boolean fOn) {
   if (!TLPrefs.syncBacklight)
      return false;

   // or, if (GHwrMiscFlags & hwrMiscFlagHasBacklight) ?
   if (PPBacklight & 1)  // test bit0
      return HwrBacklight(true, fOn);
   else
      return false;
}


static Boolean SendText(void) {
   int  i = 0;
   char text[32];             // limited by MAXCHARS in tl.rcp

   StrCopy(text, FldGetTextPtr(fldPtrText));
   do {
      // pause for the length specified by the blink interval, up to 2s
      if (i)                  // only run this delay after the first time
         SysTaskDelay(min(ticksInterval, sysTicksPerSecond * 2));

      // if user cancels in SendChar(), TLStatus != statusMorsing
      for (i = 0; i < StrLen(text) && TLStatus == statusMorsing; i++) {
         // highlight the character being sent
         FldSetSelection(fldPtrText, i, i + 1);
         FldDrawField(fldPtrText);
    
         if (i)               // only run a delay after the first character
            SysTaskDelay(morseWait[TLPrefs.morseSpeed]);
         SendChar(text[i]);
      }
   // repeat indefinitely if the repeat button is selected
   } while (TLStatus == statusMorsing && TLPrefs.morseRepeat);

   // unhilight last char
   FldSetSelection(fldPtrText, i, i);

   // user may have cancelled the send
   return (TLStatus == statusMorsing);
}


static void SendChar(char c) {
   int index;

   // this does add overhead to not place this mini-func in SendCharMain(),
   // but it allows us to send a char by its ASCII value or array index
   if (c == '.')              // ASCII 46
      index = 26;
   else
   if (c == '?')              // ASCII 63
      index = 27;
   else
   if (c == '/')              // ASCII 47
      index = 28;
   else
   if (c == '=')              // ASCII 61
      index = 29;
   else
   if (c >= '0' && c <= '9')  // ASCII 48 - 57
      index = c - 18;
   else
   if (c >= 'A' && c <= 'Z')  // ASCII 65 - 90
      index = c - 65;
   else
   if (c >= 'a' && c <= 'z')  // ASCII 97 - 122
      index = c - 97;
   else
      return;

   SendCharMain(index);
}


static void SendCharMain(int index) {
   int i,
       bit;
   short err;
   EventType event;

   // get past the initial 1 (start indicator)
   i = 7;
   bit = 0;
   while (!bit)
      bit = (1 << i--) & morseChars[index];

   for (i; i >= 0; i--) {
      // only test this if we're sending Morse
      if (TLStatus == statusMorsing) {
         // this is the most responsive place to put this code
         EvtGetEvent(&event, 0);
         if (event.eType)     // only do this if we have a real event
            if (!SysHandleEvent(&event))
               if (!MenuHandleEvent(NULL, &event, &err))
                  FrmDispatchEvent(&event);
         // TLStatus is indicative if something else was pressed/started
         if (TLStatus != statusMorsing)
            break;
      }

      InitComm();

      // grab a bit and translate it to a dash or a dot
      bit = (1 << i) & morseChars[index];
      if (TLPrefs.syncSound)
         SendSound(morseWait[TLPrefs.morseSpeed] * (bit ? DASH_DELAY : DOT_DELAY));
         // bit = 1 or 0, morseWait[] gets multiplied by the dot or dash multiplier
      else
         SysTaskDelay(morseWait[TLPrefs.morseSpeed] * (bit ? DASH_DELAY : DOT_DELAY) * sysTicksPerSecond / 1000);
         // translate millisecond values into ticks per second
      // bug fix: for some reason the following line was in the code; the
      // condition was never reached because of course TLStatus != 0 here
      //if (!TLStatus)

      CloseComm();

      if (TLStatus != statusOn && TLStatus != statusBlinking)
         SysTaskDelay(morseWait[TLPrefs.morseSpeed]);
         // this check (and the one above) is important because if
         // TLStatus == statusOn || statusBlinking then we don't
         // want these delays (so the event loop will run faster)
   }
}


static void SendSound(int delay) {
   SndCommandType sndCmd;

   sndCmd.cmd = sndCmdFreqDurationAmp;
   sndCmd.param1 = 554;    // middle C
   sndCmd.param2 = delay;
   sndCmd.param3 = sndMaxAmp;
   SndDoCmd(0, &sndCmd, true);
}


static void DrawBlinkRateText() {
   ListPtr  lst  = GetObjectPtr(listID_interval); 
   FieldPtr fld1 = GetObjectPtr(labelID_text1);
   FieldPtr fld2 = GetObjectPtr(labelID_text2);

   if (TLPrefs.blinkSeconds) {
      SetFieldText(fld1, "every");
      SetFieldText(fld2, "s");
      // there are 6 timing items
      UpdateIntervalList(blinkSeconds, 6);
      // 47 = FntCharsWidth("30 seconds", 10) + 3 -- the +3 is for spacing
      lst->bounds.extent.x = 47;
   } else {
      SetFieldText(fld1, "at");
      SetFieldText(fld2, "b/m");
      // there are 8 metronome items
      UpdateIntervalList(blinkMetronome, 8);
      // 89 = FntCharsWidth("200-208Prestissimo", 19) + 3 -- the +3 is for spacing
      lst->bounds.extent.x = 89;
   }
}


/* ----------------------------------------------------------------------------
 *
 * SPECIFIC UTILITY FUNCTIONS BELOW *
 */
static void SetBlinkInterval(void) {
/* In a way, this is a stupid function: a function that modifies globals.
   But its value is that it saves on the final PRC-size (8 bytes! Wow!)--
   the same code isn't in 2 places (StartApplication and BREventHandler).
*/
   if (TLPrefs.blinkSeconds)
      ticksInterval = sysTicksPerSecond * TLPrefs.blinkInterval;
   else
      ticksInterval = (sysTicksPerSecond * 60) / TLPrefs.blinkInterval;
}

static void StopActivity(void) {
   if (TLStatus == statusOn)
      ToggleTaleOnOff();
   else if(TLStatus == statusBlinking)
      ToggleBlinkStartStop();
   else if (TLStatus == statusMorsing)
      ToggleMorseStartStop();
}

static void ToggleBlinkStartStop(void) {
   ControlPtr ctl = GetObjectPtr(seltrigID_blinkStartStop);

   if (TLStatus == statusBlinking) {
      // when the label is "Start" the blinking should be off
      CtlSetLabel(ctl, "Start");
      TLStatus = statusOff;
   }
   else {
      CtlSetLabel(ctl, "Stop");
      TLStatus = statusBlinking;
   }
}

static void ToggleMorseStartStop(void) {
   ControlPtr ctl = GetObjectPtr(buttonID_send);

   if (TLStatus == statusMorsing) {
      // when the label is "Start" the blinking should be off
      CtlSetLabel(ctl, "Send Text");
      TLStatus = statusOff;
   }
   else {
      CtlSetLabel(ctl, "Stop");
      TLStatus = statusMorsing;
   }
}

static void ToggleTaleOnOff(void) {
   ControlPtr ctl = GetObjectPtr(seltrigID_taleOnOff);

   if (TLStatus == statusOn) {
      CloseComm();
      CtlSetLabel(ctl, "On");
      TLStatus = statusOff;
   }
   else {
      if (InitComm()) {
         CtlSetLabel(ctl, "Off");
         TLStatus = statusOn;
         if (TLPrefs.syncSound)
            // since we can't play a continuous tone this will have to do
            SendSound(morseWait[TLPrefs.morseSpeed] * DASH_DELAY);
      }
   }
}

static void UpdateIntervalList(char **listChoices, int itemCount) {
   ListPtr lst;

   lst = GetObjectPtr(listID_interval);
   LstSetListChoices(lst, listChoices, itemCount);
   LstSetSelection(lst, TLPrefs.blinkItem);
   LstSetHeight(lst, itemCount);
   return;
}

static void UpdateTextsList(void) {
   int i;
   ListPtr lst;
   // pulled this out of lottery.c (thanks Dave Kelly); We're not sure
   // why it works as opposed to char *items[MAXITEMS], but it does
   char **items = (char **) MemPtrNew(TLPrefs.itemCount * (sizeof(char *)));

   for (i = 0; i < TLPrefs.itemCount; i++)
      items[i] = TLPrefs.items[i];

   lst = GetObjectPtr(listID_text);
   LstSetListChoices(lst, items, TLPrefs.itemCount);
   LstSetHeight(lst, TLPrefs.itemCount);

   return;
}


/* --------------------------------------------------------------------------
 *
 * GENERAL UTILITY FUNCTIONS BELOW *
 */
static VoidPtr GetObjectPtr(int objectID) {
   // as many have noted, this function should've been in the API
   FormPtr frm;

   frm = FrmGetActiveForm();
   return (FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, objectID)));
}

static Boolean ItemInList(char *item, ListPtr lst) {
   int     i = 0;
   Boolean notFound = true;

   while (notFound && i < TLPrefs.itemCount) {
      notFound = StrCaselessCompare(item, LstGetSelectionText(lst, i));
      i++;
   }

   return !notFound;    // !notFound = found, of course  :)
}

static void SetFieldText(FieldPtr fld, char *text) {
/* These two lines accomplish the same thing and save 28 bytes.
   The downside is that it doesn't work on NONEDITABLE fields:
      FldDelete(fld, 0, FldGetTextLength(fld));
      FldInsert(fld, text, StrLen(text));
*/
   int      err;
   CharPtr  fldText;
   VoidHand fldHandle;

   FldFreeMemory(fld);
   fldHandle = MemHandleNew(StrLen(text) + 1);
   if (fldHandle) {
      fldText = MemHandleLock(fldHandle);
      StrCopy(fldText, text);
      MemHandleUnlock(fldHandle);
      FldSetTextHandle(fld, (Handle) fldHandle);
      FldDrawField(fld);
   }
}

/*
 *  TaleLight 1.21 public code release: December 15, 1998
 *
 * Disclaimers and other legal stuff:
 *  This code is provided AS-IS; there is NO WARRANTY, express or implied.
 *  The user assumes FULL RESPONSIBILITY AND LIABILITY for any use of
 *  this code.
 *
 *  Copyright (c)1998 Daniel McCarty
 *  All Rights Reserved by Daniel McCarty, except as noted below:
 *
 *  Permission is hereby granted to the public, free of charge, to
 *  distribute this package complete and unmodified, without restriction.
 *
 *  Modified versions may be distributed with this notice intact, provided
 *  that said versions are clearly marked as modified code.  The above
 *  provisions shall also apply to any such modified packages.
 *
 * NOTE: These terms are intended to protect _me_, not make this package any
 *       less useful to you. If you have a special need or use for this code,
 *       contact me and I'm sure we can work something out.
 *
 *       Warm regards,
 *       Daniel.
 *       mcdan@csi.com
 */