
/*
 * OutputBox.java
 *
 *
 * Copyright (c) 2000, David M. Archer
 *
 * See LICENSE file for more information.
 *
 */

package wabadma.ui;

import waba.ui.*;
import waba.fx.*;
import waba.util.*;

/**
 * A waba container providing text output.
 *
 * @author David M. Archer <dmarcher@pobox.com>
 *
 * @version 0.2, 2000 Nov 21 - added printFormattedStringArray & printFormatted.
 * @version 0.1, 2000 Nov 07 - initial version.
 *
 */
public class OutputBox extends Container {
	private	Font font;
	private	boolean displayed = false;
	private FontMetrics fm;

	/**
	 * Behaviour of OutputBox is to clear & return to top when the bottom
	 * is reached.  (Currently this is the only available behaviour.
	 */
	public static final int BEHAVIOUR_AUTOCLEAR = 1;

	/**
	 * Style of a line's text is normal (left-justified).
	 */
	public static final int STYLE_NORMAL = 1;

	/**
	 * Style of a line's text is centered.
	 */
	public static final int STYLE_CENTERED = 2;

	private int behaviour;
	private int lineheight;
	private int numlines;
	private int currentline;
	private Vector vec;


	private class LineData {
		public String text;
		public int style;
	
		public LineData(String text, int style) {
			this.text = text;
			this.style = style;
		}
	}


	/**
  	 * Construct an OutputBox with the given metrics and given font.
	 */
	public OutputBox(int x, int y, int w, int h, Font font) {
		this.setRect(x,y,w,h);
		this.font=font;
		fm=getFontMetrics(this.font);
		lineheight = fm.getHeight()+fm.getLeading();
		numlines = this.height/lineheight;
		currentline = 0;
		vec = new Vector();
		behaviour = BEHAVIOUR_AUTOCLEAR;
		displayed = true;

	//		System.err.println("new outputbox");
	//		System.err.println(this.height+" "+lineheight);
	//		System.err.println("numlines = "+numlines);
	}


	/**
	 * Construct an OutputBox with the given metrics using a default font.
	 *
	 */
	public OutputBox(int x, int y, int w, int h) {
		this(x,y,w,h,new Font("Helvetica", Font.PLAIN, 10));
	}

	/**
	 * Return true if the Dialog is currently being displayed.
	 *
	 */
	public boolean isActive() {
		return displayed;
	}


	/**
	 * Return true if s contains a \n
	 *
	 */
	private int newlineAt(String s) {
		for(int i=0;i < (s.length()-1);i++) 
			if (s.substring(i,i+1).equals("\n"))
				return i;
		return -1;
	}


	/**
	 * Print a styled line.
	 *
	 * @param string String to print
	 * @param style style to use
	 * @see #STYLE_NORMAL
	 * @see #STYLE_CENTERED
	 */
	public void printlnStyled(String string, int style) {
		String s;
		int a; // where we found a newline At

		s=string;
		while ((a = newlineAt(s)) != -1) {
			printS(s.substring(0,a),style);
			s = s.substring(a+1,s.length());
		//	printlnStyled(s.substring(a+1,s.length()),style);
		} 
		printS(s, style);
		this.repaint();
	}


	/**
	 *
	 * Word wrapping handled here.
 	 *
 	 */
	private void printS(String string, int style) {
		String s;
		int a;
		//int sa;// saved a for wordwrap (gets restored if no space)
		int ls; // last space seen

		s = string;
		while ((fm.getTextWidth(s))>this.width) {
			a=1;
			ls=-1;
			while(fm.getTextWidth(s.substring(0,a))<this.width) {
				if (s.charAt(a)==' ')
					ls=a;
				a++;

			}
			a--; // back off one, stop here if raw wrap
			//do word wrap
			//sa=a;
			/*
			while(s.charAt(a)!=' ') {
				a--;
				if (a == 0) {
					a=sa; // no space to break on
					break;
				}
			}
			*/
			if (ls!= -1)
				a=ls;
			//call self for each widthfull
			printA(s.substring(0,a),style);
			s=s.substring(a+1,s.length());			
		} 
		printA(s, style);
	}

	
	/**
	 * Line is added to vector here after being split for newlines and
	 * word wrapped.
	 *
	 */
	private void printA(String s, int style) {
		if (currentline == numlines) {
			// this is where the AUTO_CLEAR behaviour comes from	
			clear();
		}
		//System.err.println("output line "+currentline+": "+s);
		vec.add(new LineData(s,style));
		currentline++;
	}

	
	/**
	 * Print a centered line to the OutputBox.
	 *
	 */
	public void printlnCentered(String s) {
		printlnStyled(s, STYLE_CENTERED);
	}


	/**
	 * Print a line to the OutputBox.
	 *
	 */
	public void println(String s) {
		printlnStyled(s, STYLE_NORMAL);
	}


	/**
	 * Print an array of pre-formatted strings to the OutputBox using
	 * 	STYLE_NORMAL.
	 *
	 */

	public void printFormattedStringArray(String[] sa) {
		for(int i=0;i<sa.length;i++) 
			printA(sa[i], STYLE_NORMAL);
	}

		
	/**
	 * Print a pre-formatted string to the OutputBox using STYLE_NORMAL.
	 *
	 */

	public void printFormatted(String s) {
		printA(s, STYLE_NORMAL);
	}


	/**
	 * Print an empty line.
	 *
	 */
	public void println() {
		printlnStyled("",STYLE_NORMAL);
	}


	/**
	 * Clear the OutputBox.
	 *
	 */
	public void clear() {
		vec = new Vector();
		currentline = 0;
		this.repaint();
	}


	private void clearRect(Graphics g, int x, int y, int w, int h) {
                g.setColor(255,255,255);
                g.fillRect(x,y,w,h);
                g.setColor(0,0,0);
        }


	public void onPaint(Graphics g) {
		g.setFont(font);
		if (displayed) {
		//	System.err.println("OutputBox onPaint veclen="+vec.getCount());
                	int y=0;
	                for(int i=0;i<numlines;i++) {
       		                 clearRect(g,0,y,160,lineheight);
				 if (i<vec.getCount())
	       		                 if (vec.get(i) != null) {
						LineData ld = (LineData) vec.get(i);
						if (ld.text != null) {
						if (ld.style == STYLE_CENTERED) 
							g.drawText(ld.text,width/2-fm.getTextWidth(ld.text)/2,y);
						else
       	        		                 g.drawText(ld.text,0,y);
						}
					}
				y=y+lineheight;
                	}
		}
	}
}


