
/*
 * DLWRandom.java
 *
 *
 * Copyright (c) 2000, David M. Archer
 *
 * See LICENSE file for more information.
 *
 */

package wabadma.util;

import waba.sys.*;


/**
 * This class provides a non-secure pseudo-random number generator designed
 * for use with waba and PalmOS.  
 *
 * DLW stands for either Dave's Lame Waba (Random) or
 *                       Dave's LCG Waba (Random).
 *
 * @see "Applied Cryptography by Bruce Schneier (Ch.15, pg. 347, blue ed.)"
 *
 * @version 1.0, 2000 Nov 05
 * @author David M. Archer <dmarcher@pobox.com>
 */
public class DLWRandom {

	private final static int A = 4096 , B = 150889 , M = 714025;
	private int x;
	
	/**
	 * Creates a new DLWRandom object and seeds it with a value
    	 * based on the current time.  
 	 */
	public DLWRandom() {
		Time t = new Time();
		x = (t.millis+1) * (t.second+1);
	}

	/**
	 * Creates a new DLWRandom object and seeds it with the given value.
	 * @param i Value with which to seed the generator.
	 */
	public DLWRandom(int i) {
		x = i;
	}

	/**
   	 * Set the generator's register to a given value, essentially 
	 * reseeding the generator.  
       	 * @param i Value to reset (reseed) the generator with.
	 */
	public void set(int i) {
		x = i;
	}

	/**
	 * Get the generator's register without advancing it.  Repeated calls
	 * to this routine will return the same value.  Mostly only useful
	 * for test programs.
	 */
	public int get() {
		return x;
	}

	/**
	 * Get the next value in the generator's sequence.
	 */
	public int next() {
		x = ((A*x)+B) % M;
		return x;
	}

	private int abs(int i) {
		if (i >= 0)
			return i;
		else
			return -i;
	}

	/**
	 * Get a (non-secure, pseudo-)random value from 0..(i-1).
	 * @param i Modulus of random number wanted.
	 */
	public int random(int i) {
		return abs(this.next() % i);
	}
}


