
/*
 * DWCatalog.java
 *
 *
 * Copyright (c) 2000, David M. Archer
 *
 * See LICENSE file for more information.
 *
 */

package wabadma.util;

import waba.sys.*;
import waba.io.*;


/**
 * DWCatalog is an extension of waba's Catalog class for writing and 
 * reading Whipples to a PalmOS device.  This class is intended for use
 * by simple programs to save & load state when exited and restarted.
 *
 * Very little to no error checking is done.  When reading a database
 * typically one should use the inherited isOpen and getRecordCount methods 
 * after creating the DWCatalog object to check if the database is
 * existing, not empty, and sane.  When writing, one should typically 
 * clear the database after creating the DWCatalog object and before writing to
 * it.  
 *
 * @see Whipple
 *
 * @version 1.0, 2000 Nov 23
 * @author David M. Archer <dmarcher@pobox.com>
 */


public class DWCatalog extends Catalog {
	private int p; //pointer to next record


	/**
   	 *  Create a new DWCatalog object using the database name given.
	 *
	 * @param f name of database to use
	 *
	 */
	public DWCatalog(String f) {
		super(f, Catalog.CREATE);
		p = 0;
	}


	/**
	 *  Clear the current database.  All records are deleted, the database
	 *  itself is not removed.
	 *
	 */
	public void clear() {
		while ( getRecordCount() > 0 ) {
			if (setRecordPos(0)) {
				if (!deleteRecord())
					return;
			} else
				return;
		}
		p = 0;
	}


	/**
	 *  Write a Whipple to the end of the current database.
	 *
	 */
	public int writeNewWhipple(Whipple w) {
		String s = w.get();
		char[] ca = s.toCharArray();
		byte[] ba = new byte[ca.length];
		for (int i=0;i<ca.length;i++)
			ba[i] = (byte)ca[i];

		if ((p=addRecord(ba.length)) != -1)
			return writeBytes(ba, 0, ba.length);
		else
			return -1;
	}

	/**
	 *  Read the Whipple at the given position in the database.
	 *
	 * @param pos position in database of Whipple to read.
	 *
	 */
	public Whipple readWhipple(int pos) {
		setRecordPos(pos);
		byte[] ba = new byte[getRecordSize()];
		readBytes(ba, 0, getRecordSize());
		char[] ca = new char[ba.length];
		for (int i=0;i<ba.length;i++)
			ca[i] = (char)ba[i];
		p=pos+1;
		return new Whipple(new String(ca));
	}


	/**
	 *  Read the next Whipple in the database.
	 *
	 */
	public Whipple readWhipple() {
		return readWhipple(p);
	}
}

