
/*
 * Whipple.java
 *
 *
 * Copyright (c) 2000, David M. Archer
 *
 * See LICENSE file for more information.
 *
 */

package wabadma.util;

import waba.sys.*;


/**
 * This class provides a simple series of methods to pack several values into
 * a single string.   The value types are not checked upon retreival and
 * no escaping of the internal seprator char is done.  The default 
 * separator is '~', so if you use that in your strings, pick something else
 * when constructing your Whipple.
 *
 * @version 1.0, 2000 Nov 23
 * @author David M. Archer <dmarcher@pobox.com>
 */
public class Whipple {
	private StringBuffer s;
	private char sep;
	private final static char DEF_SEP='~';
	private final static int EOW=-1;
	private int p; // a pointer to the next value in the Whipple.
	

	/**
	 * Creates an empty Whipple object with given separator.
	 *
 	 */
	public Whipple(char c) {
		s = new StringBuffer();
		sep = c;
		p = EOW;
		s.append(sep);
	}


	/**
	 * Creates an empty Whipple object with a default separator.
	 *
 	 */
	public Whipple() {
		this(DEF_SEP);
	}


	/**
	 * Create a Whipple object from a passed string.
	 *
	 */
	public Whipple(String s) {
		this.s = new StringBuffer(s);
		if (s.length() == 0) {
			sep = DEF_SEP;
			this.s.append(sep);
			p = EOW;
		} else {
			sep = s.charAt(0);
			p = 0;
		}
	}
			

	/**
	 * Return the Whipple object as a string.
	 *
	 */
	public String get() {
		return s.toString();
	}


	/**
	 * Put a String value into the Whipple.
	 *
	 */
	public void put(String s) {
		this.s.append(s);
		this.s.append(sep);
	}

	/**
	 * Put an int value into the Whipple.
	 *
	 */
	public void put(int i) {
		this.put(Convert.toString(i));
	}


	/**
	 * Put a boolean value into the Whipple.
	 *
	 */
	public void put(boolean b) {
		if (b) 
			this.put("T");
		else
			this.put("F");
	}


	/**
	 * Get the next value in the Whipple as a String.
	 *
	 */
	public String nextString() {
		/* we start expecting p is on the separator preceeding the
		value we want.  we find q by looking forward for the next
		separator.  if there isn't one, we set EOW and return null.
		*/

		if (p == EOW) 
			return null;

		int q = p+1;
		String s2 = s.toString();  // waba's StringBuffer is lame.
		while (q < s2.length()) {
			if (s2.charAt(q) == sep) {
				String r = s2.substring(p+1,q);
				p = q;
				return r;
			}
			q++;
		}
		p = EOW;
		return null;
	}


	/**
	 * Get the next value in the Whipple as an int.
	 *
	 */
	public int nextInt() {
		return Convert.toInt(this.nextString());
	}


	/**
	 * Get the next value in the Whipple as a boolean.
	 *
	 */
	public boolean nextBoolean() {
		String s = this.nextString();

		if ((s.length()==1) && (s.charAt(0) == 'T'))
			return true;
		return false;
	} 
}


